# UNICADO - UNIversity Conceptual Aircraft Design and Optimization
#
# Copyright (C) 2025 UNICADO consortium
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#
# Description:
# This file is part of UNICADO.

# Import standard  library.
from numpy import nonzero


def calculate_seat_kilometer_costs(dict_ac_data, dict_operating_costs, runtime_output):
    """Calculate seat kilometer costs (SKC).

    The seat kilometer offered (SKO) is a measure of an aircraft's passenger carrying capacity or, in other words, its
    potential to generate revenue by providing available seats to passengers. They are calculated by multiplying the
    number of seats available by the range. The seat kilometer costs allow the analysis of a change in seat capacity
    and thus the assessment of the passenger kilometer potential.

    This function adds the 'seat_kilometer_costs' dict to the 'dict_operating_costs' that contains the following
    parameters:
        - array seat_kilometer_costs: Seat kilometer costs (range dependent) in EUR/SKO

    :param dict dict_ac_data: Dict containing parameters and values from aircraft exchange and module config file
    :param dict dict_operating_costs: Dict containing parameters from cost estimation calculation
    :param logging.Logger runtime_output: Logging object used for capturing log messages in the module
    :return dict dict_operating_costs: Dict containing parameters from cost estimation calculation
    """

    # Extract necessary values from 'dict_operating_costs'.
    direct_operating_costs_annual = dict_operating_costs['direct_operating_costs']['direct_operating_costs_annual']
    n_pax_for_prd = dict_operating_costs['payload_range_diagram_data']['n_pax_for_prd']
    ranges_for_prd = dict_operating_costs['payload_range_diagram_data']['ranges_for_prd']/1000

    # Calculate seat kilometer offered.
    seat_kilometer_offered = n_pax_for_prd * ranges_for_prd
    # Replace zeros to eliminate 'division by zero' warning.
    zero_sko = nonzero(seat_kilometer_offered == 0)[0]
    seat_kilometer_offered[zero_sko] = 1e-8
    # Calculate seat kilometer costs.
    seat_kilometer_costs = direct_operating_costs_annual / seat_kilometer_offered

    # Prepare output data.
    dict_operating_costs['seat_kilometer_costs'] = {}
    dict_operating_costs['seat_kilometer_costs']['seat_kilometer_costs'] = seat_kilometer_costs

    return dict_operating_costs
