/*
 * UNICADO - UNIversity Conceptual Aircraft Design and Optimization
 * 
 * Copyright (C) 2025 UNICADO consortium
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 * Description:
 * This file is part of UNICADO.
 */

#include <standardFiles/functions.h>

#include "standard_mission.h"

/**
 * @brief Construct a new StandardMission:: StandardMission object
 * 
 * @param toolName entry from toolinfo.h
 * @param toolVersion entry from toolinfo.h
 */
StandardMission::StandardMission(const std::shared_ptr<RuntimeIO>& rtIO) :
    rtIO(rtIO),
    mission_type(),
    mission_xml_name(),
    auto_select_initial_cruise_altitude(),
    auto_select_flight_level(),
    round_to_regular_flight_level(),
    no_steps(),
    /* Mission values */
    range("range", "Mission range", 0, "m", 0, 100000000),
    payload("payload", "Payload mass", 0, "kg", 0, 100000),
    pax_number("number_of_pax", "Number of passenger (Mass per PAX = " + num2Str(
        EndnodeReadOnly<double>("requirements_and_specifications/design_specification/transport_task/passenger_definition/mass_per_passenger").read(rtIO->acxml).value() +
        EndnodeReadOnly<double>("requirements_and_specifications/design_specification/transport_task/passenger_definition/luggage_mass_per_passenger").read(rtIO->acxml).value()) +
        " kg)", 0, "1", 0, 1000),
    cargo_mass("cargo_mass", "Cargo mass", 0, "kg", 0, 100000),
    initial_cruise_mach("desired_cruise_speed", "Planned cruise Mach number for fuel calculation", 0, "1", 0, 1),
    alternate_distance("alternate_distance", "Distance from destination to alternate aerodrome", 0, "m", 0, 10000000),
    taxi_time_origin("taxi_time_origin", "Taxi time at departure airport", 0, "s", 0, 10000),
    taxi_time_destination("taxi_time_destination", "Taxi time at destination", 0, "s", 0, 10000),
    taxiing_procedure("taxiing_procedure", "Taxiing procedure for start and landing."),
    engine_warmup_time("engine_warmup_time", "Running time of the engines before take-off", 0, "s", 0, 10000),
    terminal_operation_time("terminal_operation_time", "Time at the terminal for stopovers", 0, "s", 0, 10000),
    origin_airport("origin_airport", "ICAO code of origin"),
    destination_airport("destination_airport", "ICAO code of destination"),
    auto_climb_altitude_steps(),
    auto_rate_of_climb_steps(),
    cruise_step_number(),
    CAS_ATC_limit_climb(),
    CAS_ATC_limit_descent(),
    CAS_over_flight_level_100_climb(),
    CAS_over_flight_level_100_descent(),
    mach_climb(),
    initial_cruise_altitude(),
    delta_temperature_ISA(),
    delta_mach_climb_to_mach_cruise(),
    maximum_rate_of_climb(),
    approach_speed(),
    cruise_step_segments(),
    departure_steps(),
    cruise_steps(),
    approach_steps(),
    takeoff_procedure(),
    approach_procedure(),
    climb_thrust_setting(),
    use_delta_max_operating_mach(EndnodeReadOnly<bool>("program_settings/maximum_operating_mach_number/delta").read(rtIO->moduleConfig).value()),
    adapt_climb_speed(EndnodeReadOnly<bool>("program_settings/adapt_climb_speed_schedule/enable").read(rtIO->moduleConfig).value()),
    use_design_mission_mach(EndnodeReadOnly<bool>("program_settings/study_mission/copy_mach_number").read(rtIO->moduleConfig).value()),
    use_design_mission_ICO(EndnodeReadOnly<bool>("program_settings/study_mission/copy_initial_cruise_altitude").read(rtIO->moduleConfig).value()),
    delta_max_operating_mach(EndnodeReadOnly<double>("program_settings/maximum_operating_mach_number/delta").read(rtIO->moduleConfig).value()),
    crossover_altitude_specs(EndnodeReadOnly<double>("program_settings/adapt_climb_speed_schedule/crossover_altitude").read(rtIO->moduleConfig).value()) {
    }

void StandardMission::initialize() {
    myRuntimeInfo->info << "Initialize Mission Parameters ..." << std::endl;
    this->set_module_settings();
    this->set_range_and_payload();
    this->set_cruise();
    this->set_climb_and_descent();
    this->set_start_and_landing();
    this->set_performance_limits();
    // TODO(GERRIT): Still needed?
    /* Adjust maximum operating Mach number by delta to Mach_Cruise if necessary (for parameter studies with Mach variation) */
    // if (mySettingsPt->getUseDeltaMaxOperatingMachNumber()) {
    //     double maxOperatingMachNumber(designMission->getMachInitialCruise() + mySettingsPt->getDeltaMaxOperatingMachNumber());
    //     acXML.at("AcftExchangeFile/Requirements/DesignFlightEnvelope/MMO") = maxOperatingMachNumber;
    // }
}

void StandardMission::run() {
    /* Shell output */
    myRuntimeInfo->out << "***************** Gathered input *****************" << std::endl;
    myRuntimeInfo->out << "* Mission type:            " + this->mission_type << std::endl;
    /* Range */
    myRuntimeInfo->out << "* Range                    = " << convertUnit(METER, NAUTICALMILE, this->range.value()) << " NM" << std::endl;
    /* Payload */
    myRuntimeInfo->out << "* Payload                  = " << this->payload.value() << " kg" << std::endl;
    /* Initial cruise parameters */
    myRuntimeInfo->out << "* Cruise Mach number       = " << this->initial_cruise_mach.value() << std::endl; // Mach number
    myRuntimeInfo->out << "* Initial cruise altitude  = " << this->initial_cruise_altitude << " FL" << std::endl; // Altitude
    /* Speed schedule */
    myRuntimeInfo->out << "* Climb speed under FL100  = " << this->CAS_ATC_limit_climb << " m/s" << std::endl;
    myRuntimeInfo->out << "* Climb speed above FL100  = " << this->CAS_over_flight_level_100_climb << " m/s" << std::endl;
    myRuntimeInfo->out << "* Climb Mach number        = " << this->mach_climb << std::endl;
    myRuntimeInfo->out << "**************************************************" << std::endl;

    /* Set mission profiles */
    this->set_departure_profile();
    this->set_cruise_profile();
    this->set_approach_profile();
}

void StandardMission::update() {
    myRuntimeInfo->out << "Create " << this->mission_xml_name << "..." << std::endl;
    /* Check if the file directory exists */
    if (!fileExists(rtIO->getMissionDataDir())) {
        myRuntimeInfo->out << "Create directory for mission data output in project folder: " << rtIO->getMissionDataDir() << std::endl;
        createFolder(rtIO->getMissionDataDir().c_str());
    }
    /* Open file */
    std::ofstream outstream;
    outstream.open(rtIO->getMissionDataDir() + "/" + this->mission_xml_name.c_str());
    if (!outstream) { // File can still not be opened, error message
        myRuntimeInfo->err << "Could not open " << rtIO->getMissionDataDir() + "/" +  this->mission_xml_name << ". Abort progrm." << std::endl;
        exit(1);
    } else { // File can be opened, the generate mission XML output is created
        /* Version and encoding */
        outstream << "<?xml version=\"1.0\" encoding=\"utf-8\" ?>" << std::endl;
        outstream << "<mission>" << std::endl;
        outstream << "</mission>" << std::endl;
        outstream.close();
    }

    node& file(aixml::openDocument(rtIO->getMissionDataDir() + "/" + this->mission_xml_name));
    this->range.update(file.at("mission"));
    this->payload.update(file.at("mission"));
    this->pax_number.update(file.at("mission"));
    this->cargo_mass.update(file.at("mission"));
    this->initial_cruise_mach.update(file.at("mission"));
    this->alternate_distance.update(file.at("mission"));
    this->taxi_time_origin.update(file.at("mission"));
    this->taxi_time_destination.update(file.at("mission"));
    this->engine_warmup_time.update(file.at("mission"));
    this->terminal_operation_time.update(file.at("mission"));
    this->taxiing_procedure.update(file.at("mission"));
    if (!this->origin_airport.value().empty() &&  !this->origin_airport.value().empty()) {
        this->origin_airport.update(file.at("mission"));
        this->destination_airport.update(file.at("mission"));
    }

    /* Setup departure blocks */
    std::string subpath("mission/departure");
    std::string description("Departure procedure; mode Takeoff: %N1 (double), aerodynamic configuration (string).");
    description += " -- mode Climb: %N1 (double), aerodynamic configuration (string), End Point Altitude [m] (double).";
    description += " -- mode Accelerate: %N1 (double), aerodynamic configuration (string), Rate of climb [m/s] (double), End point CAS [m/s] (double)";
    description += " -- Alternatively, use the names from the ANP database";

    file[subpath];
    file.at(subpath).addAttrib("description", description);

    subpath += "/departure_step@";
    for (uint16_t i(0); i < this->departure_steps.size(); i++) {
        file[(subpath + num2Str(i))];
        file.at(subpath + num2Str(i)).addAttrib("description", "Single departure step");
        this->departure_steps.at(i).write_departure_steps(subpath + num2Str(i), &file);
    }

    /* Setup cruise blocks */
    subpath = "mission/cruise";
    description =  "Cruise procedure: mode change_speed_to_CAS:  %N1 (double), aerodynamic configuration (string), Rate of climb [m/s] (double), End point CAS [m/s] (double)";
    description += " -- mode change_speed_to_Mach: %N1 (double), aerodynamic configuration (string), Rate of climb [m/s] (double), End point CAS [m/s] (double)";
    description += " -- mode climb_to_cruise:  %N1 (double), aerodynamic configuration (string), End Point Altitude [m] (double), Mach [-] (double).";
    description += " -- mode cruise: aerodynamic configuration (string), Range [%] (double, relative distance at the end of the cruise segment without climb and descent)";
    description += " -- mode change_flight_level: %N1 (double), aerodynamic configuration (string), End Point Altitude [m] (double).";
    description += " -- mode descriptionent_to_approach: aerodynamic configuration (string), End Point Altitude [m] (double).";

    file[subpath];
    file.at(subpath).addAttrib("description", description);

    subpath += "/cruise_step@";
    for (uint16_t i(0); i < this->cruise_steps.size(); i++) {
        file[(subpath + num2Str(i))];
        file.at(subpath + num2Str(i)).addAttrib("description", "Single cruise step");
        this->cruise_steps.at(i).write_cruise_steps(subpath + num2Str(i), &file);
    }

    /* Setup approach blocks */
    subpath = "mission/approach";
    description =  "Approach procedure: mode descent: aerodynamic configuration (string), End Point Altitude [m] (double), glidepath [deg] (double).";
    description += " -- mode change_speed: aerodynamic configuration (string), Ground distance [m] (double), End point CAS [m/s] (double), glidepath [deg] (double).";
    description += " -- mode landing: aerodynamic configuration (string), End Point Altitude [m] (double), glidepath [deg] (double).";

    file[subpath];
    file.at(subpath).addAttrib("description", description);

    subpath += "/approach_step@";
    for (uint16_t i(0); i < this->approach_steps.size(); i++) {
        file[(subpath + num2Str(i))];
        file.at(subpath + num2Str(i)).addAttrib("description", "Single approach step");
        this->approach_steps.at(i).write_approach_steps(subpath + num2Str(i), &file);
    }

    aixml::saveDocument(file);
    aixml::closeDocument(file);
    // std::string subPath("AcftExchangeFile/Performance/RequirementsChecks/TLARs/"); // TODO(Gerrit): Check if still needed!
    // /* Mission type: requirement mission */
    // if (myRequirementMissionSettingsPt->getWriteMissionFile() &&
    //         (!this->myMissionXMLPt->acXML.at(subPath + "h_initialCruise_Design").getIntAttrib("Checked") ||
    //          !this->myMissionXMLPt->acXML.at(subPath + "h_maxOperating").getIntAttrib("Checked") ||
    //          !this->myMissionXMLPt->acXML.at(subPath + "h_maxOEI").getIntAttrib("Checked") ||
    //          !this->myMissionXMLPt->acXML.at(subPath + "M_initialCruise_Design").getIntAttrib("Checked") ||
    //          !this->myMissionXMLPt->acXML.at(subPath + "TTC_Design").getIntAttrib("Checked"))) {
    //     this->myMissionXMLPt->acXML.at("AcftExchangeFile/Performance/RequirementsChecks/MissionFile").setAttrib("OfftakesWritten", 0);
    //     this->generateMissionXMLFile(myRequirementMissionProfilePt, myRequirementMissionSettingsPt, myRequirementMissionPt);
    // }
}

void StandardMission::report() {
}

void StandardMission::save() {
}
