/* UNICADO - UNIversity Conceptual Aircraft Design and Optimization
 *
 * Copyright (C) 2025 UNICADO consortium
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 * Description:
 * This file is part of UNICADO.
 */

#ifndef ECOLOGICAL_ASSESSMENT_SRC_STANDARD_STRATEGY_EMISSION_CALCULATION_ECODATABASE_H_
#define ECOLOGICAL_ASSESSMENT_SRC_STANDARD_STRATEGY_EMISSION_CALCULATION_ECODATABASE_H_

#include <string>
#include <vector>
#include <unitConversion/unitConversion.h>

/* Global Constants */

// DEVELOPMENT PHASE (NEW method notes K. Schaefer (Personnel expenditure), Assumptions on different areas of work)
const double t_labourYear_dev = {1712.}; /**< Hours per month for Engineering , Tool Dev., Tool Manuf. [h] */
const double C_labourHour_dev = {150.};  /**< Costs per employee per hour for Engineering , Tool Dev., Tool Manuf [USD/h] */
const double spaceEfficiency_dev = {0.6}; /**< Space Efficiency [-] */
const double F_netFloor_dev = {14.5}; /**< Net floor per workplace for Engineering , Tool Dev., Tool Manuf [m^2]*/
const double energy_perSqmYear_dev[2] = {convertUnit(KILO, WATTHOUR, MEGA, JOULE, 106.),
     convertUnit(KILO, WATTHOUR, MEGA, JOULE, 94.)}; /**< Assumptions on electrical and heating energy consumption per square metre of working area [MJ/(m^2*yr)]*/

// PRODUCTION PHASE (Assumptions for material distribution see Arzdorf S.104 Tab.D.2) */
const int numberOfMainParts(7); /**< Number of main parts [-]*/
const int numberOfMaterials(5); /**< Number of materials [-]*/
const double matDistr[numberOfMainParts][numberOfMaterials] = { /**< Material distribution of main parts [-]*/
    //Alu ,CFRP, Stahl , Titan , Nickel, Chrome , copper, Nickel (Achim), Kunstst.
    {0.79, 0.07, 0.04, 0.09, 0.0},//, 0.0000, 0.0000, 0.0000, 0.0000},//Wing
    {0.82, 0.12, 0.02, 0.03, 0.0},//, 0.0000, 0.0000, 0.0000, 0.0000},//Fuselage
    {0.01, 0.99, 0.00, 0.00, 0.0},//, 0.0000, 0.0000, 0.0000, 0.0000},//Tailplane
    {0.03, 0.97, 0.00, 0.00, 0.0},//, 0.0000, 0.0000, 0.0000, 0.0000},//Fin
    {0.05, 0.01, 0.86, 0.08, 0.0},//, 0.0500, 0.0000, 0.0000, 0.0200},//LandingGear
    {0.05, 0.13, 0.38, 0.21, 0.16},//, 0.0500, 0.0000, 0.0000, 0.0200},//Pylon; assumed like PowerUnit, because together considered!
    {0.05, 0.13, 0.38, 0.21, 0.16},//, 0.0500, 0.0000, 0.0000, 0.0200},//PowerUnit
//    {0.800, 0.0800, 0.0200, 0.1000},//, 0.0000, 0.0000, 0.0000, 0.0000},//Pylons
//    {0.490, 0.1000, 0.2500, 0.1500},//, 0.0000, 0.0000, 0.0000, 0.0100},//PowerUnit
//    {0.667, 0.0445, 0.1677, 0.0277},//, 0.0052, 0.0145, 0.0125, 0.0609},//Systems
//    {0.580, 0.3500, 0.0200, 0.0000},//, 0.0000, 0.0000, 0.0000, 0.0500},//Furnishings
//    {0.600, 0.2500, 0.1000, 0.0000},//, 0.0000, 0.0000, 0.0000, 0.0500},//OperatorsItems
};
//Assumptions for manufacturing energy and scrap ratio s. Schaefer, specific Energy s. Arzdorf censored version (!); falsified values that may be published*/
//Alu,  Composites, Stahl, Titan, Nickel, Chrome, copper, Nickel, Plastic.
const double clippings[numberOfMaterials] = {0.85,   0.1,  0.8,  0.8, 0.5};//,  0.35,   0.35,   0.35, 0.4};  /**< Clipping of material (if buyToFly=1: all 0) [-] */
const double specificE_manuf[numberOfMaterials] = {convertUnit(KILO, WATTHOUR, MEGA, JOULE, 38.5), /**< Specific energy for manufacturing [MJ/kg] */
                                                convertUnit(KILO, WATTHOUR, MEGA, JOULE, 79.8),
                                                convertUnit(KILO, WATTHOUR, MEGA, JOULE, 25.2),
                                                convertUnit(KILO, WATTHOUR, MEGA, JOULE, 38.7),
                                                convertUnit(KILO, WATTHOUR, MEGA, JOULE, 17.5)};//,  25.2,   17.5,   17.5, 5.25};
const double recyclingEnergyRatio[numberOfMaterials] = {20, 20, 3.6, 2.6, 3.9}; /**< Ratio of recycling energy [-] */
const double recoveryRate[numberOfMaterials] = {0.9, 0., 0.95, 0.91, 0.56}; /**< Recovery rate [-] */
const double paintVOCFactor = 0.6;/**< Emission of VOC-gases by colour per m^2 of lacquered area (assumptions for lacquering (K. Schaefer) [kg/m^2] */
const double finalflight_test_hours = 6.; /**< Assumptions Flight hours Test flights s. Arzdorf S.72 [h] */
const unsigned int numberOfTransportationTypes = 4.; /**< Assumed number of transportation types for transport logistics [-]*/
const double transportationFuelCons[numberOfTransportationTypes] = {/**<Fuel consumption for aircraft(kerosene), truck(diesel), ship(heavy fuel oil), train(diesel)" [kg/t-km]*/
                                                    0.21,  0.0214, 1.2125E-3, 7.81E-3};
//Information about to be transported components
// ATTENTION: Only a part of a component will be transported, the distance has to be multiplied by the weight percentage (estimated) (emission values per kgkm, therefore ok)
// (BSP: Fuselage Midsection 1,2,3 and nose (= 4 components, (wrong assumption) all same weight, i.e. divide distance by 4)
//Pylon Production in Toulousse! Power Unit from Derby!
const unsigned int numberOfCompTransported(6);/**< Number of aircraft components to be transported [-] */
const double transportationDistance[numberOfCompTransported] = {(787. + 1181.), (493. + 110. + 493. + 493. + 1264.) / 4., 562., /**< Transportation distance of componants [km] */
            1264., 1130. / 2.*2., 1020.  }; //,1058.};
//Labour
const double energy_PerWorker_pr[2] = {convertUnit(KILO, WATTHOUR, MEGA, JOULE, 11000.), /**< Annual electrical & heating energy per worker in production [MJ/yr]*/
                                convertUnit(KILO, WATTHOUR, MEGA, JOULE, 24000.)};
const double t_labourYear_pr = 1712.; /**< Annual working hours in production [h/yr] */
const double C_labourHour_pr = 153.;/**< Hourly wage in production [USD/h] */
// TODO(franz#1#): C_labourHour_prod from "calculateCost_config" (manufacturing) sot it becomes consistent
// TODO(franz#1#): T127 maintenance labour wage in operatingCompany.xml, only used for work/material distribution
//-> make consistent (only list once in XML and then extract for this)

// OPERATION PHASE */
// NEW assumption K. Schaefer*/
const double energy_PerWorker_op[2] = {convertUnit(KILO, WATTHOUR, MEGA, JOULE, 11000.), /**< Annual electrical & heating energy per worker in operation [MJ/yr]*/
    convertUnit(KILO, WATTHOUR, MEGA, JOULE, 24000.)};
const double t_labourYear_op = 1712.;/**< Annual working hours in operation [h/a] */
const double C_labourHour_op = 153.;/**< hourly wage in operation [USD/h] */
const int numberOfSpareComponents = 7.; /**< Number of different components to be replaced during the operation phase; ATTENTION: Pylon+Nacelles in ATA 54!*/
//Attention during calculation, since separated in Production, must also be listed separately here.

// END OF LIFE PHASE
const double recyclingScen[numberOfMainParts][numberOfMaterials] = { /**< Ratio of recycled material [-] */
    //Alu   ,CFRP,  Stahl , Titan, Nickel
    {0.70,  0.00,   0.75,   0.50, 0.},//Wing
    {0.85,  0.00,   0.85,   0.50, 0.},//Fuselage
    {0.64,  0.00,   0.00,   0.00, 0.},//Tailplane
    {0.64,  0.00,   0.00,   0.00, 0.},//Fin
    {0.70,  0.00,   0.75,   0.50, 0.},//LandingGear
    {0.70,  0.00,   0.80,   0.50, 0.50},//Pylon (like Power Unit, considered together!)
    {0.70,  0.00,   0.80,   0.50, 0.50},//Power Unit (together with Pylons considered)
};
const double incinerationScen[numberOfMainParts][numberOfMaterials] = { /**< Ratio of incinerated material [-] */
    //Alu   ,CFRP,  Stahl , Titan, Nickel
    {0.00,  0.50,   0.00,   0.00, 0.},//Wing
    {0.00,  0.50,   0.00,   0.00, 0.},//Fuselage
    {0.00,  0.50,   0.00,   0.00, 0.},//Tailplane
    {0.00,  0.50,   0.00,   0.00, 0.},//Fin
    {0.00,  0.50,   0.00,   0.00, 0.},//LandingGear
    {0.00,  0.50,   0.00,   0.00, 0.},//Pylon (like Power Unit, considered together!)
    {0.00,  0.50,   0.00,   0.00, 0.},//Power Unit (together with Pylons considered)
};
const double landfillScen[numberOfMainParts][numberOfMaterials] = { /**< Ratio of landfilled material [-] */
    //Alu   ,CFRP,  Stahl , Titan, Nickel
    {0.30,  0.50,   0.25,   0.50, 0.},//Wing
    {0.15,  0.50,   0.15,   0.50, 0.},//Fuselage
    {0.36,  0.50,   0.00,   0.00, 0.},//Tailplane
    {0.36,  0.50,   0.00,   0.00, 0.},//Fin
    {0.30,  0.50,   0.25,   0.50, 0.},//LandingGear
    {0.30,  0.50,   0.20,   0.50, 0.50},//Pylon (like Power Unit, considered together!)
    {0.30,  0.50,   0.20,   0.50, 0.50},//Power Unit (together with Pylons considered)
};

/** \class ecoDatabase
 * \brief Class contains ecological values from GaBi database
 */
class ecoDatabase {
 public:
    /** \class EmissionsData
     * \brief Class contains all properties of the respective emissions database
     */
    class EmissionsData {
     public:
        std::string specification;/**< Specification/Name of Emission (e.g. Alu) */

        //Emissions
        double CO2;/**< Carbon dioxide [kg] */
        double H2O;/**< Water [kg] */
        double NOx;/**< Nitrogen oxide [kg] */
        double SO2;/**< Sulfur dioxide [kg] */
        double CO;/**< Carbon oxide [kg] */
        double CH4;/**< Methane [kg] */
        double PM10;/**< Particulate matter [kg]*/

        //Potential
        double GWP100;/**< Global Warming Potential for a 100 year time frame [1/MJ] or [1/kg]*/
        double AP;/**< Acidification Potential [1/MJ] or [1/kg]*/
        double ADP;/**< Abiotic Resource Depletion Potential [1/MJ] or [1/kg]*/
        double POCP;/**< Photochemical Ozone Creation Potential [1/MJ] or [1/kg] */
        double OLDP;/**< Ozone Layer Depletion Potential [1/MJ] or [1/kg] */
        double PMFP;/**< Particulate Matter Formation Potential [1/MJ] or [1/kg] */

        double specificEnergy;/**< Specific energy [-]*/

        /** \brief Default constructor of class EmissionsData
         */
        EmissionsData();
        /** \brief Constructor of class EmissionsData
         * \param theSpecification Specification/Name of Emission (e.g. Alu)
         */
        explicit EmissionsData(const std::string& theSpecification);
        /** \brief Destructor of class EmissionsData
         */
        virtual ~EmissionsData();
    };
    EmissionsData elecEmData;/**< Emissions from electric energy production */
    EmissionsData heatEmData;/**< Emissions from heating energy production*/
    std::vector<EmissionsData> fuelProdEmData; /**< Emissions from production of kerosene, diesel, heat oil, gasoline, natural gas*/
    std::vector<EmissionsData> matEmData; /**< Emissions from rea material extraction*/
    std::vector<EmissionsData> transpEmData; /**< Emissions from transport logistics*/
    EmissionsData aircraftEmData; /**< Emissions from aeroplane*/
    std::vector<EmissionsData> recyclingEmData;/**< Emissions from recycling*/
    std::vector<EmissionsData> incinerationEmData;/**< Emissions from incineration*/
    std::vector<EmissionsData> disposalEmData;/**< Emissions from disposal*/

    /* Specific energy for raw material extraction/depositing */
    std::vector<double> matSpecificEnergy; /**< at(0): Al, at(1): CFRP, at(2): Steel, at(3): Tit,, at(4): Ni [MJ/kg]*/
    std::vector<double> disposalSpecificEnergy; /**<at(0): Al, at(1): CFRP, at(2): Steel, at(3): Tit,, at(4): Ni [MJ/kg]*/
    std::vector<double> incinerationSpecificEnergy;/**<at(0): Al, at(1): CFRP, at(2): Steel, at(3): Tit,, at(4): Ni [MJ/kg]*/
    std::vector<double> fuelSpecificEnergy;/**<kerosene, Diesel, Fuel oil, gasoline, natural gas [MJ/kg]*/
    std::vector<double> transportSpecificEnergy;/**< at(0): Aircraft, at(1): LKW, at(2): Ship, at(3): Train; [MJ/1000kgkm]*/


    /* ICAO LTO times */
    const double taxi_time_ICAO = convertUnit(MINUTE, SECOND, 26);/**< ICAO taxi time: 26 Min [s]*/
    const double takeoff_time_ICAO = convertUnit(MINUTE, SECOND, 0.7); /**< ICAO take off time: 0.7 Min [s]*/
    const double climbout_time_ICAO = convertUnit(MINUTE, SECOND, 2.2); /**< ICAO climb out time: 2.2 Min [s]*/
    const double appTime_ICAO = convertUnit(MINUTE, SECOND, 4.0); /**< ICAO approach time: 4 Min [s] */

    /* GEMIS data; emissions [kg] of the electricity mix Germany El-KW-Park-DE-2010 DG-TREN [TJ]*/
    const double elecN2O = 5.9087673 / pow(10., 6.); /**< Nitrous oxide (laughing gas) [kg/MJ] */
    const double elecPFC = 0.000112006 / pow(10., 6.); /**< Perfluorinated hydrocarbons [kg/MJ]*/
    const double elecNMVOC = 8.6130268 / pow(10., 6.); /**< Non methane volatile organic compounds [kg/MJ] */

    /** \brief Constructor of class ecoDatabase */
    ecoDatabase();
    /** \brief Destructor of class ecoDatabase */
    virtual ~ecoDatabase();
};

#endif // ECOLOGICAL_ASSESSMENT_SRC_STANDARD_STRATEGY_EMISSION_CALCULATION_ECODATABASE_H_

