/* UNICADO - UNIversity Conceptual Aircraft Design and Optimization
 *
 * Copyright (C) 2025 UNICADO consortium
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 * Description:
 * This file is part of UNICADO.
 */

#ifndef ECOLOGICAL_ASSESSMENT_SRC_STANDARD_STRATEGY_EMISSION_CALCULATION_LIFE_CYCLE_EMISSIONS_LCA_SCHAEFER_ENDOFLIFE_H_
#define ECOLOGICAL_ASSESSMENT_SRC_STANDARD_STRATEGY_EMISSION_CALCULATION_LIFE_CYCLE_EMISSIONS_LCA_SCHAEFER_ENDOFLIFE_H_

#include <string>
#include <vector>
#include <memory>

#include "../../ecoDatabase.h"
#include "../../../emissionsClass.h"
#include "production.h"

class SchaeferIOData;
class Production;

/** \class EndOfLife
 * \brief Contains the properties of the end-of-life phase
 */
class EndOfLife {
 private:
    const ecoDatabase myEcoDatabase; /**< Object contains data from ecoDatabase*/

 public:
    /** \brief Constructor of class EndOfLife */
    EndOfLife();
    /** \brief Destructor of class EndOfLife */
    virtual ~EndOfLife();

    double cost_DD;/**< Costs for disassembly and dismantling [USD] */
    double engineering_energy_electricity_DD;/**< Electric energy for disassembly and dismantling [MJ] */

    /** \class mainComponent
     * \brief Distribution of emissions by main components (e.g. wing, tailplane)
     */
    class mainComponent {
     public:
        std::string component_name; /**< Name of the component */

        double mass;/**< Mass of the component [kg] */
        double energy_recycling;/**< Energy demand for recycling of the component's materials [MJ] */
        double energy_incineration;/**< Energy demand for incinerating the component's materials [MJ] */
        double energy_landfill;/**< Energy demand for landfill [MJ] */
        double energy_endOfLife;/**< Total energy demand for the main component in the end-of-life phase [MJ] */

        std::vector<double> rate_material;/**<Percentage of different materials. s. const double matDistr [-]*/
        std::vector<double> recovered_material_mass;/**< Mass of recycled material [kg]*/
        std::vector<double> rate_recycling;/**< Recycling rate of the various materials. s. const double recylcingScen [-]*/
        std::vector<double> rate_incineration;/**< Incineration rate of the various materials. s. const double incinerationScen [-]*/
        std::vector<double> rate_landfill;/**< Landfill rate of different materials. s. const double landfillScen [-] */

        /** \brief Constructor of class mainComponent */
        explicit mainComponent(const std::string& name);
        /** \brief Destructor of class mainComponent */
        virtual ~mainComponent();
    };
    std::vector<mainComponent> theAcComp; /**< Contains aircraft components */

    /** \class material
     * \brief Distribution of the emissions by materials (e.g. Aluminium, Copper)
     */
    class material {
     public:
        std::string material_name; /**< Name of material */

        double mass;/**< Mass of the material [kg] */
        double mass_recovered;/**< Mass share of the material that can be recycled [kg] */
        double energy_recycling;/**< Energy demand for recycling of the material's materials [MJ] */
        double energy_incineration;/**< Energy demand for incinerating the material's materials [MJ] */
        double energy_landfill;/**< Energy demand for landfill [MJ]*/
        double energy_endOfLife;/**< Total energy demand for the main material in the end-of-life phase [MJ] */

        std::vector<double> rate_recycling;/**< Recycling share of different materials, cf. const. double recylcingScen [-]*/
        std::vector<double> rate_incineration;/**<Incineration share of different materials, cf. const. double incinerationScen [-]*/
        std::vector<double> rate_landfill;/**<Landfill share of different materials, cf. const. double landfillScen [-]*/

        /** \brief Constructor of class material */
        material();
        /** \brief Destructor of class material*/
        virtual ~material();
    };
    std::vector<material> theMaterial; /**< Contains used materials */

    double energy_EoL_total;/**< Energy demand from EoL Scen (recycling, incineration, landfill) [MJ] */
    double energy_total;/**< Total energy demand of the entire end-of-ife phase [MJ] */

    Emissions endOfLifeTransport;/**< Emissions due to end of life transport */
    double transport_fuel;/**< Fuel for transport [kg] */
    Emissions endOfLifeDD;/**< Total emissions disassembly/dismantling at end of life */
    std::vector<Emissions> endOfLifeMaterial;/**< Emissions from component manufacture distributed according to the individual components */
    Emissions endOfLifeMaterialTotal;/**< Total emissions due to material at end of life */
    Emissions endOfLifeMaterialEngine;/**< Total emissions due to engine material at end of life */
    Emissions endOfLifeTotal;/**< Total EOL emissions */

    /** \brief Calculates material and energy demand of ens of life phase
     * \param data IO data
     */
    void calculate_EoL_resources(const std::shared_ptr<SchaeferIOData>& data);

    /** \brief Calculates emissions of end of life phase
     * \param data IO data
     * \param theProduction: Access to production data
     */
    void calculate_EoL_emissions(const std::shared_ptr<SchaeferIOData>& data, const Production&theProduction);

    /** \brief Calculates emissions of end of life phase
     */
    void emissions_material_mode();

    /** \brief Calculates emissions of end of life phase
     * \param data IO data
     * \param theProduction: Access to production data
     */
    void emissions_main_parts_mode(const std::shared_ptr<SchaeferIOData>& data, const Production&theProduction);
};

#endif // ECOLOGICAL_ASSESSMENT_SRC_STANDARD_STRATEGY_EMISSION_CALCULATION_LIFE_CYCLE_EMISSIONS_LCA_SCHAEFER_ENDOFLIFE_H_

