/* UNICADO - UNIversity Conceptual Aircraft Design and Optimization
 *
 * Copyright (C) 2025 UNICADO consortium
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 * Description:
 * This file is part of UNICADO.
 */

#include <algorithm>
#include "production.h"
#include <aixml/node.h>
#include <runtimeInfo/runtimeInfo.h>
#include <standardFiles/functions.h>

Production::Production()
    :
    myEcoDatabase(ecoDatabase()),
    material_energy(0.),
    test_flight_fuel(0.),
    total_RC(0.),
    energy_manufacturing_engine(0.),
    energy_manufacturing_total(0.),
    energy_engine(0.),
    energy_total(0.) {
}

void Production::calculate_production_resources(const std::shared_ptr<SchaeferIOData>& data, const std::shared_ptr<RuntimeIO>& rtIO) {
    /* Raw material and production */
    //Distribution of components
    double energyRecycling = 0;
    double energyLandfill = 0;
    recovery.clear();
    if (data->configuration.primary_material_recycling) {
        myRuntimeInfo->info << " Primary Recycling on" << std::endl;
        for (int i(0); i < numberOfMaterials; i++)
            recovery.push_back(recoveryRate[i]);
    } else {
        myRuntimeInfo->info << " No Primary Recycling" << std::endl;
        for (int i(0); i < numberOfMaterials; i++)
            recovery.push_back(0);
    }
    double wingRC(0.);
    for (int i(0); i < data->configuration.parts_number; i++) {
        theAcComp.push_back(mainComponent(data->main_parts_name[i]));
        theAcComp.back().mass = data->main_parts_mass[i];
        myRuntimeInfo->info << "Calculate RC of main part: " << data->main_parts_name[i] << std::endl;
        double CPI75(4.05); // Factor to convert 1975-USD to 2010-USD
        if (data->main_parts_name[i] == "Wing") {
            theAcComp.back().RC = CPI75 * get_NASA_costs(theAcComp.back().mass, data->configuration.number_produced_aircraft, 1730., 0.766, -0.218);
            wingRC = theAcComp.back().RC;
        } else if (data->main_parts_name[i] == "Fuselage") {
            theAcComp.back().RC = CPI75 * get_NASA_costs(theAcComp.back().mass, data->configuration.number_produced_aircraft, 2060., 0.766, -0.218);
        } else if (data->main_parts_name[i] == "Fin") {
            theAcComp.back().RC = data->fin_mass_ratio * CPI75 * get_NASA_costs(data->empennage_mass, data->configuration.number_produced_aircraft, 1820., 0.766, -0.218);
        } else if (data->main_parts_name[i] == "Tailplane") {
            theAcComp.back().RC = data->stabilizer_mass_ratio * CPI75 * get_NASA_costs(data->empennage_mass, data->configuration.number_produced_aircraft, 1820., 0.766, -0.218);
        } else if (data->main_parts_name[i] == "LandingGear") {
            const double structureWeightPercentage = 0.55;
            const double controlWeightPercentage = 0.1;
            const double wheelsBrakesWeightPercentage = 0.25;
            const double tiresPercentage = 0.1;
            double gearStructureCost = 0.;
            double gearControlsCost = 0.;
            double gearWheelsAndBrakesCost = 0.;
            double gearTiresCost = 0.;
            if ((convertUnit(KILO, GRAM, POUND, theAcComp.back().mass)* structureWeightPercentage) <= 10000.) {
                gearStructureCost = CPI75 * get_NASA_costs(structureWeightPercentage * theAcComp.back().mass, data->configuration.number_produced_aircraft, 1180., 0.766, -0.218);
            } else {
                gearStructureCost = CPI75 * get_NASA_costs(structureWeightPercentage * theAcComp.back().mass, data->configuration.number_produced_aircraft, 136., 1.0, -0.218);
            }
            gearControlsCost = CPI75 * get_NASA_costs(controlWeightPercentage * theAcComp.back().mass, data->configuration.number_produced_aircraft, 157., 1.0, -0.0896);
            gearWheelsAndBrakesCost = CPI75 * get_NASA_costs(wheelsBrakesWeightPercentage * theAcComp.back().mass, data->configuration.number_produced_aircraft, 23.8, 1.0, -0.0896);
            gearTiresCost = CPI75 * get_NASA_costs(tiresPercentage *theAcComp.back().mass, data->configuration.number_produced_aircraft, 2., 1.0, 0.0);
            theAcComp.back().RC = gearStructureCost + gearControlsCost + gearWheelsAndBrakesCost + gearTiresCost;
        } else if (data->main_parts_name[i] == "Pylons") {
            theAcComp.back().RC = data->pylon_mass_ratio * CPI75 * static_cast<double>(data->number_of_engines) * get_NASA_costs((data->nacelle_mass
                            + data->pylon_mass) / static_cast<double>(data->number_of_engines), data->configuration.number_produced_aircraft, 3470., 0.766, -0.218);
        } else if (data->main_parts_name[i] == "PowerUnit") {
            /*Engine Price after Stephan Langhans Dissertation 2013, S.50; Attention: here all cost values in Mio USD, assumption: Langhans formula for Dry Engine
                (without systems and nacelles); see values Airbus NSR-01 and calculation CSR-01 in documentation*/
            double engineWeightin1000LBS = convertUnit(KILO, GRAM, POUND, data->myEngine.get_dry_mass()) / 1000.; //mass of one engine
            double equippedEnginesRC = data->number_of_engines * (2.941 + 0.2603 * engineWeightin1000LBS + 0.04765
                                                * pow(engineWeightin1000LBS, 2)) * 1000000;
            //from here again NASA formula
            //Engine Systems: Assumption Engine Dry and Engine Systems = Equipped Engines; from here again NASA Formula
            double other_engine_systems_mass = data->engine_mass / data->number_of_engines - data->myEngine.get_dry_mass(); // Other systems mass eg. hydraulics, starter, etc.;
            equippedEnginesRC += static_cast<double>(data->number_of_engines) * CPI75 * get_NASA_costs(other_engine_systems_mass,
                                    data->configuration.number_produced_aircraft, 159., 1., -0.0896);
            //Nacelles
            double nacellesRC = data->nacelle_mass_ratio * static_cast<double>(data->number_of_engines) * CPI75 * get_NASA_costs((data->nacelle_mass
                                    + data->pylon_mass) / static_cast<double>(data->number_of_engines), data->configuration.number_produced_aircraft, 3470., 0.766, -0.218);
            //Bleed Air
            double bleedAirSystemRC(0.);
            if (convertUnit(KILO, GRAM, POUND, data->bleed_air_system_mass) <= 400.) {
                bleedAirSystemRC = CPI75 * get_NASA_costs(data->bleed_air_system_mass, data->configuration.number_produced_aircraft, 151., 1.0, -0.0896);
            } else {
                bleedAirSystemRC = CPI75 * get_NASA_costs(data->bleed_air_system_mass, data->configuration.number_produced_aircraft, 201., 1.0, -0.0896);
            }
            //Fuel System
            double fuelSystemRC = CPI75 * get_NASA_costs(data->fuel_system_mass, data->configuration.number_produced_aircraft, 61.9, 1.0, -0.0896);
            // Total engine
            theAcComp.back().RC = equippedEnginesRC + nacellesRC + bleedAirSystemRC + fuelSystemRC;
        } else {
            myRuntimeInfo->err << "No recurring cost method available for main component " << data->main_parts_name[i] << " !"<< std::endl;
        }
        myRuntimeInfo->info << "mass [kg]: " << theAcComp.back().mass << std::endl;
        myRuntimeInfo->info << "RC [$]: " << theAcComp.back().RC << std::endl;
        /* Energy calculation: Distribution according to aircraft components (Main Parts Modus) */
        //Material distribution. For numeration see production.h: const double matDistr
        double manufacturingEnergy(0.);
        for (int j(0); j < numberOfMaterials; j++) {
            theAcComp.back().rate_material.push_back(matDistr[i][j]);
            double massMaterialComp = theAcComp.back().rate_material.back() * theAcComp.back().mass;
            double massMFGMaterialComp = massMaterialComp / (1 - clippings[j]);
            double raw_material_massMaterialComp = massMFGMaterialComp * (1 - clippings[j] * recovery[j]);
            theAcComp.back().raw_material.push_back(raw_material_massMaterialComp);
            double energyMaterial = massMaterialComp * specificE_manuf[j];
            if (data->configuration.primary_material_recycling) {
                if (j != 1) {
                    //Composites will be recycled.
                    /*Note: Specific energy is wrong described; rather Primary Energy=CED; but here process energy is looked for,
                      i.e. divide by PrimaryEnergy from electricity consumption; see Notes Annex Diss Schaefer*/
                    energyRecycling = 1 / recyclingEnergyRatio[j] * myEcoDatabase.matSpecificEnergy.at(j)
                                      / myEcoDatabase.elecEmData.specificEnergy * massMFGMaterialComp * clippings[j];
                }
                energyLandfill = myEcoDatabase.disposalSpecificEnergy.at(j) / myEcoDatabase.elecEmData.specificEnergy * massMFGMaterialComp * clippings[j] * (1 - recovery[j]);
            }
            manufacturingEnergy += (energyMaterial + energyRecycling + energyLandfill);
            myRuntimeInfo->info << std::endl;
            myRuntimeInfo->info << "Material " << j + 1 << "..." << std::endl;
            myRuntimeInfo->info << "ER: " << recyclingEnergyRatio[j] << std::endl;
            myRuntimeInfo->info << "clippings[j]: " << clippings[j] << std::endl;
            myRuntimeInfo->info << "Mass share: " << theAcComp.back().rate_material.back() << std::endl;
            myRuntimeInfo->info << "Mass material: " << massMaterialComp << std::endl;
            myRuntimeInfo->info << "Mass Ore: " << theAcComp.back().raw_material.back() << std::endl;
            myRuntimeInfo->info << "spec. energy consumption: " << specificE_manuf[j] << std::endl;
            myRuntimeInfo->info << "energyMaterial: " << energyMaterial << std::endl;
            myRuntimeInfo->info << "energyRecycling: " << energyRecycling << std::endl;
            myRuntimeInfo->info << "energyLandfill: " << energyLandfill << std::endl;
            myRuntimeInfo->info << "manufacturingEnergy: " << manufacturingEnergy << std::endl;
        }
        theAcComp.back().energy_manufacturing = manufacturingEnergy;
        energy_manufacturing_total += theAcComp.back().energy_manufacturing;
        myRuntimeInfo->debug << "Mass:   " << theAcComp.back().mass << std::endl;
        myRuntimeInfo->debug << "Energy: " << theAcComp.back().energy_manufacturing << std::endl;
    }
    energy_manufacturing_engine = energy_manufacturing_total - theAcComp.back().energy_manufacturing;
    /* Energy Calculation: Distribution by Materials (Material Mode) */
    myRuntimeInfo->debug << "energy_manufacturing_total: " << energy_manufacturing_total << std::endl;
    for (int i(0); i < numberOfMaterials; i++) {
        theMaterial.push_back(Production::material());
        theMaterial.back().material_name = data->material_name[i];
        for (int j(0); j < data->configuration.parts_number; j++) {
            theMaterial.back().mass += theAcComp.at(j).rate_material.at(i) * theAcComp.at(j).mass;
            theMaterial.back().raw_material_mass += theAcComp.at(j).rate_material.at(i) * theAcComp.at(j).mass / (1 - clippings[i]) * (1 - clippings[i] * recovery[i]); //
        }
    }
    myRuntimeInfo->debug << "E_manuf_total: " << energy_manufacturing_total << std::endl;
    /* TEST Output */
    for (int i(0); i < numberOfMaterials; i++) {
        myRuntimeInfo->debug << std::endl;
        myRuntimeInfo->debug << theMaterial.at(i).material_name << std::endl;
        myRuntimeInfo->debug << "Material mass:  " << theMaterial.at(i).mass << std::endl;
        myRuntimeInfo->debug << "Raw material masse  " << theMaterial.at(i).mass*(1 + clippings[i]) << std::endl;
        myRuntimeInfo->debug << "Energy:            " << theMaterial.at(i).energy_manufacturing << std::endl;
    }
    /* personnel expenditure */
    total_RC = wingRC / 0.0627; // Assumption for single aisle aircraft: wingRC = 7% * total_RC [Lam12]
    theFinalAssembly.costs = 0.13 * total_RC; // Assumption for single aisle aircraft: final_assemblyRC = 20% * total_RC [Lam12]

//41% Labour of RC, s. Markish, s. 64 (33% Materials, Rest Others! (do not count here since also engineering work that has been neglected, s. Diss. Schaefer)
    theManufacturingWork.labour_costs = 0.41 * (total_RC - theFinalAssembly.costs);
    myRuntimeInfo->debug << "theFinalAssembly.costs: " << theFinalAssembly.costs << std::endl;
    myRuntimeInfo->debug << "total_RC: " << total_RC << std::endl;
    myRuntimeInfo->debug << "theManufacturingWork.labour_costs: " << theManufacturingWork.labour_costs << std::endl;
    theManufacturingWork.engineering_energy_electricity_eng = energy_PerWorker_pr[0] * theManufacturingWork.labour_costs / (C_labourHour_pr * t_labourYear_pr) * (data->configuration.engine_engineering_factor);
    theManufacturingWork.engineering_energy_heat_eng = energy_PerWorker_pr[1] * theManufacturingWork.labour_costs / (C_labourHour_pr * t_labourYear_pr) * (data->configuration.engine_engineering_factor);
    myRuntimeInfo->debug << "engineering_energy_heat_eng: " << theManufacturingWork.engineering_energy_heat_eng << std::endl;
    theManufacturingWork.engineering_energy_electricity = energy_PerWorker_pr[0] * theManufacturingWork.labour_costs / (C_labourHour_pr * t_labourYear_pr) * (1 + data->configuration.engine_engineering_factor);
    theManufacturingWork.engineering_energy_heat = energy_PerWorker_pr[1] * theManufacturingWork.labour_costs / (C_labourHour_pr * t_labourYear_pr) * (1 + data->configuration.engine_engineering_factor);
    myRuntimeInfo->debug << "Man years " << theManufacturingWork.labour_costs / (C_labourHour_pr * t_labourYear_pr) << std::endl;
    //final assembly
    theFinalAssembly.engineering_energy_electricity = energy_PerWorker_pr[0] * theFinalAssembly.costs / (C_labourHour_pr * t_labourYear_pr);
    theFinalAssembly.engineering_energy_heat = energy_PerWorker_pr[1] * theFinalAssembly.costs / (C_labourHour_pr * t_labourYear_pr);
    energy_total = energy_manufacturing_total + theFinalAssembly.engineering_energy_electricity + theFinalAssembly.engineering_energy_heat + theManufacturingWork.engineering_energy_electricity +
                   theManufacturingWork.engineering_energy_heat; //still without raw material and without "energy expenditure for energy production")
    energy_engine = energy_manufacturing_engine + (theManufacturingWork.engineering_energy_electricity + theManufacturingWork.engineering_energy_heat) * data->configuration.engine_engineering_factor
                        / (1 + data->configuration.engine_engineering_factor);
    myRuntimeInfo->info << "Assembly costs " << theFinalAssembly.costs << std::endl;
    myRuntimeInfo->info << "Assembly C_labourHour_pr " << C_labourHour_pr << std::endl;
    myRuntimeInfo->info << "Assembly t_labourYear_pr " << t_labourYear_pr << std::endl;
    /* Transport */
    for (unsigned int i(0.); i < numberOfTransportationTypes; i++) {
        std::string name_tmp;
        name_tmp = data->available_mean_of_transportation[i];
        theTransportation.push_back(TransportLogistics(name_tmp));
        myRuntimeInfo->debug << theTransportation.back().transport_type << "...:" << std::endl;
        for (int j(0.); j < data->configuration.transported_component_number; j++) {
            double distance_tmp(0.);
            double mass_tmp(0.);
            if (data->chosen_mean_of_transportation[j] == theTransportation.back().transport_type) {
                std::string component_name_tmp = data->transported_component_name[j];
                distance_tmp = transportationDistance[j];
                mass_tmp = data->transported_component_mass[j];
                theTransportation.back().name_component.push_back(component_name_tmp);
                theTransportation.back().distance_component.push_back(distance_tmp);
                theTransportation.back().mass_component.push_back(data->transported_component_mass[j]);
                myRuntimeInfo->debug << theTransportation.back().name_component.back() << "... " << std::endl;
                myRuntimeInfo->debug << "Distance " << theTransportation.back().distance_component.back() << std::endl;
                myRuntimeInfo->debug << "Mass (pro rata) " << theTransportation.back().mass_component.back() << std::endl;
            }
            theTransportation.back().mass_total += mass_tmp;
            theTransportation.back().distance_total += distance_tmp;
        }
        theTransportation.back().fuel = theTransportation.back().mass_total * theTransportation.back().distance_total * transportationFuelCons[i] / 1000.;
        myRuntimeInfo->debug << "Total distance: " << theTransportation.back().distance_total << std::endl;
        myRuntimeInfo->debug << "Total mass: " << theTransportation.back().mass_total << std::endl;
        myRuntimeInfo->debug << "Total fuel: " << theTransportation.back().fuel << std::endl;
    }
}

void Production::calculate_production_emissions(const std::shared_ptr<SchaeferIOData>& data) {
    /* Manufacturing and raw material */
    if (data->configuration.production_mode == "mode_0") { //in case with engines: always mode_1! s. settings
        //Material mode
        for (int i(0); i < numberOfMaterials; i++) {
            std::string materialEntry(theMaterial.at(i).material_name);
            std::vector<std::string> material_name_tmp;
            for (unsigned int k(0.); k < myEcoDatabase.matEmData.size(); k++) {
                material_name_tmp.push_back(myEcoDatabase.matEmData.at(k).specification);
            }
            std::vector<std::string>::iterator entry = find(material_name_tmp.begin(), material_name_tmp.end(), materialEntry); //Iterator to find the configuration input
            if (entry == material_name_tmp.end()) { //Input does not exist!
                myRuntimeInfo->err << "Material " << materialEntry << " not found. Abort!" << std::endl;
                exit(1);
            } else {
                unsigned int pos(0);
                pos = static_cast<int>(entry - material_name_tmp.begin());
                /* emission calculation */
                materialProduction.push_back(Emissions());
                materialProcessing.push_back(Emissions());
                materialProduction.back().CO2 = theMaterial.at(i).raw_material_mass * myEcoDatabase.matEmData.at(pos).CO2; //raw material
                materialProcessing.back().CO2 = theMaterial.at(i).energy_manufacturing * myEcoDatabase.elecEmData.CO2; //manufacturing
                materialProduction.back().H2O = theMaterial.at(i).raw_material_mass * myEcoDatabase.matEmData.at(pos).H2O;
                materialProcessing.back().H2O = theMaterial.at(i).energy_manufacturing * myEcoDatabase.elecEmData.H2O;
                materialProduction.back().NOx = theMaterial.at(i).raw_material_mass * myEcoDatabase.matEmData.at(pos).NOx;
                materialProcessing.back().NOx = theMaterial.at(i).energy_manufacturing * myEcoDatabase.elecEmData.NOx;
                materialProduction.back().SO2 = theMaterial.at(i).raw_material_mass * myEcoDatabase.matEmData.at(pos).SO2;
                materialProcessing.back().SO2 = theMaterial.at(i).energy_manufacturing * myEcoDatabase.elecEmData.SO2;
                materialProduction.back().CO = theMaterial.at(i).raw_material_mass * myEcoDatabase.matEmData.at(pos).CO;
                materialProcessing.back().CO = theMaterial.at(i).energy_manufacturing * myEcoDatabase.elecEmData.CO;
                materialProduction.back().CH4 = theMaterial.at(i).raw_material_mass * myEcoDatabase.matEmData.at(pos).CH4;
                materialProcessing.back().CH4 = theMaterial.at(i).energy_manufacturing * myEcoDatabase.elecEmData.CH4;
                materialProduction.back().PM10 = theMaterial.at(i).raw_material_mass * myEcoDatabase.matEmData.at(pos).PM10;
                materialProcessing.back().PM10 = theMaterial.at(i).energy_manufacturing * myEcoDatabase.elecEmData.PM10;
                materialProduction.back().CED = theMaterial.at(i).raw_material_mass * myEcoDatabase.matSpecificEnergy.at(pos);
            }
            materialProductionTotal.CO2  += materialProduction.back().CO2;
            materialProductionTotal.H2O  += materialProduction.back().H2O;
            materialProductionTotal.NOx  += materialProduction.back().NOx;
            materialProductionTotal.SO2  += materialProduction.back().SO2;
            materialProductionTotal.CO   += materialProduction.back().CO;
            materialProductionTotal.CH4  += materialProduction.back().CH4;
            materialProductionTotal.PM10 += materialProduction.back().PM10;
            materialProductionTotal.N2O += materialProduction.back().N2O;
            materialProductionTotal.PFC += materialProduction.back().PFC;
            materialProductionTotal.NMVOC += materialProduction.back().NMVOC;
            materialProductionTotal.HC += materialProduction.back().HC;
            materialProductionTotal.CED += materialProduction.back().CED;
            materialProcessingTotal.CO2  += materialProcessing.back().CO2;
            materialProcessingTotal.H2O  += materialProcessing.back().H2O;
            materialProcessingTotal.NOx  += materialProcessing.back().NOx;
            materialProcessingTotal.SO2  += materialProcessing.back().SO2;
            materialProcessingTotal.CO   += materialProcessing.back().CO;
            materialProcessingTotal.CH4  += materialProcessing.back().CH4;
            materialProcessingTotal.PM10 += materialProcessing.back().PM10;
            materialProcessingTotal.N2O += materialProcessing.back().N2O;
            materialProcessingTotal.PFC += materialProcessing.back().PFC;
            materialProcessingTotal.NMVOC += materialProcessing.back().NMVOC;
            materialProcessingTotal.HC += materialProcessing.back().HC;
            materialProcessingTotal.CED = energy_manufacturing_total * myEcoDatabase.elecEmData.specificEnergy;
        }
        myRuntimeInfo->info << "    Material Manufacturing Energy Total: " << materialProcessingTotal.CED + materialProductionTotal.CED << " MJ" << std::endl;
        myRuntimeInfo->info << "      Material Processing Energy: " << energy_manufacturing_total << " MJ" << std::endl;
        myRuntimeInfo->info << "      Material Processing Primary Energy: " << materialProcessingTotal.CED << " MJ" << std::endl;
        myRuntimeInfo->info << "      Material Processing CO2: " << materialProcessingTotal.CO2 << " kg" << std::endl;
        myRuntimeInfo->info << "      Material Extraction Energy: " << materialProductionTotal.CED << " MJ" << std::endl;
    } else if (data->configuration.production_mode == "mode_1") {
        //Main Parts mode
        for (int i(0); i < data->configuration.parts_number; i++) {
            /* Emission calculation */
            //manufacturing
            materialProduction.push_back(Emissions());
            materialProcessing.push_back(Emissions());
            materialProcessing.back().CO2 = theAcComp.at(i).energy_manufacturing * myEcoDatabase.elecEmData.CO2;
            materialProcessing.back().H2O = theAcComp.at(i).energy_manufacturing * myEcoDatabase.elecEmData.H2O;
            materialProcessing.back().NOx = theAcComp.at(i).energy_manufacturing * myEcoDatabase.elecEmData.NOx;
            materialProcessing.back().SO2 = theAcComp.at(i).energy_manufacturing * myEcoDatabase.elecEmData.SO2;
            materialProcessing.back().CO = theAcComp.at(i).energy_manufacturing * myEcoDatabase.elecEmData.CO;
            materialProcessing.back().CH4 = theAcComp.at(i).energy_manufacturing * myEcoDatabase.elecEmData.CH4;
            materialProcessing.back().PM10 = theAcComp.at(i).energy_manufacturing * myEcoDatabase.elecEmData.PM10;
            materialProcessing.back().CED = theAcComp.at(i).energy_manufacturing * myEcoDatabase.elecEmData.specificEnergy;
            //raw material extraction
            for (int j(0); j < numberOfMaterials; j++) {
                std::string materialEntry(theMaterial.at(j).material_name);
                std::vector<std::string> material_name_tmp;
                for (unsigned int k(0.); k < myEcoDatabase.matEmData.size(); k++) {
                    material_name_tmp.push_back(myEcoDatabase.matEmData.at(k).specification);
                }
                std::vector<std::string>::iterator entry = find(material_name_tmp.begin(), material_name_tmp.end(), materialEntry); //Iterator to find the configuration input
                if (entry == material_name_tmp.end()) { //Input not available!
                    myRuntimeInfo->err << "Material " << materialEntry << " in EcoDatabase not found. Abort!" << std::endl;
                    exit(1);
                } else {
                    unsigned int pos(0);
                    pos = static_cast<int>(entry - material_name_tmp.begin());
                    myRuntimeInfo->info << "Ore Mat. at " << j << " :" << theAcComp.at(i).raw_material.at(j) << std::endl;
                    myRuntimeInfo->info << "CED at " << j << " :" << myEcoDatabase.matSpecificEnergy.at(pos) << std::endl;
                    /* emission calculation */
                    materialProduction.back().CO2 += theAcComp.at(i).raw_material.at(j) * myEcoDatabase.matEmData.at(pos).CO2;
                    materialProduction.back().H2O += theAcComp.at(i).raw_material.at(j) * myEcoDatabase.matEmData.at(pos).H2O;
                    materialProduction.back().NOx += theAcComp.at(i).raw_material.at(j) * myEcoDatabase.matEmData.at(pos).NOx;
                    materialProduction.back().SO2 += theAcComp.at(i).raw_material.at(j) * myEcoDatabase.matEmData.at(pos).SO2;
                    materialProduction.back().CO += theAcComp.at(i).raw_material.at(j) * myEcoDatabase.matEmData.at(pos).CO;
                    materialProduction.back().CH4 += theAcComp.at(i).raw_material.at(j) * myEcoDatabase.matEmData.at(pos).CH4;
                    materialProduction.back().PM10 += theAcComp.at(i).raw_material.at(j) * myEcoDatabase.matEmData.at(pos).PM10;
                    materialProduction.back().CED += theAcComp.at(i).raw_material.at(j) * myEcoDatabase.matSpecificEnergy.at(pos);
                }
            }
            materialProductionTotal.CO2 += materialProduction.back().CO2;
            materialProductionTotal.H2O += materialProduction.back().H2O;
            materialProductionTotal.NOx += materialProduction.back().NOx;
            materialProductionTotal.SO2 += materialProduction.back().SO2;
            materialProductionTotal.CO += materialProduction.back().CO;
            materialProductionTotal.CH4 += materialProduction.back().CH4;
            materialProductionTotal.PM10 += materialProduction.back().PM10;
            materialProductionTotal.N2O += materialProduction.back().N2O;
            materialProductionTotal.PFC += materialProduction.back().PFC;
            materialProductionTotal.NMVOC += materialProduction.back().NMVOC;
            materialProductionTotal.HC += materialProduction.back().HC;
            materialProductionTotal.CED += materialProduction.back().CED;
            materialProcessingTotal.CO2 += materialProcessing.back().CO2;
            materialProcessingTotal.H2O += materialProcessing.back().H2O;
            materialProcessingTotal.NOx += materialProcessing.back().NOx;
            materialProcessingTotal.SO2 += materialProcessing.back().SO2;
            materialProcessingTotal.CO += materialProcessing.back().CO;
            materialProcessingTotal.CH4 += materialProcessing.back().CH4;
            materialProcessingTotal.PM10 += materialProcessing.back().PM10;
            materialProcessingTotal.N2O += materialProcessing.back().N2O;
            materialProcessingTotal.PFC += materialProcessing.back().PFC;
            materialProcessingTotal.NMVOC += materialProcessing.back().NMVOC;
            materialProcessingTotal.HC += materialProcessing.back().HC;
            materialProcessingTotal.CED += materialProcessing.back().CED;
        }
        if (data->configuration.engine_mode) {
            materialProductionEngine.CO2 = materialProduction.back().CO2;
            materialProductionEngine.H2O = materialProduction.back().H2O;
            materialProductionEngine.NOx = materialProduction.back().NOx;
            materialProductionEngine.SO2 = materialProduction.back().SO2;
            materialProductionEngine.CO = materialProduction.back().CO;
            materialProductionEngine.CH4 = materialProduction.back().CH4;
            materialProductionEngine.PM10 = materialProduction.back().PM10;
            materialProductionEngine.N2O = materialProduction.back().N2O;
            materialProductionEngine.PFC = materialProduction.back().PFC;
            materialProductionEngine.NMVOC = materialProduction.back().NMVOC;
            materialProductionEngine.HC = materialProduction.back().HC;
            materialProductionEngine.CED = materialProduction.back().CED;
            materialProcessingEngine.CO2 = materialProcessing.back().CO2;
            materialProcessingEngine.H2O = materialProcessing.back().H2O;
            materialProcessingEngine.NOx = materialProcessing.back().NOx;
            materialProcessingEngine.SO2 = materialProcessing.back().SO2;
            materialProcessingEngine.CO = materialProcessing.back().CO;
            materialProcessingEngine.CH4 = materialProcessing.back().CH4;
            materialProcessingEngine.PM10 = materialProcessing.back().PM10;
            materialProcessingEngine.N2O = materialProcessing.back().N2O;
            materialProcessingEngine.PFC = materialProcessing.back().PFC;
            materialProcessingEngine.NMVOC = materialProcessing.back().NMVOC;
            materialProcessingEngine.HC = materialProcessing.back().HC;
            materialProcessingEngine.CED = materialProcessing.back().CED;
        }
        myRuntimeInfo->info << "    Material Manufacturing Energy Total: " << materialProcessingTotal.CED + materialProductionTotal.CED << " MJ" << std::endl;
        myRuntimeInfo->info << "      Material Extraction Energy: " << materialProductionTotal.CED << " MJ" << std::endl;
        myRuntimeInfo->info << "      Material Processing Primary Energy: " << materialProcessingTotal.CED << " MJ" << std::endl;
        myRuntimeInfo->info << "      Material Processing Energy: " << energy_manufacturing_total << " MJ" << std::endl;
        myRuntimeInfo->info << "    Material Processing CO2: " << materialProcessingTotal.CO2 << " kg" << std::endl;
    } else {
        myRuntimeInfo->err << "Unknown production mode " << data->configuration.production_mode << ". (config-entry \"production_mode\") " << std::endl;
        myRuntimeInfo->err << "Abort program!" << std::endl;
        exit(1);
    }
    /* Personnel expenses for production */
    /*Manufacturing Labour*/
    /* emission calculation */
    labourProductionElectricity.CO2  = myEcoDatabase.elecEmData.CO2 *  (theManufacturingWork.engineering_energy_electricity);
    labourProductionElectricity.H2O  = myEcoDatabase.elecEmData.H2O *  (theManufacturingWork.engineering_energy_electricity);
    labourProductionElectricity.NOx  = myEcoDatabase.elecEmData.NOx *  (theManufacturingWork.engineering_energy_electricity);
    labourProductionElectricity.SO2  = myEcoDatabase.elecEmData.SO2 *  (theManufacturingWork.engineering_energy_electricity);
    labourProductionElectricity.CO  = myEcoDatabase.elecEmData.CO *  (theManufacturingWork.engineering_energy_electricity);
    labourProductionElectricity.CH4  = myEcoDatabase.elecEmData.CH4 *  (theManufacturingWork.engineering_energy_electricity);
    labourProductionElectricity.PM10  = myEcoDatabase.elecEmData.PM10 *  (theManufacturingWork.engineering_energy_electricity);
    labourProductionElectricity.CED = myEcoDatabase.elecEmData.specificEnergy *  (theManufacturingWork.engineering_energy_electricity);
    labourProductionHeat.CO2  = myEcoDatabase.heatEmData.CO2 *  (theManufacturingWork.engineering_energy_heat);
    labourProductionHeat.H2O  = myEcoDatabase.heatEmData.H2O *  (theManufacturingWork.engineering_energy_heat);
    labourProductionHeat.NOx  = myEcoDatabase.heatEmData.NOx *  (theManufacturingWork.engineering_energy_heat);
    labourProductionHeat.SO2  = myEcoDatabase.heatEmData.SO2 *  (theManufacturingWork.engineering_energy_heat);
    labourProductionHeat.CO  = myEcoDatabase.heatEmData.CO *  (theManufacturingWork.engineering_energy_heat);
    labourProductionHeat.CH4  = myEcoDatabase.heatEmData.CH4 *  (theManufacturingWork.engineering_energy_heat);
    labourProductionHeat.PM10  = myEcoDatabase.heatEmData.PM10 *  (theManufacturingWork.engineering_energy_heat);
    labourProductionHeat.CED = myEcoDatabase.heatEmData.specificEnergy *  (theManufacturingWork.engineering_energy_heat);
    //Total assembly emissions
    myRuntimeInfo->info << "    Labour Production Energy: " << theManufacturingWork.engineering_energy_electricity + theManufacturingWork.engineering_energy_heat << " MJ" << std::endl;
    labourProductionTotal.CO2 = labourProductionElectricity.CO2 + labourProductionHeat.CO2;
    labourProductionTotal.H2O = labourProductionElectricity.H2O + labourProductionHeat.H2O;
    labourProductionTotal.NOx = labourProductionElectricity.NOx + labourProductionHeat.NOx;
    labourProductionTotal.SO2 = labourProductionElectricity.SO2 + labourProductionHeat.SO2;
    labourProductionTotal.CO = labourProductionElectricity.CO + labourProductionHeat.CO;
    labourProductionTotal.CH4 = labourProductionElectricity.CH4 + labourProductionHeat.CH4;
    labourProductionTotal.PM10 = labourProductionElectricity.PM10 + labourProductionHeat.PM10;
    labourProductionTotal.N2O = labourProductionElectricity.N2O + labourProductionHeat.N2O;
    labourProductionTotal.PFC = labourProductionElectricity.PFC + labourProductionHeat.PFC;
    labourProductionTotal.NMVOC = labourProductionElectricity.NMVOC + labourProductionHeat.NMVOC;
    labourProductionTotal.HC = labourProductionElectricity.HC + labourProductionHeat.HC;
    labourProductionTotal.CED = labourProductionElectricity.CED + labourProductionHeat.CED;
    myRuntimeInfo->info << "    Labour Production Primary Energy: " << labourProductionTotal.CED << " MJ" << std::endl;
    if (data->configuration.engine_mode) {
        labourProductionElectricityEngine.CO2  = myEcoDatabase.elecEmData.CO2 *  (theManufacturingWork.engineering_energy_electricity_eng);
        labourProductionElectricityEngine.H2O  = myEcoDatabase.elecEmData.H2O *  (theManufacturingWork.engineering_energy_electricity_eng);
        labourProductionElectricityEngine.NOx  = myEcoDatabase.elecEmData.NOx *  (theManufacturingWork.engineering_energy_electricity_eng);
        labourProductionElectricityEngine.SO2  = myEcoDatabase.elecEmData.SO2 *  (theManufacturingWork.engineering_energy_electricity_eng);
        labourProductionElectricityEngine.CO  = myEcoDatabase.elecEmData.CO *  (theManufacturingWork.engineering_energy_electricity_eng);
        labourProductionElectricityEngine.CH4  = myEcoDatabase.elecEmData.CH4 *  (theManufacturingWork.engineering_energy_electricity_eng);
        labourProductionElectricityEngine.PM10  = myEcoDatabase.elecEmData.PM10 *  (theManufacturingWork.engineering_energy_electricity_eng);
        labourProductionElectricityEngine.CED = myEcoDatabase.elecEmData.specificEnergy *  (theManufacturingWork.engineering_energy_electricity_eng);
        labourProductionHeatEngine.CO2  = myEcoDatabase.heatEmData.CO2 *  (theManufacturingWork.engineering_energy_heat_eng);
        labourProductionHeatEngine.H2O  = myEcoDatabase.heatEmData.H2O *  (theManufacturingWork.engineering_energy_heat_eng);
        labourProductionHeatEngine.NOx  = myEcoDatabase.heatEmData.NOx *  (theManufacturingWork.engineering_energy_heat_eng);
        labourProductionHeatEngine.SO2  = myEcoDatabase.heatEmData.SO2 *  (theManufacturingWork.engineering_energy_heat_eng);
        labourProductionHeatEngine.CO  = myEcoDatabase.heatEmData.CO *  (theManufacturingWork.engineering_energy_heat_eng);
        labourProductionHeatEngine.CH4  = myEcoDatabase.heatEmData.CH4 *  (theManufacturingWork.engineering_energy_heat_eng);
        labourProductionHeatEngine.PM10  = myEcoDatabase.heatEmData.PM10 *  (theManufacturingWork.engineering_energy_heat_eng);
        labourProductionHeatEngine.CED = myEcoDatabase.heatEmData.specificEnergy *  (theManufacturingWork.engineering_energy_heat_eng);
        //Total assembly emissions
        labourProductionEngine.CO2 = labourProductionElectricityEngine.CO2 + labourProductionHeatEngine.CO2;
        labourProductionEngine.H2O = labourProductionElectricityEngine.H2O + labourProductionHeatEngine.H2O;
        labourProductionEngine.NOx = labourProductionElectricityEngine.NOx + labourProductionHeatEngine.NOx;
        labourProductionEngine.SO2 = labourProductionElectricityEngine.SO2 + labourProductionHeatEngine.SO2;
        labourProductionEngine.CO = labourProductionElectricityEngine.CO + labourProductionHeatEngine.CO;
        labourProductionEngine.CH4 = labourProductionElectricityEngine.CH4 + labourProductionHeatEngine.CH4;
        labourProductionEngine.PM10 = labourProductionElectricityEngine.PM10 + labourProductionHeatEngine.PM10;
        labourProductionEngine.N2O = labourProductionElectricityEngine.N2O + labourProductionHeatEngine.N2O;
        labourProductionEngine.PFC = labourProductionElectricityEngine.PFC + labourProductionHeatEngine.PFC;
        labourProductionEngine.NMVOC = labourProductionElectricityEngine.NMVOC + labourProductionHeatEngine.NMVOC;
        labourProductionEngine.HC = labourProductionElectricityEngine.HC + labourProductionHeatEngine.HC;
        labourProductionEngine.CED = labourProductionElectricityEngine.CED + labourProductionHeatEngine.CED;
        myRuntimeInfo->info << "    Labour Production Primary Energy Engine: " <<  labourProductionEngine.CED << " MJ" << std::endl;
    }
    /*Final Assembly*/
    /* emission calculation */
    finalAssemblyElectricity.CO2  = myEcoDatabase.elecEmData.CO2 *  (theFinalAssembly.engineering_energy_electricity);
    finalAssemblyElectricity.H2O  = myEcoDatabase.elecEmData.H2O *  (theFinalAssembly.engineering_energy_electricity);
    finalAssemblyElectricity.NOx  = myEcoDatabase.elecEmData.NOx *  (theFinalAssembly.engineering_energy_electricity);
    finalAssemblyElectricity.SO2  = myEcoDatabase.elecEmData.SO2 *  (theFinalAssembly.engineering_energy_electricity);
    finalAssemblyElectricity.CO  = myEcoDatabase.elecEmData.CO *  (theFinalAssembly.engineering_energy_electricity);
    finalAssemblyElectricity.CH4  = myEcoDatabase.elecEmData.CH4 *  (theFinalAssembly.engineering_energy_electricity);
    finalAssemblyElectricity.PM10  = myEcoDatabase.elecEmData.PM10 *  (theFinalAssembly.engineering_energy_electricity);
    finalAssemblyElectricity.CED = myEcoDatabase.elecEmData.specificEnergy *  (theFinalAssembly.engineering_energy_electricity);
    finalAssemblyHeat.CO2  = myEcoDatabase.heatEmData.CO2 *  (theFinalAssembly.engineering_energy_heat);
    finalAssemblyHeat.H2O  = myEcoDatabase.heatEmData.H2O *  (theFinalAssembly.engineering_energy_heat);
    finalAssemblyHeat.NOx  = myEcoDatabase.heatEmData.NOx *  (theFinalAssembly.engineering_energy_heat);
    finalAssemblyHeat.SO2  = myEcoDatabase.heatEmData.SO2 *  (theFinalAssembly.engineering_energy_heat);
    finalAssemblyHeat.CO  = myEcoDatabase.heatEmData.CO *  (theFinalAssembly.engineering_energy_heat);
    finalAssemblyHeat.CH4  = myEcoDatabase.heatEmData.CH4 *  (theFinalAssembly.engineering_energy_heat);
    finalAssemblyHeat.PM10  = myEcoDatabase.heatEmData.PM10 *  (theFinalAssembly.engineering_energy_heat);
    finalAssemblyHeat.CED = myEcoDatabase.heatEmData.specificEnergy *  (theFinalAssembly.engineering_energy_heat);
    //Final Assembly emissions
    myRuntimeInfo->info << "    Final Assembly Energy: " << theFinalAssembly.engineering_energy_electricity + theFinalAssembly.engineering_energy_heat << " MJ" << std::endl;
    finalAssemblyTotal.CO2 = finalAssemblyElectricity.CO2 + finalAssemblyHeat.CO2;
    finalAssemblyTotal.H2O = finalAssemblyElectricity.H2O + finalAssemblyHeat.H2O;
    finalAssemblyTotal.NOx = finalAssemblyElectricity.NOx + finalAssemblyHeat.NOx;
    finalAssemblyTotal.SO2 = finalAssemblyElectricity.SO2 + finalAssemblyHeat.SO2;
    finalAssemblyTotal.CO = finalAssemblyElectricity.CO + finalAssemblyHeat.CO;
    finalAssemblyTotal.CH4 = finalAssemblyElectricity.CH4 + finalAssemblyHeat.CH4;
    finalAssemblyTotal.PM10 = finalAssemblyElectricity.PM10 + finalAssemblyHeat.PM10;
    finalAssemblyTotal.N2O = finalAssemblyElectricity.N2O + finalAssemblyHeat.N2O;
    finalAssemblyTotal.PFC = finalAssemblyElectricity.PFC + finalAssemblyHeat.PFC;
    finalAssemblyTotal.NMVOC = finalAssemblyElectricity.NMVOC + finalAssemblyHeat.NMVOC;
    finalAssemblyTotal.HC = finalAssemblyElectricity.HC + finalAssemblyHeat.HC;
    finalAssemblyTotal.CED = finalAssemblyElectricity.CED + finalAssemblyHeat.CED;
    myRuntimeInfo->info << "    Final Assembly Primary Energy: " << finalAssemblyTotal.CED << " MJ" << std::endl;
    /* transport logistics */
/*TODO(franz#1#): T141 here assumption that values WITHOUT fuel prod., therefore additional addition of fuel prod. here!
 (not completely comprehensible in GaBi, but probably correct)*/
    int fuelTypeNo(99);
    for (unsigned int i(0.); i < numberOfTransportationTypes; i++) {
        for (unsigned int j(0.); j < theTransportation.at(i).mass_component.size(); j++) {
            double mass_tmp(theTransportation.at(i).mass_component.at(j));
            double distance_tmp(theTransportation.at(i).distance_component.at(j));
            /* emission calculation */
            transport.CO2 += mass_tmp * distance_tmp * myEcoDatabase.transpEmData.at(i).CO2 / 1000.;
            transport.H2O += mass_tmp * distance_tmp * myEcoDatabase.transpEmData.at(i).H2O / 1000.;
            transport.NOx += mass_tmp * distance_tmp * myEcoDatabase.transpEmData.at(i).NOx / 1000.;
            transport.SO2 += mass_tmp * distance_tmp * myEcoDatabase.transpEmData.at(i).SO2 / 1000.;
            transport.CO += mass_tmp * distance_tmp * myEcoDatabase.transpEmData.at(i).CO / 1000.;
            transport.CH4 += mass_tmp * distance_tmp * myEcoDatabase.transpEmData.at(i).CH4 / 1000.;
            transport.PM10 += mass_tmp * distance_tmp * myEcoDatabase.transpEmData.at(i).PM10 / 1000.;
            transport.CED += mass_tmp * distance_tmp * myEcoDatabase.transportSpecificEnergy.at(i) / 1000.;
        }
        // TODO(franz#1#): T140 should be programmed more generically... settings must be set differently, that means of transport+spec.consumption+fuel type belong together
        if  (theTransportation.at(i).transport_type == "Aircraft") {
            fuelTypeNo = 0;
        } else if  (theTransportation.at(i).transport_type == "Truck" || theTransportation.at(i).transport_type == "Train") {
            fuelTypeNo = 1;
        } else if  (theTransportation.at(i).transport_type == "Ship") {
            fuelTypeNo = 2;
        } else {
            myRuntimeInfo->err << "No transportation type for transport of the production materials specified."
                                  " Please specify 0 for aircraft, 1 for truck, 2 for ship." << std::endl;
            myRuntimeInfo->err << "Program termination!" << std::endl;
            exit(1);
        }
        transport.CO2 += theTransportation.at(i).fuel * myEcoDatabase.fuelProdEmData.at(fuelTypeNo).CO2;
        transport.H2O += theTransportation.at(i).fuel * myEcoDatabase.fuelProdEmData.at(fuelTypeNo).H2O;
        transport.NOx += theTransportation.at(i).fuel * myEcoDatabase.fuelProdEmData.at(fuelTypeNo).NOx;
        transport.SO2 += theTransportation.at(i).fuel * myEcoDatabase.fuelProdEmData.at(fuelTypeNo).SO2;
        transport.CO += theTransportation.at(i).fuel * myEcoDatabase.fuelProdEmData.at(fuelTypeNo).CO;
        transport.CH4 += theTransportation.at(i).fuel * myEcoDatabase.fuelProdEmData.at(fuelTypeNo).CH4;
        transport.PM10 += theTransportation.at(i).fuel * myEcoDatabase.fuelProdEmData.at(fuelTypeNo).PM10;
        transport.CED += theTransportation.at(i).fuel * myEcoDatabase.fuelSpecificEnergy.at(fuelTypeNo);
    }
    myRuntimeInfo->info << "    Transport to Final Assembly Primary Energy: " << transport.CED << " MJ" << std::endl;
    /* test flights incl. Fuel Prod. */
    /* emission calculation */
    double testFlightFactor = finalflight_test_hours / data->flight_time_design;
    test_flight_fuel = (data->blockfuel_design) * testFlightFactor;
    myRuntimeInfo->info << "    Final Test Flight Fuel: " << test_flight_fuel << " kg" << std::endl;
    finalFlightTests.CO2 = data->theTestMission.LTO.CO2 + data->theTestMission.cruise.CO2 * testFlightFactor + test_flight_fuel * myEcoDatabase.fuelProdEmData.at(0).CO2;;
    finalFlightTests.H2O = data->theTestMission.LTO.H2O + data->theTestMission.cruise.H2O * testFlightFactor + test_flight_fuel * myEcoDatabase.fuelProdEmData.at(0).H2O;
    finalFlightTests.NOx = data->theTestMission.LTO.NOx + data->theTestMission.cruise.NOx * testFlightFactor + test_flight_fuel * myEcoDatabase.fuelProdEmData.at(0).NOx;
    finalFlightTests.SO2 = data->theTestMission.LTO.SO2 + data->theTestMission.cruise.SO2 * testFlightFactor + test_flight_fuel * myEcoDatabase.fuelProdEmData.at(0).SO2;
    finalFlightTests.CO = data->theTestMission.LTO.CO + data->theTestMission.cruise.CO * testFlightFactor + test_flight_fuel * myEcoDatabase.fuelProdEmData.at(0).CO;
    finalFlightTests.CH4 = test_flight_fuel * myEcoDatabase.fuelProdEmData.at(0).CH4;
    finalFlightTests.PM10 = data->theTestMission.LTO.soot + data->theTestMission.cruise.soot * testFlightFactor + test_flight_fuel * myEcoDatabase.fuelProdEmData.at(0).PM10;
    finalFlightTests.HC = data->theTestMission.LTO.HC + data->theTestMission.cruise.HC * testFlightFactor;
    finalFlightTests.CED = test_flight_fuel * myEcoDatabase.fuelSpecificEnergy.at(0);
    myRuntimeInfo->info << "  Production Energy (w/o Material): " << energy_total << " MJ" << std::endl;
    myRuntimeInfo->info << "  Production Fuel (w/o Transport): " << test_flight_fuel << " kg" << std::endl;
    productionTotal.CO2 = materialProductionTotal.CO2 + materialProcessingTotal.CO2 + labourProductionTotal.CO2 + finalAssemblyTotal.CO2 + transport.CO2 +
                           finalFlightTests.CO2;
    productionTotal.H2O = materialProductionTotal.H2O + materialProcessingTotal.H2O + labourProductionTotal.H2O + finalAssemblyTotal.H2O + transport.H2O +
                           finalFlightTests.H2O;
    productionTotal.NOx = materialProductionTotal.NOx + materialProcessingTotal.NOx + labourProductionTotal.NOx + finalAssemblyTotal.NOx + transport.NOx +
                           finalFlightTests.NOx;
    productionTotal.SO2 = materialProductionTotal.SO2 + materialProcessingTotal.SO2 + labourProductionTotal.SO2 + finalAssemblyTotal.SO2 + transport.SO2 +
                           finalFlightTests.SO2;
    productionTotal.CO = materialProductionTotal.CO + materialProcessingTotal.CO + labourProductionTotal.CO
                          + finalAssemblyTotal.CO + transport.CO + finalFlightTests.CO;
    productionTotal.CH4 = materialProductionTotal.CH4 + materialProcessingTotal.CH4 + labourProductionTotal.CH4 + finalAssemblyTotal.CH4 + transport.CH4 +
                           finalFlightTests.CH4;
    productionTotal.PM10 = materialProductionTotal.PM10 + materialProcessingTotal.PM10 + labourProductionTotal.PM10 + finalAssemblyTotal.PM10 + transport.PM10 +
                            finalFlightTests.PM10;
    productionTotal.N2O = materialProductionTotal.N2O + materialProcessingTotal.N2O + labourProductionTotal.N2O + finalAssemblyTotal.N2O + transport.N2O +
                           finalFlightTests.N2O;
    productionTotal.PFC = materialProductionTotal.PFC + materialProcessingTotal.PFC + labourProductionTotal.PFC + finalAssemblyTotal.PFC + transport.PFC +
                           finalFlightTests.PFC;
    productionTotal.NMVOC = materialProductionTotal.NMVOC + materialProcessingTotal.NMVOC + labourProductionTotal.NMVOC + finalAssemblyTotal.NMVOC + transport.NMVOC +
                             finalFlightTests.NMVOC;
    productionTotal.HC = materialProductionTotal.HC + materialProcessingTotal.HC + labourProductionTotal.HC
                          + finalAssemblyTotal.HC + transport.HC + finalFlightTests.HC;
    productionTotal.CED = materialProductionTotal.CED + materialProcessingTotal.CED + labourProductionTotal.CED
                           + finalAssemblyTotal.CED + transport.CED + finalFlightTests.CED;
    myRuntimeInfo->info << "  ProductionPrimary Energy: " << productionTotal.CED << " MJ" << std::endl;
    //For deduction at static_AC
    if (data->configuration.engine_mode) {
        productionEngine.CO2 = materialProductionEngine.CO2 + materialProcessingEngine.CO2 + labourProductionEngine.CO2;
        productionEngine.H2O = materialProductionEngine.H2O + materialProcessingEngine.H2O + labourProductionEngine.H2O;
        productionEngine.NOx = materialProductionEngine.NOx + materialProcessingEngine.NOx + labourProductionEngine.NOx;
        productionEngine.SO2 = materialProductionEngine.SO2 + materialProcessingEngine.SO2 + labourProductionEngine.SO2;
        productionEngine.CO = materialProductionEngine.CO + materialProcessingEngine.CO + labourProductionEngine.CO;
        productionEngine.CH4 = materialProductionEngine.CH4 + materialProcessingEngine.CH4 + labourProductionEngine.CH4;
        productionEngine.PM10 = materialProductionEngine.PM10 + materialProcessingEngine.PM10 + labourProductionEngine.PM10;
        productionEngine.N2O = materialProductionEngine.N2O + materialProcessingEngine.N2O + labourProductionEngine.N2O;
        productionEngine.PFC = materialProductionEngine.PFC + materialProcessingEngine.PFC + labourProductionEngine.PFC;
        productionEngine.NMVOC = materialProductionEngine.NMVOC + materialProcessingEngine.NMVOC + labourProductionEngine.NMVOC;
        productionEngine.HC = materialProductionEngine.HC + materialProcessingEngine.HC + labourProductionEngine.HC;
        productionEngine.CED = materialProductionEngine.CED + materialProcessingEngine.CED + labourProductionEngine.CED;
    }
}

double Production::get_NASA_costs(double Mass, double Quantity, double A0, double A1, double A2) {
    //Beltramo: 10% Earnings are already included in the cost calculation, S. 2-3
    //Beltramo: also system-level assembly (up to 21%) included (-> 1.1*1.21 = 1.33; for Tail, Nacelle, Thrust Reservers and Engine System only 1.21!), S. 2-13 f. S.4-4, 4-14, 4-18
    return 0.9 * A0 * pow(convertUnit(KILO, GRAM, POUND, Mass), A1) * pow(static_cast<double>(Quantity), A2);
}

/* Constuctors and destructors */
Production::~Production() {}

Production::mainComponent::mainComponent(const std::string& name)
    :
    component_name(name),
    mass(0.),
    energy_manufacturing(0.),
    RC(0.) {
}
Production::mainComponent::~mainComponent() {}

Production::material::material()
    :
    mass(0.),
    raw_material_mass(0),
    energy_manufacturing(0.) {
}

Production::material::~material() {}

Production::manufacturingWork::manufacturingWork()
    :
    labour_costs(0.),
    engineering_energy_electricity(0.),
    engineering_energy_electricity_eng(0.),
    engineering_energy_heat(0.),
    engineering_energy_heat_eng(0.) {
}

Production::manufacturingWork::~manufacturingWork() {
}

Production::Assembly::Assembly()
    :
    costs(0.),
    engineering_energy_electricity(0.),
    engineering_energy_heat(0.) {
}

Production::Assembly::~Assembly() {
}

Production::TransportLogistics::TransportLogistics(const std::string& name)
    :
    transport_type(name),
    mass_total(0.),
    distance_total(0.),
    fuel(0.) {
}

Production::TransportLogistics::~TransportLogistics() {
}
