/* UNICADO - UNIversity Conceptual Aircraft Design and Optimization
 *
 * Copyright (C) 2025 UNICADO consortium
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 * Description:
 * This file is part of UNICADO.
 */

#include "schaeferConfig.h"
#include "../../ecoDatabase.h"
#include <aixml/node.h>
#include <runtimeInfo/runtimeInfo.h>
#include <standardFiles/functions.h>

using std::string;

SchaeferConfig::SchaeferConfig()
    :
    engine_mode(true),
    /*Development*/
    development_emission_setting(""),
    test_ETOPS(true),
    engine_options_number(0),
    wind_tunnel_test_hours(0),
    structural_test_cycles(0),
    system_integration_test_hours(0),
    iron_bird_test_hours(0),
    do_engine_tests(true),
    number_new_engines(0),
    number_test_engines(0),
    rig_test_hours(0),
    include_ETOPS(true),
    max_cont_thrust_percentage(0.),
    flying_test_bed_hours(0),
    flying_test_bed_engine_number(0),
    flying_test_bed_fuel_consumption(0),
    flight_test_hours(0),
    include_engine_options(true),
    structural_test_aircraft_number(0.),
    flight_tests_aircraft_number(0.),
    engine_engineering_factor(0.),
    /*Production*/
    production_mode(""),
    primary_material_recycling(true),
    number_produced_aircraft(0),
    /*End of life*/
    distance_to_EoL_site(0.),
    /*Derived settings*/
    parts_number(0),
    transported_component_number(0),
    spare_parts_number(0) {}

void SchaeferConfig::read_all_config_data(const node& configXML) {
    engine_mode = EndnodeReadOnly<bool>("program_settings/standard_strategy/emission_calculation/engine_mode_switch").read(configXML).value();
    engine_engineering_factor = EndnodeReadOnly<double>(std::string("program_settings/standard_strategy/emission_calculation/life_cycle_emissions_methods/")
                                                        + std::string("schaefer/engine_engineering_ratio")).read(configXML).value();
    /* LCA phases */
    read_development_config_data(configXML.at("program_settings/standard_strategy/emission_calculation/development_phase/"));
    read_production_config_data(configXML.at("program_settings/standard_strategy/emission_calculation/production_phase/"));
    read_operation_config_data(configXML.at("program_settings/standard_strategy/emission_calculation/operating_phase/"));
    read_end_of_life_config_data(configXML.at("program_settings/standard_strategy/emission_calculation/end_of_life_phase/"));
    /* Settings depending on engine mode */
    if (engine_mode) { //do_engine_tests may be false if no new engines will be developed
        parts_number = numberOfMainParts;
        transported_component_number = numberOfCompTransported;
        spare_parts_number = numberOfSpareComponents;
        //production_mode = 2; //because engines from static test ac must be subtracted
    } else {
        do_engine_tests = false;
        parts_number = numberOfMainParts - 1;
        transported_component_number = numberOfCompTransported - 1;
        spare_parts_number = numberOfSpareComponents - 1;
    }
    if (number_new_engines > engine_options_number) {
        myRuntimeInfo->err << "number_new_engines is bigger than engine_options_number. Program abort!" << std::endl;
        exit(1);
    }
}

void SchaeferConfig::read_development_config_data(const node& development) {
    development_emission_setting = EndnodeReadOnly<string>("test_phase/development_emission_setting").read(development).value();
    test_ETOPS = EndnodeReadOnly<bool>("test_phase/aircraft_specs/ETOPS_switch").read(development).value();
    engine_options_number = EndnodeReadOnly<int>("test_phase/aircraft_specs/engine_options").read(development).value();
    wind_tunnel_test_hours = EndnodeReadOnly<int>("test_phase/wind_tunnel_test/test_hours").read(development).value_xml();
    structural_test_cycles = EndnodeReadOnly<int>("test_phase/structural_test/test_cycles").read(development).value();
    structural_test_aircraft_number = EndnodeReadOnly<int>("test_phase/structural_test/number_of_structural_test_aircraft").read(development).value();
    system_integration_test_hours = EndnodeReadOnly<int>("test_phase/system_test/system_integration_test_hours").read(development).value_xml();
    iron_bird_test_hours = EndnodeReadOnly<int>("test_phase/system_test/iron_bird_test_hours").read(development).value_xml();
    do_engine_tests = EndnodeReadOnly<bool>("test_phase/engine_test/enable").read(development).value();
    number_new_engines = EndnodeReadOnly<int>("test_phase/engine_test/number_of_new_engines").read(development).value();
    number_test_engines = EndnodeReadOnly<int>("test_phase/engine_test/number_of_test_engines").read(development).value();
    rig_test_hours = EndnodeReadOnly<int>("test_phase/engine_test/rig_test/test_hours").read(development).value_xml();
    include_ETOPS = EndnodeReadOnly<bool>("test_phase/engine_test/rig_test/incl_ETOPS_switch").read(development).value();
    max_cont_thrust_percentage = EndnodeReadOnly<double>("test_phase/engine_test/rig_test/max_continuous_thrust_percentage").read(development).value();
    flying_test_bed_hours = EndnodeReadOnly<int>("test_phase/engine_test/flying_testbed/test_hours").read(development).value_xml();
    flying_test_bed_engine_number = EndnodeReadOnly<int>("test_phase/engine_test/flying_testbed/flying_testbed_engines/engine_number").read(development).value();
    flying_test_bed_fuel_consumption = EndnodeReadOnly<double>(string("test_phase/engine_test/flying_testbed/flying_testbed_engines/") +
                                    string("flying_testbed_fuel_consumption_per_engine")).read(development).value();
    flight_test_hours = EndnodeReadOnly<int>("test_phase/flight_test/test_hours").read(development).value_xml();
    include_engine_options = EndnodeReadOnly<bool>("test_phase/flight_test/test_hours/incl_engine_options").read(development).value();
    flight_tests_aircraft_number = EndnodeReadOnly<int>("test_phase/flight_test/number_of_flight_test_aircraft").read(development).value();
}

void SchaeferConfig::read_production_config_data(const node& production) {
    production_mode = EndnodeReadOnly<string>("production_mode").read(production).value();
    primary_material_recycling = EndnodeReadOnly<bool>("primary_material_recycling_switch").read(production).value();
    number_produced_aircraft = EndnodeReadOnly<int>("number_produced_aircraft").read(production).value();
}

void SchaeferConfig::read_operation_config_data(const node& operation) {
}

void SchaeferConfig::read_end_of_life_config_data(const node& endOfLife) {
    distance_to_EoL_site = EndnodeReadOnly<double>("distance_to_end_of_life_site").read(endOfLife).value();
}

SchaeferConfig::~SchaeferConfig() {
    //dtor
}

