/* UNICADO - UNIversity Conceptual Aircraft Design and Optimization
 *
 * Copyright (C) 2025 UNICADO consortium
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 * Description:
 * This file is part of UNICADO.
 */

#ifndef ECOLOGICAL_ASSESSMENT_SRC_STANDARD_STRATEGY_EMISSION_CALCULATION_LIFE_CYCLE_EMISSIONS_LCA_SCHAEFER_SCHAEFERCONFIG_H_
#define ECOLOGICAL_ASSESSMENT_SRC_STANDARD_STRATEGY_EMISSION_CALCULATION_LIFE_CYCLE_EMISSIONS_LCA_SCHAEFER_SCHAEFERCONFIG_H_

#include <string>
#include <unordered_map>
#include <utility>
#include <vector>
#include <array>
#include <aixml/endnode.h>
#include "../../../emissionsClass.h"

/** \class SchaeferConfig
 * \brief Class for the config settings needed for schaefer strategy
*/
class SchaeferConfig {
 public:
    /*Program parameter */
    bool engine_mode; /**<true: with Engine Production, false: without */
    /*Development phase inputs*/
    std::string development_emission_setting;/**<mode_0: no development emissions, mode_1: calculate w/o production/recycling, mode_2: include production/recycling*/
    bool test_ETOPS;/**<true: Engine are tested for ETOPS approval, additional 3000 test cycles (a 0.5h), false: no test for ETOPS */
    int engine_options_number;/**<Number of engine options for aircraft [-]*/
    int wind_tunnel_test_hours;/**<Number of wind tunnel hours in aircraft development [h]*/
    int structural_test_cycles;/**<Tested flight cycles in aircraft development; twice the number of flight cycles for which the aircraft will be certified [-]*/
    int system_integration_test_hours;/**<Number of test hours at the system integration test rig [h]*/
    int iron_bird_test_hours;/**<Number of test hours at Iron-Bird test rig [h]*/
    bool do_engine_tests;/**<true: new engine will be certified for this aircraft, false: only engine tests */
    int number_new_engines;/**<Number of new to be certified engines [-]*/
    int number_test_engines;/**<Number of engies tested (rig + flight test) [-]*/
    int rig_test_hours;/**<Number of test hours on the engine-Rig in aircraft development [h]*/
    bool include_ETOPS;/**<true: ETOPS hours already in rig_test_hours, false: not included */
    double max_cont_thrust_percentage;/**<Percentage of test hours on the engine rig with Maximum Continuous Thrust in aircraft development [-]*/
    int flying_test_bed_hours;/**<Number of test hours on Flying Test Bed aircraft development [h]*/
    int flying_test_bed_engine_number;/**<Number of engines on Flying Test Bed (without tested engine) [-]*/
    double flying_test_bed_fuel_consumption;/**<Fuel Consumption per hour von Test Bed AC (pro Engine) [kg/s]*/
    int flight_test_hours; /**<Number of flight test hours in aircraft development [h]*/
    bool include_engine_options;/**< true: flight_test_hours incl. all Engine options, false: only for one. */
    int structural_test_aircraft_number;/**<Number of aircraft destroyed during structural tests [-]*/
    int flight_tests_aircraft_number;  /**<Number of aircraft used only for flight tests [-]*/
    double engine_engineering_factor;/**< Ratio of engine engineering effort to whole aircraft, derived from Micado CSR-02 costs [-] */
    /*Production phase inputs*/
    std::string production_mode; /**<mode_0: material mode, mode_1: main parts mode*/
    bool primary_material_recycling;/**<'true': primary material will be recycled, 'false': no recyling*/
    int number_produced_aircraft; /**<Number of aircraft produced per program*/
    /*End of life phase inputs*/
    double distance_to_EoL_site;/**<Distance to be flown to the wreck location [m] */
    /*Additional settings*/
    int parts_number;/**< Number of main part [-]*/
    int transported_component_number; /**< Number of components transported [-]*/
    int spare_parts_number;/**< Number of spare components [-]*/


    /** \brief Default constructor of class SchaeferConfig
     */
    SchaeferConfig();


    /** \brief Constructor of class SchaeferConfig
     */
    virtual ~SchaeferConfig();

    /** \brief Function to read schaefer specific data from ecological_assessment.xml
     * \param config Node of the data structure defined by config file
     */
    void read_all_config_data(const node& config);

 private:
    /** \brief Function to read development data from config.xml
     * \param development Node of the development data structure defined in config file
     */
    void read_development_config_data(const node& development);

    /** \brief Function to read production data from config.xml
     * \param production Node of the production data structure defined by config file
     */
    void read_production_config_data(const node& production);

    /** \brief Function to read operation data from config.xml
     * \param operation Node of the operation data structure defined by config file
     */
    void read_operation_config_data(const node& operation);

    /** \brief Function to read enf of life data from config.xml
     * \param endOfLife node of the endOfLife data structure defined by config file
     */
    void read_end_of_life_config_data(const node& endOfLife);
};

#endif // ECOLOGICAL_ASSESSMENT_SRC_STANDARD_STRATEGY_EMISSION_CALCULATION_LIFE_CYCLE_EMISSIONS_LCA_SCHAEFER_SCHAEFERCONFIG_H_
