/*
 * UNICADO - UNIversity Conceptual Aircraft Design and Optimization
 *
 * Copyright (C) 2025 UNICADO consortium
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 * Description:
 * This file is part of UNICADO.
 */

#include "lowBwbVerticalTailsDesign.h"

#include <aircraftGeometry2/processing/measure.h>
#include <cmath>

#include "lib/design_methods/volumeCoefficientMethod.h"
#include "lib/mass_methods/flopsMassMethod.h"

namespace bwb {
namespace low {

VerticalTails::VerticalTails(const std::shared_ptr<RuntimeIO>& rtIO)
    : rtIO(rtIO),
      config(std::make_shared<bwb::low::VerticalTailsConfig>()),
      data(std::make_shared<bwb::low::VerticalTailsData>()),
      design_mode({{"mode_0", [this]() { design(); }}, {"mode_1", [this]() { redesign(); }}}),
      mass_mode({{"mode_0", [this]() { flops_mass(); }}}),
      reporter(rtIO) {}

void VerticalTails::initialize() {
  // Setup methods for mode
  myRuntimeInfo->out << "Initializing empennage" << std::endl;
  myRuntimeInfo->out << "Empennage type -> vertical tails (multiple vertical tails - symmetric)" << std::endl;

  /* Read configuration data */
  config->read(rtIO->moduleConfig);

  /* Create airfoils directory in project path (only if non-existing -> e.g. if standalone usage w. given wing and fuselage)*/

  /* Read acxml data - non aircraftgeometry 2 */
  data->read(rtIO->acxml);

  /* Read acxml data - aircraftgeometry2 */
  data->read_wing(rtIO->acxmlAccess, rtIO->getAirfoilDataDir());

  /* Create airfoils library */
  airfoils_library = std::make_shared<Airfoils>(static_cast<std::filesystem::path>(config->common_airfoil_data_path.value()));
  airfoils_library->add_directory_airfoils(rtIO->getAirfoilDataDir());
}

void VerticalTails::run() {
  // run
  myRuntimeInfo->out << "Running empennage -> vertical tails" << std::endl;
  design_mode[config->design_mode.value()]();
  mass_mode[config->mass_mode.value()]();
}

void VerticalTails::update() {
  // update
  myRuntimeInfo->out << "Updating empennage -> vertical tails" << std::endl;
  data->update(rtIO->acxml);

}

void VerticalTails::report() {
  // report sections
  myRuntimeInfo->out << "Reporting empennage -> vertical tails" << std::endl;

  /* Set html report data and plots */
  set_html_body();

  /* Generate reports */
  reporter.generateReports();
  for (auto& tail : data->tails) {
    geom2::Mesh mesh = geom2::transform::to_mesh(tail);
    if (tail.origin.y() < 0) {
      CGAL::IO::write_STL(tail.name + "_negative.stl", mesh);
    } else {
      CGAL::IO::write_STL(tail.name + "_positive.stl", mesh);
    }
  }
}

void VerticalTails::save() {
  // save data if necessary
}

void VerticalTails::design() {
  myRuntimeInfo->out << "Sizing mode active Vertical Tails" << '\n';

  double wing_reference_area = geom2::measure::reference_area(data->wing);
  double wing_span = geom2::measure::span(data->wing);
  double wing_sweep = -geom2::measure::sweep(data->wing, -wing_span * 0.5 * 0.9, 0.25);
  double wing_taper_ratio = geom2::measure::taper_ratio(data->wing);
  double wing_aspect_ratio = geom2::measure::aspect_ratio(data->wing);
  double wing_mac = geom2::measure::mean_aerodynamic_chord(data->wing);
  double span_position_mac = geom2::measure::mean_aerodynamic_chord_position(data->wing);

  geom2::Point_3 wing_neutral_position =
      geom2::Point_3(data->wing_reference_position.x.value() +
                         geom2::measure::offset_LE(data->wing, span_position_mac).x() + 0.25 * wing_mac,
                     0., 0.);

  /* Design vertical tail(s) - each vertical tail will be computed symmetric - so one vertical tail from config results
   * in two vertical tails on the left and right wing*/
  for (auto& vertical_tail : config->vertical_tails) {
    VolumeCoefficientMethod volume_coefficient_method;
    const double volume_coefficient = vertical_tail.volume_coefficient.value();
    const double sweep = vertical_tail.delta_sweep.value() + wing_sweep;
    const double taper_ratio = vertical_tail.factor_taper_ratio.value() * wing_taper_ratio;
    const double aspect_ratio = vertical_tail.factor_aspect_ratio.value() * wing_aspect_ratio;
    const int number_of_tails = 2;
    const geom2::Direction_3 orientation = geom2::Direction_3(0, 0, -1);
    geom2::Point_3 offset_position(vertical_tail.rear_x_offset.value(), vertical_tail.centerline_y_offset.value(),
                                   vertical_tail.centerline_z_offset.value());
    double wing_length = geom2::measure::offset_LE(data->wing, -vertical_tail.centerline_y_offset.value()).x() +
                         geom2::measure::chord(data->wing, -vertical_tail.centerline_y_offset.value()) +
                         data->wing_reference_position.x.value();

    std::string inner_profile_name = vertical_tail.inner_profile.value();
    std::string outer_profile_name = vertical_tail.outer_profile.value();

    volume_coefficient_method.compute_geometry(number_of_tails, volume_coefficient, sweep, taper_ratio, aspect_ratio,
                                               wing_reference_area, wing_span, wing_length, wing_neutral_position,
                                               offset_position, orientation, false);

    /* Read */
    geom2::AirfoilSection profile_root(airfoils_library->get_airfoil(inner_profile_name));
    airfoils_library->copy_available_airfoil(inner_profile_name, rtIO->getAirfoilDataDir());
    profile_root.name = inner_profile_name;

    /* Read outer profile */
    geom2::AirfoilSection profile_tip(airfoils_library->get_airfoil(inner_profile_name));
    airfoils_library->copy_available_airfoil(inner_profile_name, rtIO->getAirfoilDataDir());
    profile_tip.name = outer_profile_name;

    /* Vertical tail */
    data->tails.push_back(volume_coefficient_method.get_design(profile_root, profile_tip, 0.0));
    data->empennage_position.x = data->tails.front().origin.x();
    data->tails.back().origin =
          geom2::Point_3(0, data->tails.front().origin.y(), data->tails.front().origin.z());
    data->tails.back().name = vertical_tail.name.value() + "_positive";

    /* Vertical tail - controls */
    for (auto device : vertical_tail.control_devices) {
      data->controls.push_back(control_device(device.type.value(), device.position.inner_position_chord_from.value(),
      device.position.inner_position_chord_to.value(), device.position.inner_position_spanwise.value(),
      device.position.outer_position_chord_from.value(), device.position.outer_position_chord_to.value(),
      device.position.outer_position_spanwise.value()));
      data->control_devices_deflections.push_back(
        {device.deflection_full_negative.value(), device.deflection_full_positive.value()});
    }
    /* Vertical tail - spars */
    for (auto spar_element : vertical_tail.spars) {
      data->spars.push_back(spar(
        spar_element.name.value(), spar_element.position.inner_position_chord_from.value(),
        spar_element.position.inner_position_chord_to.value(), spar_element.position.inner_position_spanwise.value(),
        spar_element.position.outer_position_chord_from.value(), spar_element.position.outer_position_chord_to.value(),
        spar_element.position.outer_position_spanwise.value()));
    }



    /* Vertical tail (mirrored) */
    data->tails.push_back(data->tails.back());
    data->tails.back().origin =
          geom2::Point_3(0, -data->tails.front().origin.y(), data->tails.front().origin.z());
    data->tails.back().name = vertical_tail.name.value() + "_negative";



    // set empennage position
    data->empennage_position.y = 0;
    data->empennage_position.z = 0;
  }

}

void VerticalTails::redesign() {
  myRuntimeInfo->out << "Resizing mode active Vertical Tails" << '\n';
  // VolumeCoefficientMethod volume_coefficient_method;
  // double wing_sweep = data...;
  // double wing_reference_area = data...;
  // double wing_span = data...;
  // double wing_mac = data...;
  // double wing_taper_ratio = data...;
  // double wing_aspect_ratio = data...;
  // Vector3d wing_neutral_position = ...;
  // double fuselage_length = data...;

  // for (auto& vertical_tail : config_->vertical_tails) {
  //   const double vs_volume_coefficient = vertical_tail.volume_coefficient;
  //   const double vs_sweep = vertical_tail.delta_sweep + wing_sweep;
  //   const double vs_taper_ratio = vertical_tail.factor_taper_ratio * wing_taper_ratio;
  //   const double vs_aspect_ratio = vertical_tail.factor_aspect_ratio * wing_aspect_ratio;
  //   const Vector3d vs_offset_position(vertical_tail.rear_x_offset, 0.0, 0.0);
  //   const Alignment vs_alignment = vertical;
  //   const int number_of_vs = 1;
  //   volume_coefficient_method.Compute(number_of_vs, vs_volume_coefficient, vs_sweep, vs_taper_ratio, vs_aspect_ratio,
  //                                     wing_reference_area, wing_span, fuselage_length, wing_neutral_position,
  //                                     vs_offset_position, vs_alignment);
  // }
}

// == Spars and Controls ==============================================================================================

geom2::MultisectionSurface<geom2::PolygonSection> VerticalTails::control_device(
  const std::string &name, double inner_chord_rel_pos_from, double inner_chord_rel_pos_to,
  double inner_spanwise_rel_pos, double outer_chord_rel_pos_from, double outer_chord_rel_pos_to,
  double outer_spanwise_rel_pos) {
geom2::MultisectionSurface<geom2::PolygonSection> control_surface;
geom2::Polygon_2 inner_chord;
inner_chord.push_back(geom2::Point_2(inner_chord_rel_pos_from, 0.0));
inner_chord.push_back(geom2::Point_2(inner_chord_rel_pos_to, 0.0));

geom2::Polygon_2 outer_chord;
outer_chord.push_back(geom2::Point_2(outer_chord_rel_pos_from, 0.0));
outer_chord.push_back(geom2::Point_2(outer_chord_rel_pos_to, 0.0));

control_surface.sections.emplace_back(inner_chord);
control_surface.sections.emplace_back(outer_chord);

control_surface.sections.front().origin = geom2::Point_3(0.0, 0.0, inner_spanwise_rel_pos);
control_surface.sections.back().origin = geom2::Point_3(0.0, 0.0, outer_spanwise_rel_pos);

control_surface.name = name;
return control_surface;
}

geom2::MultisectionSurface<geom2::PolygonSection> VerticalTails::spar(
  const std::string &name, double inner_chord_rel_pos_from, double inner_chord_rel_pos_to,
  double inner_spanwise_rel_pos, double outer_chord_rel_pos_from, double outer_chord_rel_pos_to,
  double outer_spanwise_rel_pos) {
geom2::MultisectionSurface<geom2::PolygonSection> spar;
geom2::Polygon_2 inner_chord;
inner_chord.push_back(geom2::Point_2(inner_chord_rel_pos_from, 0.0));
inner_chord.push_back(geom2::Point_2(inner_chord_rel_pos_to, 0.0));

geom2::Polygon_2 outer_chord;
outer_chord.push_back(geom2::Point_2(outer_chord_rel_pos_from, 0.0));
outer_chord.push_back(geom2::Point_2(outer_chord_rel_pos_to, 0.0));

spar.sections.emplace_back(inner_chord);
spar.sections.emplace_back(outer_chord);

spar.sections.front().origin = geom2::Point_3(0.0, 0.0, inner_spanwise_rel_pos);
spar.sections.back().origin = geom2::Point_3(0.0, 0.0, outer_spanwise_rel_pos);
spar.name = name;

return spar;
}

// == Mass ============================================================================================================
void VerticalTails::flops_mass() {
  myRuntimeInfo->out << "Mode active -> Mass computation - FLOPS" << std::endl;
  const double maximum_takeoff_mass = data->maximum_takeoff_mass.data["mass"].value();

  /* Initialize empennage mass */
  data->empennage_mass.data["mass"].set_value(0.0);

  /* Iterate mass over tails as vertical stabilizers */
  double overall_empennage_mass = 0.0;
  for (size_t i = 0; i < data->tails.size(); ++i) {
    data->tails_mass.push_back(
        MassPropertiesIO(std::format("component_design/empennage/specific/geometry/aerodynamic_surface@{}/", i),
                         std::format("vertical_tail_{}", i)));
    data->tails_mass.back().data["mass"] =
        flops::vertical_stabilizer_mass(config->mass_mode_flops_technology_factor.value(), 1, geom2::measure::reference_area(data->tails[i]),
                                        geom2::measure::taper_ratio(data->tails[i]), maximum_takeoff_mass);
    geom2::Point_3 cg = geom2::measure::centroid(data->tails[i]);

    data->tails_mass.back().data["x"] = cg.x() + data->empennage_position.x.value() + data->tails[i].origin.x();
    data->tails_mass.back().data["y"] = cg.y() + data->empennage_position.y.value() + data->tails[i].origin.y();
    data->tails_mass.back().data["z"] = -cg.z() + data->empennage_position.z.value() + data->tails[i].origin.z();
    /* Assume symmetric gravity*/
    if (std::fabs(data->tails_mass.back().data["y"].value()) < 1E-2 ) {
      data->tails_mass.back().data["y"] = 0.0;
    }
    overall_empennage_mass += data->tails_mass.back().data["mass"].value();
  }
  data->empennage_mass.data["mass"] = overall_empennage_mass;
  double tmp_x(0), tmp_y(0), tmp_z(0);
  for (auto& tail_mass : data->tails_mass) {
    tmp_x +=  tail_mass.data["mass"].value()*(tail_mass.data["x"].value());
    tmp_y +=  tail_mass.data["mass"].value()*(tail_mass.data["y"].value());
    tmp_z +=  tail_mass.data["mass"].value()*(tail_mass.data["z"].value());
  }
  data->empennage_mass.data["x"] = tmp_x / data->empennage_mass.data["mass"].value();
  data->empennage_mass.data["y"] = tmp_y / data->empennage_mass.data["mass"].value();
  data->empennage_mass.data["z"] = tmp_z / data->empennage_mass.data["mass"].value();

  myRuntimeInfo->out << "Mass and CG Position for vertical tails" << std::endl;
  myRuntimeInfo->out << std::format("Mass   : {:.3f}", data->empennage_mass.data["mass"].value()) << std::endl;
  myRuntimeInfo->out << std::format("x-cg   : {:.3f}", data->empennage_mass.data["x"].value()) << std::endl;
  myRuntimeInfo->out << std::format("y-cg   : {:.3f}", data->empennage_mass.data["y"].value()) << std::endl;
  myRuntimeInfo->out << std::format("z-cg   : {:.3f}", data->empennage_mass.data["z"].value()) << std::endl;
}

}  // namespace low
}  // namespace bwb
