/*
 * UNICADO - UNIversity Conceptual Aircraft Design and Optimization
 *
 * Copyright (C) 2025 UNICADO consortium
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 * Description:
 * This file is part of UNICADO.
 */

#ifndef LOW_VERTICAL_TAILS_BWB_H_
#define LOW_VERTICAL_TAILS_BWB_H_

#include <airfoils/airfoils.h>
#include <moduleBasics/module.h>
#include <moduleBasics/report.h>
#include <moduleBasics/strategySelector.h>

#include <functional>
#include <map>

#include "bwb/vertical_tails/bwbVerticalTailsDesignConfig.h"
#include "bwb/vertical_tails/bwbVerticalTailsDesignData.h"

namespace bwb {
namespace low {

class VerticalTails : public Strategy {
 public:
  VerticalTails(const std::shared_ptr<RuntimeIO>& rtIO);
  ~VerticalTails() = default;

  /**
   * \brief Initialize method
   *
   */
  void initialize();

  /**
   * \brief Run method
   *
   */
  void run();


  void update();

  /**
   * \brief Report generation
   *
   */
  void report();

  /**
   * \brief Save results
   *
   */
  void save();

  /**
   * \brief Design method
   *
   */
  void design();

  /**
   * \brief Redesign method for empennage
   *
   */
  void redesign();

  /**
   * \brief Method to create shape of a control surface
   *
   * \param name name of the control surface
   * \param inner_chord_rel_pos_from
   * \param inner_chord_rel_pos_to
   * \param inner_spanwise_rel_pos
   * \param outer_chord_rel_pos_from
   * \param outer_chord_rel_pos_to
   * \param outer_spanwise_rel_pos
   * \return geom2::MultisectionSurface<geom2::PolygonSection>
   */
  geom2::MultisectionSurface<geom2::PolygonSection> control_device(
    const std::string& name, double inner_chord_rel_pos_from, double inner_chord_rel_pos_to,
    double inner_spanwise_rel_pos, double outer_chord_rel_pos_from, double outer_chord_rel_pos_to,
    double outer_spanwise_rel_pos);

  /**
   * \brief Method to create shape of a predefined spar
   *
   * \param name Name of the spar
   * \param inner_chord_rel_pos_from
   * \param inner_chord_rel_pos_to
   * \param inner_spanwise_rel_pos
   * \param outer_chord_rel_pos_from
   * \param outer_chord_rel_pos_to
   * \param outer_spanwise_rel_pos
   * \return geom2::MultisectionSurface<geom2::PolygonSection> spar element as a polygonsection
   */
  geom2::MultisectionSurface<geom2::PolygonSection> spar(const std::string& name, double inner_chord_rel_pos_from,
                                                             double inner_chord_rel_pos_to,
                                                             double inner_spanwise_rel_pos,
                                                             double outer_chord_rel_pos_from,
                                                             double outer_chord_rel_pos_to,
                                                             double outer_spanwise_rel_pos);

  /**
   * \brief Calculate empennage mass (vertical tails) and associated center of gravity by
   *        Flops method
   *
   */
  void flops_mass();

  /**
   * \brief Set the html body report
   *
   */
  void set_html_body();

  /**
   * \brief Generates plots for stabilizer planform and returns relative path to plot (svg)
   *
   * \param surface stabilizer
   * \param spars spars vector
   * \param devices control devices vector
   * \return fs::path relative path between report and plot
   */
  fs::path plot_stabilizer_planform(const geom2::MultisectionSurface<geom2::AirfoilSection>& surface,
    const std::vector<geom2::MultisectionSurface<geom2::PolygonSection>>& spars,
    const std::vector<geom2::MultisectionSurface<geom2::PolygonSection>>& devices);

  /**
  * \brief Generate unique plots of airfoils
  *
  * \param surface stabilizer
  * \return std::vector<fs::path> vector of relative paths between report and plots
  */
  std::vector<fs::path> plot_airfoils(const geom2::MultisectionSurface<geom2::AirfoilSection>& surface);

  /**
  * \brief Generate plot of thickness and twist distribution
  *
  * \param surface stabilizer
  * \return fs::path relative path between report and plot
  */
  fs::path plot_thickness_and_twist_distribution(const geom2::MultisectionSurface<geom2::AirfoilSection>& surface);

  public:
  std::map<std::string, std::function<void(void)>> design_mode;
  std::map<std::string, std::function<void(void)>> mass_mode;
  std::shared_ptr<Airfoils> airfoils_library;

  const std::shared_ptr<bwb::low::VerticalTailsConfig> config;
  const std::shared_ptr<bwb::low::VerticalTailsData> data;
  const std::shared_ptr<RuntimeIO>& rtIO;
  Report reporter;

};

}  // namespace low
}  // namespace bwb

#endif  // LOW_VERTICAL_TAILS_H_
