/*
 * UNICADO - UNIversity Conceptual Aircraft Design and Optimization
 *
 * Copyright (C) 2025 UNICADO consortium
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 * Description:
 * This file is part of UNICADO.
 */

#include <aircraftGeometry2/processing/measure.h>
#include <aircraftGeometry2/processing/transform.h>
#include <matplot/matplot.h>

#include <array>
#include <cstdlib>
#include <deque>
#include <map>
#include <utility>
#include <algorithm>

#include "lowBwbVerticalTailsDesign.h"
#include "lib/plot_methods/plot_utility.h"
namespace bwb {
namespace low {

fs::path VerticalTails::plot_stabilizer_planform(
    const geom2::MultisectionSurface<geom2::AirfoilSection>& surface,
    const std::vector<geom2::MultisectionSurface<geom2::PolygonSection>>& spars,
    const std::vector<geom2::MultisectionSurface<geom2::PolygonSection>>& devices) {
  /* Plot active */
  if (!rtIO->plotOn) {
    return "";
  }
  auto coordinates_xy = geom2::transform::outline_2d(surface, geom2::Direction_3(0, 0, 1));
  std::deque<double> spanwise;
  std::deque<double> chordwise;
  auto span = geom2::measure::span(surface);
  auto half_span = surface.is_symmetric ? span * 0.5 : span;
  auto sections = surface.sections;

  /* Check if control surface is oriented vertical */
  bool is_vertical = std::fabs(surface.normal.dz()) > 0.95;

  /* Loop coordinates from out to inside for LE and TE */
  for (auto it = sections.rbegin(); it != sections.rend(); ++it) {
    spanwise.push_front(-it->origin.z());
    spanwise.push_back(-it->origin.z());
    chordwise.push_front(it->origin.x());
    chordwise.push_back(it->origin.x() + it->get_chord_length());
  }

  /* Generate plot */
  std::string name = surface.name;
  std::replace(name.begin(), name.end(), '_', ' ');
  std::vector<std::string> legend_strings;
  auto f = matplot::figure(true);
  f->title(std::format("{} planform", name));
  auto ax = f->add_axes(false);
  if (is_vertical) {
    ax->plot(std::vector<double>(chordwise.begin(), chordwise.end()),
             std::vector<double>(spanwise.begin(), spanwise.end()))
        ->color("black")
        .line_width(1.5);
  } else {
    ax->plot(std::vector<double>(spanwise.begin(), spanwise.end()),
             std::vector<double>(chordwise.begin(), chordwise.end()))
        ->color("black")
        .line_width(1.5);
  }

  legend_strings.push_back("contour");
  matplot::hold(true);

  /* Plot spars */
  for (auto& spar : spars) {
    const auto [x, y, _] = gen_polygon(surface, spar);
    (void)_;
    if (is_vertical) {
      ax->plot(x, y)->color(get_spar_colors()).line_width(1).line_style("--");
    } else {
      ax->plot(y, x)->color(get_spar_colors()).line_width(1).line_style("--");
    }
    legend_strings.push_back("spar");
  }
  for (auto& device : devices) {
    const auto [x, y, name] = gen_polygon(surface, device);
    if (is_vertical) {
      ax->plot(x, y)->color(get_control_device_colors(name)).line_width(1.5);
    } else {
      ax->plot(y, x)->color(get_control_device_colors(name)).line_width(1.5);
    }
    legend_strings.push_back(device.name);
  }

  /* Select if vertical or horizontal element */
  if (std::fabs(surface.normal.dz()) > 0.7) {
    ax->xlabel("x [m]");
    ax->ylabel("z [m]");
  } else {
    ax->xlabel("y [m]");
    ax->ylabel("x [m]");
  }

  ax->limits_mode(matplot::manual);
  ax->axis(matplot::equal);
  if (is_vertical) {
    ax->ylim({0, *std::max_element(spanwise.begin(), spanwise.end()) * 1.1});
  } else {
    ax->xlim({0, *std::max_element(spanwise.begin(), spanwise.end()) * 1.1});
  }
  ax->legend(legend_strings);
  ax->legend()->location(matplot::legend::general_alignment::bottomright);
  ax->legend()->inside(true);
  ax->legend()->box(true);
  ax->legend()->font_size(8);

  fs::path plot_path = rtIO->getPlotDir() + "/empennage_design_" + surface.name + "_planform.svg";
  matplot::save(f, plot_path.string(), "svg");
  std::this_thread::sleep_for(std::chrono::milliseconds(500));
  return fs::relative(plot_path, fs::path(rtIO->getHtmlDir()));
}

fs::path VerticalTails::plot_thickness_and_twist_distribution(
    const geom2::MultisectionSurface<geom2::AirfoilSection>& surface) {
  /* Plot active */
  if (!rtIO->plotOn) {
    return "";
  }
  std::vector<double> eta;
  std::vector<double> thickness_to_chord_ratio;
  std::vector<double> twist;
  double half_span = surface.is_symmetric ? geom2::measure::span(surface) * 0.5 : geom2::measure::span(surface);
  for (auto section : surface.sections) {
    eta.push_back(-section.origin.z() / half_span);
    thickness_to_chord_ratio.push_back(geom2::measure::thickness_max(section) / section.get_chord_length());
    twist.push_back(section.get_twist_angle());
  }
  const auto [t_to_c_min,t_to_c_max] = std::ranges::minmax_element(thickness_to_chord_ratio);
  const auto [twist_min,twist_max] = std::ranges::minmax_element(twist);

  auto f = matplot::figure(true);

  auto ax1 = f->add_subplot(2, 1, 0);
  ax1->plot(eta, thickness_to_chord_ratio, "g-")->color("black").line_width(1.5);
  // ax1->xlabel("\u03B7");
  ax1->x_axis().visible(true);
  ax1->grid(true);
  ax1->ylabel("t / c [1]");
  ax1->ylim({*t_to_c_min - 0.01,*t_to_c_max + 0.01});

  auto ax2 = f->add_subplot(2, 1, 1);
  ax2->plot(eta, twist, "b-")->color("black").line_style("--").line_width(1.5);
  ax2->ylabel("\u03F5 [\u00b0]");
  ax2->xlabel("\u03B7 [1]");
  ax2->grid(true);
  ax2->ylim({*twist_min - 0.01, *twist_max + 0.01});

  fs::path plot_path = rtIO->getPlotDir() + "/empennage_design_" + surface.name + "_thickness_and_twist_distribution.svg";
  matplot::save(f, plot_path.string(), "svg");
  std::this_thread::sleep_for(std::chrono::milliseconds(500));
  return fs::relative(plot_path, fs::path(rtIO->getHtmlDir()));
}

/**
 * \brief Plot airfoils
 *
 * \return fs::path
 */
std::vector<fs::path> VerticalTails::plot_airfoils(const geom2::MultisectionSurface<geom2::AirfoilSection>& surface) {
  /* Plot active */
  if (!rtIO->plotOn) {
    return {""};
  }
  std::map<std::string, std::vector<geom2::Point_2>> airfoils_data{};
  std::vector<fs::path> airfoil_plots_path;

  for (auto section : surface.sections) {
    if (airfoils_data.find(section.name) != airfoils_data.end()) {
      continue;
    } else {
      airfoils_data.insert({section.name, section.get_contour().vertices()});
    }
  }
  /* loop over airfoils */

  for (const auto& airfoil_data : airfoils_data) {
    std::vector<double> x{}, y{};
    for (auto coord : airfoil_data.second) {
      x.push_back(coord.x());
      y.push_back(coord.y());
    }

    {
      auto f = matplot::figure(true);
      auto h = f->current_axes()->plot(x, y);
      h->line_width(1.5).color("black");

      auto ax = f->current_axes();
      ax->title(std::format("Airfoil - {}", airfoil_data.first));
      matplot::xlabel("\u03B7");
      ax->ylabel("z/c");
      ax->xlabel("\u03B7 [1]");
      ax->grid(true);
      ax->ylim({-0.2, 0.2});
      ax->xlim({0, 1});
      ax->axis(matplot::equal);

      fs::path plot_path = rtIO->getPlotDir() + "/empennage_design_" + surface.name + "_" + airfoil_data.first + "_airfoil.svg";
      matplot::save(f, plot_path.string(), "svg");
      airfoil_plots_path.push_back(fs::relative(plot_path, fs::path(rtIO->getHtmlDir())));
      std::this_thread::sleep_for(std::chrono::milliseconds(500));
    }
  }
  return airfoil_plots_path;
}

}  // namespace low
}  // namespace bwb
