# UNICADO - UNIversity Conceptual Aircraft Design and Optimization
#
# Copyright (C) 2025 UNICADO consortium
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#
# Description:
# This file is part of UNICADO.

def cabin_width_calculation(dict_ac_exchange, dict_mod_config, runtime_output):
    """ Calculates cabin width from given class definition.

    :param dict dict_ac_exchange: Dict containing parameter and according values from aircraft exchange file
    :param dict dict_mod_config: Dict containing parameter and according values from module configuration file
    :param logging.Logger runtime_output: Logging object used for capturing log messages in the module
    :return dict passenger_cabin_width: Value contains the cabin width of the payload tubes
            dict interior_description: Dict containing the interior description of each passenger class and payload tube
    """
    # initialize local parameter
    interior_description = {}
    passenger_cabin_width = {}

    # Loop across all selected passenger classes to estimate the required payload tube width per class.
    for class_name in dict_ac_exchange['relevant_classes']['passenger_classes']['class_names']:
        passenger_cabin_width[class_name] = []
        interior_description[class_name] = {}
        interior_description[class_name]['width_of_seat_rows_per_tube'] = []
        # Loop across all payload tubes to estimate the required tube width of current class for each payload tube.
        for tube in range(0, dict_ac_exchange['number_of_payload_tubes']):
            width_of_seat_rows_per_tube = []
            # Estimate the number of seats per row of current payload tube and current passenger class.
            number_of_seats_per_row = \
                [int(x) for x in dict_ac_exchange['relevant_classes']['passenger_classes'][class_name][
                    'seat_abreast'][tube].split('_')]
            if number_of_seats_per_row[0] == 0:
                number_of_inner_armrests = 0
                number_of_outer_armrests = 0
            elif number_of_seats_per_row[0] == 1:
                number_of_inner_armrests = 0
                number_of_outer_armrests = 2
            else:
                number_of_inner_armrests = number_of_seats_per_row[0] - 1
                number_of_outer_armrests = 2
            # Loop across all entries of "number of seats per row" to estimate seat row widths of current payload tube.
            width_of_seat_rows_per_tube.append(
                number_of_seats_per_row[0]
                * float(dict_ac_exchange['relevant_classes']['passenger_classes'][class_name]['passenger_seats'][
                                             'seat_left_side'][tube]['width']['value'])
                + (number_of_inner_armrests
                    * float(dict_ac_exchange['relevant_classes']['passenger_classes'][class_name]['passenger_seats'][
                                             'seat_left_side'][tube]['width_inner_armrest']['value']))
                + (number_of_outer_armrests
                   * float(dict_ac_exchange['relevant_classes']['passenger_classes'][class_name]['passenger_seats'][
                                             'seat_left_side'][tube]['width_outer_armrest']['value'])))

            if number_of_seats_per_row[-1] == 0:
                number_of_inner_armrests = 0
                number_of_outer_armrests = 0
            elif number_of_seats_per_row[-1] == 1:
                number_of_inner_armrests = 0
                number_of_outer_armrests = 2
            else:
                number_of_inner_armrests = number_of_seats_per_row[-1] - 1
                number_of_outer_armrests = 2
            width_of_seat_rows_per_tube.append(
                number_of_seats_per_row[-1]
                * float(dict_ac_exchange['relevant_classes']['passenger_classes'][class_name]['passenger_seats'][
                          'seat_right_side'][tube]['width']['value'])
                + (number_of_inner_armrests
                   * float(dict_ac_exchange['relevant_classes']['passenger_classes'][class_name]['passenger_seats'][
                               'seat_right_side'][tube]['width_inner_armrest']['value']))
                + (number_of_outer_armrests
                   * float(dict_ac_exchange['relevant_classes']['passenger_classes'][class_name]['passenger_seats'][
                               'seat_right_side'][tube]['width_outer_armrest']['value']))
            )
            passenger_aisle_width = dict_mod_config['passenger_aisle_width']
            # Check if the selected fuselage type is a wide body aircraft -> if true: -> append width of center row
            if len(number_of_seats_per_row) == 3:
                if number_of_seats_per_row[1] == 0:
                    number_of_inner_armrests = 0
                    number_of_outer_armrests = 0
                elif number_of_seats_per_row[1] == 1:
                    number_of_inner_armrests = 0
                    number_of_outer_armrests = 2
                else:
                    number_of_inner_armrests = number_of_seats_per_row[1] - 1
                    number_of_outer_armrests = 2
                width_of_seat_rows_per_tube.append(
                    number_of_seats_per_row[1]
                    * float(dict_ac_exchange['relevant_classes']['passenger_classes'][class_name]['passenger_seats'][
                              'seat_in_the_center'][tube]['width']['value'])
                    + (number_of_inner_armrests
                       * float(dict_ac_exchange['relevant_classes']['passenger_classes'][class_name]['passenger_seats'][
                                   'seat_in_the_center'][tube]['width_inner_armrest']['value']))
                    + (number_of_outer_armrests
                       * float(dict_ac_exchange['relevant_classes']['passenger_classes'][class_name]['passenger_seats'][
                                   'seat_in_the_center'][tube]['width_outer_armrest']['value']))
                    )
                # reorder the list "width_of_seat_rows_per_class" for wide body aircraft (left_side, center, right_side)
                center = width_of_seat_rows_per_tube[-1]
                width_of_seat_rows_per_tube[-1] = width_of_seat_rows_per_tube[1]
                width_of_seat_rows_per_tube[1] = center
                passenger_aisle_width = 2 * passenger_aisle_width

            passenger_cabin_width[class_name].append(
                round(sum(width_of_seat_rows_per_tube) + passenger_aisle_width
                      + (2 * dict_ac_exchange['relevant_classes']['passenger_classes'][
                                              class_name]['seat_space'][tube]), 3))

            interior_description[class_name]['width_of_seat_rows_per_tube'].append(width_of_seat_rows_per_tube)

    lowest_passenger_class = list(passenger_cabin_width.items())[-1]
    for key, value in passenger_cabin_width.items():
        for tube in range(0, dict_ac_exchange['number_of_payload_tubes']):
            if value[tube] > lowest_passenger_class[-1][tube]:
                runtime_output.warning('Attention! The cabin width required for ' + key + ' class of payload tube '
                                       + str(tube) + ' is greater than the cabin width required for the lowest selected'
                                                     ' passenger class. An optimal cabin layout is not possible under'
                                                     ' these conditions.')

    ''' debug print '''
    runtime_output.debug('Debug: The cabin_width_calculation function was successfully executed!')

    return passenger_cabin_width, interior_description
