/*
 * UNICADO - UNIversity Conceptual Aircraft Design and Optimization
 *
 * Copyright (C) 2025 UNICADO consortium
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 * Description:
 * This file is part of UNICADO.
 */

#ifndef INITIALSIZING_SRC_BWB_BWBINITIALSIZINGIODATA_H_
#define INITIALSIZING_SRC_BWB_BWBINITIALSIZINGIODATA_H_

#include <tuple>
#include <map>
#include <memory>
#include <string>
#include <vector>
#include <aixml/node.h>
#include <aixml/endnode.h>
#include "BwbInitialSizingData.h"
#include "BwbInitialSizingConfig.h"
#include <moduleBasics/runtimeIO.h>

/**
 * \brief BewbInitialSizingIOData class
 */
class BwbInitialSizingIOData {
protected:
    std::shared_ptr<RuntimeIO> rtIO_;

public:
    std::string aircraftType;  /** Aircraft type (e.g. CeRAS) */
    std::string aircraftModel; /** Aircraft model and version */
    std::string aircraftName;  /** Aircraft name = "type" - "model" */
    const node& configFile;    /** Nodes of the configuration file */

    BwbInitialSizingData data;
    BwbInitialSizingConfig configuration;

    /**
     * \brief Construct a new systems IO data object
     * \param rtIO rtIO pointer
     * \param config Node of configuration xml
     */
    BwbInitialSizingIOData(std::shared_ptr<RuntimeIO> rtIO, const node& config);

    /**
     * \brief Reads accommodation data from the acxml
     * \param acxml Aircraft XML
     * \return void
     */
    void setGlobalReferencePoint();
    
    /** 
     * \brief Update function for the maximum takeoff mass (MTOM) in the acxml.
     * \param mo_iteration Maximum takeoff mass estimation
     * \return void
     */
    void updateMTOM(double mo_iteration);

    /** 
     * \brief Update function for the operating mass empty (OME) in the acxml.
     * \param final_mass_estimation Final mass estimation
     * \return void
     */
    void updateOME(double final_mass_estimation);

    /** 
     * \brief Update function for the wing loading in the acxml.
     * \param designWS Wing loading design value
     * \return void
     */
    void updateWS(double designWS);

    /** 
     * \brief Update function for the thrust-to-weight ratio in the acxml.
     * \param designTW Thrust-to-weight ratio design value
     * \return void
     */
    void updateTW(double designTW);
};

#endif // INITIALSIZING_SRC_BWB_BWBINITIALSIZINGIODATA_H_
