/*
 * UNICADO - UNIversity Conceptual Aircraft Design and Optimization
 *
 * Copyright (C) 2025 UNICADO consortium
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 * Description:
 * This file is part of UNICADO.
 */

#include <string>
#include <memory>
#include <utility>
#include <stdexcept>
#include <algorithm>
#include "TawInitialSizingIOData.h"
#include <iostream>
#include <vector>
#include <sstream>
#include <limits>
#include <cmath>

TawInitialSizingIOData::TawInitialSizingIOData(std::shared_ptr<RuntimeIO> rtIO, const node& config)
    :
    rtIO_(rtIO),
    aircraftType(EndnodeReadOnly<std::string>("aircraft_exchange_file/general/type").read(rtIO_->acxml).value()),
    aircraftModel(EndnodeReadOnly<std::string>("aircraft_exchange_file/general/model").read(rtIO_->acxml).value()),
    configFile(config),
    configuration(TawInitialSizingConfig()) {
    // Call readAll method here
    configuration.readAll(config);
}




void TawInitialSizingIOData::readAccomodation(const node& acxml) {
    data.Accommodation.numberPassengers =
        EndnodeReadOnly<double>("/aircraft_exchange_file/requirements_and_specifications/design_specification/transport_task/passenger_definition/total_number_passengers")
        .read(acxml).value();
    data.Accommodation.massPassenger =
        EndnodeReadOnly<double>("/aircraft_exchange_file/requirements_and_specifications/design_specification/transport_task/passenger_definition/mass_per_passenger")
        .read(acxml).value();
    data.Accommodation.massLuggage =
        EndnodeReadOnly<double>("/aircraft_exchange_file/requirements_and_specifications/design_specification/transport_task/passenger_definition/luggage_mass_per_passenger")
        .read(acxml).value();
    data.Accommodation.massCargo =
        EndnodeReadOnly<double>("/aircraft_exchange_file/requirements_and_specifications/design_specification/transport_task/cargo_definition/additional_cargo_mass")
        .read(acxml).value();
}


void TawInitialSizingIOData::readMissionInfo(const node& acxml) {
    data.MissionInfo.range =
        EndnodeReadOnly<double>("/aircraft_exchange_file/requirements_and_specifications/requirements/top_level_aircraft_requirements/design_mission/range")
        .read(acxml).value();
    data.MissionInfo.contingency =
        EndnodeReadOnly<double>("/aircraft_exchange_file/requirements_and_specifications/requirements/top_level_aircraft_requirements/design_mission/contingency")
        .read(acxml).value();
    data.MissionInfo.distalt =
        EndnodeReadOnly<double>("/aircraft_exchange_file/requirements_and_specifications/requirements/top_level_aircraft_requirements/design_mission/alt_dist")
        .read(acxml).value();
    data.MissionInfo.loitertime =
        EndnodeReadOnly<double>("/aircraft_exchange_file/requirements_and_specifications/requirements/top_level_aircraft_requirements/design_mission/loiter_time")
        .read(acxml).value();
    data.MissionInfo.vApproach =
        EndnodeReadOnly<double>("/aircraft_exchange_file/requirements_and_specifications/requirements/top_level_aircraft_requirements/flight_envelope/maximum_approach_speed")
        .read(acxml).value();
    data.MissionInfo.sTOFL =
        EndnodeReadOnly<double>("/aircraft_exchange_file/requirements_and_specifications/requirements/top_level_aircraft_requirements/takeoff_distance")
        .read(acxml).value();
    data.MissionInfo.machNumber =
        EndnodeReadOnly<double>("/aircraft_exchange_file/requirements_and_specifications/design_mission/initial_cruise_mach_number")
        .read(acxml).value();
    data.MissionInfo.altitude =
        EndnodeReadOnly<double>("/aircraft_exchange_file/requirements_and_specifications/design_mission/initial_cruise_altitude")
        .read(acxml).value();
    data.MissionInfo.Gamma =
        EndnodeReadOnly<double>
        ("/aircraft_exchange_file/requirements_and_specifications/requirements/top_level_aircraft_requirements/flight_envelope/climb_or_descent_segment_gradient")
        .read(acxml).value();
}


void TawInitialSizingIOData::readEnergyCarrier(const node& acxml) {
    data.EnergyCarrier.energyCarrier =
        EndnodeReadOnly<std::string>("/aircraft_exchange_file/requirements_and_specifications/design_specification/energy_carriers/energy_carrier/type").read(acxml).value();
}

void TawInitialSizingIOData::updateMTOM(double mo_iteration) {
    std::string basePath = "aircraft_exchange_file/analysis/masses_cg_inertia/maximum_takeoff_mass/mass_properties";

    // Ensure mass_properties node exists
    if (!(rtIO_->acxml.find(basePath))) {
        rtIO_->acxml[basePath];
        std::cout << "Created mass_properties node." << std::endl;
    }

    // Mass
    Endnode<double> maximumTakeoffMass = data.create_endnode(mo_iteration, "kg", basePath + "/mass", "mass");
    maximumTakeoffMass.set_boundaries(0, INFINITY);
    maximumTakeoffMass.update(rtIO_->acxml);

    // Inertia
    std::string inertiaPath = basePath + "/inertia";

    if (!(rtIO_->acxml.find(inertiaPath))) {
        rtIO_->acxml[inertiaPath];
    }

    // Inertia components
    Endnode<double> j_xxNode = data.create_endnode(0.0, "kgm^2", inertiaPath + "/j_xx", "inertia in x");
    j_xxNode.set_boundaries(-std::numeric_limits<double>::infinity(), std::numeric_limits<double>::infinity());
    j_xxNode.update(rtIO_->acxml);

    Endnode<double> j_yyNode = data.create_endnode(0.0, "kgm^2", inertiaPath + "/j_yy", "inertia in y");
    j_yyNode.set_boundaries(-std::numeric_limits<double>::infinity(), std::numeric_limits<double>::infinity());
    j_yyNode.update(rtIO_->acxml);

    Endnode<double> j_zzNode = data.create_endnode(0.0, "kgm^2", inertiaPath + "/j_zz", "inertia in z");
    j_zzNode.set_boundaries(-std::numeric_limits<double>::infinity(), std::numeric_limits<double>::infinity());
    j_zzNode.update(rtIO_->acxml);

    Endnode<double> j_xyNode = data.create_endnode(0.0, "kgm^2", inertiaPath + "/j_xy", "inertia in xy");
    j_xyNode.set_boundaries(-std::numeric_limits<double>::infinity(), std::numeric_limits<double>::infinity());
    j_xyNode.update(rtIO_->acxml);

    Endnode<double> j_xzNode = data.create_endnode(0.0, "kgm^2", inertiaPath + "/j_xz", "inertia in xz");
    j_xzNode.set_boundaries(-std::numeric_limits<double>::infinity(), std::numeric_limits<double>::infinity());
    j_xzNode.update(rtIO_->acxml);

    Endnode<double> j_yxNode = data.create_endnode(0.0, "kgm^2", inertiaPath + "/j_yx", "inertia in yx");
    j_yxNode.set_boundaries(-std::numeric_limits<double>::infinity(), std::numeric_limits<double>::infinity());
    j_yxNode.update(rtIO_->acxml);

    Endnode<double> j_yzNode = data.create_endnode(0.0, "kgm^2", inertiaPath + "/j_yz", "inertia in yz");
    j_yzNode.set_boundaries(-std::numeric_limits<double>::infinity(), std::numeric_limits<double>::infinity());
    j_yzNode.update(rtIO_->acxml);

    Endnode<double> j_zxNode = data.create_endnode(0.0, "kgm^2", inertiaPath + "/j_zx", "inertia in zx");
    j_zxNode.set_boundaries(-std::numeric_limits<double>::infinity(), std::numeric_limits<double>::infinity());
    j_zxNode.update(rtIO_->acxml);

    Endnode<double> j_zyNode = data.create_endnode(0.0, "kgm^2", inertiaPath + "/j_zy", "inertia in zy");
    j_zyNode.set_boundaries(-std::numeric_limits<double>::infinity(), std::numeric_limits<double>::infinity());
    j_zyNode.update(rtIO_->acxml);

    // Center of Gravity
    std::string cogPath = basePath + "/center_of_gravity";

    if (!(rtIO_->acxml.find(cogPath))) {
        rtIO_->acxml[cogPath];
    }

    // Center of gravity components
    Endnode<double> x_cgNode = data.create_endnode(0.0, "m", cogPath + "/x", "x component");
    x_cgNode.set_boundaries(-std::numeric_limits<double>::infinity(), std::numeric_limits<double>::infinity());
    x_cgNode.update(rtIO_->acxml);

    Endnode<double> y_cgNode = data.create_endnode(0.0, "m", cogPath + "/y", "y component");
    y_cgNode.set_boundaries(-std::numeric_limits<double>::infinity(), std::numeric_limits<double>::infinity());
    y_cgNode.update(rtIO_->acxml);

    Endnode<double> z_cgNode = data.create_endnode(0.0, "m", cogPath + "/z", "z component");
    z_cgNode.set_boundaries(-std::numeric_limits<double>::infinity(), std::numeric_limits<double>::infinity());
    z_cgNode.update(rtIO_->acxml);
}


void TawInitialSizingIOData::updateOME(double final_mass_estimation) {
    std::string basePath = "aircraft_exchange_file/analysis/masses_cg_inertia/operating_mass_empty/mass_properties";
    if (!(rtIO_->acxml.find(basePath))) {
        rtIO_->acxml[basePath];
    }

    Endnode<double> operatingMassEmpty = data.create_endnode(final_mass_estimation, "kg", basePath + "/mass", "mass");
    operatingMassEmpty.set_boundaries(0, 300000);
    operatingMassEmpty.update(rtIO_->acxml);

    // Inertia
    std::string inertiaPath = basePath + "/inertia";

    if (!(rtIO_->acxml.find(inertiaPath))) {
        rtIO_->acxml[inertiaPath];
    }

    // Inertia components
    Endnode<double> j_xxNode = data.create_endnode(0.0, "kgm^2", inertiaPath + "/j_xx", "inertia in x");
    j_xxNode.set_boundaries(-std::numeric_limits<double>::infinity(), std::numeric_limits<double>::infinity());
    j_xxNode.update(rtIO_->acxml);

    Endnode<double> j_yyNode = data.create_endnode(0.0, "kgm^2", inertiaPath + "/j_yy", "inertia in y");
    j_yyNode.set_boundaries(-std::numeric_limits<double>::infinity(), std::numeric_limits<double>::infinity());
    j_yyNode.update(rtIO_->acxml);

    Endnode<double> j_zzNode = data.create_endnode(0.0, "kgm^2", inertiaPath + "/j_zz", "inertia in z");
    j_zzNode.set_boundaries(-std::numeric_limits<double>::infinity(), std::numeric_limits<double>::infinity());
    j_zzNode.update(rtIO_->acxml);

    Endnode<double> j_xyNode = data.create_endnode(0.0, "kgm^2", inertiaPath + "/j_xy", "inertia in xy");
    j_xyNode.set_boundaries(-std::numeric_limits<double>::infinity(), std::numeric_limits<double>::infinity());
    j_xyNode.update(rtIO_->acxml);

    Endnode<double> j_xzNode = data.create_endnode(0.0, "kgm^2", inertiaPath + "/j_xz", "inertia in xz");
    j_xzNode.set_boundaries(-std::numeric_limits<double>::infinity(), std::numeric_limits<double>::infinity());
    j_xzNode.update(rtIO_->acxml);

    Endnode<double> j_yxNode = data.create_endnode(0.0, "kgm^2", inertiaPath + "/j_yx", "inertia in yx");
    j_yxNode.set_boundaries(-std::numeric_limits<double>::infinity(), std::numeric_limits<double>::infinity());
    j_yxNode.update(rtIO_->acxml);

    Endnode<double> j_yzNode = data.create_endnode(0.0, "kgm^2", inertiaPath + "/j_yz", "inertia in yz");
    j_yzNode.set_boundaries(-std::numeric_limits<double>::infinity(), std::numeric_limits<double>::infinity());
    j_yzNode.update(rtIO_->acxml);

    Endnode<double> j_zxNode = data.create_endnode(0.0, "kgm^2", inertiaPath + "/j_zx", "inertia in zx");
    j_zxNode.set_boundaries(-std::numeric_limits<double>::infinity(), std::numeric_limits<double>::infinity());
    j_zxNode.update(rtIO_->acxml);

    Endnode<double> j_zyNode = data.create_endnode(0.0, "kgm^2", inertiaPath + "/j_zy", "inertia in zy");
    j_zyNode.set_boundaries(-std::numeric_limits<double>::infinity(), std::numeric_limits<double>::infinity());
    j_zyNode.update(rtIO_->acxml);

    // Center of Gravity
    std::string cogPath = basePath + "/center_of_gravity";

    if (!(rtIO_->acxml.find(cogPath))) {
        rtIO_->acxml[cogPath];
    }

    // Center of gravity components
    Endnode<double> x_cgNode = data.create_endnode(0.0, "m", cogPath + "/x", "x component");
    x_cgNode.set_boundaries(-std::numeric_limits<double>::infinity(), std::numeric_limits<double>::infinity());
    x_cgNode.update(rtIO_->acxml);

    Endnode<double> y_cgNode = data.create_endnode(0.0, "m", cogPath + "/y", "y component");
    y_cgNode.set_boundaries(-std::numeric_limits<double>::infinity(), std::numeric_limits<double>::infinity());
    y_cgNode.update(rtIO_->acxml);

    Endnode<double> z_cgNode = data.create_endnode(0.0, "m", cogPath + "/z", "z component");
    z_cgNode.set_boundaries(-std::numeric_limits<double>::infinity(), std::numeric_limits<double>::infinity());
    z_cgNode.update(rtIO_->acxml);
}


void TawInitialSizingIOData::updateWS(double designWS) {
    std::string subPath = "aircraft_exchange_file/sizing_point";
    if (!(rtIO_->acxml.find(subPath))) {
        rtIO_->acxml[subPath];
    }
    Endnode<double> wingLoading = data.create_endnode(designWS, "kg/m^2", subPath + "/wing_loading", "Maximum takeoff mass (MTOM) divided by wing area (Sref)");
    wingLoading.set_boundaries(0, INFINITY);
    wingLoading.update(rtIO_->acxml);
}


void TawInitialSizingIOData::updateTW(double designTW) {
    std::string subPath = "aircraft_exchange_file/sizing_point";
    if (!(rtIO_->acxml.find(subPath))) {
        rtIO_->acxml[subPath];
    }
    Endnode<double> thrustToweightRatio = data.create_endnode(designTW, "1", subPath + "/thrust_to_weight", "Total thrust (kN) divided by maximum aircraft weight (kN)");
    thrustToweightRatio.set_boundaries(0, 1);
    thrustToweightRatio.update(rtIO_->acxml);
}

void TawInitialSizingIOData::setGlobalReferencePoint() {
    std::string componentDesignPath = "aircraft_exchange_file/component_design/global_reference_point";

    // Check if the global_reference_point node exists under component_design
    if (!(rtIO_->acxml.find(componentDesignPath))) {
        rtIO_->acxml[componentDesignPath];
    }

    // Set reference component based on aircraft configuration
    std::string ACconfiguration = rtIO_->aircraft_configuration_type();
    std::string referenceComponent = (ACconfiguration == "tube_and_wing") ? "fuselage" : "wing";

    // Create Endnode for reference_component
    Endnode<std::string> refCompNode =
        data.create_endnode<std::string>(referenceComponent, componentDesignPath + "/reference_component", "Global reference point of aircraft geometry");
    refCompNode.update(rtIO_->acxml);

    // Create Endnode for x coordinate
    Endnode<double> xNode = data.create_endnode(0.0, "m", componentDesignPath+ "/x", "X coordinate of the global reference point");
    xNode.set_boundaries(-std::numeric_limits<double>::infinity(), std::numeric_limits<double>::infinity());
    xNode.update(rtIO_->acxml);

    // Create Endnode for y coordinate
    Endnode<double> yNode = data.create_endnode(0.0, "m", componentDesignPath + "/y", "Y coordinate of the global reference point");
    yNode.set_boundaries(-std::numeric_limits<double>::infinity(), std::numeric_limits<double>::infinity());
    yNode.update(rtIO_->acxml);

    // Create Endnode for z coordinate
    Endnode<double> zNode = data.create_endnode(0.0, "m", componentDesignPath + "/z", "Z coordinate of the global reference point");
    zNode.set_boundaries(-std::numeric_limits<double>::infinity(), std::numeric_limits<double>::infinity());
    zNode.update(rtIO_->acxml);
}



