# UNICADO - UNIversity Conceptual Aircraft Design and Optimization
#
# Copyright (C) 2025 UNICADO consortium
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#
# Description:
# This file is part of UNICADO.

def set_acn_flex_input_parameter(paths_and_names, landing_gear_description, dict_ac_exchange, dict_mod_config,
                                 runtime_output, dict_acn_flex_parameter, dict_constants):
    """ This function prepares all input parameter to estimate the aircraft classification number in accordance to
        ICAO COMFAA tool.

    :param dict paths_and_names: Dictionary containing system paths and ElementTrees
    :param dict landing_gear_description: Dict containing parameter and values of current landing gear estimation
    :param dict dict_ac_exchange: Dict containing parameter and according values from aircraft exchange file
    :param dict dict_mod_config: Dict containing parameter and according values from module configuration file
    :param logging.Logger runtime_output: Logging object used for capturing log messages in the module
    :param dict dict_acn_flex_parameter: Dict containing initial parameter and values for acn estimation
    :param dict dict_constants: Dict containing constants for acn estimation
    :return dict dict_acn_flex_parameter: Dict containing parameter and values of current acn estimation
    """

    # initialize constant parameter
    conv_inch = float(dict_constants['inch_to_cm'])
    conv_press = dict_constants['mpa_to_psi']
    # read name of current aircraft project
    aircraft_name = paths_and_names['name_of_project']

    # read user input of Coverages from configuration file of landing gear design module
    dict_acn_flex_parameter['coverages'] = int(dict_mod_config['coverages'])

    # read bool of alpha curve switch from configuration file of landing gear design module
    dict_acn_flex_parameter['select_new_alpha_from_curve'] = dict_mod_config['select_new_alpha_from_curve']

    # read status flag to use alpha defines from boeing proposal for new alpha factor from curve method
    dict_acn_flex_parameter['boeing_proposals_flag'] = dict_mod_config['use_boeing_proposals']

    # read number of tires of one main landing gear strut
    number_of_wheels = \
        landing_gear_description[
            'main_gear_description']['main_gear_arrangement']['number_of_main_gear_tires_per_outer_strut']

    # read x coordinates of wheel center of main landing gear strut in meter and convert to inch
    #  -> origin is in the center of the rearmost axis
    x_coord_of_wheels_from_input = \
        landing_gear_description['main_gear_description']['main_gear_arrangement']['local_tire_position_in_x_for_acn']
    x_coord_of_wheels = [round(x_coord * 100 / conv_inch, 4) for x_coord in x_coord_of_wheels_from_input]

    # read y coordinates of wheel center of main landing gear strut in meter and convert to inch
    #  -> origin is in the center of the rearmost axis
    y_coord_of_wheels_from_input = \
        landing_gear_description['main_gear_description']['main_gear_arrangement']['local_tire_position_in_y_for_acn']
    y_coord_of_wheels = [round(y_coord * 100 / conv_inch, 4) for y_coord in y_coord_of_wheels_from_input]

    # read grid coordinates and number of grid points
    #  -> grid_origin_in_x and grid_origin_in_y related to rearmost axis in inch
    grid_origin_in_x = 0.0 * 100 / conv_inch
    x_max_grid = round(x_coord_of_wheels_from_input[-1]/2 * 100 / conv_inch, 4)
    number_of_grid_points_in_x = 15
    grid_origin_in_y = 0.0 * 100 / conv_inch
    y_max_grid = round(y_coord_of_wheels_from_input[0] * 100 / conv_inch, 4)
    number_of_grid_points_in_y = 10

    # delta between grid points of x and y direction in inch
    delta_grid_in_x = round((x_max_grid - grid_origin_in_x) / (number_of_grid_points_in_x - 1), 4)
    delta_grid_in_y = round((y_max_grid - grid_origin_in_y) / (number_of_grid_points_in_y - 1), 4)

    # check if the given number of grid point in x direction is equal to zero
    #  -> if true: -> estimate grid maximum in x direction
    if number_of_grid_points_in_x <= 0:
        x_max_grid = max([max_x for max_x in x_coord_of_wheels if abs(max_x) > 0]) if number_of_wheels > 0 else 0.0
        if x_max_grid > 0:
            x_max_grid *= -1
        grid_origin_in_x = 0.0
        number_of_grid_points_in_x = 10
        delta_grid_in_x = (x_max_grid - grid_origin_in_x) / 2 / (number_of_grid_points_in_x - 1)

    # check if the given number of grid point in y direction is equal to zero
    #  -> if true: -> estimate grid maximum in y direction
    if number_of_grid_points_in_y <= 0:
        y_max_grid = max([max_y for max_y in y_coord_of_wheels if abs(max_y) > 0]) if number_of_wheels > 0 else 0.0
        if y_max_grid > 0:
            y_max_grid *= -1

        grid_origin_in_y = 0
        number_of_grid_points_in_y = 10
        delta_grid_in_y = (y_max_grid - grid_origin_in_y) / 2 / (number_of_grid_points_in_y - 1)

    # read aircraft gross mass in kg
    gross_mass = round(landing_gear_description['landing_gear_loads']['maximum_ramp_weight'] / dict_constants['g0'], 4)

    # read aircraft gross weight in kg
    if dict_ac_exchange['manufacturer_mass_empty'] is not None:
        manufacture_empty_mass = dict_ac_exchange['manufacturer_mass_empty']
    else:
        manufacture_empty_mass = 0.55 * gross_mass

    # read tire pressure and convert from psi to kPa
    rated_pressure_list = []
    for tire_name, sub_dict in landing_gear_description['main_gear_description']['main_gear_tires'].items():
        for key, value in sub_dict.items():
            if key == 'rated_pressure_in_psi':
                rated_pressure_list.append(round(float(value), 2))

    tire_pressure = round((max(rated_pressure_list) / conv_press) * 1000, 4)

    # read percentage load on main landing gear in percent
    percentage_load_on_main_gear_struts = \
        (landing_gear_description['landing_gear_loads']['main_gear_maximum_total_load']
         / landing_gear_description['landing_gear_loads']['maximum_ramp_weight'] * 100)

    # read number of main lading gear struts
    number_of_main_landing_struts = \
        landing_gear_description['main_gear_description']['main_gear_arrangement']['number_of_main_gear_struts']

    # read input alpha from user input and set to alternate -> default is 0.0
    input_alpha = 0.0
    if dict_mod_config['use_specific_alpha_value']:
        alternate_alpha = dict_mod_config['specific_alpha_value']
    else:
        alternate_alpha = input_alpha

    # read input CBR from user input -> default is 0.0
    if dict_mod_config['use_specific_california_bearing_ratio_for_flexible_pavement']:
        input_cbr = dict_mod_config['specific_california_bearing_ratio_for_flexible_pavement']
    else:
        input_cbr = 0.0

    # save parameter to dictionary of acn flex estimation according to ICAO
    dict_acn_flex_parameter['KTITLE'] = aircraft_name
    dict_acn_flex_parameter['NW'] = number_of_wheels
    dict_acn_flex_parameter['X'] = x_coord_of_wheels
    dict_acn_flex_parameter['Y'] = y_coord_of_wheels
    dict_acn_flex_parameter['GX'] = grid_origin_in_x
    dict_acn_flex_parameter['XMax'] = x_max_grid
    dict_acn_flex_parameter['XK'] = number_of_grid_points_in_x
    dict_acn_flex_parameter['GY'] = grid_origin_in_y
    dict_acn_flex_parameter['YMax'] = y_max_grid
    dict_acn_flex_parameter['YK'] = number_of_grid_points_in_y
    dict_acn_flex_parameter['DGX'] = delta_grid_in_x
    dict_acn_flex_parameter['DGY'] = delta_grid_in_y
    dict_acn_flex_parameter['KX'] = number_of_grid_points_in_x
    dict_acn_flex_parameter['KY'] = number_of_grid_points_in_y
    dict_acn_flex_parameter['IPR'] = int(1)
    dict_acn_flex_parameter['IPRS'] = int(1)
    dict_acn_flex_parameter['AMASS'] = gross_mass
    dict_acn_flex_parameter['MEM'] = manufacture_empty_mass
    dict_acn_flex_parameter['PRSW'] = tire_pressure
    dict_acn_flex_parameter['PMMG'] = percentage_load_on_main_gear_struts
    dict_acn_flex_parameter['AMLG'] = number_of_main_landing_struts
    dict_acn_flex_parameter['InputAlpha'] = input_alpha
    dict_acn_flex_parameter['AlternateAlpha'] = alternate_alpha
    dict_acn_flex_parameter['InputCBR'] = input_cbr

    runtime_output.debug('set_acn_flex_input_parameter successfully executed!')

    return dict_acn_flex_parameter
