/*
 * UNICADO - UNIversity Conceptual Aircraft Design and Optimization
 * 
 * Copyright (C) 2024 UNICADO consortium
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 * Description:
 * This file is part of UNICADO.
 */

#include <algorithm>

#include "flight_conditions.h"

/* Initialization of attributes of the FlightConditions object with the constructor */
FlightConditions::FlightConditions(std::vector<MissionFileDataHandler::MissionDataSet>* mission_data,
                                    OperatingConditions* oc, aerodynamics* aero,
                                    atmosphere* atm, MultiEnginePropulsion* engine, const size_t& variable_camber_segment_quantity)
    :
    /* Initialization of mission and operating condition pointer */
    mission_data(mission_data),
    operating_condition(oc),
    atm(atm),
    engine(engine),
    aero(aero),
    /* Initialization of thrust */
    thrust(0.),
    end_thrust(0.),
    /* Initialization of masses */
    aircraft_mass(0.),
    end_aircraft_mass(0.),
    /* Initialization of fuel masses and fuel mass flows */
    fuel_mass_flow(0.),
    mean_fuel_mass_flow(0.),
    end_fuel_mass_flow(0.),
    step_fuel_mass(0.),
    /* Initialization of CL */
    CL(0.),
    /* Initialization of drag */
    drag(0.),
    end_drag(0.),
    /* Initialization of Reynolds number */
    reynolds(0.),
    end_reynolds(0.),
    /* Initialization of spoiler factors */
    spoiler_factor(1.),
    end_spoiler_factor(1.),
    /* Initialization of true airspeed */
    TAS(0.),
    end_TAS(0.),
    /* Initialization of altitudes */
    altitude(0.),
    end_altitude(0.),
    /* Initialization of change parameters in segment */
    delta_h(0.),
    delta_x(0.),
    delta_t(0.),
    /* Initialization of rate of climb */
    ROC(0.),
    /* Initialization of path angle parameters */
    glide_path_angle(0.),
    mean_glide_path_angle(0.),
    end_glide_path_angle(0.),
    /* Initialization of acceleration factors */
    acceleration(0.),
    mean_acceleration(0.),
    end_acceleration(0.),
    /* Initialization of parameters at the end of the segment */
    end_angle_of_attack(0.),
    end_CL(0.),
    end_L_over_D(0.),
    end_specific_air_range(0.),
    /* Initialization of parameters for engine parameter calculation */
    rel_engine_availability(1.),
    thrust_efficiency(1.),
    bleed(0.),
    shaft_power(0.),
    origin_polar(),
    end_origin_polar{2, std::vector<double>(variable_camber_segment_quantity)},
    stabilizer_incidence(0.),
    end_stabilizer_incidence(0.),
    end_aircraft_cg(0.),
    first_run_switch(false),
    start_cg(false),
    /* Initialization of engine operation limit */
    engine_limit_reached(false) {
}

/* Destructor of the flight conditions object */
FlightConditions::~FlightConditions() {
}

void FlightConditions::update_operating_condition(const OperatingConditions& oc) {
    this->operating_condition = &oc;
}

/* Set Initial flight conditions */
void FlightConditions::set_initial_flight_track() {
    switch (this->mission_data->at(this->operating_condition->mission_step).mode) {
        case ACCELERATE: {
            this->ROC = this->mission_data->at(this->operating_condition->mission_step).ROC;
            this->glide_path_angle = fabs(this->ROC) < 1.e-3 ? 0. : asin(this->ROC / this->TAS);
            break;
        }
        case CLIMBCONSTCAS:
        case CLIMBTOCRUISE:
        case CLIMBTOSERVICECEILING:
        case DESCENT:
        case FLCHANGE:
        case FLCHANGESIMULATION:
        case FREEFLIGHT: { // use kinetic energy to climb/descend
            this->glide_path_angle = asin((std::min(1., (this->thrust - this->drag) / this->aircraft_mass / G_FORCE)));
            this->ROC = sin(this->glide_path_angle) * this->TAS;
            // If highest possible ROC (all thrust used) is above maximum ROC, set to maximum ROC
            if (this->mission_data->at(this->operating_condition->mission_step).ROC > 0.
                    && this->ROC > this->mission_data->at(this->operating_condition->mission_step).ROC) {
                this->ROC = this->mission_data->at(this->operating_condition->mission_step).ROC; //fix ROC for climb to maximum given climb rate
                this->glide_path_angle = asin(this->ROC / this->TAS);
            }
            break;
        }
        case DECELERATE:
        case APPROACH:
        case LANDING: {
            this->glide_path_angle = this->mission_data->at(this->operating_condition->mission_step).glide_path;
            this->ROC = sin(this->glide_path_angle) * this->TAS;
            break;
        }
        default: //use kinetic energy to accelerate/decelerate
            this->glide_path_angle = 0.;
            this->ROC = sin(this->glide_path_angle) * this->TAS;
    }
}

void FlightConditions::set_initial_acceleration_factors(const double& braking_coefficient, const double& friction_coefficient, const double& wing_surface_area) {
    switch (this->mission_data->at(this->operating_condition->mission_step).mode) {
        case TAKEOFF:
        case LANDING: {
            double minimum_acceleration(NAN);
            if (this->mission_data->at(this->operating_condition->mission_step).mode == LANDING) {
                minimum_acceleration = -braking_coefficient * G_FORCE;
            }
            double lift_force(this->CL * this->atm->getDensity(this->altitude) / 2. * pow(this->TAS, 2.) * wing_surface_area);
            this->acceleration = this->get_ground_run_acceleration(this->thrust, this->drag, lift_force,
                                                                    this->aircraft_mass * G_FORCE, minimum_acceleration, friction_coefficient);
            break;
        }
        case CRUISE:
        case CRUISECI:
        case GSINTERCEPTION:
            this->acceleration = 0.;
            break;
        default:
            //according to Newton's law F = m * a. Acceleration force F_a = (Thrust - drag - m * sin(glide_path_angle) * g) = m * a
            this->acceleration = (this->thrust - this->drag - this->aircraft_mass * sin(this->glide_path_angle) * G_FORCE) / this->aircraft_mass;
    }
}

double FlightConditions::get_ground_run_acceleration(const double& thrust, const double& drag, const double& lift,
                                                    const double& weight, const double& minimum_acceleration, const double& friction_coefficient) {
    double acceleration_tmp(G_FORCE * ((thrust / weight - friction_coefficient) + (lift * friction_coefficient - drag) / weight));
    if (!std::isnan(minimum_acceleration) && fabs(acceleration_tmp) < fabs(minimum_acceleration)) {// If braking can provide "more" deceleration, use the minimum_acceleration
        // Thrust reversers are still on, but some of the deceleration is done by the brakes. See: Flight Saftey Foundation, Briefing Note 8.4 - Braking Devices, Figure 3
        acceleration_tmp = minimum_acceleration; // LINK: https://flightsafety.org/files/alar_bn8-4-braking.pdf
    }
    return acceleration_tmp;
}

void FlightConditions::set_initial_thrust() {
    switch (this->mission_data->at(this->operating_condition->mission_step).mode) {
        case CRUISE:
        case CRUISECI:
        case APPROACH:
        case GSINTERCEPTION: {
            double required_thrust(sin(this->glide_path_angle) * this->aircraft_mass * G_FORCE + this->aircraft_mass * this->acceleration + this->drag);
            /* Set engine rating to required thrust:
             * if required thrust is smaller than idle thrust, the engine library throws engine_status = 3, which is caught here and the rating is then set to idle
             * in the almost impossible case that in descent or cruise the required thrust exceeds maximum available thrust, the engine library throws engine_status = 2
             */
            try {
                this->engine->calculate_N1_with_thrustlimit(this->mission_data->at(this->operating_condition->mission_step).energy_carrier_id,
                                                                this->altitude, convertUnit(TRUEAIRSPEED, MACH, this->altitude, *this->atm, this->TAS),
                                                                *this->atm, this->operating_condition->derate, this->operating_condition->rating,
                                                                this->bleed, this->shaft_power, required_thrust);
            } catch (const int& engine_status) {
                switch (engine_status) {
                    case 2: {
                        std::stringstream error_message;
                        error_message << "Required thrust (" << convertUnit(NOPREFIX, NEWTON, KILO, NEWTON, required_thrust)
                                        << " kN) cannot be delivered by engines in current status. ";
                        std::string new_rating("climb");
                        if (this->operating_condition->rating == "climb") {
                            new_rating = "maximum_continuous";
                        } else if (this->operating_condition->rating == "maximum_continuous") {
                            error_message << "Reduce drag.";
                            throwError(__FILE__, __func__, __LINE__, error_message.str());
                        } else {}
                        const OperatingConditions* oc_backup(this->operating_condition);
                        OperatingConditions oc_tmp(*(this->operating_condition));
                        oc_tmp.rating = new_rating;
                        this->update_operating_condition(oc_tmp);
                        this->set_initial_thrust();
                        this->operating_condition = oc_backup;
                        break;
                    }
                    case 3: {
                        myRuntimeInfo->info << "Set engine rating to idle ..." << std::endl;
                        this->engine_limit_reached = true;
                        this->engine->calculate_N1_with_penalties(this->mission_data->at(this->operating_condition->mission_step).energy_carrier_id,
                                                                    this->altitude, convertUnit(TRUEAIRSPEED, MACH, this->altitude, *this->atm, this->TAS),
                                                                    *this->atm, this->operating_condition->derate, "idle", this->bleed, this->shaft_power);
                        break;
                    }
                    default:
                        throwError(__FILE__, __func__, __LINE__, "Engine could not be set to requested status.");
                }
            }
            break;
        }
        default:
            this->engine->calculate_N1_with_penalties(this->mission_data->at(this->operating_condition->mission_step).energy_carrier_id, this->altitude,
                                                        convertUnit(TRUEAIRSPEED, MACH, this->altitude, *this->atm, this->TAS), *this->atm,
                                                        this->operating_condition->derate, this->operating_condition->rating, this->bleed, this->shaft_power);
    }
    this->thrust = this->engine->get_thrust_aircraft(this->mission_data->at(this->operating_condition->mission_step).energy_carrier_id) *
                        this->rel_engine_availability * this->thrust_efficiency;
}

void FlightConditions::set_flight_track() {
    switch (this->mission_data->at(this->operating_condition->mission_step).mode) {
        case ACCELERATE:
        case FLCHANGE:
        case FLCHANGESIMULATION: { //constant ROC
            this->ROC = this->mission_data->at(this->operating_condition->mission_step).ROC;
            this->end_glide_path_angle = fabs(this->ROC) < 1.e-3 ? 0. : asin(this->ROC / this->end_TAS);
            this->mean_glide_path_angle = (this->glide_path_angle + this->end_glide_path_angle) / 2.;
            this->ROC = sin(this->mean_glide_path_angle) * (this->TAS + this->end_TAS) / 2.;
            break;
        }
        case CLIMBCONSTCAS:
        case CLIMBTOCRUISE:
        case CLIMBTOSERVICECEILING:
        case DESCENT:
        case FREEFLIGHT: { // use kinetic energy to climb/descend
            this->end_glide_path_angle = asin((std::min(1., (this->end_thrust - this->end_drag) / this->end_aircraft_mass / G_FORCE)));
            this->ROC = sin(this->end_glide_path_angle) * this->end_TAS;
            if (this->mission_data->at(this->operating_condition->mission_step).ROC > 0.
                    && this->ROC > this->mission_data->at(this->operating_condition->mission_step).ROC) {
                this->ROC = this->mission_data->at(this->operating_condition->mission_step).ROC; //fix ROC for climb to maximum given climb rate
                this->end_glide_path_angle = asin(this->ROC / this->end_TAS);
            }
            this->mean_glide_path_angle = (this->glide_path_angle + this->end_glide_path_angle) / 2.;
            this->ROC = sin(this->mean_glide_path_angle) * (this->TAS + this->end_TAS) / 2.;
            break;
        }
        case DECELERATE:
        case APPROACH:
        case LANDING: {
            this->end_glide_path_angle = this->mission_data->at(this->operating_condition->mission_step).glide_path;
            this->mean_glide_path_angle = (this->glide_path_angle + this->end_glide_path_angle) / 2.;
            this->ROC = sin(this->mean_glide_path_angle) * (this->TAS + this->end_TAS) / 2.;
            break;
        }
        default:
            this->mean_glide_path_angle = (this->glide_path_angle + this->end_glide_path_angle) / 2.;
            this->ROC = sin(this->mean_glide_path_angle) * (this->TAS + this->end_TAS) / 2.;
    }
}

/* Update fligh conditions */
void FlightConditions::set_acceleration_factors(const double& braking_coefficient, const double& friction_coefficient, const double& wing_surface_area) {
    if ((this->mission_data->at(this->operating_condition->mission_step).mode == TAKEOFF ||
         this->mission_data->at(this->operating_condition->mission_step).mode == LANDING) &&
                (accuracyCheck(this->delta_h, 0., ACCURACY_HIGH))) { //ground run
            double minimum_acceleration(NAN);
            if (this->mission_data->at(this->operating_condition->mission_step).mode == LANDING) {
                minimum_acceleration = -braking_coefficient * G_FORCE;
            }
            double lift_force(this->end_CL * atm->getDensity(this->end_altitude) / 2. * pow(this->end_TAS, 2.) * wing_surface_area);
            this->end_acceleration = this->get_ground_run_acceleration(this->end_thrust, this->end_drag, lift_force, this->end_aircraft_mass * G_FORCE,
                                                                        minimum_acceleration, friction_coefficient);
    } else {
        // Acceleration according to Newton's law F = m * a. Acceleration force F_a = (Thrust - drag - m * sin(glide_path_angle) * g) = m * a
        this->end_acceleration = (this->end_thrust - this->end_drag - this->end_aircraft_mass * sin(this->end_glide_path_angle) * G_FORCE) / this->end_aircraft_mass;
    }
    if (accuracyCheck(this->mean_glide_path_angle, 0., ACCURACY_HIGH) || accuracyCheck(this->delta_h, 0., ACCURACY_HIGH)) {
        this->mean_acceleration = (this->acceleration + this->end_acceleration) / 2.;
    } else {
        this->mean_acceleration = (pow(this->end_TAS, 2.) - pow(this->TAS, 2.)) / (2. * this->delta_h / sin(this->mean_glide_path_angle));
    }
    if (accuracyCheck(this->mean_acceleration, 0., ACCURACY_LOW)) { //set numerical instability to zero
        this->mean_acceleration = 0.;
    }
}

void FlightConditions::set_thrust() {
    switch (this->mission_data->at(this->operating_condition->mission_step).mode) {
        case CRUISE:
        case CRUISECI:
        case APPROACH:
        case GSINTERCEPTION: {
            /* Set engine rating to required thrust:
             * if required thrust is smaller than idle thrust, the engine library throws engine_status = 3, which is caught here and the rating is then set to idle
             * in the almost impossible case that in descent or cruise the required thrust exceeds maximum available thrust, the engine library throws engine_status = 2
             */
            double required_thrust(sin(this->mean_glide_path_angle) * this->end_aircraft_mass * G_FORCE + this->end_aircraft_mass * this->mean_acceleration + this->end_drag);
            try {
                this->engine->calculate_N1_with_thrustlimit(this->mission_data->at(this->operating_condition->mission_step).energy_carrier_id, this->end_altitude,
                                                                convertUnit(TRUEAIRSPEED, MACH, this->end_altitude, *this->atm, this->end_TAS),
                                                                *this->atm, this->operating_condition->derate, this->operating_condition->rating,
                                                                this->bleed, this->shaft_power, required_thrust);
            } catch (const int& engine_status) {
                switch (engine_status) {
                    case 2: {
                        std::stringstream error_message;
                        error_message << "Required thrust (" << convertUnit(NOPREFIX, NEWTON, KILO, NEWTON, required_thrust)
                                        << " kN) cannot be delivered by engines in current status. ";
                        std::string new_rating("climb");
                        if (this->operating_condition->rating == "climb") {
                            new_rating = "maximum_continuous";
                        } else if (this->operating_condition->rating == "maximum_continuous") {
                            error_message << "Reduce drag.";
                            throwError(__FILE__, __func__, __LINE__, error_message.str());
                        } else {}
                        const OperatingConditions* oc_backup(this->operating_condition);
                        OperatingConditions oc_tmp(*(this->operating_condition));
                        oc_tmp.rating = new_rating;
                        this->update_operating_condition(oc_tmp);
                        this->set_thrust();
                        this->operating_condition = oc_backup;
                        break;
                    }
                    case 3: {
                        myRuntimeInfo->info << "Set engine rating to idle ..." << std::endl;
                        this->engine_limit_reached = true;
                        this->engine->calculate_N1_with_penalties(this->mission_data->at(this->operating_condition->mission_step).energy_carrier_id, this->end_altitude,
                                                                    convertUnit(TRUEAIRSPEED, MACH, this->end_altitude, *this->atm, this->end_TAS),
                                                                    *this->atm, this->operating_condition->derate, "idle", this->bleed, this->shaft_power);
                        break;
                    }
                    default:
                        throwError(__FILE__, __func__, __LINE__, "Engine could not be set to requested status.");
                }
            }
            break;
        }
        default:
            this->engine->calculate_N1_with_penalties(this->mission_data->at(this->operating_condition->mission_step).energy_carrier_id, this->end_altitude,
                                                        convertUnit(TRUEAIRSPEED, MACH, this->end_altitude, *this->atm, this->end_TAS), *this->atm,
                                                        this->operating_condition->derate, this->operating_condition->rating, this->bleed, this->shaft_power);
    }
    this->end_thrust = this->engine->get_thrust_aircraft(this->mission_data->at(this->operating_condition->mission_step).energy_carrier_id) *
                            this->rel_engine_availability * this->thrust_efficiency;
}

void FlightConditions::activate_spoilers() {
    switch (this->mission_data->at(this->operating_condition->mission_step).mode) {
        case CRUISE:
        case CRUISECI:
        case APPROACH:
        case GSINTERCEPTION: {
            if (!this->aero->useSpoiler((this->end_thrust - sin(this->mean_glide_path_angle) * this->end_aircraft_mass * G_FORCE -
                                            this->end_aircraft_mass * this->mean_acceleration), &this->end_drag))
                throwError(__FILE__, __func__, __LINE__, "drag too low to maintain requested flight state.");
            myRuntimeInfo->info << "end_drag after spoiler: " << this->end_drag << std::endl;
            break;
        }
        case LANDING: {
            this->end_drag *= this->aero->maxSpoilerFactor; //activate ground spoilers
            break;
        }
        default: break;
    }
}

void FlightConditions::update_end_mass(const bool& use_breguet) {
    this->end_fuel_mass_flow =
        this->engine->get_aircraft_fuelflow(this->mission_data->at(this->operating_condition->mission_step).energy_carrier_id) * this->rel_engine_availability;
    this->mean_fuel_mass_flow = (this->fuel_mass_flow + this->end_fuel_mass_flow) / 2.;

    if (use_breguet) {
        double Breguetfactor_levelCruise(this->end_specific_air_range * this->end_aircraft_mass);
        this->step_fuel_mass = this->aircraft_mass * (1. - exp(-this->delta_x / Breguetfactor_levelCruise));
    } else {
        this->step_fuel_mass = this->delta_t * this->mean_fuel_mass_flow;
    }
    this->end_aircraft_mass = this->aircraft_mass - this->step_fuel_mass;
}

void FlightConditions::set_way_and_time() {
    /* Accelerated motion:
     * NOTE: all velocity vectors are directed in flight direction (and with this also the final flight path vector)
     *       --> in the end the flight path vector must be corrected by track angle to get the range back
     * in general (from standstill): v = a * t; s = 1/2 * a * t^2
     * here: end_TAS = TAS + mean_acceleration * delta_t; delta_x = TAS * delta_t + 1/2 * mean_acceleration * delta_t^2 (constant motion + accelerated motion)
     * Climb/Descent: delta_x = delta_h / sin(mean_glide_path_angle) --> delta_t^2 + 2 * TAS / mean_acceleration - 2 * delta_h / sin(mean_glide_path_angle) / mean_acceleration = 0
     *  delta_t = -TAS / mean_acceleration +/- sqrt((TAS/mean_acceleration)^2 + 2 * delta_h / sin(mean_glide_path_angle) / mean_acceleration)
     *  mean_acceleration = (TAS_end^2 - TAS^2) / (2 * delta_h / sin(mean_glide_path_angle))
     */
    if (accuracyCheck(this->mean_acceleration, 0., 0.01)) { //constant motion
        if (!accuracyCheck(this->TAS, 0., ACCURACY_LOW) && !accuracyCheck(this->mean_glide_path_angle, 0., ACCURACY_LOW)) {
            this->delta_t = this->delta_h / sin(this->mean_glide_path_angle) / this->TAS;
        } else { // this condition is only met if CRUISE and, there, delta_x is preset
            this->delta_t = this->delta_x / this->TAS;
        }
    } else if (!accuracyCheck(this->mean_acceleration, 0., ACCURACY_LOW)) {
        if (accuracyCheck(this->mean_glide_path_angle, 0., ACCURACY_LOW)) { // horizontal acceleration
            this->delta_t = (this->end_TAS - this->TAS) / this->mean_acceleration;
        } else if (this->mean_acceleration < 0.) { // deceleration in climb/descent
            this->delta_t = -sqrt(pow(this->TAS / this->mean_acceleration, 2.) +
                            2. * this->delta_h / sin(this->mean_glide_path_angle) / this->mean_acceleration) - this->TAS / this->mean_acceleration;
        } else { // acceleration in climb/descent
            this->delta_t = sqrt(pow(this->TAS / this->mean_acceleration, 2.) +
                            2. * this->delta_h / sin(this->mean_glide_path_angle) / this->mean_acceleration) - this->TAS / this->mean_acceleration;
        }
    } else {
        myRuntimeInfo->err << "TAS: " << this->TAS << std::endl;
        myRuntimeInfo->err << "end_TAS: " << this->end_TAS << std::endl;
        myRuntimeInfo->err << "mean_acceleration: " << this->mean_acceleration << std::endl;
        myRuntimeInfo->err << "mean_glide_path_angle: " << this->mean_glide_path_angle << std::endl;
        myRuntimeInfo->err << "delta_h: " << this->delta_h << std::endl;
        throwError(__FILE__, __func__, __LINE__, "BUG in program. Abort program.");
    }
    this->delta_x = this->TAS * this->delta_t + 0.5 * this->mean_acceleration * pow(this->delta_t, 2.); // s = v*t + 1/2*a*t^2 = constant motion + accelerated motion
    /* Shift vector in ground distance direction */
    this->delta_x *= cos(this->mean_glide_path_angle);
}
