/*
 * UNICADO - UNIversity Conceptual Aircraft Design and Optimization
 * 
 * Copyright (C) 2025 UNICADO consortium
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 * Description:
 * This file is part of UNICADO.
 */

#include <algorithm>
#include <tuple>

#include "standard_mission.h"
#include "../output/standard_mission_output.h"

/**
 * @brief Construct a new StandardMission:: StandardMission object
 * 
 * @param rtIO pointer to the configuration file and the aircraft xml data
 */
StandardMission::StandardMission(const std::shared_ptr<RuntimeIO>& rtIO) :
    rtIO(rtIO),
    mission_file(nullptr),
    /* Variables */
    mission_type(EndnodeReadOnly<std::string>("program_settings/mode/mission_methods/mission_type").read(rtIO->moduleConfig).value()),
    low_fidelity(false),
    interpolate_aero(EndnodeReadOnly<bool>("program_settings/mode/mission_methods/interpolate_polars").read(rtIO->moduleConfig).value()),
    /* Relative mass values */
    rel_fuel_overshoot(EndnodeReadOnly<bool>("program_settings/program_specific/exit_if_fuel_limit_reached/enable").read(rtIO->moduleConfig).value() ?
        (EndnodeReadOnly<double>("program_settings/program_specific/exit_if_fuel_limit_reached/allowed_relative_overshoot").read(rtIO->moduleConfig).value())
        :
        (0.0)),
    rel_MTOM_overshoot(EndnodeReadOnly<bool>("program_settings/program_specific/exit_if_maximum_takeoff_mass_limit_reached/enable").read(rtIO->moduleConfig).value() ?
        (EndnodeReadOnly<double>("program_settings/program_specific/exit_if_maximum_takeoff_mass_limit_reached/allowed_relative_overshoot").read(rtIO->moduleConfig).value())
        :
        (0.0)),
    /* Masses */
    estimate_MTOM(this->mission_type == "design_mission" ? true : false), // estimate MTOM only for design mission
    max_payload_mass(EndnodeReadOnly<double>( // Maximum payload requirement
        "requirements_and_specifications/requirements/top_level_aircraft_requirements/maximum_structrual_payload_mass").read(rtIO->acxml).value()),
    OME(EndnodeReadOnly<double>("analysis/masses_cg_inertia/operating_mass_empty/mass_properties/mass").read(rtIO->acxml).value()), // Operating Empty Mass
    MTOM(checkBoundaries(rtIO->acxml.at("analysis/masses_cg_inertia/maximum_takeoff_mass/mass_properties/mass/value"), OME, true, // Check if OME boundary is crossed
        EndnodeReadOnly<double>("analysis/masses_cg_inertia/maximum_takeoff_mass/mass_properties/mass").read(rtIO->acxml).upper_boundary(), true)), // Maximum Takeoff Mass
    max_fuel_mass(0.), // Maximum fuel mass
    takeoff_mass(this->mission_type == "design_mission" ? // Check if a new MTOM shall be written (only for design mission)
        Endnode<double>("masses_cg_inertia/maximum_takeoff_mass/mass_properties/mass", "mass", 0, "kg", 0, 1.e6)
        :
        Endnode<double>("mission/" + this->mission_type + "/takeoff_mass", "Takeoff mass", 0, "kg", 0, 1.e6)),
    trip_fuel_mass(),
    previous_trip_fuel_mass(),
    loaded_fuel_mass(),
    previous_loaded_fuel_mass(),
    takeoff_fuel_mass(),
    landing_fuel_mass(),
    zero_fuel_mass(0.),
    taxi_fuel_mass_origin(),
    taxi_fuel_mass_destination(),
    contingency_fuel_mass(),
    alternate_fuel_mass(),
    final_reserve_fuel_mass(),
    extra_fuel_mass(),
    additional_fuel_mass(),
    top_of_climb_mass("top_of_climb_mass", "Total aircraft mass at top of climb (= start of initial cruise altitude (ICA))", 0, "kg", 0, 1.e6), // Mass at top of climb
    top_of_descent_mass("top_of_descent_mass", "Total aircraft mass at top of descent (TOD)", 0, "kg", 0, 1.e6),
    top_of_descent_mass_next_FL(0.),
    touchdown_mass(NAN),
    touchdown_mass_next_FL(NAN),
    /* Fueling */
    fuel_planning(fuel_planning_method(Modeselector(EndnodeReadOnly<std::string>("requirements_and_specifications/requirements/top_level_aircraft_requirements/" +
        std::string("design_mission/fuel_planning/fuel_estimation_selector")).read(rtIO->acxml).value()).get_mode_numeric())),
    cg_shift(),
    /* Flight performance parameters */
    max_operating_mach(EndnodeReadOnly<double>
            ("requirements_and_specifications/requirements/top_level_aircraft_requirements/flight_envelope/maximum_operating_mach_number").read(rtIO->acxml).value()),
    max_operating_speed(EndnodeReadOnly<double>
            ("requirements_and_specifications/requirements/top_level_aircraft_requirements/flight_envelope/maximum_operating_velocity").read(rtIO->acxml).value()),
    initial_cruise_mach(0.),
    initial_cruise_mach_required(EndnodeReadOnly<double>(
        "requirements_and_specifications/requirements/top_level_aircraft_requirements/design_mission/initial_cruise_mach_number").read(rtIO->acxml).value()),
    initial_cruise_altitude(0.),
    initial_cruise_altitude_required(EndnodeReadOnly<double>(
        "requirements_and_specifications/requirements/top_level_aircraft_requirements/design_mission/initial_cruise_altitude").read(rtIO->acxml).value()),
    max_operating_altitude(NAN),
    max_operating_altitude_required(EndnodeReadOnly<double>(
        "requirements_and_specifications/requirements/top_level_aircraft_requirements/maximum_operating_altitude").read(rtIO->acxml).value()),
    max_ceiling_altitude(NAN),
    max_altitude_one_engine_inoperative(NAN),
    approach_speed("approach_speed", "Approach speed when reaching FL15.", NAN, "m/s", 0, 200),
    approach_speed_required(EndnodeReadOnly<double>(
        "requirements_and_specifications/requirements/top_level_aircraft_requirements/flight_envelope/maximum_approach_speed").read(rtIO->acxml).value()),
    climb_speed_below_FL100(this->mission_type == "study_mission" ? // study mission has it's own node while design and requirements  mission get the same
        EndnodeReadOnly<double>("requirements_and_specifications/requirements/top_level_aircraft_requirements/" + this->mission_type +
        "/climb_speed_schedule/climb_speed_below_FL100").read(rtIO->acxml).value()
        :
        EndnodeReadOnly<double>(
            "requirements_and_specifications/requirements/top_level_aircraft_requirements/design_mission/climb_speed_schedule/climb_speed_below_FL100").read(rtIO->acxml).value()),
    climb_speed_above_FL100(this->mission_type == "study_mission" ? // study mission has it's own node while design and requirements  mission get the same
        EndnodeReadOnly<double>("requirements_and_specifications/requirements/top_level_aircraft_requirements/" + mission_type +
            "/climb_speed_schedule/climb_speed_above_FL100").read(rtIO->acxml).value()
        :
        EndnodeReadOnly<double>(
            "requirements_and_specifications/requirements/top_level_aircraft_requirements/design_mission/climb_speed_schedule/climb_speed_above_FL100").read(rtIO->acxml).value()),
    delta_mach_above_crossover(this->mission_type == "study_mission" ? // study mission has it's own node while design and requirements  mission get the same
        EndnodeReadOnly<double>("requirements_and_specifications/requirements/top_level_aircraft_requirements/" + mission_type +
            "/climb_speed_schedule/delta_mach_climb_cruise").read(rtIO->acxml).value()
        :
        EndnodeReadOnly<double>(
            "requirements_and_specifications/requirements/top_level_aircraft_requirements/design_mission/climb_speed_schedule/delta_mach_climb_cruise").read(rtIO->acxml).value()),
    descent_speed_below_FL100(this->mission_type == "study_mission" ? // study mission has it's own node while design and requirements  mission get the same
        EndnodeReadOnly<double>("requirements_and_specifications/requirements/top_level_aircraft_requirements/" + mission_type +
            "/descent_speed_schedule/descent_speed_below_FL100").read(rtIO->acxml).value()
        :
        EndnodeReadOnly<double>("requirements_and_specifications/requirements/top_level_aircraft_requirements" +
            std::string("/design_mission/descent_speed_schedule/descent_speed_below_FL100")).read(rtIO->acxml).value()),
    descent_speed_above_FL100(this->mission_type == "study_mission" ? // study mission has it's own node while design and requirements  mission get the same
        EndnodeReadOnly<double>("requirements_and_specifications/requirements/top_level_aircraft_requirements/" + mission_type +
            "/descent_speed_schedule/descent_speed_above_FL100").read(rtIO->acxml).value()
        :
        EndnodeReadOnly<double>("requirements_and_specifications/requirements/top_level_aircraft_requirements/" +
            std::string("design_mission/descent_speed_schedule/descent_speed_above_FL100")).read(rtIO->acxml).value()),
    lift_off_speed(Endnode<double>("lift_off_speed", "Speed at lift-off", 0, "m/s", 0, 350)),
    lift_off_lift_coefficient(Endnode<double>("lift_off_coefficient", "Lift coefficient at lift-off", 0, "1", 0, 100)),
    lift_off_angle_of_attack(Endnode<double>("lift_off_angle_of_attack", "Angle of attack at lift-off", 0, "rad", 0, Rounding(PI/2., 2))),
    takeoff_climb_angle(Endnode<double>("takeoff_climb_angle", "Climb angle towards screen height (35 ft)", 0, "rad", 0, Rounding(PI/2., 2))),
    takeoff_stall_speed(Endnode<double>("takeoff_stall_speed", "Stall speed until screen height (35 ft)", 0, "m/s", 0, 350)),
    maximum_landing_angle_off_attack(Endnode<double>("maximum_landing_angle_of_attack", "Maximum angle of attack while landing (between 50 ft and touchdown) ",
                                                        0, "rad", 0, Rounding(PI/2., 2))),
    /* Propulsion */
    engine(rtIO),
    /* Aerodynamic - Polar calculation */
    atm(),
    polar_file(rtIO->getAeroDataDir() + "/" + rtIO->acxml.at("analysis/aerodynamics/polar/polar_file/value")),
    polar_xml(aixml::openDocument(polar_file)),
    aero(rtIO->acxml, polar_xml),
    aero_forward(nullptr),
    aero_aft(nullptr),
    cg_forward(EndnodeReadOnly<double>("analysis/masses_cg_inertia/most_forward_mass/mass_properties/center_of_gravity/x").read(rtIO->acxml).value()),
    cg_design(EndnodeReadOnly<double>("analysis/masses_cg_inertia/design_mass/mass_properties/center_of_gravity/x").read(rtIO->acxml).value()),
    cg_aft(EndnodeReadOnly<double>("analysis/masses_cg_inertia/most_afterward_mass/mass_properties/center_of_gravity/x").read(rtIO->acxml).value()),
    /* Landing Gear Properties */
    friction_coefficient(EndnodeReadOnly<double>(
        "requirements_and_specifications/requirements/additional_requirements/landing_gear/friction_coefficient").read(rtIO->acxml).value()),
    braking_coefficient(EndnodeReadOnly<double>(
        "requirements_and_specifications/requirements/additional_requirements/landing_gear/braking_coefficient").read(rtIO->acxml).value()),
    /* Geometry */
    MAC(EndnodeReadOnly<double>("analysis/aerodynamics/reference_values/MAC").read(rtIO->acxml).value()), // Wing MAC
    wing_surface_area(EndnodeReadOnly<double>("analysis/aerodynamics/reference_values/S_ref").read(rtIO->acxml).value()),
    variable_camber_segment_quantity(0),
    variable_camber_switch(false),
    /* Requirement check */
    requirement_switch(false),
    initial_cruise_altitude_check(false),
    /* Mission data */
    auto_climb_switch(false),
    optimum_ROC(false),
    delta_ISA(this->mission_type == "study_mission" ? // study mission has it's own delta_ISA while design and requirements  mission get the same
        EndnodeReadOnly<double>("requirements_and_specifications/requirements/top_level_aircraft_requirements/" + this->mission_type + "/delta_ISA").read(rtIO->acxml).value()
        :
        EndnodeReadOnly<double>("requirements_and_specifications/requirements/top_level_aircraft_requirements/design_mission/delta_ISA").read(rtIO->acxml).value()),
    previous_block_time(),
    previous_flight_time(),
    average_cruise_bleed(0.),
    average_shaft_bleed(0.),
    payload_mass("payload_mass", "Total payload carried by the aircraft", 0, "kg", 0, 1.e6),
    top_of_climb_index(0),
    top_of_descent_index(0),
    end_of_descent_index(0),
    step_height_cruise(0.),
    cruise_profile_start(0),
    cruise_profile_end(0),
    top_of_climb_range("top_of_climb_range", "Flown range from takeoff to top of climb (= start of initial cruise altitude (ICA))", 0, "m", 0, 1.e9),
    top_of_descent_range("top_of_descent_range", "Flown range from takeoff to top of descent", 0, "m", 0, 1.e9),
    descent_range(0.),
    descent_range_next_FL(0.),
    top_of_descent_altitude(0.),
    time_start_to_1500ft(NAN),
    time_1500ft_to_ICA(NAN),
    time_start_to_ICA(NAN),
    range_start_to_ICA(NAN),
    fuel_start_to_ICA(NAN),
    takeoff_distance("takeoff_distance_normal_safety",
        "Takeoff distance at Sea Level for MTOM and (ISA + deltaISA)-Conditions(calculated by mission_analysis using missionDesign.xml settings) with all engines operating (AEO)",
        0, "m", 0, 10000),
    landing_distance("needed_runway_length", "Needed runway length with all engines operating and maximum landing mass", 0, "m", 0, 10000),
    CL_average(0.),
    CL_cruise_average("average_lift_coefficient", "Lift coefficient CL_average: Arithmetic mean over the entire cruise flight", 0, "1", -0.01, 1), // Average
    CL_cruise_max("maximum_lift_coefficient", "Maximum cruise flight lift coefficient CL_max", 0, "1", -0.01, 1), // Maximum
    CL_cruise_min("minimum_lift_coefficient", "Minimum cruise flight lift coefficient CL_min", 1, "1", -0.01, 1) { // Minimum (Don't initilize with 0 -> Min would stick to it)
}

void StandardMission::initialize() {
    /* Set mission's atmosphere and get its data from the mission file */
    atm.setAtmosphere(0., ISA_TEMPERATURE + delta_ISA, ISA_PRESSURE);
    mission_file = new MissionFileDataHandler(rtIO->getMissionDataDir() + "/" +
                            EndnodeReadOnly<std::string>("requirements_and_specifications/mission_files/" + mission_type + "_file").read(rtIO->acxml).value());

    /* Check if trimmed polars were calculated */
    if (interpolate_aero) {
        if (!rtIO->acxml.find("analysis/masses_cg_inertia/design_cg_shift")) {
            myRuntimeInfo->warn << "No shift for center of gravity (CoG) was provided by Weight and Balance Analysis: Will only use design mass CoG!" << std::endl;
        } else { // If a cg shift was done, fill the map
            for (node* cg_pair : rtIO->acxml.at("analysis/masses_cg_inertia/design_cg_shift").getVector("cg_point")) {
                std::map<std::string, double> map_tmp; // Erstelle eine neue Map
                map_tmp["mass"] = EndnodeReadOnly<double>("total_mass").read(*cg_pair).value();
                map_tmp["cg"] = EndnodeReadOnly<double>("center_of_gravity").read(*cg_pair).value();
                cg_shift.push_back(map_tmp);
            }
        }

        std::string polar_file_base = rtIO->getAeroDataDir() + "/" + rtIO->acxml.at("analysis/aerodynamics/polar/polar_file/value");
        std::string eraseXML = "_design_";
        std::string::size_type i = polar_file_base.find(eraseXML);
        if (i != std::string::npos) {
            polar_file_base.erase(i, polar_file_base.length());
        }
        std::string polar_file_forward = polar_file_base + "_forward_trimmed_polar.xml";
        std::string polar_file_aft = polar_file_base + "_aft_trimmed_polar.xml";
        if (!(fileExists(polar_file_forward)) || !(fileExists(polar_file_aft))) { // If polars do not exist, don't use trimmed polars
            interpolate_aero = false;
            myRuntimeInfo->warn << polar_file_forward  << " and/or "  <<  polar_file_aft  << " not existing!" << std::endl;
            myRuntimeInfo->warn << "To interpolate between design, forward, and aft polar, "
                                << "turn on 'doTrim' and 'TrimWithAdditionalCoGPositions' when executing Aerodynamic Analysis." << std::endl;
            myRuntimeInfo->warn << "The center of gravity shift will not be taken into account for aerodynamics!" << std::endl;
        } else {
            /* Open forward polar xml file and overwrite nullptr */
            node& polar_xml_forward(aixml::openDocument(polar_file_forward));
            aerodynamics forward_tmp(rtIO->acxml, polar_xml_forward);
            this->aero_forward = new aerodynamics(forward_tmp);

            /* Open aft polar xml file and overwrite nullptr */
            node& polar_xml_aft(aixml::openDocument(polar_file_aft));
            aerodynamics aft_tmp(rtIO->acxml, polar_xml_aft);
            this->aero_aft = new aerodynamics(aft_tmp);
        }
    }

    /* Variable Camber */
    this->variable_camber_switch = EndnodeReadOnly<bool>("requirements_and_specifications/design_specification/technology/variable_camber").read(rtIO->acxml).value();
    /* Get number of variable camber segments */
    if (this->variable_camber_switch) {
        std::string origin_polar_string = aero.getConfiguration(0).polars.at(0).vOrigin_polar.at(0);
        std::string string_tmp1 = origin_polar_string;
        this->variable_camber_segment_quantity = count(string_tmp1.begin(), string_tmp1.end(), '_');
    } else {
        this->variable_camber_segment_quantity = 0;
    }

    /* Get energy provider and setup tank capacities */
    std::string subpath("requirements_and_specifications/design_specification/configuration/tank_definition/tank@");
    const node& tank_definition(rtIO->acxml.at("requirements_and_specifications/design_specification/configuration/tank_definition"));
    const node& tank_design(rtIO->acxml.at("component_design/tank/specific"));
    for (uint16_t tank_id(0); tank_id < tank_design.getVector("tank").size(); tank_id++) {
        uint16_t energy_carrier_id(EndnodeReadOnly<int>("energy_carrier_ID").read(tank_definition.at("tank@" + num2Str(tank_id))).value());
        max_fuel_mass += EndnodeReadOnly<double>("maximum_energy_capacity").read(tank_design.at("tank@" + num2Str(tank_id))).value() /
                           EnergyCarrier(rtIO->get_fuel_type(energy_carrier_id), rtIO->get_fuel_density(energy_carrier_id)).gravimetric_energy_density;
    }

    for (const auto& [energy_carrier_id, energy_carrier] : rtIO->aircraft_energy_carriers()) {
        try {
            engine.get_energy_carrier_capacity(energy_carrier_id);
        } catch (...) {
            myRuntimeInfo->warn << "Energy carrier with ID = " << energy_carrier_id
                                << " @ requirements_and_specifications/design_specification/energy_carriers/energy_carrier is unused!" << std::endl;
        }
    }

    /* Check if performance maxima were analyzed previously */
    subpath = "assessment/performance/speed/";
    if (rtIO->acxml.find(subpath + "maximum_operating_mach_number")) {
        this->max_operating_mach = EndnodeReadOnly<double>("maximum_operating_mach_number").read(rtIO->acxml).value();
    }
    if (rtIO->acxml.find(subpath + "maximum_operating_velocity")) {
        this->max_operating_speed = EndnodeReadOnly<double>("maximum_operating_velocity").read(rtIO->acxml).value();
    }

    //TODO(GErrit): Overshoot must consider Multi-Fuel
    /* Check if further values from previous mission loops can be initialized */
    subpath = "analysis/mission/" + this->mission_type;
    if (rtIO->acxml.find(subpath)) {
        this->previous_trip_fuel_mass = this->fuel_map_setup(subpath, "in_flight_energy/trip_energy");
            // if (this->max_fuel_mass * (1 + this->rel_fuel_overshoot) < this->loaded_fuel_mass) {
            //     throwError(__FILE__, __func__, __LINE__, std::format("The trip fuel mass exceeded the maximum capacity by {} kg. Abort program!",
            //                 ceil(this->previous_trip_fuel_mass - this->max_fuel_mass * (1 + this->rel_fuel_overshoot))));
            // }
        this->previous_loaded_fuel_mass = this->fuel_map_setup(subpath, "loaded_mission_energy/mission_energy");
            // if (this->max_fuel_mass * (1 + this->rel_fuel_overshoot) < this->loaded_fuel_mass) {
            //     throwError(__FILE__, __func__, __LINE__, std::format("The loaded fuel mass exceeded the maximum capacity by {} kg. Abort program!",
            //                 ceil(this->loaded_fuel_mass - this->max_fuel_mass * (1 + this->rel_fuel_overshoot))));
            // }
        if (rtIO->acxml.find(subpath + "/block_time")) {
            this->previous_block_time = EndnodeReadOnly<double>(subpath + "/block_time").read(rtIO->acxml).value();
        }
        if (rtIO->acxml.find(subpath + "/flight_time")) {
            this->previous_flight_time = EndnodeReadOnly<double>(subpath + "/flight_time").read(rtIO->acxml).value();
        }
        if (rtIO->acxml.find(subpath + "/average_lift_coefficient")) {
            this->CL_cruise_average = EndnodeReadOnly<double>(subpath + "/cruise/average_lift_coefficient").read(rtIO->acxml).value();
        }
    }

    /* Check if the maximum spoiler factor is invalid; i.e. smaller than 1 or bigger than 4. If so, error message: abort program */
    if (this->aero.maxSpoilerFactor < 1. || this->aero.maxSpoilerFactor > 4.) {
        myRuntimeInfo->err << "Invalid value for spoiler factor (cf. AC-XML: 'Aerodynamics/MaxSpoilerFactor'): " << aero.maxSpoilerFactor;
        myRuntimeInfo->err << " (valid range: [1:4]). Abort program!" << std::endl;
        exit(1);
    }

    /* Set payloads and check if maxmimum was analyzed previously */
    if ((rtIO->acxml.find("analysis/masses_cg_inertia/maximum_payload_mass"))) { // if actual max payload mass was analyzed, use it
        max_payload_mass = EndnodeReadOnly<double>("analysis/masses_cg_inertia/maximum_payload_mass/mass_properties/mass").read(rtIO->acxml).value();
    }
    this->payload_mass = this->mission_file->payload_mass;

    /* Initialization of mission XML parameters */
    const node& energy_provider(rtIO->acxml.at("requirements_and_specifications/design_specification/propulsion/energy_provider"));
    mission_file->read_departure_steps(this->max_operating_speed, energy_provider);
    mission_file->read_flight_steps(&this->average_shaft_bleed, &this->average_cruise_bleed, &this->auto_climb_switch,
                                    &this->step_height_cruise, this->max_operating_mach, this->max_operating_speed, energy_provider);
    if (!(this->requirement_switch)) { // Requirements mission only flys until ceiling; no approach needed!
        mission_file->read_approach_steps(this->max_operating_speed, energy_provider);
    }

    /* Initialize mission */
    bool last_segment_free_flight_switch(false);
    this->analyze_mission_profile(&last_segment_free_flight_switch); // Analyze mission profile
    /* Guess mass at top of descent if initial mass at TOD is zero in AiXML file */
    if (accuracyCheck(this->top_of_descent_mass.value(), 0., ACCURACY_LOW)) {
        this->top_of_descent_mass = (this->MTOM * 0.15 + this->payload_mass.value() * .85 + this->OME * .85);
        this->top_of_descent_mass_next_FL = this->top_of_descent_mass.value();
    }
    if (!low_fidelity) { // Not needed in low fidelity method
        this->estimate_descent(last_segment_free_flight_switch); // Estimate descent distance and top of descent altitude
    }
    this->tanking_and_load_control(); // Control tanking and load masses
    this->takeoff_mass = this->zero_fuel_mass + this->get_fuel_sum(this->loaded_fuel_mass) - this->taxi_fuel_mass_origin.second; //Set current takeoff mass
    /* Output additional important masses */
    myRuntimeInfo->info << "m_ZF: " << this->zero_fuel_mass << std::endl;
    myRuntimeInfo->info << "m_TO: " << this->takeoff_mass.value() << std::endl;
    myRuntimeInfo->info << "MTOM: " << this->MTOM << std::endl;
}

void StandardMission::run() {
    /* Set conditions and descent step values */
    OperatingConditions oc(&this->mission_file->mission_data); // Initialization of operating conditions object
    oc.config = this->mission_file->mission_data.at(0).config;
    oc.rating = this->mission_file->mission_data.at(0).rating;
    FlightConditions fc(&this->mission_file->mission_data, &oc,
                                        &this->aero, &this->atm, &this->engine, this->variable_camber_segment_quantity); // Initialization of flight conditions object

    // fill an empty map with needed energy carrier IDs and initialize them with 0 J consumed energy
    std::map<int, double> initial_energy;
    for (const auto& [energy_carrier_id, energy_carrier] : rtIO->aircraft_energy_carriers()) {
        initial_energy[energy_carrier_id] = 0.;
    }
    std::map<int, double> fuel_mass_map; // initiliaze fuel map with all used energy carriers
    for (const auto& [energy_carrier_id, energy_carrier] : rtIO->aircraft_energy_carriers()) {
        fuel_mass_map[energy_carrier_id] = 0.;
    }
    std::map<uint16_t, double> n1_map; // Fill map with N1 for each used propulsor and its ID
    for (node* propulsor : rtIO->acxml.at("requirements_and_specifications/design_specification/propulsion").getVector("propulsor")) {
        n1_map[propulsor->getIntAttrib("ID")] = 0.;
    }
    // Set center of gravity for start
    if (this->interpolate_aero) {
        fc.end_aircraft_cg = this->get_current_cg(this->takeoff_mass.value());
    }
    // Push back first mission profile using additional input values (trigger alternative constructor for first segment of takeoff)
    this->mission_profile.push_back(MissionSegment(fc, oc, n1_map, fuel_mass_map, this->takeoff_mass.value()));

    /* Starting calculations */
    myRuntimeInfo->out << "Starting mission calculation with a TOM = " << this->mission_profile.back().end_total_mass << " kg." << std::endl;
    fc.first_run_switch = true;
    fc.start_cg = true;
    /* Takeoff according to ECAC Doc. 29 */
    for (size_t i(0); i < this->mission_file->mission_data.size(); i++) {
        /* Output: calculation mode and current parameters */
        myRuntimeInfo->out << "*********************************************************" << std::endl;
        myRuntimeInfo->out << "Calculation mode:     " << this->mission_file->mission_data.at(i).mode << " in mission segment: " << i + 1 << std::endl;
        myRuntimeInfo->out << "Current altitude:     " << convertUnit(METER, FOOT, this->mission_profile.back().end_altitude) << " ft." << std::endl;
        myRuntimeInfo->out << "Current range:        " << convertUnit(METER, NAUTICALMILE, this->mission_profile.back().end_range) << " NM." << std::endl;
        myRuntimeInfo->out << "Current Mach number:  " << this->mission_profile.back().end_mach << std::endl;
        myRuntimeInfo->out << "Current CL:           " << this->mission_profile.back().end_CL << std::endl;
        myRuntimeInfo->out << "Current L/D:          " << this->mission_profile.back().end_L_over_D << std::endl;
        switch (this->mission_file->mission_data.at(i).mode) {
            /* Mode: take off */
            case FlightConditions::TAKEOFF:
                this->takeoff_calculation(&oc, &fc, i);
                break;
            /* Mode: accelerate while maintaining a constant rate of climb */
            case FlightConditions::ACCELERATE:
                this->change_speed_at_constant_ROC(&oc, &fc, i);
                break;
            /* Mode: climb while maintaining a constant calibrated airspeed */
            case FlightConditions::CLIMBCONSTCAS:
                this->change_altitude_at_constant_speed(&oc, &fc, i);
                break;
            /* Mode: climb to cruise altitude */
            case FlightConditions::CLIMBTOCRUISE:
                this->climb_to_cruise_altitude(&oc, &fc, i);
                break;
            /* Mode: cruise with a constant Mach number */
            case FlightConditions::CRUISE:
                this->cruise_at_constant_mach(&oc, &fc, i);
                break;
            /* Mode: descend to approach */
            case FlightConditions::DESCENT:
                myRuntimeInfo->out << "Descend to 10.000 ft ... " << std::endl;
                this->cruise_profile_end = this->mission_profile.size() - 1; // Set index
                this->change_altitude_at_constant_speed(&oc, &fc, i);
                break;
            /* Mode: approach to landing */
            case FlightConditions::APPROACH:
                myRuntimeInfo->out << "Approach to landing ... " << std::endl;
                this->change_altitude_at_constant_speed(&oc, &fc, i);
                /* Set the approach speed if it was not set yet and the plane reaches FL15.
                 * Only needed to be set at the highest approach point, since the CAS is the highest there (for later requirements check).
                 */
                if (std::isnan(approach_speed.value()) && fc.altitude <= convertUnit(FOOT, METER, 1500.)) {
                        approach_speed = convertUnit(TRUEAIRSPEED, CALIBRATEDAIRSPEED, fc.altitude, this->atm, fc.TAS);
                    }
                break;
            /* Mode: decelerate during approach */
            case FlightConditions::DECELERATE:
                this->change_speed_at_constant_ROC(&oc, &fc, i);
                break;
            /* Mode: landing */
            case FlightConditions::LANDING:
                this->landing(&oc, &fc, i);
                break;
            /* Mode: glide slope interception */
            case FlightConditions::GSINTERCEPTION:
                this->glide_slope_interception(&oc, &fc, i);
                break;
            /* Mode: climb (change flight level) with a constant calibrated airspeed or Mach number at a fixed rate of climb */
            case FlightConditions::FLCHANGE:
                this->change_altitude_at_constant_speed(&oc, &fc, i);
                break;
            /* Mode: climb to service ceiling */
            case FlightConditions::CLIMBTOSERVICECEILING:
                this->climb_to_service_ceiling(&oc, &fc, i);
                i = INT_MAX; // Break for loop after service ceiling (requirement check). Other steps are not needed anymore
                break;
            /* Mode: air to air fueling */
            case FlightConditions::AIRTOAIRFUELLING:
                this->air_to_air_fuelling(&oc, &fc, i);
                break;
            /* Mode: free flight */
            case FlightConditions::FREEFLIGHT:
                this->free_flight(&oc, &fc, i);
                break;
            /* No valid mission mode, error message */
            default:
                myRuntimeInfo->err << "No valid mission mode given. Abort program." << std::endl;
                exit(1);
        }
        if (this->mission_profile.back().end_engine_rating != oc.rating) {
            myRuntimeInfo->warn << "Engine rating was changed in this mission step from '" << oc.rating
                                << "' to '" << this->mission_profile.back().end_engine_rating << "'" << std::endl;
        }
    }
    /* Update relevant mass values */
    this->update_mass_values();

    /* Analyze the cruise profile to write them into aircraft xml */
    if (!this->requirement_switch) {
        this->analyze_cruise_profile();
    }
}

void StandardMission::update() {
    /* Create path to mission and build node if none exists yet */
    std::string subpath("analysis/mission");
    check_and_add_nodes(subpath, "Mission data", -1, 0);
    subpath += "/" + this->mission_type;
    check_and_add_nodes(subpath, "Data of " + replaceAll(this->mission_type, "_", " "));
    std::string subpath_backup(subpath);

    /* Non-global endnode initialization */
    EndnodeReadOnly<double> energy_carrier_tmp(EndnodeReadOnly<double>("consumed_energy"));
    EndnodeReadOnly<double> range_tmp( // Get endnode boundaries from TLARs
        EndnodeReadOnly<double>("requirements_and_specifications/requirements/top_level_aircraft_requirements/design_mission/range").read(rtIO->acxml));
    Endnode<double> relative_cruise_step_end("relative_end_of_cruise_step", "End of cruise step relative to total cruise length", 0 , "1", 0, 1);
    Endnode<double> altitude("altitude", "Altitude of cruise step", 0 , "m", 0, 15000);
    Endnode<double> block_time("block_time", "Block time for the whole mission: Time from break release to end of taxiing after landing",
        this->mission_profile.back().end_time + this->mission_file->taxi_time_origin + this->mission_file->taxi_time_destination, "s", 0, 1.e6);
    Endnode<double> takeoff_engine_derate("takeoff_engine_derate", "Engine power demand", this->mission_file->mission_data.at(0).derate, "1", 0, 1);
    Endnode<double> flight_time("flight_time", "Flight time for the whole mission", this->mission_profile.back().end_time, "s", 0, 1.e6);
    Endnode<double> total_range("range", "Traveled range from break release to end of taxi at destination",
                        this->mission_profile.back().end_range, range_tmp.unit(), range_tmp.lower_boundary(), range_tmp.upper_boundary());
    Endnode<int> energy_carrier_id("energy_carrier_ID", "See energy carrier specification node", 0, "1", 0, 5); // TODO(GErrit): Auf Multifuel anpassen hier und drunter
    Endnode<double> consumed_energy("consumed_energy", "Energy amount", 0, "J", 0, 1.e13);

    /* Aircraft parameters */
    // Since takeoff_mass can be put into mission node for study mission and into masses_cg_inertia node for design mission,
    // only the analysis node is passed (see Endnode declaration)
    this->takeoff_mass.update(rtIO->acxml.at("analysis"));
    total_range.update(rtIO->acxml.at(subpath));
    // this->payload_mass.update(rtIO->acxml.at(subpath));  ggf. max_payload_mass bei design
    // rtIO->acxml.at(subpath + "number_of_pax/value") = this->PAX; // TODO (Gerrit): Soll doch bestimmt nach createMissionXML o.ae.
    // rtIO->acxml.at(subpath + "cargo_mass/value")    = this->cargo_mass;

    /* Energy parameters */
    /* Mission energy */
    std::string description("Amount of energy loaded into tanks (including reserves) for the mission");
    this->check_and_add_nodes(subpath + "/loaded_mission_energy", description);
    description = "Amount of energy loaded into tanks (including reserves) for specified energy carrier";
    this->update_energy_nodes(this->loaded_fuel_mass, subpath + "/loaded_mission_energy/mission_energy", description, &(energy_carrier_id), &(consumed_energy));

    /* In-flight energy */
    description = "Amount of energy needed for in-flight segments (all segments from takeoff to landing)";
    this->check_and_add_nodes(subpath + "/in_flight_energy", description);
    description = "Amount of energy needed for trip segments (all segments from takeoff to landing) for specified energy carrier";
    this->update_energy_nodes(this->trip_fuel_mass, subpath + "/in_flight_energy/trip_energy", description, &(energy_carrier_id), &(consumed_energy));
    description = "Amount of energy needed for takeoff segment for specified energy carrier";
    this->update_energy_nodes(this->takeoff_fuel_mass, subpath + "/in_flight_energy/takeoff_energy", description, &(energy_carrier_id), &(consumed_energy));
    description = "Amount of energy needed for landing segment for specified energy carrier";
    this->update_energy_nodes(this->landing_fuel_mass, subpath + "/in_flight_energy/landing_energy", description, &(energy_carrier_id), &(consumed_energy));

    /* Taxiing */
    description = "Amount of energy needed for taxiing specified energy carrier";
    energy_carrier_id = taxi_fuel_mass_origin.first; // origin and destination have the same energy carrier
    this->check_and_add_nodes(subpath + "/taxi_energy", description);

    description = "Amount of energy needed for taxiing at origin for specified energy carrier";
    this->check_and_add_nodes(subpath + "/taxi_energy/taxi_out_energy", description, 0);
    consumed_energy = this->taxi_fuel_mass_origin.second *
        EnergyCarrier(rtIO->get_fuel_type(energy_carrier_id.value()), rtIO->get_fuel_density(energy_carrier_id.value())).gravimetric_energy_density;
    consumed_energy.update(rtIO->acxml.at(subpath + "/taxi_energy/taxi_out_energy@0"));
    energy_carrier_id.update(rtIO->acxml.at(subpath + "/taxi_energy/taxi_out_energy@0"));

    description = "Amount of energy needed for taxiing at destination for specified energy carrier";
    this->check_and_add_nodes(subpath + "/taxi_energy/taxi_in_energy", description, 0);
    consumed_energy =  this->taxi_fuel_mass_destination.second *
        EnergyCarrier(rtIO->get_fuel_type(energy_carrier_id.value()), rtIO->get_fuel_density(energy_carrier_id.value())).gravimetric_energy_density;
    consumed_energy.update(rtIO->acxml.at(subpath + "/taxi_energy/taxi_in_energy@0"));
    energy_carrier_id.update(rtIO->acxml.at(subpath + "/taxi_energy/taxi_in_energy@0"));

    // Block time
    block_time.update(rtIO->acxml.at(subpath));

    /* Mission parameters */
    flight_time.update(rtIO->acxml.at(subpath));
    takeoff_engine_derate.update(rtIO->acxml.at(subpath));

    /* Takeoff and landing performance*/
    subpath += "/takeoff_performance_parameters";
    this->check_and_add_nodes(subpath, "Performance parameters for takeoff steps from 0 ft to screen height (35 ft).");
    this->lift_off_speed.update(rtIO->acxml.at(subpath));
    this->takeoff_stall_speed.update(rtIO->acxml.at(subpath));
    this->lift_off_lift_coefficient.update(rtIO->acxml.at(subpath));
    this->lift_off_angle_of_attack.update(rtIO->acxml.at(subpath));
    this->takeoff_climb_angle.update(rtIO->acxml.at(subpath));

    subpath = subpath_backup + "/landing_performance_parameters";
    this->check_and_add_nodes(subpath, "Performance parameters for landing steps from 50 ft to touchdown.");
    myRuntimeInfo->out << this->maximum_landing_angle_off_attack.value() << std::endl;
    this->maximum_landing_angle_off_attack.update(rtIO->acxml.at(subpath));

    /* Cruise lift coefficients */
    subpath = subpath_backup + "/cruise";
    this->check_and_add_nodes(subpath, "Characteristics of the cruise segment");
    this->CL_cruise_average.update(rtIO->acxml.at(subpath));
    this->CL_cruise_min.update(rtIO->acxml.at(subpath));
    this->CL_cruise_max.update(rtIO->acxml.at(subpath));

    /* Cruise masses */
    this->top_of_climb_mass.update(rtIO->acxml.at(subpath));
    this->top_of_descent_mass.update(rtIO->acxml.at(subpath));

    /* Cruise profile */
    this->top_of_climb_range.update(rtIO->acxml.at(subpath));
    this->top_of_descent_range.update(rtIO->acxml.at(subpath));
    subpath += "/cruise_steps";
    this->check_and_add_nodes(subpath, "Cruise step information");
    for (size_t i(0); i < this->cruise_profile.size(); i++) {
        this->check_and_add_nodes(subpath + "/cruise_step", "Data of a cruise step", i);
        relative_cruise_step_end   =  this->cruise_profile.at(i).rel_step_length;
        altitude =  convertUnit(HECTO, FOOT, NOPREFIX, METER, this->cruise_profile.at(i).FL);
        relative_cruise_step_end.update(rtIO->acxml.at(subpath + "/cruise_step@" + num2Str(i)));
        altitude.update(rtIO->acxml.at(subpath + "/cruise_step@" + num2Str(i)));
    }

    /* Runway length */
    if (this->estimate_MTOM) { // estimate_MTOM = true -> design mission -> set runway lengths and approach speed
        // Set nodes if not existing
        subpath = "aircraft_exchange_file/assessment";
        this->check_and_add_nodes(subpath, "Assessment data");
        subpath += "/performance";
        this->check_and_add_nodes(subpath, "Assessment of performance values");
        this->check_and_add_nodes(subpath + "/takeoff", "Takeoff assessment");

        this->takeoff_distance.update(rtIO->acxml.at(subpath + "/takeoff"));
    }
    this->check_requirements();
}

void StandardMission::report() {
    StandardMissionOutput StandardMissionOutput(this);
}

void StandardMission::save() {
}

// Analysis methods
void StandardMission::analyze_cruise_profile() {
    bool FL_change(false);
    /* Initialization of a cruise step object */
    CruiseStep step;
    /* Initialization of cruise step parameters */
    step.rel_step_length = 0.; // Relative step length
    step.FL = 0.; // Flight level
    step.mass = 0.; // Mass

    int cruise_profile_count(0);
    for (size_t i(this->cruise_profile_start); i < this->mission_profile.size() - 1 && i < this->cruise_profile_end; i++) {
        if (fabs(this->mission_profile.at(i).end_altitude - this->mission_profile.at(i + 1).end_altitude) < ACCURACY_HIGH) {
            /* No climbing or sinking */
            FL_change = false;
        } else {
            /* Climbing or sinking */
            if (!FL_change) {
                FL_change = true;
                /* Calculate the relative step length */
                step.rel_step_length = (this->mission_profile.at(i).end_range - this->mission_profile.at(this->cruise_profile_start).end_range) /
                                        (this->mission_profile.at(this->cruise_profile_end).end_range - this->mission_profile.at(this->cruise_profile_start).end_range);
                /* Calculate the flight level */
                step.FL = convertUnit(METER, FOOT, this->mission_profile.at(i).end_altitude) / 100.;
                /* Set the mass at the end of the step */
                step.mass = this->mission_profile.at(i).end_total_mass;
                this->cruise_profile.push_back(step);
            }
        }
        /* Gather lift coefficients */
        this->CL_cruise_average  = this->CL_cruise_average.value() + this->mission_profile.at(i).end_CL; // Average
        this->CL_cruise_max = std::max(this->CL_cruise_max.value(), this->mission_profile.at(i).end_CL); // Maximum
        this->CL_cruise_min = std::min(this->CL_cruise_min.value(), this->mission_profile.at(i).end_CL); // Minimum
        cruise_profile_count++;
    }
    this->CL_cruise_average /= cruise_profile_count; // Devide the gathered lift coefficients by the counter to get the arethmetic mean
    step.rel_step_length = 1.; // Set the relative step length to 1, because the cruise end was reached by the for-loop
    step.FL = convertUnit(NOPREFIX, METER, HECTO, FOOT, this->mission_profile.at(this->cruise_profile_end).end_altitude); // Calculation of the flight level
    /* Set the mass at the end of the step */
    step.mass = this->mission_profile.at(this->cruise_profile_end).end_total_mass;
    this->cruise_profile.push_back(step);
    /* Output: current parameters */
    myRuntimeInfo->out << "Current altitude:    " << convertUnit(METER, FOOT, this->mission_profile.back().end_altitude) << " ft." << std::endl;
    myRuntimeInfo->out << "Current range:       " << convertUnit(METER, NAUTICALMILE, this->mission_profile.back().end_range) << " NM." << std::endl;
    myRuntimeInfo->out << "Cruise Step Profile: " << std::endl;
    myRuntimeInfo->out << "  FL   | rel. Step Length | End Mass [kg]" << std::endl;
    for (size_t i(0); i < this->cruise_profile.size(); ++i) {
        myRuntimeInfo->out  << std::setprecision(2) << std::fixed  << this->cruise_profile.at(i).FL << " |        "
                            << this->cruise_profile.at(i).rel_step_length << "      | " << this->cruise_profile.at(i).mass << std::endl;
    }
    myRuntimeInfo->out << "CL_cruise: Average = " << this->CL_cruise_average.value()
                        << ", Max = " << this->CL_cruise_max.value() << ", Min = " << this->CL_cruise_min.value() << std::endl;
    /* Set the mass and range at the top of climb */
    this->top_of_climb_mass = this->mission_profile.at(this->cruise_profile_start).end_total_mass;
    this->top_of_climb_range = this->mission_profile.at(this->cruise_profile_start).end_range;
    /* Set the mass and range at the top of descent */
    this->top_of_descent_mass = this->cruise_profile.back().mass;
    this->top_of_descent_range = this->mission_profile.at(this->cruise_profile_end).end_range;
    /* Output of the masses at the top of climb and top of descent */
    myRuntimeInfo->out << "Mass (TOC) = " << this->top_of_climb_mass.value() << " kg; Mass (TOD) = " << this->top_of_descent_mass.value() << " kg" << std::endl;
    myRuntimeInfo->out << "(TOC reached after " << convertUnit(METER, NAUTICALMILE, this->top_of_climb_range.value())<< " NM; TOD reached after " <<
                            convertUnit(METER, NAUTICALMILE, this->top_of_descent_range.value())  << " NM)" << std::endl;
}

void StandardMission::analyze_mission_profile(bool* last_segment_free_flight_switch) {
    /* Evaluate mission data */
    std::vector<int> modes;
    uint16_t position(0);
    std::vector<uint16_t> positions;
    /* Check if the mission data is empty. If so, error message */
    if (this->mission_file->mission_data.empty()) throwError(__FILE__, __func__, __LINE__, "No mission data available.");
    for (size_t i(0); i < mission_file->mission_data.size(); i++) {
        modes.push_back(mission_file->mission_data.at(i).mode);
    }
    /* Iterators to find the configuration entries */
    std::vector<int>::iterator entryCruise = find(modes.begin(), modes.end(), FlightConditions::CRUISE); // Iterator to find the configuration entry
    std::vector<int>::iterator entryFF = find(modes.begin(), modes.end(), FlightConditions::FREEFLIGHT); // Iterator to find the configuration entry
    std::vector<int>::iterator descent_entry = find(modes.begin(), modes.end(), FlightConditions::DESCENT); // Iterator to find the configuration entry

    if (entryCruise == modes.end() && this->requirement_switch) { // No cruise phase given? Check whether the requirements missions works (if turned on)
        entryCruise = find(modes.begin(), modes.end(), FlightConditions::CLIMBTOSERVICECEILING); // CLIMTOSERVICECEILING is only used in requirement mission
        if (entryCruise == modes.end()) { // Undefined mission type
            /* Error message */
            throwError(__FILE__, __func__, __LINE__, "Undefined mission type. No Cruise phase for normal mission or climb_to_service_ceiling phase for requirement mission given.");
        }
    } else {
       this->top_of_climb_index = static_cast<uint16_t>(entryCruise - modes.begin());
    }
    while (entryCruise != modes.end() || entryFF != modes.end()) { // Until the search iterator reaches its end
        if (entryCruise != modes.end()) { // Already at the end?
            position = static_cast<uint16_t>(entryCruise - modes.begin());
            positions.push_back(position); // Save positions with cruise entries
            entryCruise = find(entryCruise + 1, modes.end(), FlightConditions::CRUISE); // Let the iterator search further
        }
        if (entryFF != modes.end()) { // Already at the end?
            position = static_cast<uint16_t>(entryFF - modes.begin());
            positions.push_back(position); // Save positions with cruise entries
            entryFF = find(entryFF + 1, modes.end(), FlightConditions::FREEFLIGHT); // Let the iterator search further
            *last_segment_free_flight_switch = true;
        }
    }
    /* Set mission data indexes */
    this->top_of_descent_index = position; // Top of descent
    while (descent_entry != modes.end()) { // Until the search iterator reaches its end
        this->end_of_descent_index = static_cast<uint16_t>(descent_entry - modes.begin());
        descent_entry = find(descent_entry + 1, modes.end(), FlightConditions::DESCENT); // Let the iterator search further
    }
    /* Output of segment distances */
    for (size_t i(0); i < positions.size(); i++) {
        myRuntimeInfo->info << "Relative segment length in section " << positions.at(i) << ": "
            << this->mission_file->mission_data.at(positions.at(i)).rel_segment_length << std::endl;
    }
    /* Check if the top of climb point is out of mission data. If so, error message */
    if (this->top_of_climb_index >= this->mission_file->mission_data.size()) throwError(__FILE__, __func__, __LINE__, "Top of climb point is out of mission data.");
    /* Check if the top of descent point is out of mission data. If so, error message */
    if (this->top_of_descent_index >= this->mission_file->mission_data.size()) throwError(__FILE__, __func__, __LINE__, "Top of descent point is out of mission data.");
}

void StandardMission::check_requirements() {
    std::string performance_path      = "assessment/performance";
    std::string requirements_check_path = "requirement_compliance/top_level_aircraft_requirements";
    /* If performance requirements are checked, save them for the output plots */
    if (rtIO->acxml.find(requirements_check_path + "/maximum_operating_altitude") && // Look up check if it exists
            EndnodeReadOnly<bool>(requirements_check_path + "/maximum_operating_altitude/checked").read(rtIO->acxml).value()) {
        this->max_operating_altitude = EndnodeReadOnly<double>(performance_path + "/altitude/maximum_operating_altitude").read(rtIO->acxml).value();
        this->max_ceiling_altitude =  EndnodeReadOnly<double>(performance_path + "/altitude/maximum_ceiling_altitude").read(rtIO->acxml).value();
    }
    if (rtIO->acxml.find(requirements_check_path + "/maximum_altitude_one_engine_inoperative") && // Look up check if it exists
            EndnodeReadOnly<bool>(requirements_check_path + "/maximum_altitude_one_engine_inoperative/checked").read(rtIO->acxml).value()) {
        this->max_altitude_one_engine_inoperative = EndnodeReadOnly<double>(performance_path + "/altitude/maximum_altitude_one_engine_inoperative").read(rtIO->acxml).value();
    }
    if (rtIO->acxml.find(requirements_check_path + "/design_time_to_climb") && // Look up check if it exists
            EndnodeReadOnly<bool>(requirements_check_path + "/design_time_to_climb/checked").read(rtIO->acxml).value()) {
        this->time_start_to_1500ft = EndnodeReadOnly<double>(performance_path + "/climb/time_to_1500ft_ISA").read(rtIO->acxml).value();
        this->time_1500ft_to_ICA = EndnodeReadOnly<double>(performance_path + "/climb/design_time_to_climb").read(rtIO->acxml).value();
        this->time_start_to_ICA = EndnodeReadOnly<double>(performance_path + "/climb/time_to_ICA_ISA").read(rtIO->acxml).value();
        this->range_start_to_ICA = EndnodeReadOnly<double>(performance_path + "/climb/range_to_ICA_ISA").read(rtIO->acxml).value();
        this->fuel_start_to_ICA = EndnodeReadOnly<double>(performance_path + "/climb/fuel_to_ICA_ISA").read(rtIO->acxml).value();
    }
}

// Error handling
void StandardMission::throw_mission_error(int error_code, const uint16_t &mission_step) {
    std::stringstream error_message;
    switch (error_code) {
        /* Error: negative flight route leading to an insufficient thrust */
        case 1:
            error_message << "Error: Negative flight distance! -> Thrust not sufficient.";
            break;
        /* Error: insufficient acceleration potential */
        case 2:
            error_message << "Error: Acceleration not sufficient!";
            break;
        /* Error: resistance too low to maintain the glide path */
        case 3:
            error_message << "Error: Drag too low to maintain the glide path!";
            break;
        /* Error: insufficient thrust to maintain trajectory */
        case 4:
            error_message << "Error: Thrust not sufficient to maintain trajectory!";
            break;
        /* Error: divergence too low */
        case 5:
            error_message << "Error: Divergence too large!";
            break;
        /* Error: climb angle for iteration too small leading to a too large or too small thrust */
        case 6:
            error_message << "Climb angle for Iteration too small - Thrust too high or too low!";
            break;
        /* Unknown error */
        default:
            error_message << "Unkown Error";
            break;
    }
    error_message << " Occured in mode step " << mission_step + 1;
    error_message << " Mode: " << this->mission_file->mission_data.at(mission_step).mode_name << ", ";
    error_message << " Index in mission file: " << this->mission_file->mission_data.at(mission_step).mission_file_index << std::endl;
    throw error_message.str();
}

// Add new nodes
void StandardMission::check_and_add_nodes(std::string nodepath, const std::string& description, const int& id, const int& tool_level) {
    /* If a nodepath is given without 'aircraft_exchange_file' at front, add it to avoid bugs */
    std::string::size_type i = nodepath.rfind("aircraft_exchange_file/");
    if (i == std::string::npos) {
        nodepath = "aircraft_exchange_file/" + nodepath;
    }
    /* Check if the wanted node is already set */
    if (!rtIO->acxml.find(nodepath) || (id >= 0 && (!rtIO->acxml.find(nodepath + "@" + num2Str(id))))) {
        if (id >= 0) { // If ID is needed, manually append child with (why? []operator fails with 'invalid map<K, T> key')
            /* Extract subpath and node to create a child node */
            std::string target_node(nodepath);
            std::string subpath(nodepath);
            i = nodepath.rfind("/");
            if (i != std::string::npos) {
                target_node.erase(0, i + 1);
                subpath.erase(i, nodepath.length());
            }
            node child(target_node);
            child.addAttrib("ID", num2Str(id));
            rtIO->acxml.at(subpath).appendChild(child);
            nodepath += "@" + num2Str(id); // alter nodepath to take id into account for description and tool level
        } else {
            rtIO->acxml[nodepath];
        }
        rtIO->acxml.at(nodepath).addAttrib("description", description);
        if (tool_level >= 0) {
            rtIO->acxml.at(nodepath).addAttrib("tool_level", num2Str(tool_level));
        }
    }
}

// Adapted methods for requirements mission
namespace requirements_mission {
void StandardMission::requirements_mission_update() {
    /* Outputs */
    myRuntimeInfo->out << std::endl;
    myRuntimeInfo->out << "******************* Requirement Mission Results ********************" << std::endl;
    myRuntimeInfo->out << "ICA: " <<  convertUnit(METER, FOOT, this->initial_cruise_altitude) << " ft." << std::endl;
    myRuntimeInfo->out << "ICA reached: " << ((this->initial_cruise_altitude_check) ? "Yes" : "No") << std::endl;
    myRuntimeInfo->out << "time_1500ft_to_ICA: " << convertUnit(SECOND, MINUTE, this->time_1500ft_to_ICA) << " min." << std::endl;
    myRuntimeInfo->out << "time_start_to_ICA: " << convertUnit(SECOND, MINUTE, this->time_start_to_ICA) << " min." << std::endl;
    myRuntimeInfo->out << "range_start_to_ICA: " << convertUnit(METER, NAUTICALMILE, this->range_start_to_ICA) << " NM" << std::endl;
    myRuntimeInfo->out << "Fuel to ICA: " << this->fuel_start_to_ICA << " kg" << std::endl;
    myRuntimeInfo->out << "max_operating_altitude: " << convertUnit(METER, FOOT, this->max_operating_altitude) << " ft." << std::endl;
    myRuntimeInfo->out << "max_ceiling_altitude: " << convertUnit(METER, FOOT, this->max_ceiling_altitude) << " ft." << std::endl;
    myRuntimeInfo->out << "max_altitude_one_engine_inoperative: " << convertUnit(METER, FOOT, this->max_altitude_one_engine_inoperative) << " ft." << std::endl;
    myRuntimeInfo->out << "********************************************************************" << std::endl;

    std::string subpath("analysis/mission/design_mission/cruise/cruise_steps");

    /* Reference top of descent from design mission */
    double design_mission_top_of_descent_altitude =
        rtIO->acxml.at(subpath +"/cruise_step@" + num2Str(rtIO->acxml.at(subpath).getVector("cruise_step").size() - 1) + "/altitude/value");
    /* Read in requirements for altitudes, time to climb and Mach number */
    subpath = "requirements_and_specifications/requirements/top_level_aircraft_requirements";
    double max_operating_altitude_required = rtIO->acxml.at(subpath + "/flight_envelope/maximum_operating_altitude/value");
    double max_altitude_one_engine_inoperative_required = rtIO->acxml.at(subpath + "/flight_envelope/maximum_altitude_one_engine_inoperative/value");
    double time_to_climb_requirement = rtIO->acxml.at(subpath + "/design_mission/time_to_climb/value");

    /* Define endnodes */
    Endnode<bool> maintainable_switch("maintainable", "Value shows if the requirement can be achieved by the aircraft.", false);
    Endnode<bool> checked_switch("checked", "Indicates if the value has been checked against the requirement.", true);
    Endnode<double> max_operating_altitude_node("maximum_operating_altitude", "Maximum operating altitude (100 ft/min rate of climb)",
                                                    this->max_operating_altitude, "m", 4500, 20000);
    Endnode<double> max_ceiling_altitude_node("maximum_ceiling_altitude", "Maximum operating altitude (50 ft/min rate of climb)", this->max_ceiling_altitude, "m", 4500, 20000);
    Endnode<double> max_altitude_one_engine_inoperative_node("maximum_altitude_one_engine_inoperative",
                                                                "Maximum operating altitude with one engine inoperative (100 ft/min rate of climb)",
                                                                this->max_altitude_one_engine_inoperative, "m", 1500, 12000);
    Endnode<double> time_1500ft_to_ICA_node("design_time_to_climb", "Time needed to climb from 1500ft to ICA (ISA + deltaISA)", this->time_1500ft_to_ICA, "s", 0, 7200);
    Endnode<double> time_start_to_1500ft_node("time_to_1500ft_ISA", "Time needed to climb from break release to 1500ft (ISA + deltaISA)", this->time_start_to_1500ft, "s", 0, 7200);
    Endnode<double> time_start_to_ICA_node("time_to_ICA_ISA", "Time needed to climb from break release to ICA (ISA + deltaISA)", this->time_start_to_ICA, "s", 0, 7200);
    Endnode<double> range_start_to_ICA_node("range_to_ICA_ISA", "Range needed to climb from break release to ICA (ISA + deltaISA)", this->range_start_to_ICA, "m", 0, 500000);
    Endnode<double> fuel_start_to_ICA_node("fuel_to_ICA_ISA", "Fuel mass needed to climb from break release to ICA (ISA + deltaISA)", this->fuel_start_to_ICA, "kg", 0, 10000);

    /* Check if compliance is already set */
    // Set up compliance nodes
    std::vector<std::string> node_names {
        "initial_cruise_altitude",
        "initial_cruise_mach_number",
        "maximum_operating_altitude",
        "maximum_altitude_one_engine_inoperative",
        "design_time_to_climb",
        "design_takeoff_distance",
        "design_landing_field_length",
        "design_approach_speed"
    };

    bool takeoff_check(EndnodeReadOnly<double>("requirements_and_specifications/requirements/top_level_aircraft_requirements/takeoff_distance").read(rtIO->acxml).value() >
                        EndnodeReadOnly<double>("assessment/takeoff/takeoff_distance_normal_safety").read(rtIO->acxml).value());
    bool landing_check(EndnodeReadOnly<double>("requirements_and_specifications/requirements/top_level_aircraft_requirements/landing_field_length").read(rtIO->acxml).value() >
                        EndnodeReadOnly<double>("assessment/landing/needed_runway_length").read(rtIO->acxml).value());
    bool approach_check(approach_speed_required + ACCURACY_LOW >= // ACCURACY_LOW added to avoid floating point issues for equal values
        EndnodeReadOnly<double>("assessment/performance/approach/approach_speed").read(rtIO->acxml).value());

    std::vector<bool> maintainable_vector {
        this->initial_cruise_altitude_check,
        (this->initial_cruise_mach + ACCURACY_LOW >= this->initial_cruise_mach_required), // Added ACCURACY_LOW to catch floating point errors when the values are actually equal
        (this->max_operating_altitude >= this->max_operating_altitude_required && max_operating_altitude_required >= design_mission_top_of_descent_altitude),
        (max_altitude_one_engine_inoperative_required <= this->max_altitude_one_engine_inoperative),
        (time_to_climb_requirement >= convertUnit(SECOND, MINUTE, this->time_1500ft_to_ICA) && this->initial_cruise_altitude_check),
        takeoff_check,
        landing_check,
        approach_check
    };

    /* Add requirement_compliance nodes if not existing */
    subpath = "requirement_compliance";
    this->check_and_add_nodes(subpath, "Compliance of given requirements");
    subpath += "/top_level_aircraft_requirements";
    this->check_and_add_nodes(subpath, "Compliance of top level aircraft requirements");

    /* Loop through the checked nodes */
    for (uint8_t i(0); i < maintainable_vector.size(); i++) {
        this->check_and_add_nodes(subpath + "/" + node_names[i], "Compliance check of the " + replaceAll(node_names[i], "_", " ") + ".");
        maintainable_switch = maintainable_vector[i];
        maintainable_switch.update(rtIO->acxml.at(subpath + "/" + node_names[i]));
        checked_switch.update(rtIO->acxml.at(subpath + "/" + node_names[i]));
    }

    /* Subpath: altitudes */
    subpath = "assessment/performance/altitude";
    this->check_and_add_nodes(subpath, "Assessment of altitude related values");
    max_operating_altitude_node.update(rtIO->acxml.at(subpath));
    max_ceiling_altitude_node.update(rtIO->acxml.at(subpath));
    max_altitude_one_engine_inoperative_node.update(rtIO->acxml.at(subpath));

    /* Subpath: climb */
    subpath = "assessment/performance/climb";
    this->check_and_add_nodes(subpath, "Assessment of the climb segment");
    time_1500ft_to_ICA_node.update(rtIO->acxml.at(subpath));
    time_start_to_1500ft_node.update(rtIO->acxml.at(subpath));
    time_start_to_ICA_node.update(rtIO->acxml.at(subpath));
    range_start_to_ICA_node.update(rtIO->acxml.at(subpath));
    fuel_start_to_ICA_node.update(rtIO->acxml.at(subpath));
}
} // namespace requirements_mission

namespace breguet {
void  StandardMission::breguet_update() {/* Create path to mission and build node if none exists yet */
    std::string subpath("analysis/mission");
    check_and_add_nodes(subpath, "Mission data", -1, 0);
    subpath += "/" + this->mission_type;
    check_and_add_nodes(subpath, "Data of " + replaceAll(this->mission_type, "_", " "));

    /* Non-global endnode initialization */
    EndnodeReadOnly<double> energy_carrier_tmp(EndnodeReadOnly<double>("consumed_energy"));
    EndnodeReadOnly<double> range_tmp( // Get endnode boundaries from TLARs
        EndnodeReadOnly<double>("requirements_and_specifications/requirements/top_level_aircraft_requirements/design_mission/range").read(rtIO->acxml));
    Endnode<double> total_range("range", "Traveled range from break release to end of taxi at destination", // take required range for 1st assumption
        EndnodeReadOnly<double>("requirements_and_specifications/requirements/top_level_aircraft_requirements/design_mission/range").read(rtIO->acxml).value(),
                                range_tmp.unit(), range_tmp.lower_boundary(), range_tmp.upper_boundary());
    Endnode<int> energy_carrier_id("energy_carrier_ID", "See energy carrier specification node", 0, "1", 0, 5); // TODO(GErrit): Auf Multifuel anpassen hier und drunter
    Endnode<double> consumed_energy("consumed_energy", "Energy amount", 0, "J", 0, 1.e13);

    /* Aircraft parameters */
    // Since takeoff_mass can be put into mission node for study mission and into masses_cg_inertia node for design mission,
    // only the analysis node is passed (see Endnode declaration)
    this->takeoff_mass.update(rtIO->acxml.at("analysis"));
    total_range.update(rtIO->acxml.at(subpath));

    /* Energy parameters */
    /* Mission energy */
    std::string description("Amount of energy loaded into tanks (including reserves) for the mission");
    this->check_and_add_nodes(subpath + "/loaded_mission_energy", description);
    description = "Amount of energy loaded into tanks (including reserves) for specified energy carrier";
    this->update_energy_nodes(this->loaded_fuel_mass, subpath + "/loaded_mission_energy/mission_energy", description, &(energy_carrier_id), &(consumed_energy));

    /* In-flight energy */
    description = "Amount of energy needed for in-flight segments (all segments from takeoff to landing)";
    this->check_and_add_nodes(subpath + "/in_flight_energy", description);
    description = "Amount of energy needed for trip segments (all segments from takeoff to landing) for specified energy carrier";
    this->update_energy_nodes(this->trip_fuel_mass, subpath + "/in_flight_energy/trip_energy", description, &(energy_carrier_id), &(consumed_energy));

    /* Taxiing */
    description = "Amount of energy needed for taxiing specified energy carrier";
    energy_carrier_id = taxi_fuel_mass_origin.first; // origin and destination have the same energy carrier
    this->check_and_add_nodes(subpath + "/taxi_energy", description);

    description = "Amount of energy needed for taxiing at origin for specified energy carrier";
    this->check_and_add_nodes(subpath + "/taxi_energy/taxi_out_energy", description, 0);
    consumed_energy = this->taxi_fuel_mass_origin.second *
        EnergyCarrier(rtIO->get_fuel_type(energy_carrier_id.value()), rtIO->get_fuel_density(energy_carrier_id.value())).gravimetric_energy_density;
    consumed_energy.update(rtIO->acxml.at(subpath + "/taxi_energy/taxi_out_energy@0"));
    energy_carrier_id.update(rtIO->acxml.at(subpath + "/taxi_energy/taxi_out_energy@0"));

    description = "Amount of energy needed for taxiing at destination for specified energy carrier";
    this->check_and_add_nodes(subpath + "/taxi_energy/taxi_in_energy", description, 0);
    consumed_energy =  this->taxi_fuel_mass_destination.second *
        EnergyCarrier(rtIO->get_fuel_type(energy_carrier_id.value()), rtIO->get_fuel_density(energy_carrier_id.value())).gravimetric_energy_density;
    consumed_energy.update(rtIO->acxml.at(subpath + "/taxi_energy/taxi_in_energy@0"));
    energy_carrier_id.update(rtIO->acxml.at(subpath + "/taxi_energy/taxi_in_energy@0"));
}
} // namespace breguet
