/*
 * UNICADO - UNIversity Conceptual Aircraft Design and Optimization
 * 
 * Copyright (C) 2025 UNICADO consortium
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 * Description:
 * This file is part of UNICADO.
 */

/* List of references
    [1] Airbus customer services, Flight Operations Support & Line Assistance: Getting to grips with the cost index, Issue II, May 1998, p.7
*/

#ifndef MISSION_ANALYSIS_SRC_STANDARD_MISSION_STANDARD_MISSION_STANDARD_MISSION_H_
#define MISSION_ANALYSIS_SRC_STANDARD_MISSION_STANDARD_MISSION_STANDARD_MISSION_H_

#include <map>
#include <memory>
#include <string>
#include <unordered_map>
#include <utility>
#include <vector>

#include <aerodynamics/aerodynamics.h>
#include <aixml/endnode.h>
#include <engine/engine.h>

#include "../../mission_analysis.h"
#include "../../libs/condition_methods/flight_conditions.h"
#include "../../libs/condition_methods/operating_conditions.h"
#include "../../libs/multi_engine_propulsion/multi_engine_propulsion.h"
#include "../../libs/mission_file_data_handler/mission_file_data_handler.h"


/** \class StandardMission
 *  \brief Standard method to calculate a mission
 */
class StandardMission : public Strategy {
 public:
    explicit StandardMission(const std::shared_ptr<RuntimeIO>& rtIO);
    ~StandardMission() = default;

    void initialize();
    void run();
    void update();
    void report();
    void save();

    /* Pointer to external objects */
    const std::shared_ptr<RuntimeIO>& rtIO;
    MissionFileDataHandler* mission_file;

    /* Constants */
    enum fuel_planning_method {JAR, FAR_DOMESTIC, FAR_FLAG};

    /* Variables */
    std::string mission_type; /**< Type of mission */
    bool low_fidelity;        /**< Switch to use low fidelity Breguet-based estimations for the trip fuel consumption */
    bool interpolate_aero;    /**< Switch to interpolate the polars for different center of gravity positions */

    /* Relative mass values */
    double rel_fuel_overshoot; /**< Relative overshoot when maximum fuel mass is reached*/
    double rel_MTOM_overshoot; /**< Relative overshoot when MTOM is reached*/

    /* Masses */
    bool estimate_MTOM;                         /**< Switch for MTOM estimation */
    double max_payload_mass;                    /**< Maximum payload mass MP [kg] */
    double OME;                                 /**< Operating empty mass OME [kg] */
    double MTOM;                                /**< Maximum takeoff mass MTOM [kg] */
    double max_fuel_mass;                       /**< Maximum fuel mass MFM [kg] */
    Endnode<double> takeoff_mass;               /**< Aircraft mass at takeoff TOM [kg] */
    std::map<int, double> trip_fuel_mass;             /**< Trip fuel mass [kg] */
    std::map<int, double> previous_trip_fuel_mass;    /**< Trip fuel mass from previous loop or tool [kg] */
    std::map<int, double> loaded_fuel_mass;           /**< Calculated fuel mass [kg] */
    std::map<int, double> previous_loaded_fuel_mass;  /**< Loaded fuel mass from previous loop or tool [kg] */
    std::map<int, double> takeoff_fuel_mass;          /**< Calculated takeoff fuel mass [kg] */
    std::map<int, double> landing_fuel_mass;          /**< Calculated landing fuel mass [kg] */
    double zero_fuel_mass;                            /**< Aircraft zero fuel mass (OME + PL) [kg] */
    std::pair<int, double> taxi_fuel_mass_origin;     /**< Taxi fuel mass for origin airport [kg] */
    std::pair<int, double> taxi_fuel_mass_destination;/**< Taxi fuel mass for destination airport [kg] */
    std::pair<int, double> contingency_fuel_mass;     /**< Contingency fuel mass [kg] */
    std::pair<int, double> alternate_fuel_mass;       /**< Alternate fuel mass [kg] */
    std::pair<int, double> final_reserve_fuel_mass;   /**< Final reserve fuel mass [kg] */
    std::pair<int, double> extra_fuel_mass;     /**< Extra fuel mass [kg] */
    std::pair<int, double> additional_fuel_mass;/**< Additional fuel mass [kg] */
    Endnode<double> top_of_climb_mass;          /**< Mass at top of climb [kg] */
    Endnode<double> top_of_descent_mass;        /**< Mass at top of descent [kg] */
    double top_of_descent_mass_next_FL;         /**< Mass at top of descent from next flight level above (current altitude + altitude_change_step) [kg] */
    double touchdown_mass;                      /**< Mass at touchdown [kg] */
    double touchdown_mass_next_FL;              /**< Mass at touchdown for descent from next flight level above (current altitude + altitude_change_step) [kg] */


    /* Fueling */
    fuel_planning_method fuel_planning;     /**< Fuel planning procedure from fuel_planning_method: 0: EASA-JAR; 1: FAA-FAR_DOMESTIC; 2: FAA-FAR_FLAG [-] */
    std::vector<std::map<std::string, double>> cg_shift; /**< Map containing the cg position in relation to total aircraft mass */

    /* Flight performance parameters */
    double max_operating_mach;                  /**< Maximum Mach Operating Speed MMO [-] */
    double max_operating_speed;                 /**< Maximum Operating (calibrated air-)speed VMO [m/s] */
    double initial_cruise_mach;                 /**< Optimum Mach Operating Speed [-] */
    double initial_cruise_mach_required;        /**< Initial cruise Mach number from XML Requirements Design Mission (TOC) [-] */
    double initial_cruise_altitude;             /**< Initial cruise altitude [m] */
    double initial_cruise_altitude_required;    /**< First Cruise-Altitude from XML Requirements DesignMission (TOC) [m] */
    double max_operating_altitude;              /**< Service ceiling (100 ft/min climbing criterion) = maximum operating altitude achieved by requirements mission [m] */
    double max_operating_altitude_required;     /**< Maximum Operating Altitude required from aircraft XML requirements block [m] */
    double max_ceiling_altitude;                /**< Absolute ceiling (50 ft/min climbing criterion) at MTOM [m] */
    double max_altitude_one_engine_inoperative; /**< Maximum altitude (100 ft/min climbing criterion) with one engine inoperative and at MTOM [m] */
    Endnode<double> approach_speed;             /**< Maximum approach speed needed for the mission [m] */
    double approach_speed_required;             /**< Maximum approach speed that shall not be exceeded [m/s] */
    double climb_speed_below_FL100;             /**< Climb calibrated airspeed below flight level 100 [m/s] */
    double climb_speed_above_FL100;             /**< Climb calibrated airspeed above flight level 100 [m/s] */
    double delta_mach_above_crossover;          /**< Delta to desired cruise Mach number above crossover [-] */
    double descent_speed_below_FL100;           /**< Descent calibrated airspeed below flight level 100 [m/s] */
    double descent_speed_above_FL100;           /**< Descent calibrated airspeed above flight level 100 [m/s] */
    Endnode<double> lift_off_speed;             /**< Speed at lift-off [m/s] */
    Endnode<double> lift_off_lift_coefficient;  /**< Lift coefficient at lift-off [-] */
    Endnode<double> lift_off_angle_of_attack;   /**< Angle of attack at lift-off [rad] */
    Endnode<double> takeoff_climb_angle;        /**< Climb angle towards screen height (35 ft) [rad] */
    Endnode<double> takeoff_stall_speed;        /**< Stall speed until screen height (35 ft) [m/s] */
    Endnode<double> maximum_landing_angle_off_attack; /**< Maximum angle of attack while landing (between 50 ft and touchdown) [rad] */


    /* Propulsion */
    MultiEnginePropulsion engine;          /**< Engine object */

    /* Aerodynamic - Polar calculation */
    atmosphere atm;             /**< Standard atmosphere */
    std::string polar_file;     /**< Polar file */
    node& polar_xml;            /**< Node object, polar xml */
    aerodynamics aero;          /**< Aerodynamics object */
    aerodynamics* aero_forward; /**< Aerodynamics object (forward CoG) */
    aerodynamics* aero_aft;     /**< Aerodynamics object (aft CoG) */
    double cg_forward;          /**< Center of grivity forward limit [] */
    double cg_design;           /**< Center of grivity design position [] */
    double cg_aft;              /**< Center of grivity aft limit [] */

    /* Landing Gear Properties */
    double friction_coefficient; /**< Friction coefficient between tyres and pavement */
    double braking_coefficient;  /**< Braking coefficient (decelaration due to braking) */

    /* Geometry */
    double MAC;                              /**< Mean aerodynamic chord [m] */
    double wing_surface_area;                /**< Wing surface [m^2] */
    size_t variable_camber_segment_quantity; /**< Number of variable camber segments */
    bool variable_camber_switch;             /**< variable camber switch */

    /* Requirement check */
    bool requirement_switch;            /**< Switch for turning the requirement check on */
    bool initial_cruise_altitude_check; /**< Switch for ICA was achieved */

    /* Mission data */
    /** \struct CruiseStep
     *  \brief Definition of cruise steps within the mission
     */
    struct CruiseStep {
        double rel_step_length; /**< Relative step length [-] */
        double FL;              /**< Flight level [100 ft] */
        double mass;            /**< Mass [kg] */
    };

    bool auto_climb_switch;                 /**< Switch for auto climb */
    bool optimum_ROC;                       /**< Switch for using optimum rate of climb in "climb to cruise from 1500 ft" segment */
    double delta_ISA;                       /**< Deviation from ISA [K] */
    double previous_block_time;               /**< Flight time + taxiing from previous loop or tool [s] */
    double previous_flight_time;              /**< Flight time read from previous loop or tool [s] */
    double average_cruise_bleed;            /**< Average cruise bleed for tank calculation [kg/s] */
    double average_shaft_bleed;             /**< Average shaft power for tank calculation [W] */
    Endnode<double> payload_mass;           /**< Payload [kg] */
    uint16_t top_of_climb_index;            /**< Mission data index at the top of climb [-] */
    uint16_t top_of_descent_index;          /**< Mission data index at the top of descent [-] */
    uint16_t end_of_descent_index;          /**< Mission data index at the end of descent [-] */
    double step_height_cruise;              /**< Step climb cruise [ft] */
    std::vector<CruiseStep> cruise_profile; /**< Cruise profile */
    unsigned int cruise_profile_start;      /**< Start of cruise TOC */
    unsigned int cruise_profile_end;        /**< End of cruise TOD */
    Endnode<double> top_of_climb_range;     /**< Range at top of climb [m] */
    Endnode<double> top_of_descent_range;   /**< Range at top of descent [m] */
    double descent_range;                   /**< Range of descent [m]*/
    double descent_range_next_FL;           /**< Range of descent from next flight level above (current altitude + altitude_change_step) [m]*/
    double top_of_descent_altitude;         /**< Altitude at the top of descent [m] */
    double time_start_to_1500ft;            /**< Time for takeoff procedure to 1500 ft [s] */
    double time_1500ft_to_ICA;              /**< Time to climb in seconds 1500 ft to ICA [s] */
    double time_start_to_ICA;               /**< Time from 0 ft to ICA [s] */
    double range_start_to_ICA;              /**< Range to ICA from start [m] */
    double fuel_start_to_ICA;               /**< Fuel to ICA from start [kg] */
    Endnode<double> takeoff_distance;       /**< Take-off distance needed (from BR until screen height of 35ft) [m] */
    Endnode<double> landing_distance;       /**< Take-off distance needed (from screen height of 50ft until runway exit at taxi-speed (not necessarily full stop) [m] */
    double CL_average;                      /**< Average lift coefficient [-] */
    Endnode<double> CL_cruise_average;      /**< Average cruise lift coefficient [-] */
    Endnode<double> CL_cruise_max;          /**< Maximum cruise lift coefficient [-] */
    Endnode<double> CL_cruise_min;          /**< Minimum cruise lift coefficient [-] */

    /** \class MissionSegment
     *  \brief A mission segment saves all relevant values of a point within the flown trajectory
     */
    class MissionSegment {
     public:
        std::map<int, double> end_total_fuel_mass;     /**< Cumulative consumption at the end of the segment [kg] */
        double end_fuel_flow;           /**< Fuel flow at the end of the segment [kg/s] */
        double end_time;                /**< Covered time at the end of the segment [s] */
        double end_range;               /**< Covered range at the end of the segment [m] */
        double end_altitude;            /**< Altitude at the end of the segment [m] */
        double end_total_mass;          /**< Aircraft mass at the end of the segment [kg] */
        double end_CL;                  /**< Lift coefficient at the end of the segment [-] */
        double end_L_over_D;            /**< Lift over drag coefficient at the end of the segment [-] */
        double end_specific_air_range;  /**< Specific air range at the end of the segment [m/kg] */
        std::string end_aero_config;    /**< Aerodynamic configuration at the end of the segment */
        double end_glide_path_angle;     /**< Path angle at the end of the segment [rad] */
        double end_TAS;                 /**< True airspeed at the end of the segment [m/s] */
        double end_mach;                /**< Mach number at the end of the segment [-] */
        double end_CAS;                 /**< Calibrated airspeed at the end of the segment [m/s] */
        double end_ground_speed;        /**< Ground speed at the end of the segment [m/s] */
        double end_thrust;              /**< Thrust at the end of the segment [N] */
        double end_bleed;               /**< Bleed at the end of the segment [kg/s] */
        double end_shaft_power;         /**< Shaft power at the end of the segment [W] */
        std::string end_engine_rating;  /**< Engine rating at the end of the segment */
        const std::map<uint16_t, double> end_n1; /**< rotational speed of the engines (N1) at the end of the segment [-] */
        double end_angle_of_attack;     /**< Current angle of attack AOA at the end of the segment [deg] */
        double end_spoiler_factor;      /**< Spoiler factor at the end of the segment [-] */
        std::string end_mode_name;      /**< Mode name at the end of the segment */
        double end_reynolds_number;     /**< Reynolds number at MAC at the end of the segment [-] */
        double end_aircraft_cg;         /**< Aircraft center of gravity at the end of the segment [m] */
        double end_stabilizer_incidence;              /**< incidence angle of stabilizer at the end of the segment [deg] */
        std::vector<std::vector<double>> end_origin_polar; /**< Origin polar with variable camber infos at the end of the segment */
        int energy_carrier_id;          /* Energy carrier used for the profile step */
        /* Constructor and destructor */
        /** \brief Constructor for the class MissionSegment
         * \param fc const FlightConditions&: Pointer to the flight conditions of the segment
         * \param oc const OperatingConditions& oc: Pointer to the operating conditions of the segment
         * \param previous_segment const MissionSegment&: pointer to the previous mission segment
         * \param an_end_n1: rotational speed of the engine (N1) at the end of the segment [-]
         */
        MissionSegment(const FlightConditions& fc, const OperatingConditions& oc, const MissionSegment& previous_segment, const std::map<uint16_t, double>& an_end_n1);
        /** \brief Constructor for the first segment (takeoff) using the class MissionSegment
         * \param fc const FlightConditions&: Pointer to the flight conditions of the segment
         * \param oc const OperatingConditions& oc: Pointer to the operating conditions of the segment
         * \param previous_segment const MissionSegment&: pointer to the previous mission segment
         * \param an_end_n1 const std::map<uint16_t, double>&: rotational speed of the engine (N1) at the end of the segment [-]
         * \param fuel_mass_map const std::map<int, double>&: map of all fuels used during the mission [kg]
         * \param takeoff_mass double: takeoff mass [kg]
         * \param cg_design double: center of gravity at takeoff [m]
         * \param cg_wing double: center of gravity of the wing to initialize the fuel's center of gravity at takeoff [m]
         */
        MissionSegment(const FlightConditions& fc, const OperatingConditions& oc, const std::map<uint16_t, double>& an_end_n1,
                        const std::map<int, double>& fuel_mass_map, double takeoff_mass);
        /** \brief Destructor for the class MissionSegment
         */
        virtual ~MissionSegment();
    };
    std::vector<MissionSegment> mission_profile; /**< Profile of the mission */

    /* Auxillary functions */
    /** \brief Function checks if the needed node is existing. If not, it's created
     * \param nodepath: path leading to the wanted node
     * \param description: description of the new node
     * \param id: if multiple same nodes are set, given ID for it
     * \param tool_level: if a tool level is needed, given tool level
     */
    void check_and_add_nodes(std::string nodepath, const std::string& description, const int& id = -1, const int& tool_level = -1);
    /** \brief Function returns the whole fuel mass within the map
     * \param fuel_map: A fuel map with int keys corresponding to the energy carrier IDs <[-], [kg]>
     * \return double value of the fuel mass [kg]
     */
    double get_fuel_sum(std::map<int, double> fuel_map);
    /** \brief Function checks if the needed node is existing. If not, it's created
     * \param subpath: Path to energy node
     * \param energy_nodes: Name of the final energy node itself
     * \return A fuel map with int keys corresponding to the energy carrier IDs <[-], [kg]>  
     */
    std::map<int, double> fuel_map_setup(const std::string& subpath, const std::string& energy_nodes);
    /** \brief Function checks if the needed node is existing. If not, it's created
     * \param fuel_map: A fuel map with int keys corresponding to the energy carrier IDs <[-], [kg]>
     * \param subpath: Path to energy node
     * \param description: description of the new node
     * \param energy_carrier_id: Energy carrier ID of the used fuel
     * \param consumed_energy: Energy which was consumed during the respective fligth segment
     */
    void update_energy_nodes(const std::map<int, double>& fuel_map, const std::string& subpath,
                                const std::string& description, Endnode<int>* energy_carrier_id, Endnode<double>* consumed_energy);
    /** \brief Function gets current center of gravity from given CG shifts due to defueling
     *  \param current_total_mass: Current aircraft mass for which a center of gravity shall be found [kg]
     *  \return x coordinate of the center of gravity [m]
     */
    double get_current_cg(const double& current_total_mass);
    /* Inline auxillary functions */
    /** \brief Function calculates the specific air range [m/kg] for a preset static flight condition.
     * \param mach: reference to Mach number [-]
     * \param altitude: reference to altitude [m]
     * \param aircraft_mass: reference to mass of aircraft [kg]
     * \param config: reference to configuration
     * \param rating: reference to rating
     * \param bleed: reference to bleed air offtakes [kg/s]
     * \param shaft_power: reference to shaft power [W]
     * \return specific air range [m/kg]
     */
    inline double static_flight_specific_air_range(const int& energy_carrier_id, const double &mach, const double &altitude, const double &aircraft_mass, const std::string &config,
                                            const std::string &rating, const double &bleed = 0., const double &shaft_power = 0.) {
        const double& required_thrust(this->aero.getCruiseDrag(mach, altitude, 0., aircraft_mass, config, this->atm));
        this->engine.calculate_N1_with_thrustlimit(energy_carrier_id, altitude, mach, atm, 1., rating, bleed, shaft_power, required_thrust);
        return this->aero.getCruiseLoverD(mach, altitude, 0., aircraft_mass, config, atm) * convertUnit(MACH, TRUEAIRSPEED, altitude, this->atm,
                mach) / this->engine.get_aircraft_fuelflow(energy_carrier_id) * this->engine.get_thrust_aircraft(energy_carrier_id) / aircraft_mass / G_FORCE;
    }

    /** \brief Find the minimum of a function between two adjacent values using a start value
     * \details Given a function f, and given a bracketing triplet of abscissas ax, bx, cx (such that bx is between ax and cx,and f(bx) is less than both f(ax) and f(cx)),
     *          this routine performs a golden section search for the minimum, isolating it to a fractional precision of a certain accuracy.
     *          The abscissa of the minimum is returned.
     * \param lower_value: Lower limit
     * \param upper_value: Upper limit
     * \param start_value: Start value
     */
    inline std::vector<double> get_golden_section_starting_points(const double& lower_value, const double& upper_value, const double& start_value) {
        const double golden_section_ratio(1. / ((1 + sqrt(5)) / 2.));// The golden ratio
        double x0, x1, x2, x3;
        x0 = lower_value; // At any given time we will keep track of four points, x0,x1,x2,x3.
        x3 = upper_value;
        if (fabs(upper_value - start_value) > fabs(start_value - lower_value)) { // Make x0 to x1 the smaller segment,
            x1 = start_value;
            x2 = start_value + (1. - golden_section_ratio) * (upper_value - start_value); // and fill in the new point to be tried.
        } else {
            x2 = start_value;
            x1 = start_value - (1. - golden_section_ratio) * (start_value - lower_value);
        }
        return {x0, x1, x2, x3};
    }

    /** \brief Updates the abscissa (x-axis) values for optimization using the Golden Section Ratio.
     * \details This function updates the abscissa values based on the given cost functions. It employs the Golden Section Ratio to determine the new values
     *          depending on the outcome of the cost comparison.
     * \param abszissa_values A pointer to a vector containing the abscissa values.
     * \param cost_functions A reference to a vector containing cost function values.
     *
     * \note This function assumes that the input vectors are properly sized and initialized.
     */
    inline void update_abszissa_values(std::vector<double>* abszissa_values, const std::vector<double>& cost_functions) {
        const double golden_section_ratio(1. / ((1 + sqrt(5)) / 2.));// The golden ratio
        if (cost_functions.at(1) < cost_functions.at(0)) {
            /* First possible outcome: costFunction minimum within upper boundaries */
            (*abszissa_values).at(0) = (*abszissa_values).at(1);
            (*abszissa_values).at(1) = (*abszissa_values).at(2);
            (*abszissa_values).at(2) = golden_section_ratio * (*abszissa_values).at(1) + (1. - golden_section_ratio) * (*abszissa_values).at(3);
        } else {
            /* Other possible outcome: costFunction minimum within lower boundaries */
            (*abszissa_values).at(3) = (*abszissa_values).at(2);
            (*abszissa_values).at(2) = (*abszissa_values).at(1);
            (*abszissa_values).at(1) = golden_section_ratio * (*abszissa_values).at(2) + (1. - golden_section_ratio) * (*abszissa_values).at(0);
        }
    }

 private:
    /* Methods */
    /* StandardMission.cpp */
    /** \brief Function analyzes the cruise profile to write relevant values into Aircraft Exchange file
     */
    void analyze_cruise_profile();
    /** \brief Function analyzes the mission profile
     *  \details the class members requirement_switch, top_of_climb_index, top_of_descent_index and end_of_descent_index are set
     *  \param last_segment_free_flight_switch bool*: Pointer to a switch if the last mission segment is a free_flight segment
     */
    void analyze_mission_profile(bool* last_segment_free_flight_switch);
    /** \brief Function for checking and reading performance requirements
     */
    void check_requirements();
    /** \brief Function throws error
     * \param error_code: error code [-]. Coded error message: 1 = negative flight route, 2 = Acceleration potential
     * \param mission_step: reference to a mission step [-]
     */
    void throw_mission_error(int error_code, const uint16_t &mission_step);

    /* flight_segments.cpp */
    /** \brief Function to calculate the length of the ground run part of the takeoff distance
     * \param oc: pointer to an operating condition object
     * \param fc: pointer to a flight condition object
     * \param a_mission_step: reference to a mission step [-]
     */
    void calculate_ground_run(OperatingConditions* oc, FlightConditions* fc, const uint16_t& a_mission_step);
    /** \brief Function calculates take off
     * \param oc: pointer to an operating condition object
     * \param fc: pointer to a flight condition object
     * \param a_mission_step: reference to a mission step [-]
     */
    void takeoff_calculation(OperatingConditions* oc, FlightConditions* fc, const uint16_t& a_mission_step);
    /** \brief Function climbs to cruise altitude
     * \param oc: pointer to an operating condition object
     * \param fc: pointer to a flight condition object
     * \param a_mission_step: reference to a mission step [-]
     */
    void climb_to_cruise_altitude(OperatingConditions* oc, FlightConditions* fc, const uint16_t& a_mission_step);
    /** \brief Function to climb to service ceiling
     * \param oc: pointer to an operating condition object
     * \param fc: pointer to a flight condition object
     * \param a_mission_step: reference to a mission step [-]
     */
    void climb_to_service_ceiling(OperatingConditions* oc, FlightConditions* fc, const uint16_t& a_mission_step);
    /** \brief Function for climb iteration
     * \param oc OperatingConditions*: pointer to an operating condition object
     * \param fc FlightConditions*: pointer to a flight condition object
     * \param mission_step uint16_t: constant reference to a mission step [-]
     * \param minimum_ROC constant reference to a minimum rate of climb value that should be maintained during climb iteration [m/s]
     * \param save_data constant reference for a switch if the calculated flight condition data should be saved to the mission_profile vector [-]
     */
    void climb_iteration(OperatingConditions* oc, FlightConditions* fc, const uint16_t &mission_step,
                        const double& minimum_ROC = convertUnit(FOOTPERMINUTE, METERPERSECOND, 100.), const bool& save_data = true);
    /** \brief Function estimates the aircraft mass at top of descent using Breguet Range equation with current aircraft mass and current SAR
     * \param current_aircraft_mass const double&: Mass of aircraft at begin of cruise range until start of descent [kg]
     * \param current_specific_air_range const double&: Current SAR at cruise starting point [m/kg]
     * \param range_until_descent const double&: cruise range until start of descent [m]
     * \return top_of_descent_aircraft_mass the mass of the aircraft at top of descent [kg]
     */
    double calculate_top_of_descent_aircraft_mass(const double& current_aircraft_mass, const double& current_specific_air_range, const double& range_until_descent);
    /** \brief calculate the climb/descent steps from a start to an end altitude
     * \details Function loops over height steps and executes a climb iteration for the given altitude as well as velocity (Mach or CAS, dependent on transition altitude);
     *          If autoselectFlightlevel is active, the optimum end height end_altitude is re-evaluated;
     * \param oc: pointer to an operating condition object
     * \param fc: pointer to a flight condition object
     * \param a_mission_step: reference to a mission step [-]
     * \param simulation_mode Constant reference to a switch if the function should run the estimation mode or the real mode; default is the real mode
     * \return range_for_descent if simulation_mode is active, the needed range to descend to approach is returned [m]
     */
    double calculate_climb_or_descent(OperatingConditions* oc, FlightConditions* fc,
                                    const uint16_t& a_mission_step, const bool& simulation_mode = false);
    /** \brief Function estimates the values for top of descent, bottom of descent and with this the range for descent
     *  \param last_segment_free_flight_switch Constant reference to a switch if the last mission segment is a free_flight segment
     */
    void estimate_descent(const bool& last_segment_free_flight_switch);
    /** \brief Function estimates the descent distance of the approach route
     * \param oc: reference to an operating condition object
     * \param fc: reference to a flight condition object
     * \param top_of_descent_mass: mass at top of descent [kg]
     * \return (Range, Fuel) pair<double, double>: Returns the descent distance of the approach route [m] and the consumed fuel [kg]
     */
    std::pair<double, double> calculate_descent(const OperatingConditions& oc, const FlightConditions& fc, double top_of_descent_mass);
    /** \brief Function updates the descent range current_descent_distance and with this also the end of current cruise segment as well as the aircraft mass at top of descent
     * \param segmend_end_range double*: pointer to the cruise end of current segment
     * \param top_of_descent_aircraft_mass double*: pointer to the current aircraft mass at top of descent
     * \param current_descent_distance const double&: constant reference to the current descent distance
     * \param oc const operatingCondition&: constant reference to an operating condition object
     * \param fc const FlightConditions&: constant reference to a flight condition object
     * \param a_mission_step: reference to a mission step [-]
     * \return (Range, Fuel) pair<double, double>: Returns the descent distance of the approach route [m] and the consumed fuel [kg]
     */
    std::pair<double, double> iterate_descent(double* segmend_end_range, double* top_of_descent_aircraft_mass, const double& current_descent_distance,
                                const OperatingConditions& oc, const FlightConditions& fc, const uint16_t& a_mission_step);
    /** \brief Function to calculate glide_slope interception
     * \param oc: pointer to an operating condition object
     * \param fc: pointer to a flight condition object
     * \param a_mission_step: reference to a mission step [-]
     * \param simulation_mode Constant reference to a switch if the function should run the estimation mode or the real mode; default is the real mode
     * \return range_for_descent Needed range to descend to perform a glide slope landing [m]
     */
    double glide_slope_interception(OperatingConditions* oc, FlightConditions* fc, const uint16_t& a_mission_step,
                                    const bool& simulation_mode = false);
    /** \brief Function for landing
     * \param oc: pointer to an operating condition object
     * \param fc: pointer to a flight condition object
     * \param a_mission_step: reference to a mission step [-]
     * \param simulation_mode Constant reference to a switch if the function should run the estimation mode or the real mode; default is the real mode
     * \return range_for_descentToTouchdown the needed range to descend from screen height (50ft) to touchdown is returned [m]
     */
    double landing(OperatingConditions* oc, FlightConditions* fc, const uint16_t& a_mission_step, const bool& simulation_mode = false);
    /** \brief Function changes speed at constant rate of climb
     * \param oc: pointer to an operating condition object
     * \param fc: pointer to a flight condition object
     * \param a_mission_step: reference to a mission step [-]
     * \param simulation_mode Constant reference to a switch if the function should run the estimation mode or the real mode; default is the real mode
     * \return range_for_descent if simulation_mode is active, the needed range to descend to approach is returned [m]
     */
    double change_speed_at_constant_ROC(OperatingConditions* oc, FlightConditions* fc,
                                        const uint16_t& a_mission_step, const bool& simulation_mode = false);
    /** \brief Function changes altitude with constant speed
     * \details if missionData.mach_switch is active, the Mach number is set constant during climb, else the CAS is taken (e.g. for CLIMBWITHCONSTCAS)
     * \param oc: pointer to an operating condition object
     * \param fc: pointer to a flight condition object
     * \param a_mission_step: reference to a mission step [-]
     * \param simulation_mode Constant reference to a switch if the function should run the estimation mode or the real mode; default is the real mode
     * \return descent_range if simulation_mode is active, the needed range to descend to approach is returned [m]
     */
    double change_altitude_at_constant_speed(OperatingConditions* oc, FlightConditions* fc,
                                                const uint16_t& a_mission_step, const bool& simulation_mode = false);
    /** \brief Function calculates cruise with constant Mach number
     * \param oc: pointer to an operating condition object
     * \param fc: pointer to a flight condition object
     * \param a_mission_step: reference to a mission step [-]
     */
    void cruise_at_constant_mach(OperatingConditions* oc, FlightConditions* fc, const uint16_t& a_mission_step);
    /** \brief Function for air to air fuelling
     * \param oc: pointer to an operating condition object
     * \param fc: pointer to a flight condition object
     * \param a_mission_step: reference to a mission step [-]
     */
    void air_to_air_fuelling(OperatingConditions* oc, FlightConditions* fc, const uint16_t& a_mission_step);
    /** \brief Function for free flight
     * \param oc: pointer to an operating condition object
     * \param fc: pointer to a flight condition object
     * \param a_mission_step: reference to a mission step [-]
     */
    void free_flight(OperatingConditions* oc, FlightConditions* fc, const uint16_t& a_mission_step);
    /** \brief Function calculates ECON cruise speed. The Mach number and the flight altitude h are changed to the maximum range values.
     * \param econ_mach_number: pointer to ECON Mach number [-]
     * \param altitude: reference to altitude [m]
     * \param CI: cost index [kg/min], scaled 0 to 999 according to Sperry/Honeywell [1]
     * \param aircraft_mass: reference to mass of the aircraft [kg]
     * \param config: reference to configuration
     * \param rating: reference to rating
     */
    void calculate_econ_cruise_speed(double* econ_mach_number, const double& altitude, const double CI,
                                        const double& aircraft_mass, const std::string& config, const std::string& rating);
    /** \brief Function simulates a flight level change to check if it is possible
     * \param fc: pointer to a flight conditions
     * \param oc: constant reference to object of operating conditions
     * \param altitude_change_step: altitude change within the step [m]
     * \param a_mission_step: reference to mission_step
     */
    void simulate_FL_change(FlightConditions* fc, const OperatingConditions& oc,
                                const double& altitude_change_step, const uint16_t& a_mission_step);
    /** \brief Function iterates fixed rate of climb acceleration
     * \param oc: pointer to an operating condition object
     * \param fc: pointer to a flight condition object
     * \param min_acceleration: minimum acceleration [m/s^2]
     * \param mission_step: reference to a mission step [-]
     * \param save_data constant reference for a switch if the calculated flight condition data should be saved to the mission_profile vector [-]
     */
    void fixed_ROC_acceleration_iteration(OperatingConditions* oc,  FlightConditions* fc, const double& min_acceleration,
                                        const uint16_t &mission_step, const bool& save_data);
    /** \brief Function checks the conditions after iteration
     * \details flight time, flight direction, minimum ROC, and flight track angle are checked for valid values; program is aborted if values are not valid
     * \param oc OperatingConditions*: pointer to an operating condition object
     * \param fc FlightConditions*: pointer to a flight condition object
     * \param mission_step uint16_t: constant reference to a mission step [-]
     * \param minimum_ROC constant reference to a minimum rate of climb value that should be maintained during climb iteration [ft/min]
     */
    void check_iteration_conditions(OperatingConditions* oc, FlightConditions* fc, const uint16_t &mission_step, const double& minimum_ROC);
    /** \brief Function for cruise iteration
     * \param oc: pointer to an operating condition object
     * \param fc: pointer to a flight condition object
     * \param mission_step: reference to a mission step [-]
     * \param save_data constant reference for a switch if the calculated flight condition data should be saved to the mission_profile vector [-]
     */
    void cruise_iteration(OperatingConditions* oc, FlightConditions* fc, const uint16_t &mission_step, bool save_data);
    /** \brief Function for free flight iteration
     * \param oc: pointer to an operating condition object
     * \param fc: pointer to a flight condition object
     * \param mission_step: reference to a mission step [-]
     * \param save_data constant reference for a switch if the calculated flight condition data should be saved to the mission_profile vector [-]
     */
    void free_flight_iteration(OperatingConditions* oc, FlightConditions* fc, const uint16_t &mission_step, bool save_data);
    /** \brief Function sets segment initial conditions
     *  \param fc: pointer to a flight condition object
     *  \param use_mission_condition constant reference to decide whether the last mission step results should be used for m, altitude, and TAS or not; default: true [-]
     *  \param an_end_CL constant reference to an end lift coefficient of the segment; if not preset, it is calculated from constant level flight [-]
     */
    void set_segment_initial_conditions(FlightConditions* fc, const bool& use_mission_condition = true, const double& an_end_CL = NAN);
    /** \brief Function sets initial segment iteration start values
     *  \details values for glide_path_angle, a, liftcoefficient, drag and thrust are set
     *  \param fc: pointer to a flight condition object
     *  \param an_end_CL constant reference to an end lift coefficient of the segment; if not preset, it is calculated from constant level flight [-]
     */
    void set_initial_iteration_start_values(FlightConditions* fc, const double& an_end_CL);
    /** \brief Functions sets initial segment value for CL and Drag
     *  \details if needed, drag is updated by using spoilers
     *  \param fc: pointer to a flight condition object
     *  \param an_end_CL constant reference to an end lift coefficient of the segment; if not preset, it is calculated from constant level flight [-]
     */
    void set_initial_aerodynamic_values(FlightConditions* fc, const double& an_end_CL);
    /** \brief Function sets iteration start values
     *  \details values for are set
     *  \param fc: pointer to a flight condition object
     *  \param end_track_angle constant reference to an end track angle of the segment [rad]
     *  \param end_acceleration constant reference to an end acceleration of the segment [m/s^2]
     *  \param an_end_CL constant reference to an end lift coefficient of the segment; if not preset, it is calculated from constant level flight [-]
     */
    void set_iteration_start_values(FlightConditions* fc, const double& end_track_angle, const double& end_acceleration, const double& an_end_CL);
    /** \brief Functions sets values for CLend and DragEnd
     *  \details if needed, DragEnd is updated by using spoilers
     *  \param fc: pointer to a flight condition object
     *  \param an_end_CL constant reference to an end lift coefficient of the segment; if not preset, it is calculated from constant level flight [-]
     */
    void set_aerodynamic_values(FlightConditions* fc, const double& an_end_CL);
    /** \brief Function sets a value for set_specific_air [m/kg]
     *  \details the current end conditions for Lift over Drag, true airspeed, TSFC, and aircraft mass are used to calculate specific air range as L/D * end_TAS / (TSFC * aircraft mass)
     *  \param fc constant reference to an end lift coefficient of the segment; if not preset, it is calculated from constant level flight [-]
     */
    void set_specific_air_range(FlightConditions* fc);
    /** \brief Function sets segment end conditions
     *  \param fc: pointer to a flight condition object
     *  \param end_TAS constant reference to an end velocity of the segment given as true airspeed [m/s]
     *  \param end_altitude constant reference to an end altitude of the segment [m]
     *  \param end_track_angle constant reference to an end track angle of the segment [rad]
     *  \param end_acceleration constant reference to an end acceleration of the segment [m/s^2]
     *  \param an_end_CL constant reference to an end lift coefficient of the segment; if not preset, it is calculated from constant level flight [-]
     *  \param use_breguet switch if Breguet equation is used for fuel mass calculation (true) or current engine fuel flow (false - default) [-]
     */
    void set_segment_end_conditions(FlightConditions* fc, const double& end_TAS, const double& end_altitude, const double& end_track_angle,
                                    const double& end_acceleration, const double& an_end_CL = NAN, const bool& use_breguet = false);
    /** \brief Function calculates the CoG at the end of each segment
     *  \param fc: pointer to a flight condition object
     */
    void set_segment_end_cg(FlightConditions *fc);
    /** \brief Function saves segment data
     * \param fc: reference to a flight condition
     * \param oc: reference to an operating condition
     * \param mission_step: mission step [-]
     */
    void save_segment_data(FlightConditions* fc, const OperatingConditions &oc, unsigned int mission_step);
    /** \brief Function interpolates the stabilizer angle i_stab between CoGs
     *  \param fc pointer to FlightConditions object
     */
    void interpolate_stabilizer_incidence(FlightConditions *fc);
    /** \brief Function decides whether the flight level is changed.
     * \param cruise_flight_condition: constant reference to object of cruise flight conditions
     * \param next_FL_flight_condition: constant reference to object of flight level change flight conditions
     * \param cruise_range: reference to the cruise range from reference mission point (end_range) to end of constant level cruise segment [m]
     * \return true or false [-]
     */
    bool decide_FL_change(const FlightConditions& cruise_flight_condition, const FlightConditions& next_FL_flight_condition, const double& cruise_range);
    /** \brief Function calculates the optimum flight level
     * \param optimum_altitude: pointer to altitude [m]
     * \param mach: reference to a Mach number [-]
     * \param CI: cost index [kg/min], scaled 0 to 999 according to Sperry/Honeywell [1]
     * \param aircraft_mass: reference to mass of the aircraft [kg]
     * \param config: reference to configuration
     * \param derate: reference to relative derate to the selected rating [-]
     * \param rating: reference to rating
     * \param rounding_switch: switch to turn on/off rounding
     */
    void calculate_optimum_FL(double* optimum_altitude, const double& mach, const double& CI, const double& aircraft_mass,
                                const std::string& config, const double& derate, const std::string& rating, const bool& rounding_switch);
    /** \brief Function calculates maximum rate of climb speed.
     * \param mach: pointer to a Mach number [-]
     * \param altitude: reference to altitude [m]
     * \param aircraft_mass: reference to mass of the aircraft [kg]
     * \param config: reference to configuration
     * \param derate: reference to relative derate to the selected rating [-]
     * \param rating: reference to rating
     */
    void calculate_max_ROC_speed(double* mach, const FlightConditions& fc, const double& derate);
    /** \brief Function calculates optimum speed and flight level.
     * \param econ_mach_number: pointer to ECON Mach number [-]
     * \param altitude: pointer to altitude [m]
     * \param CI: cost index [kg/min], scaled 0 to 999 according to Sperry/Honeywell [1]
     * \param aircraft_mass: reference to mass of the aircraft [kg]
     * \param config: reference to configuration
     * \param derate: reference to relative derate to the selected rating [-]
     * \param rating: reference to rating
     */
    void calcOptSpeedAndFL(double* econ_mach_number, double* altitude, const double CI, const double& aircraft_mass,
                            const std::string& config, const double& derate, const std::string& rating, const int& energy_carrier_id);

    /* fuel_handling.cpp */
    /** \brief Function to set the required fuel masses as start values and check the load conditions
     *  \details the member variables zero_fuel_mass, taxi_fuel_mass_origin, taxi_fuel_mass_destination, and loaded_fuel_mass are set
     */
    void tanking_and_load_control();
    /** \brief Function calculates the taxi fuels at origin and destination
     */
    void set_taxi_fuel();
    /** \brief Function calculates the extra fuel
     *  \details Captain Discretion Fuel - "at the captain's discretion" - according to the pilot's "good feeling" - e.g. weather forecast, experiences with delays.
     *  \param aircraft_mass double*: Pointer to the aircraft mass to update the final aircraft mass with current fuel component [kg]
     */
    void set_extra_fuel(double *aircraft_mass);
    /** \brief Function sets trip fuel
     *  \details if trip fuel is used from previous calculations or defined by the user; if the value is equal to zero, it is estimated with estimate_trip_fuel
     *  \param aircraft_mass double*: Pointer to the aircraft mass to update the final aircraft mass with current fuel component [kg]
     */
    void set_trip_fuel(double *aircraft_mass);
    /** \brief Function sets the contingency fuel component of the reserve fuel
     *  \param aircraft_mass double*: Pointer to the aircraft mass to update the final aircraft mass with current fuel component [kg]
     */
    void set_contingency_fuel(double *aircraft_mass);
    /** \brief Function sets the additional fuel component of the reserve fuel
     *  \param aircraft_mass double*: Pointer to the aircraft mass to update the final aircraft mass with current fuel component [kg]
     */
    void set_additional_fuel(double *aircraft_mass);
    /** \brief Function sets the final reserve fuel component of the reserve fuel
     *  \param aircraft_mass double*: Pointer to the aircraft mass to update the final aircraft mass with current fuel component [kg]
     */
    void set_final_reserve_fuel(double *aircraft_mass);
    /** \brief Function sets the alternate fuel component of the reserve fuel
     *  \param aircraft_mass double*: Pointer to the aircraft mass to update the final aircraft mass with current fuel component [kg]
     */
    void set_alternate_fuel(double *aircraft_mass);
    /** \brief Function calculates the components of the reserve fuel and updated them accordingly
     *  \details The methods distinguishes between the different fuel tanking procedures of EASA and FAA
     */
    void calculate_reserve_fuel();
    /** \brief Function calculates the fuel consumption during holding
     *  \details The fuel consumption has to be calculated with ISA-Conditions in 1500 ft holding altitude for the given holding time and aircraft mass and a econ holding speed
     *           (if holding Mach number is above 250KCAS, it is set to 250KCAS);
     *  \param aircraft_mass Constant reference to the aircraft mass [kg]
     *  \param holding_time Constant reference to a holding time [s]
     *  \return holdingFuel fuel consumed during holding [kg]
     */
    double calculate_holding_fuel(const double& aircraft_mass, const double& holding_time);
    /** \brief Function estimates the fuel consumption especially to estimate trip fuel as a start value for mission calculation
     *  \param aircraft_mass double&: Constant reference to the aircraft mass [kg]
     *  \return theTripFuelMass Sum of the consumed trip fuel of all flight segments [kg]
     */
    void estimate_trip_fuel(const double& aircraft_mass);
    /** \brief Function updates the relevant mass values of the aircraft after mission calculation
     *  \details updates trip_fuel_mass and loaded_fuel_mass to derive the takeoff_mass 
     */
    void update_mass_values();
    /** \brief Function updates the loaded fuel
     *  \details Summing up reserve, taxi and trip fuels
     */
    void update_loaded_fuel();
    /** \brief Function controls the loading condition of the aircraft if it fits the boundaries for maximum payload, maximum fuel, and maximum takeoff mass
     *  \details Program aborts if loading conditions do not fit
     */
    void load_control();
    /** \brief Function calculates the fuel flow at mid-cruise point
     *  \details the mid-cruise point is estimated at a mean flight level between that at begin and end of cruise and an aircraft mass where half of trip fuel is burned;
     *           the drag is calculated in this point to set the engine for a leveled flight; the fuelFlow is then returned
     *  \return this->acft->engine.get_aircraft_fuelflow() the current engine fuel flow summed for all engines [kg/s]
     */
    double get_mean_cruise_fuel_flow();

    /* Private inline auxiliary functions */
    /** \brief Function gets transition altitude
     * \param CAS: calibrated air speed [m/s]
     * \param mach: Mach number [-]
     * \param atm: reference to standard atmosphere
     * \return atm.getAltitudeAtPressureRatio(pressure_ratio) Altitude [m]
     */
    inline double get_transition_altitude(double CAS, double mach, const atmosphere &atm) {
        double ref_altitude(0.);
        double pressure_ratio = (pow(1 + .4 / 2. * pow(convertUnit(CALIBRATEDAIRSPEED, MACH, ref_altitude, atm, CAS), 2.), ISENTROPIC_EXPONENT_DRY_AIR / .4) - 1) /
                            (pow(1 + .4 / 2. * pow(mach, 2.), ISENTROPIC_EXPONENT_DRY_AIR / .4) - 1);
        return atm.getAltitudeAtPressureRatio(pressure_ratio);
    }
    /** \brief Function calculates the constFunction [kg/m] for a preset static flight condition and a preset cost_index.
     * \param mach: reference to Mach number [-]
     * \param altitude: reference to altitude [m]
     * \param aircraft_mass: reference to mass of aircraft [kg]
     * \param config: reference to configuration
     * \param rating: reference to rating
     * \param bleed: reference to bleed air offtakes [kg/s]
     * \param shaft_power: reference to shaft power [W]
     * \param cost_index: preset cost index in FMS [kg/min]
     * \return cost function [kg/m]
     */
    inline double calculate_cost_function(const int &energy_carrier_id, const double &mach, const double &altitude,
                                            const double &aircraft_mass, const std::string& config, const std::string &rating,
                                            const double &bleed, const double &shaft_power, const double& cost_index) {
    double the_cost_function;
    the_cost_function = 1. / this->static_flight_specific_air_range(energy_carrier_id, mach, altitude, aircraft_mass, config, rating) +
                      cost_index / convertUnit(MINUTE, SECOND, 1.) / convertUnit(MACH, TRUEAIRSPEED, altitude, this->atm, mach);
    return the_cost_function;
    }
    /** \brief Function to find number of elements having the same keywords in their names
     * \param xml xml file
     * \param path_to_node path to node in which you search
     * \param keywords the keyword to search
     * \param subpath additional subpaths for the node
     */
    inline int find_element_quantity(const node& xml, const std::string& path_to_node, const std::vector<std::string>& keywords, const std::string& subpath) {
        int element_quantity = 0;
        for (int i = 0; i < xml.getVector(path_to_node).size(); i++) {
            for (int j = 0; j < keywords.size(); j++) {
            if (std::string(xml.at(path_to_node + "@" + num2Str(i) + subpath)) ==  keywords[j]) {
                element_quantity++;
                continue;
                }
            }
        }
        return element_quantity;
    }
    /** \brief Function to calulate CoG interpolations
     * \param current_cg Current center of gravity
     * \param design_value design value for interpolation [N; if used with drag or lift]
     * \param partner_value partner value for interpolation [N; if used with drag or lift]
     */
    inline double interpolateViaCoG(const double current_cg, const double design_value, const double partner_value) {
        if (current_cg >= this->cg_design) {
            double leverarm = (current_cg - this->cg_design) / (this->cg_aft - this->cg_design);
            return (design_value + (partner_value - design_value) * leverarm);
        } else {
            double leverarm = (current_cg - this->cg_forward) / (this->cg_design - this->cg_forward);
            return partner_value + (design_value - partner_value) * leverarm;
        }
    }
    /** \brief Return the aerodynamics object depending on the current center of gravity 
     *  \details: If current CoG is behind the design CoG get aft aerodynamics; otherwise forward aerodynamics
     *  \param current_cg
     *  \return aerodynamics object
     */
    inline aerodynamics* getAlternativeAero(const double& current_cg) {
        if (current_cg >= this->cg_design) { //aft and design
            return (this->aero_aft);
        } else { //forward and design
            return (this->aero_forward);
        }
    }
};

namespace breguet {
/** \class breguet::StandardMission
 *  \brief Standard method to calculate a mission (first loop design mission namespace for robust assumptions in early design)
 */
class StandardMission : public ::StandardMission {
 public:
    // Constructor that calls the base class constructor
    explicit StandardMission(const std::shared_ptr<RuntimeIO>& rtIO)
        : ::StandardMission(rtIO) {}

    /* Overwrite method of StandardMission */
    void initialize() override {
        this->low_fidelity = true;
        ::StandardMission::initialize();
    }
    void run() override {};
    void update() override {
        this->breguet_update();
    };
    void report() override {};

    /* Own methods to replace within/add to StandardMission methods */
    /** \brief Function to only update mission values of the first loop design mission
     */
    void breguet_update();
};
} // namespace breguet

/* Namespace for design mission (not needed explicitly, but good for uniformity and doxygen) */
namespace design_mission {
/** \class design_mission::StandardMission
 *  \brief Standard method to calculate a mission (design mission namespace)
 */
class StandardMission : public ::StandardMission {
 public:
    // Constructor that calls the base class constructor
    explicit StandardMission(const std::shared_ptr<RuntimeIO>& rtIO)
        : ::StandardMission(rtIO) {}

    /* Overwrite method of StandardMission */

    /* Own methods to replace within/add to StandardMission methods */
};
} // namespace design_mission

/* Namespace for study mission (not needed explicitly, but good for uniformity and doxygen) */
namespace study_mission {
/** \class study_mission::StandardMission
 *  \brief Standard method to calculate a mission (study mission namespace)
 */
class StandardMission : public ::StandardMission {
 public:
    // Constructor that calls the base class constructor
    explicit StandardMission(const std::shared_ptr<RuntimeIO>& rtIO)
        : ::StandardMission(rtIO) {}

    /* Overwrite method of StandardMission */

    /* Own methods to replace within/add to StandardMission methods */
};
} // namespace study_mission

/* Same methods with minor changes for requirements mission */
namespace requirements_mission {
/** \class requirements_mission::StandardMission
 *  \brief Standard method to calculate a mission (requirements mission namespace)
 */
class StandardMission : public ::StandardMission {
 public:
    // Constructor that calls the base class constructor
    explicit StandardMission(const std::shared_ptr<RuntimeIO>& rtIO)
        : ::StandardMission(rtIO) {}

    /* Overwrite method of StandardMission */
    void initialize() override {
        this->requirement_switch = true;
        this->aero.switchOnReqSwitch();
        this->optimum_ROC = EndnodeReadOnly<bool>("program_settings/mode/rate_of_climb_switch").read(this->rtIO->moduleConfig).value(),
        this->mission_type = "requirements_mission";
        ::StandardMission::initialize();
    }
    void update() override {
        this->aero.switchOffReqSwitch();
        this->requirements_mission_update();
    }

    /* Own methods to replace within/add to StandardMission methods */
    /** \brief Function to only update requirements mission values
     */
    void requirements_mission_update();
};
} // namespace requirements_mission

#endif // MISSION_ANALYSIS_SRC_STANDARD_MISSION_STANDARD_MISSION_STANDARD_MISSION_H_
