/*
 * UNICADO - UNIversity Conceptual Aircraft Design and Optimization
 * 
 * Copyright (C) 2024 UNICADO consortium
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 * Description:
 * This file is part of UNICADO.
 */
#ifndef MISSION_ANALYSIS_SRC_LIBS_MULTI_ENGINE_PROPULSION_MULTI_ENGINE_PROPULSION_H_
#define MISSION_ANALYSIS_SRC_LIBS_MULTI_ENGINE_PROPULSION_MULTI_ENGINE_PROPULSION_H_

#include <map>
#include <memory>
#include <set>
#include <string>
#include <tuple>
#include <utility>
#include <vector>

#include <aixml/endnode.h>
#include <engine/engine.h>
#include <moduleBasics/module.h>
#include <energyCarriers/energyCarriers.h>

/** \brief Class for flight conditions
 */
class Multi_engine_propulsion {
 public:
    /* Constructor and destructor */
    /** \brief Constructor for class Multi_engine_propulsion
     */
    explicit Multi_engine_propulsion(const std::shared_ptr<RuntimeIO>& rtIO);
    /** \brief Destructor for class Multi_engine_propulsion
     */
    virtual ~Multi_engine_propulsion();

    size_t unique_engines_per_id(const int& energy_carrier_id);
    std::vector<std::string> get_engine_names(const int& energy_carrier_id);
    std::vector<Engine> get_engines(const int& energy_carrier_id);
    std::vector<double> get_thrust_shares(const int& energy_carrier_id);
    uint16_t get_number_of_engines(const int& energy_carrier_id);

    std::vector<OperatingPoint> get_operating_points(const int& energy_carrier_id);
    void set_operating_points(const int& energy_carrier_id, std::vector<OperatingPoint> operating_points);

    double get_aircraft_fuelflow(const int& energy_carrier_id);
    double get_aircraft_energyflow(const int& energy_carrier_id);
    double get_aircraft_energyflow(const int& energy_carrier_id, const double& aircraft_fuelflow);
    double get_thrust_aircraft(const int& energy_carrier_id);

    void calculate_N1_with_penalties(const int& energy_carrier_id,
                                     const double& altitude,
                                     const double& mach_number,
                                     const atmosphere& atm,
                                     const double& derate,
                                     const std::string& thrust_rating,
                                     double bleed_air_offtake,
                                     double shaft_power_offtake);
    void calculate_N1_with_thrustlimit(const int& energy_carrier_id,
                                       const double& altitude,
                                       const double& mach_number,
                                       const atmosphere& atm,
                                       const double& derate,
                                       const std::string& thrust_rating,
                                       double bleed_air_offtake,
                                       double shaft_power_offtake,
                                       double thrust_limit);

    class Propulsor {
     public:
        std::string name;
        Engine engine;
        std::vector<uint16_t> engine_id;
        double thrust_share;

        /** \brief Constructor of used_components.
         */
        Propulsor(const std::string& a_name, Engine an_engine, const uint16_t& an_id, const double& a_thrust_share);

        /** \brief Destructor of used_components
         */
        virtual ~Propulsor();
    };

    class Tank {
     public:
        uint16_t tank_id;
        double capacity;
        double relative_capacity;
        double consumed_energy;

        /** \brief Constructor of used_components.
         */
        Tank(uint16_t a_tank_id, double a_capacity);

        /** \brief Destructor of used_components
         */
        virtual ~Tank();
    };

    class Energy_consumption {
     public:
        const EnergyCarrier& energy_carrier;
        std::vector<Tank> tanks;

        /** \brief Constructor of used_components.
         */
        explicit Energy_consumption(const EnergyCarrier& an_energy_carrier);

        /** \brief Destructor of used_components
         */
        virtual ~Energy_consumption();
    };

    std::map<int, double> get_consumed_energy_map();
    double get_consumed_energy(const int& energy_carrier_id);
    std::map<int, double> get_energyflow_map();
    std::vector<Tank> get_tanks_by_id(const int& energy_carrier_id);
    void tank_setup(const node& tank_definition, const node& tank_design);
    void drain_tanks(const int& energy_carrier_id, const double& delta_t, const double& fuelflow = NAN);
    double get_consumed_fuel_mass();
    double get_energy_carrier_capacity(const int& energy_carrier_id);
    std::vector<Propulsor> get_propulsors(const int& energy_carrier_id);

 private:
    /* variables */
    std::map<int, std::vector<Propulsor>> propulsors;
    std::map<int, Energy_consumption> energy_map;

    /* functions */
    bool vector_check(const std::vector<int>& v1, const std::vector<int>& v2);
};

#endif // MISSION_ANALYSIS_SRC_LIBS_MULTI_ENGINE_PROPULSION_MULTI_ENGINE_PROPULSION_H_
