/*
 * UNICADO - UNIversity Conceptual Aircraft Design and Optimization
 * 
 * Copyright (C) 2025 UNICADO consortium
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 * Description:
 * This file is part of UNICADO.
 */
#include "payloadRange.h"

namespace low {

payloadRange::payloadRange(aircraft *theAircraft, const performance_assessment_settings &theSettings)
    :
    theAcftPt(theAircraft),
    theSettingsPt(&theSettings) {
}

payloadRange::~payloadRange() {
    // dtor
}

void payloadRange::calcPayloadRange() {
    // Determine Breguet Factors = (V*L/D)/g*SFC
    double breguetFactor(0.);
    double cruiseHeight(0.);
    // Average of cruise altitude for Breguet as sum of step altitudes weighted with segment length
    for (unsigned int i(0); i < theAcftPt->myMission.cruiseSegLength.size(); i++) {
        if (i == 0)
            cruiseHeight = theAcftPt->myMission.cruiseSegLength.at(i) * theAcftPt->myMission.cruiseAlt.at(i);
        else
            cruiseHeight += (theAcftPt->myMission.cruiseSegLength.at(i) - theAcftPt->myMission.cruiseSegLength.at(i - 1)) * theAcftPt->myMission.cruiseAlt.at(i);
    }
    double theLoverD = theAcftPt->aero.getCruiseLoverD(theAcftPt->myMission.crMach, theAcftPt->hInitialCruise, 0.0, theAcftPt->MTOM, "clean", theAcftPt->atm);
    double theDrag = theAcftPt->aero.getCruiseDrag(theAcftPt->myMission.crMach, theAcftPt->hInitialCruise, 0.0, theAcftPt->MTOM, "clean", theAcftPt->atm);
    theAcftPt->myEngines.calculate_N1_with_thrustlimit(theAcftPt->myEnergyProviders.energyCarrierID_cruise, theAcftPt->hInitialCruise, theAcftPt->myMission.crMach, theAcftPt->atm,
        1.0, "maximum_continuous", 0., 0., theDrag);
    double theSFC = theAcftPt->myEngines.get_aircraft_fuelflow(theAcftPt->myEnergyProviders.energyCarrierID_cruise)
        / theAcftPt->myEngines.get_thrust_aircraft(theAcftPt->myEnergyProviders.energyCarrierID_cruise);
    breguetFactor = (theAcftPt->myMission.crMach * theAcftPt->atm.getSpeedOfSound(cruiseHeight) * theLoverD) / (G_FORCE * theSFC);
    /* Determination of point 0 */
    Payload.push_back(theAcftPt->maxPayload);
    Fuel.push_back(0.);
    TOM.push_back(theAcftPt->OME + theAcftPt->maxPayload);
    Range.push_back(0.);
    /* Point 1 masses; range will be determined later */
    Payload.push_back(theAcftPt->maxPayload);
    Fuel.push_back(theAcftPt->MTOM - theAcftPt->OME - Payload.back());
    TOM.push_back(theAcftPt->MTOM);
    Range.push_back(0.);
    /* Determination of the design point according to Breguet and deviation from the given (real) design range */
    double designRangeBreguet = -breguetFactor * log((theAcftPt->OME + theAcftPt->designPayload) / (0.9065 * theAcftPt->MTOM));  // TODO(Hansman): Fuel consumed during takeoff
        // and climb  shall not be hardcoded
    double calibrateFactor = theAcftPt->designRange / designRangeBreguet;
    myRuntimeInfo->out << "calibrateFactor Breguet: " << calibrateFactor << std::endl;
    /* Determination of the 2nd point */
    Fuel.push_back(theAcftPt->MFM);
    Payload.push_back(theAcftPt->MTOM - theAcftPt->OME - Fuel.back());
    TOM.push_back(theAcftPt->MTOM);
    Range.push_back((-breguetFactor * log((theAcftPt->OME + Payload.back()) / (0.9065 * TOM.back())))*calibrateFactor);
    if (Range.at(2) > theAcftPt->designRange) {
        /* Determination of range of 1st point, from extrapolation between design point and 2nd point */
        Range.at(1) = theAcftPt->designRange
                      + (Range.at(2) - theAcftPt->designRange) / (Payload.at(2) - theAcftPt->designPayload)
                      * (Payload.at(1) - theAcftPt->designPayload);
    } else {
        myRuntimeInfo->out << "ERROR: Range_maxFuelatMTOM smaller DesignRange! " << std::endl;
        Range.at(1) = 0.;  // not extrapolatable
    }
    /* 3. point: no payload and maximum refuelled */
    if (theAcftPt->OME + theAcftPt->MFM < theAcftPt->MTOM) {
        Payload.push_back(0.);
        Fuel.push_back(theAcftPt->MFM);
        TOM.push_back(theAcftPt->OME + theAcftPt->MFM);
        double range = (-breguetFactor * log((theAcftPt->OME + Payload.back()) / (0.9065 * TOM.back())));
        if (range >= Range.back()) {
            Range.push_back(-breguetFactor * log((theAcftPt->OME + Payload.back()) / (0.9065 * TOM.back())));
        } else {
            Range.push_back((-breguetFactor * log((theAcftPt->OME + Payload.back()) / (0.9065 * TOM.back())))*calibrateFactor);
            myRuntimeInfo->warn << "Range_maxFuelEmpty smaller Range_maxFuelatMTOM. Range_maxFuelEmpty will also be calibrated!" << std::endl;
        }
    } else {
        // 3. point = 2. point
        Payload.push_back(Payload.at(2));
        Fuel.push_back(Fuel.at(2));
        TOM.push_back(TOM.at(2));
        Range.push_back(Range.at(2));
    }
}

}  // namespace low
