/*
 * UNICADO - UNIversity Conceptual Aircraft Design and Optimization
 * 
 * Copyright (C) 2025 UNICADO consortium
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 * Description:
 * This file is part of UNICADO.
 */
#include "performance_assessment_factory.h"
#include "performance_assessment_output.h"
#include <moduleBasics/html.h>


performance_assessment_output::performance_assessment_output(
        std::shared_ptr<abstractPayloadRange> abstractPayloadRangePt,
        std::shared_ptr<abstractEnginePerformance> abstractEnginePerformancePt,
        std::shared_ptr<abstractCeilingPerformance> abstractCeilingPerformancePt,
        std::shared_ptr<abstractFlightEnvelopes> abstractFlightEnvelopesPt,
        std::shared_ptr<abstractSARPerformance> abstractSARPerformancePt,
        std::shared_ptr<abstractTakeOffPerformance> abstractTakeOffPerformancePt,
        std::shared_ptr<abstractLandingPerformance> abstractLandingPerformancePt,
        std::shared_ptr<abstractVnDiagram> abstractVnDiagramPt,
        const std::shared_ptr<aircraft> anAcftPt)
    :
    myAcftPt(anAcftPt),
    mySettingsPt(myAcftPt->mySettingsPt),
    rtIO(myAcftPt->rtIO),
    abstractPayloadRangePt(abstractPayloadRangePt),
    abstractEnginePerformancePt(abstractEnginePerformancePt),
    abstractCeilingPerformancePt(abstractCeilingPerformancePt),
    abstractFlightEnvelopesPt(abstractFlightEnvelopesPt),
    abstractSARPerformancePt(abstractSARPerformancePt),
    abstractTakeOffPerformancePt(abstractTakeOffPerformancePt),
    abstractLandingPerformancePt(abstractLandingPerformancePt),
    abstractVnDiagramPt(abstractVnDiagramPt),
    theReport(myAcftPt->rtIO) {
    // ctor
    if (mySettingsPt->useStudyMissionForAnalysis) {
        theProgramName = rtIO->programname + "_study";
    } else {
        theProgramName = rtIO->programname;
    }
}

performance_assessment_output::~performance_assessment_output() {
    // dtor
}

void performance_assessment_output::generateHtmlBody() {
    // Creating the table for the display of the data from each payload-range diagram point.
    theReport.htmlReportStream() << "<div class=\"box data\">\n";
    theReport.htmlReportStream() << html::h2("Points of the payload-range diagram")
                << html::table_begin({"Point", "Payload", "Range"})
                << html::table_row("Point 1", std::to_string(static_cast<int>(abstractPayloadRangePt->Payload.at(0))) + " kg" ,
                std::to_string(static_cast<int>(abstractPayloadRangePt->Range.at(0) / 1000)) + " km")
                << html::table_row("Point 2", std::to_string(static_cast<int>(abstractPayloadRangePt->Payload.at(1))) + " kg" ,
                std::to_string(static_cast<int>(abstractPayloadRangePt->Range.at(1) / 1000)) + " km")
                << html::table_row("Point 3", std::to_string(static_cast<int>(abstractPayloadRangePt->Payload.at(2))) + " kg" ,
                std::to_string(static_cast<int>(abstractPayloadRangePt->Range.at(2) / 1000)) + " km")
                << html::table_row("Point 4", std::to_string(static_cast<int>(abstractPayloadRangePt->Payload.at(3))) + " kg" ,
                std::to_string(static_cast<int>(abstractPayloadRangePt->Range.at(3) / 1000)) + " km")
                << html::table_end;
    theReport.htmlReportStream() << "</div>\n";

    // Add the plots to the HTML report
    theReport.htmlReportStream() << "<div class=\"box data\">\n";
    theReport.htmlReportStream() << "<div class=\"plot-container\">\n";
    /* Payload Range Diagram */
    if (mySettingsPt->check_sucessfully_executed.doPayloadRange == true) {
        theReport.htmlReportStream() << "<img class=\"image-plot\" style=\"display: block; margin-bottom: 10px;\" src=\"" + relativePath(rtIO->getPlotDir(), rtIO->getReportDir())
        + "/" + std::string(TOOL_NAME) + "_payload_range.svg\">" << std::endl;
    }
    /* BFL */
    if (mySettingsPt->check_sucessfully_executed.doBLFLPerformance == true) {
        theReport.htmlReportStream() << "<img class=\"image-plot\" style=\"display: block; margin-bottom: 10px;\" src=\"" + relativePath(rtIO->getPlotDir(), rtIO->getReportDir())
        + "/" + std::string(TOOL_NAME) + "_balanced_field_length_performance.svg\">" << std::endl;
    }
    /* SAR */
    if (mySettingsPt->check_sucessfully_executed.doSARCalculation == true) {
        theReport.htmlReportStream() << "<img class=\"image-plot\" style=\"display: block; margin-bottom: 10px;\" src=\"" + relativePath(rtIO->getPlotDir(), rtIO->getReportDir())
        + "/" + std::string(TOOL_NAME) + "_best_SAR.svg\">" << std::endl;
    }
    /* Flight Envelope & SAR */
    if (mySettingsPt->check_sucessfully_executed.doFlightEnvelopePerformance == true) {
        theReport.htmlReportStream() << "<img class=\"image-plot\" style=\"display: block; margin-bottom: 10px;\" src=\"" + relativePath(rtIO->getPlotDir(), rtIO->getReportDir())
        + "/" + std::string(TOOL_NAME) + "_flight_envelope.svg\">" << std::endl;
    }
    /* Climb/Ceiling Performance */
    if (mySettingsPt->check_sucessfully_executed.doClimbPerformance == true) {
        theReport.htmlReportStream() << "<img class=\"image-plot\" style=\"display: block; margin-bottom: 10px;\" src=\"" + relativePath(rtIO->getPlotDir(), rtIO->getReportDir())
        + "/" + std::string(TOOL_NAME) + "_ceiling.svg\">" << std::endl;
    }
    /* Engine Performance */
    if (mySettingsPt->check_sucessfully_executed.doEnginePerformance == true) {
        for (unsigned int i(0); i < engine_performance_plot_names.size(); i++) {
            theReport.htmlReportStream() << "<img class=\"image-plot\" style=\"display: block; margin-bottom: 10px;\" src=\"" +
            relativePath(rtIO->getPlotDir(), rtIO->getReportDir()) + "/" + engine_performance_plot_names.at(i) + "\">" << std::endl;
        }
    }
    /* Landing Performance */
    if (mySettingsPt->check_sucessfully_executed.doLandingPerformance == true) {
        theReport.htmlReportStream() << "<img class=\"image-plot\" style=\"display: block; margin-bottom: 10px;\" src=\"" + relativePath(rtIO->getPlotDir(), rtIO->getReportDir())
        + "/" + std::string(TOOL_NAME) + "_landing_performance.svg\">" << std::endl;
    }
    /* Takeoff Performance */
    if (mySettingsPt->check_sucessfully_executed.doTOPerformance == true) {
        theReport.htmlReportStream() << "<img class=\"image-plot\" style=\"display: block; margin-bottom: 10px;\" src=\"" + relativePath(rtIO->getPlotDir(), rtIO->getReportDir())
        + "/" + std::string(TOOL_NAME) + "_takeoff_performance.svg\">" << std::endl;
    }
    /* Vn-Diagram */
    if (mySettingsPt->check_sucessfully_executed.doVnDiagram == true) {
        theReport.htmlReportStream() << "<img class=\"image-plot\" style=\"display: block; margin-bottom: 10px;\" src=\"" + relativePath(rtIO->getPlotDir(), rtIO->getReportDir())
        + "/" + std::string(TOOL_NAME) + "_Vn_diagram.svg\">" << std::endl;
    }
    /* SAR Detail Plots */
    if (mySettingsPt->check_sucessfully_executed.doFlightEnvelopePerformance) {
        for (unsigned int i(0); i < SAR_plot_names.size(); i++) {
            theReport.htmlReportStream() << "<img class=\"image-plot\" style=\"display: block; margin-bottom: 10px;\" src=\"" +
            relativePath(rtIO->getPlotDir(), rtIO->getReportDir()) + "/" + SAR_plot_names.at(i) + "\">" << std::endl;
        }
    }
    theReport.htmlReportStream() << "</div>" << std::endl;
    theReport.htmlReportStream() << "</div>\n";

    theReport.generateReports();
}

void performance_assessment_output::writeXmlOutput() {
    string subPath;
    /* MTOM */
    if (mySettingsPt->check_sucessfully_executed.doFlightEnvelopePerformance == true) {
        /* Speeds */
        subPath = "aircraft_exchange_file/assessment/performance/speed";
        Endnode<double> MMO(subPath + "/maximum_operating_mach_number", "Maximum operating mach number");
        MMO.set_value(myAcftPt->MMO);
        MMO.set_boundaries(0., 1.);
        MMO.set_unit("1");
        MMO.update(rtIO->acxml);
        Endnode<double> VMO(subPath + "/maximum_operating_velocity", "Maximum operating velocity");
        VMO.set_value(myAcftPt->VMO);
        VMO.set_boundaries(0., 400.);
        VMO.set_unit("m/s");
        VMO.update(rtIO->acxml);
        Endnode<double> MDive(subPath + "/dive_mach_number", "Maximum dive mach number");
        MDive.set_value(myAcftPt->MDive);
        MDive.set_boundaries(0., 1.);
        MDive.set_unit("1");
        MDive.update(rtIO->acxml);
        Endnode<double> VDive(subPath + "/dive_velocity", "Dive velocity");
        VDive.set_value(myAcftPt->VDive);
        VDive.set_boundaries(0., 400.);
        VDive.set_unit("m/s");
        VDive.update(rtIO->acxml);
        Endnode<double> v1gsClean(subPath + "/one_g_stall_speed_velocity", "One g stall speed");
        v1gsClean.set_value(myAcftPt->v1gS_Clean(myAcftPt->MTOM, 0.0));
        v1gsClean.set_boundaries(0., 400.);
        v1gsClean.set_unit("m/s");
        v1gsClean.update(rtIO->acxml);
    }
    /* Start */
    if (mySettingsPt->check_sucessfully_executed.doTOPerformance == true) {
        subPath = "aircraft_exchange_file/assessment/performance/takeoff";
        // Endnode<double> todns(subPath + "/takeoff_distance_normal_safety", "Takeoff distance normal safety");
        // todns.set_value(abstractTakeOffPerformancePt->myTOFLField.at(0).myTOFL.back().TODR_AEO);
        // todns.set_lower_boundary(0);
        // todns.set_upper_boundary(20000);
        // todns.set_unit("m");
        // todns.update(rtIO->acxml);
        Endnode<double> vlof(subPath + "/lift_off_speed_velocity", "Lift off velocity");
        vlof.set_value(myAcftPt->vLOF(myAcftPt->MTOM, 0.0));
        vlof.set_boundaries(0, 400);
        vlof.set_unit("m/s");
        vlof.update(rtIO->acxml);
        Endnode<double> v2(subPath + "/takeoff_safety_speed", "Take off safety speed (V2)");
        v2.set_value(myAcftPt->v2(myAcftPt->MTOM, 0.0));
        v2.set_boundaries(0, 200);
        v2.set_unit("m/s");
        v2.update(rtIO->acxml);
        Endnode<double> vfto(subPath + "/final_takeoff_speed", "Final take off speed");
        vfto.set_value(myAcftPt->vFTO(myAcftPt->MTOM, 0.0));
        vfto.set_boundaries(0, 200);
        vfto.set_unit("m/s");
        vfto.update(rtIO->acxml);
        Endnode<double> ttsh(subPath + "/time_to_screen_height", "Time to screen height");
        ttsh.set_value(abstractTakeOffPerformancePt->myTOFLField.at(0).myTOFL.back().Time);
        ttsh.set_boundaries(0, 3600);
        ttsh.set_unit("s");
        ttsh.update(rtIO->acxml);
        Endnode<double> climbDescendecondSegment(subPath + "/climb_or_descend_segment_climb_gradient", "Climb or descend in second segment");
        climbDescendecondSegment.set_value(abstractTakeOffPerformancePt->scdSegmentGradient);
        climbDescendecondSegment.set_boundaries(0, 100);
        climbDescendecondSegment.set_unit("m/s");
        climbDescendecondSegment.update(rtIO->acxml);
        Endnode<double> finalSegmentGradient(subPath + "/final_segment_climb_gradient", "Climb gradient in final segment");
        finalSegmentGradient.set_value(abstractTakeOffPerformancePt->finalSegmentGradient);
        finalSegmentGradient.set_boundaries(0, 100);
        finalSegmentGradient.set_unit("m/s");
        finalSegmentGradient.update(rtIO->acxml);
    }
    /* BLFL */
    if (mySettingsPt->check_sucessfully_executed.doBLFLPerformance == true) {
        subPath = "aircraft_exchange_file/assessment/performance/takeoff";
        Endnode<double> BLFL(subPath + "/balanced_field_length", "Balanced field length");
        BLFL.set_value(abstractTakeOffPerformancePt->myBLFL.BLFL);
        BLFL.set_boundaries(0, 100000);
        BLFL.set_unit("m");
        BLFL.update(rtIO->acxml);
        Endnode<double> decision_speed(subPath + "/decision_speed", "Take off decision speed (V1)");
        decision_speed.set_value(abstractTakeOffPerformancePt->myBLFL.v1);
        decision_speed.set_lower_boundary(0);
        decision_speed.set_upper_boundary(200);
        decision_speed.set_unit("m/s");
        decision_speed.update(rtIO->acxml);
    }
    /* Landing */
    if (mySettingsPt->check_sucessfully_executed.doLandingPerformance == true) {
        subPath = "aircraft_exchange_file/assessment/performance/landing";
        Endnode<double> needed_runway_length(subPath + "/needed_runway_length", "Needed runway length");
        needed_runway_length.set_value(abstractLandingPerformancePt->myLDNField.at(0).myLDN.back().ALD_AEO);
        needed_runway_length.set_boundaries(0, 100000);
        needed_runway_length.set_unit("m");
        needed_runway_length.update(rtIO->acxml);
        Endnode<double> vappr(subPath + "/approach_speed", "Approach speed");
        vappr.set_value(myAcftPt->vAPPR(myAcftPt->MLM, 0.0));
        vappr.set_boundaries(0, 1000);
        vappr.set_unit("m/s");
        vappr.update(rtIO->acxml);
    }
    if (mySettingsPt->check_sucessfully_executed.doPayloadRange == true) {
        /* Ranges */
        subPath = "aircraft_exchange_file/assessment/performance/range";
        Endnode<double> range_max_payload_at_maximum_takeoff_mass(subPath + "/range_max_payload_at_maximum_takeoff_mass", "Range at max payload at maximum take off mass");
        Endnode<double> range_max_fuel_at_maximum_takeoff_mass(subPath + "/range_max_fuel_at_maximum_takeoff_mass", "Range at max fuel at maximum take off mass");
        Endnode<double> payload_maximum_fuel_at_maximum_takeoff_mass(subPath + "/payload_maximum_fuel_at_maximum_takeoff_mass", "Payload at maximum fuel at maximum takeoff mass");
        Endnode<double> range_maximum_fuel_empty(subPath + "/range_maximum_fuel_empty", "Range at maximum fuel empty");
        if (abstractPayloadRangePt->Range.at(2) > myAcftPt->designRange) {
            range_max_payload_at_maximum_takeoff_mass.set_value(abstractPayloadRangePt->Range.at(1));
            range_max_fuel_at_maximum_takeoff_mass.set_value(abstractPayloadRangePt->Range.at(2));
            payload_maximum_fuel_at_maximum_takeoff_mass.set_value(abstractPayloadRangePt->Payload.at(2));
            range_maximum_fuel_empty.set_value(abstractPayloadRangePt->Range.at(3));
        }
        range_max_payload_at_maximum_takeoff_mass.set_boundaries(0, INFINITY);
        range_max_payload_at_maximum_takeoff_mass.set_unit("m");
        range_max_payload_at_maximum_takeoff_mass.update(rtIO->acxml);
        range_max_fuel_at_maximum_takeoff_mass.set_boundaries(0, INFINITY);
        range_max_fuel_at_maximum_takeoff_mass.set_unit("m");
        range_max_fuel_at_maximum_takeoff_mass.update(rtIO->acxml);
        payload_maximum_fuel_at_maximum_takeoff_mass.set_boundaries(0, INFINITY);
        payload_maximum_fuel_at_maximum_takeoff_mass.set_unit("kg");
        payload_maximum_fuel_at_maximum_takeoff_mass.update(rtIO->acxml);
        range_maximum_fuel_empty.set_boundaries(0, INFINITY);
        range_maximum_fuel_empty.set_unit("m");
        range_maximum_fuel_empty.update(rtIO->acxml);
    }
    if (mySettingsPt->check_sucessfully_executed.doFuelPlanning == true) {
        /* Mission Fuels */
        if (!mySettingsPt->designForMTOM && mySettingsPt->useStudyMissionForAnalysis) {
            subPath = "aircraft_exchange_file/analysis/mission/study_mission";
        } else {
            subPath = "aircraft_exchange_file/analysis/mission/design_mission";
        }
        Endnode<double> missionRange(subPath + "/range", "Traveled range from break release to end of taxi at destination");
        missionRange.set_value(myAcftPt->myMission.missionRange);
        missionRange.update(rtIO->acxml);
        // Rest

        Endnode<double> payload_mass(subPath + "/payload_mass", "Mission payload mass");
        payload_mass.set_value(myAcftPt->myMission.missionPayload);
        payload_mass.set_boundaries(0, INFINITY);
        payload_mass.set_unit("kg");
        payload_mass.update(rtIO->acxml);
        Endnode<int> missionPax(subPath + "/number_of_pax", "Number of passengers considered for the mission");
        missionPax.set_value(myAcftPt->myMission.missionPAX);
        missionPax.set_boundaries(0, std::numeric_limits<int>::max());
        missionPax.set_unit("1");
        missionPax.update(rtIO->acxml);
        Endnode<double> cargo_mass(subPath + "/cargo_mass", "Cargo mass considered for the mission");
        cargo_mass.set_value(myAcftPt->myMission.missionCargo);
        cargo_mass.set_boundaries(0, INFINITY);
        cargo_mass.set_unit("kg");
        cargo_mass.update(rtIO->acxml);
        Endnode<double> takeoff_engine_derate(subPath + "/takeoff_engine_derate", "Engine power demand");
        takeoff_engine_derate.set_value(myAcftPt->myMission.Steps.at(0).derating);
        takeoff_engine_derate.set_unit("1");
        takeoff_engine_derate.update(rtIO->acxml);
    }
    /* Requirements Checks */
    // TLARS
    subPath = "aircraft_exchange_file/requirement_compliance/top_level_aircraft_requirements";
    Endnode<bool> design_takeoff_distance_checked(subPath + "/design_takeoff_distance/checked", "Indicates if the value has been checked against the requirement.");
    design_takeoff_distance_checked.set_value(false);
    if (mySettingsPt->check_sucessfully_executed.doTOPerformance == true) {
        Endnode<bool> design_takeoff_distance_maintainable(subPath + "/design_takeoff_distance/maintainable",
            "Switch indicating if design takeoff distance can be maintained.");
        design_takeoff_distance_maintainable.set_value(myAcftPt->myChecks.TOFL);
        design_takeoff_distance_maintainable.update(rtIO->acxml);
        design_takeoff_distance_checked.set_value(true);
    }
    design_takeoff_distance_checked.update(rtIO->acxml);

    Endnode<bool> design_landing_field_length_checked(subPath + "/design_landing_field_length/checked", "Indicates if the value has been checked against the requirement.");
    design_landing_field_length_checked.set_value(false);
    Endnode<bool> design_approach_speed_maintainable_checked(subPath + "/design_approach_speed/checked", "Indicates if the value has been checked against the requirement.");
    design_approach_speed_maintainable_checked.set_value(false);
    if (mySettingsPt->check_sucessfully_executed.doLandingPerformance == true) {
        Endnode<bool> design_landing_field_length_maintainable(subPath + "/design_landing_field_length/maintainable",
            "Switch indicating if landing field length can be maintained.");
        design_landing_field_length_maintainable.set_value(myAcftPt->myChecks.LDN);
        design_landing_field_length_maintainable.update(rtIO->acxml);
        design_landing_field_length_checked.set_value(true);
        Endnode<bool> design_approach_speed_maintainable(subPath + "/design_approach_speed/maintainable", "Switch indicating if approach speed can be maintained.");
        design_approach_speed_maintainable.set_value(myAcftPt->myChecks.vApproach);
        design_approach_speed_maintainable.update(rtIO->acxml);
        design_approach_speed_maintainable_checked.set_value(true);
    }
    design_landing_field_length_checked.update(rtIO->acxml);
    design_approach_speed_maintainable_checked.update(rtIO->acxml);

    // Span limit
    Endnode<bool> span_limit(subPath + "/span_limit/maintainable", "Switch indicating if the span limit according to ICAO aerodrome ref code is maintained.");
    span_limit.set_value(myAcftPt->myChecks.span_limit);
    span_limit.update(rtIO->acxml);
    Endnode<bool> span_limit_checked(subPath + "/span_limit/checked", "Indicates if the value has been checked against the requirement.");
    span_limit_checked.set_value(true);
    span_limit_checked.update(rtIO->acxml);

    // Certification
    subPath = "aircraft_exchange_file/requirement_compliance/certification";
    Endnode<bool> climb_gradient_of_second_takeoff_segment_checked(subPath + "/climb_gradient_of_second_takeoff_segment/checked",
        "Indicates if the value has been checked against the requirement.");
    climb_gradient_of_second_takeoff_segment_checked.set_value(false);
    Endnode<bool> climb_gradient_of_final_takeoff_segment_checked(subPath + "/climb_gradient_of_final_takeoff_segment/checked",
        "Indicates if the value has been checked against the requirement.");
    climb_gradient_of_final_takeoff_segment_checked.set_value(false);

    if (mySettingsPt->check_sucessfully_executed.doTOPerformance == true) {
        Endnode<bool> climb_gradient_of_second_takeoff_segment(subPath + "/climb_gradient_of_second_takeoff_segment/maintainable",
            "Switch indicating if climb gradient of second takeoff segment can be maintained.");
        climb_gradient_of_second_takeoff_segment.set_value(myAcftPt->myChecks.climbGradientSecondSegment);
        climb_gradient_of_second_takeoff_segment.update(rtIO->acxml);
        climb_gradient_of_second_takeoff_segment_checked.set_value(true);
        Endnode<bool> climb_gradient_of_final_takeoff_segment(subPath + "/climb_gradient_of_final_takeoff_segment/maintainable",
            "Switch indicating if climb gradient of final takeoff segment can be maintained.");
        climb_gradient_of_final_takeoff_segment.set_value(myAcftPt->myChecks.climbGradientFinalSegment);
        climb_gradient_of_final_takeoff_segment.update(rtIO->acxml);
        climb_gradient_of_final_takeoff_segment_checked.set_value(true);
    }
    climb_gradient_of_second_takeoff_segment_checked.update(rtIO->acxml);
    climb_gradient_of_final_takeoff_segment_checked.update(rtIO->acxml);

    Endnode<bool> climb_gradient_approach_one_engine_inoperative_checked(subPath + "/climb_gradient_approach_one_engine_inoperative/checked",
        "Indicates if the value has been checked against the requirement.");
    climb_gradient_approach_one_engine_inoperative_checked.set_value(false);
    Endnode<bool> climb_gradient_all_engines_operative_checked(subPath + "/climb_gradient_all_engines_operative/checked",
        "Indicates if the value has been checked against the requirement.");
    climb_gradient_all_engines_operative_checked.set_value(false);
    if (mySettingsPt->check_sucessfully_executed.doLandingPerformance) {
        Endnode<bool> climb_gradient_approach_one_engine_inoperative(subPath + "/climb_gradient_approach_one_engine_inoperative/maintainable",
            "Switch indicating if climb gradient approach one engine inoperative can be maintained.");
        climb_gradient_approach_one_engine_inoperative.set_value(myAcftPt->myChecks.climbGradientApproachOEI);
        climb_gradient_approach_one_engine_inoperative.update(rtIO->acxml);
        climb_gradient_approach_one_engine_inoperative_checked.set_value(true);
        Endnode<bool> climb_gradient_all_engines_operative(subPath + "/climb_gradient_all_engines_operative/maintainable",
            "Switch indicating if climb gradient with all engines operative can be maintained.");
        climb_gradient_all_engines_operative.set_value(myAcftPt->myChecks.climbGradientApproachAEO);
        climb_gradient_all_engines_operative.update(rtIO->acxml);
        climb_gradient_all_engines_operative_checked.set_value(true);
    }
    climb_gradient_approach_one_engine_inoperative_checked.update(rtIO->acxml);
    climb_gradient_all_engines_operative_checked.update(rtIO->acxml);
}


void performance_assessment_output::generatePlotData() {
    if (mySettingsPt->check_sucessfully_executed.doPayloadRange == true) {
        generatePlotDataPayloadRange();
    }
    if (mySettingsPt->check_sucessfully_executed.doEnginePerformance == true) {
        generate_engine_performance_data();
    }
    if (mySettingsPt->check_sucessfully_executed.doClimbPerformance == true) {
        generateCeilingPlotData();
    }
    if (mySettingsPt->check_sucessfully_executed.doFlightEnvelopePerformance == true) {
        generateFlightEnvelopeData();
    }
    if (mySettingsPt->check_sucessfully_executed.doSARCalculation == true) {
        generateSARPlotData();
        generate_best_SAR_Data();
    }
    if (mySettingsPt->check_sucessfully_executed.doTOPerformance == true) {
        generateTOFLPlotData();
    }
    if (mySettingsPt->check_sucessfully_executed.doBLFLPerformance == true) {
        generateBLFLPlotData();
    }
    if (mySettingsPt->check_sucessfully_executed.doLandingPerformance == true) {
        generateLDNPlotData();
    }
    if (mySettingsPt->check_sucessfully_executed.doVnDiagram == true) {
        generateVnPlotData();
    }
}

void performance_assessment_output::generatePlotDataPayloadRange() {
std::ofstream csv;
    string csvDataFile(mySettingsPt->rtIO->getCsvFilesDirTool() + "/" + theProgramName + "_payloadRange.csv");
    csv.open(csvDataFile.c_str());
    if (!csv) {
        myRuntimeInfo->err << string(theProgramName) + "_payloadRange.csv could not be opened!" << std::endl;
        exit(1);
    } else {
        csv << "# " + mySettingsPt->rtIO->getCsvFilesDirTool() + theProgramName + "_missionFuel.csv" << std::endl;
        // Adjust data structure:
        csv << "# (1) Range [m]; (2) Payload [kg]; (3) Fuel [kg]; (4) TOM [kg]; (5) OME [kg]" << std::endl;
        for (unsigned int i(0); i <= abstractPayloadRangePt->Range.size() - 1; i++) {
            csv << abstractPayloadRangePt->Range.at(i) << "; ";
            csv << abstractPayloadRangePt->Payload.at(i) << "; ";
            csv << abstractPayloadRangePt->Fuel.at(i) << "; ";
            csv << abstractPayloadRangePt->TOM.at(i) << "; ";
            // csv << myPerfPt->abstractFuelPlanning.myPayloadRange.at(i).Fuel << "; ";
            // csv << myPerfPt->abstractFuelPlanning.myPayloadRange.at(i).TOM << "; ";
            // csv << myAcftPt->OME << "; ";
            csv << std::endl;
        }
        csv << std::endl;
        csv << std::endl;
        csv << "# Design Point " << std::endl;
        csv << myAcftPt->designRange << "; ";
        csv << myAcftPt->designPayload << "; ";
        // csv << "N/A" << "; ";
        // csv << myAcftPt->MTOM << "; ";
        // csv << myAcftPt->OME << "; ";
        csv << std::endl;
        csv << std::endl;
        csv << std::endl;
        csv << "# Mission Point" << std::endl;
        csv << myAcftPt->myMission.missionRange << "; ";
        csv << myAcftPt->myMission.missionPayload << "; ";
        // csv << myPerfPt->abstractFuelPlanning.myFuel.tripFuel << "; ";
        // csv << myAcftPt->MTOM << "; ";
        // csv << myAcftPt->OME << "; ";
        csv.close();
    }
}

void performance_assessment_output::generate_engine_performance_data() {
    for (unsigned int i(0); i <= abstractEnginePerformancePt->theRatingEngine.size() - 1; i++) {
        std::ofstream csv;
        string csvDataFile(mySettingsPt->rtIO->getCsvFilesDirTool() + "/" + theProgramName + "_EnginePerformance_" +
            abstractEnginePerformancePt->theRatingEngine.at(i).Rating + ".csv");
        csv.open(csvDataFile.c_str());
        if (!csv) {
            myRuntimeInfo->err << string(theProgramName) + "_EnginePerformance" + abstractEnginePerformancePt->theRatingEngine.at(i).Rating
                + ".csv could not be opened!" << std::endl;
            exit(1);
        } else {
            csv << "# " + mySettingsPt->rtIO->getCsvFilesDirTool() + theProgramName + "_EnginePerformance" +
                abstractEnginePerformancePt->theRatingEngine.at(i).Rating + ".csv" << std::endl;
            csv << "# Rating: " + abstractEnginePerformancePt->theRatingEngine.at(i).Rating << std::endl;
            /* Adjust data structure: */
            csv << "# (1) Mach; (2) Altitude [m]; (3) Thrust per Engine [N]; (4) SFC [g/Ns]" << std::endl;
            for (unsigned int j(0); j <= abstractEnginePerformancePt->theRatingEngine.at(i).theParameter.size() - 1; j++) {
                for (unsigned int n(0); n <= abstractEnginePerformancePt->theRatingEngine.at(i).theParameter.at(j).Mach.size() - 1; n++) {
                    csv << abstractEnginePerformancePt->theRatingEngine.at(i).theParameter.at(j).Mach.at(n) << "; ";
                    csv << abstractEnginePerformancePt->theRatingEngine.at(i).theParameter.at(j).Altitude << "; ";
                    csv << abstractEnginePerformancePt->theRatingEngine.at(i).theParameter.at(j).Thrust.at(n) << "; ";
                    csv << abstractEnginePerformancePt->theRatingEngine.at(i).theParameter.at(j).SFC.at(n);
                    csv << std::endl;
                }
                csv << std::endl;
                csv << std::endl;
            }
        }
        csv.close();
    }
}

void performance_assessment_output::generateCeilingPlotData() {
    std::ofstream csv;
    string csvDataFile(mySettingsPt->rtIO->getCsvFilesDirTool() + "/" + theProgramName + "_Ceiling.csv");
    csv.open(csvDataFile.c_str());
    if (!csv) {
        myRuntimeInfo->err << string(theProgramName) + "_Ceiling.csv could not be opened!" << std::endl;
        exit(1);
    } else {
        csv << "# " + mySettingsPt->rtIO->getCsvFilesDirTool() + theProgramName + "_Ceiling.csv" << std::endl;
        // Adjust data structure:
        csv << "# (1) Mach; (2) Gross Mass [kg]; (3) Service Ceiling [m]; (4) Max Ceiling [m] ; (5) OEI max Ceiling [m]; (6) Initial Cruise Altitude [m]" << std::endl;
        for (unsigned int i(0); i <= abstractCeilingPerformancePt->theCeiling.size() - 1; i++) {
            csv << abstractCeilingPerformancePt->theCeiling.at(i).Mach << "; ";
            csv << abstractCeilingPerformancePt->theCeiling.at(i).Mass << "; ";
            csv << abstractCeilingPerformancePt->theCeiling.at(i).serviceCeiling << "; ";
            csv << abstractCeilingPerformancePt->theCeiling.at(i).maxCeiling << "; ";
            csv << abstractCeilingPerformancePt->theCeiling.at(i).maxCeilingOEI << "; ";
            csv << myAcftPt->hInitialCruise << std::endl;
        }
        csv.close();
    }
}

void performance_assessment_output::generateFlightEnvelopeData() {
    std::ofstream csv;
    string csvDataFile(mySettingsPt->rtIO->getCsvFilesDirTool() + "/" + theProgramName + "_FlightEnvelope.csv");
    csv.open(csvDataFile.c_str());
    if (!csv) {
        myRuntimeInfo->err << string(theProgramName) + "_FlightEnvelope.csv could not be opened!" << std::endl;
        exit(1);
    } else {
        csv << "# " + string(theProgramName) + "_FlightEnvelope.csv" << std::endl;
        // Adjust data structure:
        csv << "# (1) Altitude [m]; (2) TAS [m/s]; (3) Gross Mass [kg]" <<std::endl;
        csv << "# Stall Limit" <<std::endl;
        for (unsigned int i(0); i <= abstractFlightEnvelopesPt->myFlightEnvelope.back().vStall.size() - 1; i++) {
            csv << abstractFlightEnvelopesPt->myFlightEnvelope.back().vStall.at(i).height  << "; ";
            csv << abstractFlightEnvelopesPt->myFlightEnvelope.back().vStall.at(i).speed << "; ";
            csv << abstractFlightEnvelopesPt->myFlightEnvelope.back().grossMass << "; ";
            csv <<std::endl;
        }
        csv << std::endl;
        csv << std::endl;
        csv << "# VMO Limit" << std::endl;
        for (unsigned int i(0); i <= abstractFlightEnvelopesPt->myFlightEnvelope.back().vMO.size() - 1; i++) {
            csv << abstractFlightEnvelopesPt->myFlightEnvelope.back().vMO.at(i).height << "; ";
            csv << abstractFlightEnvelopesPt->myFlightEnvelope.back().vMO.at(i).speed << "; ";
            csv << abstractFlightEnvelopesPt->myFlightEnvelope.back().grossMass << "; ";
            csv <<std::endl;
        }
        csv << std::endl;
        csv << std::endl;
        csv << "# MMO Limit" << std::endl;
        for (unsigned int i(0); i <= abstractFlightEnvelopesPt->myFlightEnvelope.back().MMO.size() - 1; i++) {
            csv << abstractFlightEnvelopesPt->myFlightEnvelope.back().MMO.at(i).height << "; ";
            csv << abstractFlightEnvelopesPt->myFlightEnvelope.back().MMO.at(i).speed << "; ";
            csv << abstractFlightEnvelopesPt->myFlightEnvelope.back().grossMass << "; ";
            csv << std::endl;
        }
        csv <<std::endl;
        csv <<std::endl;
        csv << "# VDive Limit" <<std::endl;
        for (unsigned int i(0); i <= abstractFlightEnvelopesPt->myFlightEnvelope.back().vDive.size() - 1; i++) {
            csv << abstractFlightEnvelopesPt->myFlightEnvelope.back().vDive.at(i).height  << "; ";
            csv << abstractFlightEnvelopesPt->myFlightEnvelope.back().vDive.at(i).speed << "; ";
            csv << abstractFlightEnvelopesPt->myFlightEnvelope.back().grossMass << "; ";
            csv <<std::endl;
        }
        csv <<std::endl;
        csv <<std::endl;
        csv << "# MDive Limit" <<std::endl;
        for (unsigned int i(0); i <= abstractFlightEnvelopesPt->myFlightEnvelope.back().MDive.size() - 1; i++) {
            csv << abstractFlightEnvelopesPt->myFlightEnvelope.back().MDive.at(i).height << "; ";
            csv << abstractFlightEnvelopesPt->myFlightEnvelope.back().MDive.at(i).speed << "; ";
            csv << abstractFlightEnvelopesPt->myFlightEnvelope.back().grossMass << "; ";
            csv <<std::endl;
        }
        csv <<std::endl;
        csv <<std::endl;
        csv << "#Altitude Limit" <<std::endl;
        csv << abstractFlightEnvelopesPt->myFlightEnvelope.back().vStall.back().height  << "; ";
        csv << abstractFlightEnvelopesPt->myFlightEnvelope.back().vStall.back().speed << "; ";
        csv << abstractFlightEnvelopesPt->myFlightEnvelope.back().grossMass << "; ";
        csv <<std::endl;
        csv << abstractFlightEnvelopesPt->myFlightEnvelope.back().MDive.back().height << "; ";
        csv << abstractFlightEnvelopesPt->myFlightEnvelope.back().MDive.back().speed << "; ";
        csv << abstractFlightEnvelopesPt->myFlightEnvelope.back().grossMass << "; ";
        csv.close();
    }
}

void performance_assessment_output::generateSARPlotData() {
    std::ofstream csv;
    for (unsigned int i(0); i < abstractSARPerformancePt->theSAR.size(); i++) {
        std::stringstream Number;
        Number << i;
        string csvDataFile(mySettingsPt->rtIO->getCsvFilesDirTool() + "/" + theProgramName + "_SAR_Mass" + Number.str() + ".csv");
        csv.open(csvDataFile.c_str());
        if (!csv) {
            myRuntimeInfo->err << csvDataFile + " could not be opened!" <<std::endl;
            exit(1);
        } else {
            csv << "# " + csvDataFile <<std::endl;
            // Adjust data structure:
            csv << "# (1) Mach; (2) Gross Mass [kg]; (3) Altitude [m]; (4) SAR [m/kg]" <<std::endl;
            for (unsigned int j(0); j < abstractSARPerformancePt->theSAR.at(i).tabSAR.size(); j++) {
                // csv << "# " << floor(myPerfPt->theSAR.at(i).Mass/1000)*1000 << " kg, " << floor(3.2808 * myPerfPt->theSAR.at(i).tabSAR.at(j).Altitude/100)*100<<std::endl;
                for (unsigned int n(0); n < abstractSARPerformancePt->theSAR.at(i).tabSAR.at(j).Mach.size(); n++) {
                    csv << abstractSARPerformancePt->theSAR.at(i).tabSAR.at(j).Mach.at(n) << "; ";
                    csv << abstractSARPerformancePt->theSAR.at(i).Mass << "; ";
                    csv << abstractSARPerformancePt->theSAR.at(i).tabSAR.at(j).Altitude << "; ";
                    csv << abstractSARPerformancePt->theSAR.at(i).tabSAR.at(j).SAR.at(n) <<std::endl;
                }
                csv <<std::endl;
                csv <<std::endl;
            }
        }
        csv.close();
    }
}

void performance_assessment_output::generate_best_SAR_Data() {
    std::ofstream csv;
    string csvDataFile(mySettingsPt->rtIO->getCsvFilesDirTool() + "/" + theProgramName + "_bestSAR.csv");
    csv.open(csvDataFile.c_str());
    if (!csv) {
        myRuntimeInfo->err << string(theProgramName) + "_bestSAR.csv could not be opened!" << std::endl;
        exit(1);
    } else {
        csv << "# " + string(theProgramName) + "_bestSAR.csv" << std::endl;
        // Adjust data structure:
        csv << "# (1) Mach; (2) Gross Mass [kg]; (3) Opt. Altitude [m]; (4) Opt. SAR [m/kg]" << std::endl;
        for (unsigned int i(0); i < abstractSARPerformancePt->theSAR.size(); i++) {
            csv << abstractSARPerformancePt->theSAR.at(i).bestSARMach << "; ";
            csv << abstractSARPerformancePt->theSAR.at(i).Mass << "; ";
            csv << abstractSARPerformancePt->theSAR.at(i).bestSARAlt << "; ";
            csv << abstractSARPerformancePt->theSAR.at(i).bestSAR << std::endl;
        }
        csv.close();
    }
}

void performance_assessment_output::generateTOFLPlotData() {
std::ofstream csv;
    string csvDataFile(mySettingsPt->rtIO->getCsvFilesDirTool() + "/" + theProgramName + "_TOFL.csv");
    csv.open(csvDataFile.c_str());
    if (!csv) {
        myRuntimeInfo->err << string(theProgramName) + "_TOFL.csv could not be opened!" << std::endl;
        exit(1);
    } else {
        csv << "# " + string(theProgramName) + "_TOFL.csv" << std::endl;
        csv << "# (1) Altitude [m]; (2) Gross Mass [kg]; (3) TODR_AEO [m]" << std::endl;
        for (unsigned int i(0); i <= abstractTakeOffPerformancePt->myTOFLField.size() - 1; i++) {
            for (unsigned int j(0); j <= abstractTakeOffPerformancePt->myTOFLField.at(i).myTOFL.size() - 1; j++) {
                csv << abstractTakeOffPerformancePt->myTOFLField.at(i).elevation << "; ";
                csv << abstractTakeOffPerformancePt->myTOFLField.at(i).myTOFL.at(j).grossMass << "; ";
                csv << abstractTakeOffPerformancePt->myTOFLField.at(i).myTOFL.at(j).TODR_AEO << std::endl;
            }
            csv << std::endl;
            csv << std::endl;
        }
        // display TOFL
        csv << abstractTakeOffPerformancePt->myTOFLField.at(0).elevation << "; ";
        csv << abstractTakeOffPerformancePt->myTOFLField.at(0).myTOFL.back().grossMass << "; ";
        csv << abstractTakeOffPerformancePt->myTOFLField.at(0).myTOFL.back().TODR_AEO << std::endl;
        csv << std::endl;
        csv << std::endl;
        // Req TOFL
        csv << 0.0 << "; ";
        csv << myAcftPt->MTOM << "; ";
        csv << myAcftPt->myReqs.TOFL << std::endl;
        csv.close();
    }
}

void performance_assessment_output::generateBLFLPlotData() {
    std::ofstream csv;
    string csvDataFile(mySettingsPt->rtIO->getCsvFilesDirTool() + "/" + theProgramName + "_BLFL.csv");
    csv.open(csvDataFile.c_str());
    if (!csv) {
        myRuntimeInfo->err << string(theProgramName) + "_BLFL.csv could not be opened!" << std::endl;
        exit(1);
    } else {
        csv << "# " + string(theProgramName) + "_BLFL.csv" << std::endl;
        csv << "# (1) Altitude [m]; (2) Gross Mass [kg]; (3) TODR_AEO [m]; (4) TODR_OEI [m]; (5) ASDR_OEI [m]; (6) v1 [m/s]" << std::endl;
        for (unsigned int i(0); i <= abstractTakeOffPerformancePt->myBLFL.vEF.size() - 1; i++) {
            csv << "0.0" << "; ";
            csv << abstractTakeOffPerformancePt->myBLFL.grossMass << "; ";
            csv << abstractTakeOffPerformancePt->myBLFL.TODR_AEO << "; ";
            csv << abstractTakeOffPerformancePt->myBLFL.TODR_OEI.at(i) << "; ";
            csv << abstractTakeOffPerformancePt->myBLFL.ASDR_OEI.at(i) << "; ";
            csv << abstractTakeOffPerformancePt->myBLFL.vEF.at(i) << std::endl;
        }
        csv.close();
    }
}

void performance_assessment_output::generateLDNPlotData() {
    std::ofstream csv;
    string csvDataFile(mySettingsPt->rtIO->getCsvFilesDirTool() + "/" + theProgramName + "_LDN.csv");
    csv.open(csvDataFile.c_str());
    if (!csv) {
        myRuntimeInfo->err << string(theProgramName) + "_LDN.csv could not be opened!" << std::endl;
        exit(1);
    } else {
        csv << "# " + mySettingsPt->rtIO->getCsvFilesDirTool() + theProgramName + "_LDN.csv" << std::endl;
        csv << "# (1) Altitude [m]; (2) Gross Mass [kg]; (3) ALD_AEO [m]" << std::endl;
        for (unsigned int i(0); i <= abstractLandingPerformancePt->myLDNField.size() - 1; i++) {
            for (unsigned int j(0); j <= abstractLandingPerformancePt->myLDNField.at(i).myLDN.size() - 1; j++) {
                csv << abstractLandingPerformancePt->myLDNField.at(i).elevation << "; ";
                csv << abstractLandingPerformancePt->myLDNField.at(i).myLDN.at(j).grossMass << "; ";
                csv << abstractLandingPerformancePt->myLDNField.at(i).myLDN.at(j).ALD_AEO << std::endl;
            }
            csv << std::endl;
            csv << std::endl;
        }
        // display TOFL
        csv << abstractLandingPerformancePt->myLDNField.at(0).elevation << "; ";
        csv << abstractLandingPerformancePt->myLDNField.at(0).myLDN.back().grossMass << "; ";
        csv << abstractLandingPerformancePt->myLDNField.at(0).myLDN.back().ALD_AEO << std::endl;
        csv << std::endl;
        csv << std::endl;
        // Req TOFL
        csv << 0.0 << "; ";
        csv << myAcftPt->MLM << "; ";
        csv << myAcftPt->myReqs.LDN << std::endl;
        csv.close();
    }
}

void performance_assessment_output::generateVnPlotData() {
    std::ofstream csv;
    string csvDataFile(mySettingsPt->rtIO->getCsvFilesDirTool() + "/" + theProgramName + "_VnDiagram.csv");
    csv.open(csvDataFile.c_str());
    if (!csv) {
        myRuntimeInfo->err << string(theProgramName) + "_VnDiagram.csv could not be opened!" << std::endl;
        getchar();
        exit(1);
    } else {
        csv << "# " + mySettingsPt->rtIO->getCsvFilesDirTool() + theProgramName + "_VnDiagram.csv" << std::endl;
        csv << "# (1) v [m]; (2) max n (pos.) [-]; (3) min n (neg.) [-]" << std::endl;
        csv << "#Clean configuration" << std::endl;
        for (unsigned int i(0); i <= abstractVnDiagramPt->myCleanDiagram.size() - 1; i++) {
            csv << abstractVnDiagramPt->myCleanDiagram.at(i).speed << "; ";
            csv << abstractVnDiagramPt->myCleanDiagram.at(i).n_upperBound << "; ";
            csv << abstractVnDiagramPt->myCleanDiagram.at(i).n_lowerBound << std::endl;
        }
        csv << std::endl;
        csv << std::endl;
        csv << "#Takeoff configuration" << std::endl;
        for (unsigned int i(0); i <= abstractVnDiagramPt->myCleanDiagram.size() - 1; i++) {
            csv << abstractVnDiagramPt->myTakeoffDiagram.at(i).speed << "; ";
            csv << abstractVnDiagramPt->myTakeoffDiagram.at(i).n_upperBound << "; ";
            csv << abstractVnDiagramPt->myTakeoffDiagram.at(i).n_lowerBound << std::endl;
        }
        csv << std::endl;
        csv << std::endl;
        csv << "#Landing configuration" << std::endl;
        for (unsigned int i(0); i <= abstractVnDiagramPt->myCleanDiagram.size() - 1; i++) {
            csv << abstractVnDiagramPt->myLandingDiagram.at(i).speed << "; ";
            csv << abstractVnDiagramPt->myLandingDiagram.at(i).n_upperBound << "; ";
            csv << abstractVnDiagramPt->myLandingDiagram.at(i).n_lowerBound << std::endl;
        }
        csv.close();
    }
}

void performance_assessment_output::generatePlots() {
    if (mySettingsPt->check_sucessfully_executed.doPayloadRange == true) {
        generate_plot_payload_range();
    }
    if (mySettingsPt->check_sucessfully_executed.doEnginePerformance == true) {
        generate_plot_engine_performance();
    }
    if (mySettingsPt->check_sucessfully_executed.doClimbPerformance == true) {
        generate_plot_ceiling_performance();
    }
    if (mySettingsPt->check_sucessfully_executed.doFlightEnvelopePerformance == true) {
        generate_plot_flight_envelope();
    }
    if (mySettingsPt->check_sucessfully_executed.doSARCalculation == true) {
        generate_plot_SAR_performance();
    }
    if (mySettingsPt->check_sucessfully_executed.doTOPerformance == true) {
        generate_plot_takeoff_field_length();
    }
    if (mySettingsPt->check_sucessfully_executed.doBLFLPerformance == true) {
        generate_plot_balanced_field_length();
    }
    if (mySettingsPt->check_sucessfully_executed.doLandingPerformance == true) {
        generate_plot_landing_performance();
    }
    if (mySettingsPt->check_sucessfully_executed.doVnDiagram == true) {
        generate_plot_v_n_diagram();
    }
}


void performance_assessment_output::generate_plot_payload_range() {
    // Gather data
    vector<double> x_range;
    vector<double> y_payload;
    // Case: Payload range data was generated by payload range class
    for (unsigned int i(0); i < abstractPayloadRangePt->Range.size(); i++) {
        x_range.push_back(abstractPayloadRangePt->Range.at(i) / 1000);
        y_payload.push_back(abstractPayloadRangePt->Payload.at(i));
    }
    // Plot
    matplot::figure_handle fig = matplot::figure(true);
    fig->quiet_mode(true);
    fig->size(600, 500);
    matplot::axes_handle ax = fig->current_axes();
    matplot::plot(x_range, y_payload);
    matplot::hold(matplot::on);
    matplot::grid(matplot::on);
    matplot::title("Payload Range Diagram ");
    ax->xlabel("Range [km]");
    ax->ylabel("Payload [kg]");
    // Design point
    ax->scatter({myAcftPt->designRange / 1000}, {myAcftPt->designPayload});
    ax->legend({"", "Design point"});
    ax->legend()->location(matplot::legend::general_alignment::bottomleft);
    // Save
    std::string plot_file_path = rtIO->getPlotDir() + "/" + std::string(TOOL_NAME) + "_payload_range.svg";
    matplot::save(plot_file_path);
}

void performance_assessment_output::generate_plot_ceiling_performance() {
    // Gather data
    vector<double> x;
    vector<double> service_ceiling_y;
    vector<double> max_ceiling_y;
    vector<double> max_oei_ceiling_y;
    for (unsigned int i(0); i < abstractCeilingPerformancePt->theCeiling.size(); i++) {
        service_ceiling_y.push_back(abstractCeilingPerformancePt->theCeiling.at(i).serviceCeiling);
        max_ceiling_y.push_back(abstractCeilingPerformancePt->theCeiling.at(i).maxCeiling);
        max_oei_ceiling_y.push_back(abstractCeilingPerformancePt->theCeiling.at(i).maxCeilingOEI);
        x.push_back(abstractCeilingPerformancePt->theCeiling.at(i).Mass);
    }
    double max_ylim = max_ceiling_y.at(0) + 1000;
    // Plot
    matplot::figure_handle fig = matplot::figure(true);
    fig->quiet_mode(true);
    fig->size(600, 500);
    matplot::axes_handle ax = fig->current_axes();
    ax->plot(x, service_ceiling_y)->line_style("-");
    matplot::hold(matplot::on);
    ax->plot(x, max_ceiling_y)->line_style("--");
    ax->plot(x, max_oei_ceiling_y);
    ax->grid(matplot::on);
    ax->ylim({0, max_ylim});
    ax->legend({"Service ceiling, Rate of climb 500 ft/min", "Max. ceiling, Rate of climb 0 ft/min", "OEI ceiling, Rate of climb 500 ft/min"});
    ax->legend()->location(matplot::legend::general_alignment::bottomright);
    ax->xlabel("Gross mass [kg]");
    ax->ylabel("Altitude [m]");
    ax->title("Ceiling Performance");
    // Save plot
    std::string plot_file_path = rtIO->getPlotDir() + "/" + std::string(TOOL_NAME) + "_ceiling.svg";
    matplot::save(plot_file_path);
}

void performance_assessment_output::generate_plot_flight_envelope() {
    // Gather data
    vector<double> x_vs1g;
    vector<double> y_vs1g;
    vector<double> x_vmo;
    vector<double> y_vmo;
    vector<double> x_vdive;
    vector<double> y_vdive;
    vector<double> x_mmo;
    vector<double> y_mmo;
    vector<double> x_mdive;
    vector<double> y_mdive;
    vector<double> x_ceiling;
    vector<double> y_ceiling;
    for (unsigned int i(0); i < abstractFlightEnvelopesPt->myFlightEnvelope.back().vStall.size(); i++) {
        x_vs1g.push_back(abstractFlightEnvelopesPt->myFlightEnvelope.back().vStall.at(i).speed);  // Last element in myFlightEnvelope contains m=MTOM data
        y_vs1g.push_back(abstractFlightEnvelopesPt->myFlightEnvelope.back().vStall.at(i).height);
        x_vmo.push_back(abstractFlightEnvelopesPt->myFlightEnvelope.back().vMO.at(i).speed);
        y_vmo.push_back(abstractFlightEnvelopesPt->myFlightEnvelope.back().vMO.at(i).height);
        x_vdive.push_back(abstractFlightEnvelopesPt->myFlightEnvelope.back().vDive.at(i).speed);
        y_vdive.push_back(abstractFlightEnvelopesPt->myFlightEnvelope.back().vDive.at(i).height);
        x_mmo.push_back(abstractFlightEnvelopesPt->myFlightEnvelope.back().MMO.at(i).speed);
        y_mmo.push_back(abstractFlightEnvelopesPt->myFlightEnvelope.back().MMO.at(i).height);
        x_mdive.push_back(abstractFlightEnvelopesPt->myFlightEnvelope.back().MDive.at(i).speed);
        y_mdive.push_back(abstractFlightEnvelopesPt->myFlightEnvelope.back().MDive.at(i).height);
    }
    x_ceiling.push_back(x_vs1g.back());
    x_ceiling.push_back(x_mdive.back());
    y_ceiling.push_back(y_vs1g.back());
    y_ceiling.push_back(y_mdive.back());
    // Create plot
    matplot::figure_handle fig = matplot::figure(true);
    fig->quiet_mode(true);
    fig->size(600, 500);
    matplot::axes_handle ax = fig->current_axes();
    ax->plot(x_vs1g, y_vs1g);
    matplot::hold(matplot::on);
    ax->plot(x_ceiling, y_ceiling);
    ax->plot(x_vmo, y_vmo);
    ax->plot(x_mmo, y_mmo);
    ax->plot(x_vdive, y_vdive);
    ax->plot(x_mdive, y_mdive);
    ax->grid(matplot::on);
    ax->ylim({0, y_ceiling.back() + 1000});
    ax->legend({"v_{s,1g}", "Ceiling", "v_{MO}", "M_{MO}", "V_{Dive}", "M_{Dive}"});
    ax->legend()->location(matplot::legend::general_alignment::topleft);
    ax->xlabel("TAS [m/s]");
    ax->ylabel("Altitude [m]");
    ax->title("Flight Envelope");
    // Save plot
    std::string plot_file_path = rtIO->getPlotDir() + "/" + std::string(TOOL_NAME) + "_flight_envelope.svg";
    matplot::save(plot_file_path);
}

void performance_assessment_output::generate_plot_takeoff_field_length() {
    matplot::figure_handle fig = matplot::figure(true);
    fig->quiet_mode(true);
    fig->size(600, 500);
    matplot::axes_handle ax = fig->current_axes();
    matplot::hold(matplot::on);
    matplot::grid(matplot::on);
    vector<string> plot_names;
    for (unsigned int i(0); i < abstractTakeOffPerformancePt->myTOFLField.size(); i++) {  // Iterates over elevations
        vector<double> x;
        vector<double> y;
        plot_names.push_back(std::format("{:.0f}", convertUnit(METER, FOOT, abstractTakeOffPerformancePt->myTOFLField.at(i).elevation)) + " ft (" +
            std::format("{:.1f}", abstractTakeOffPerformancePt->myTOFLField.at(i).elevation) + " m)");  // keeping only 2 decimal digits
        for (unsigned int j(0); j < abstractTakeOffPerformancePt->myTOFLField.at(i).myTOFL.size(); j++) {  // Iterates over masses
            x.push_back(abstractTakeOffPerformancePt->myTOFLField.at(i).myTOFL.at(j).grossMass / 1000);
            y.push_back(abstractTakeOffPerformancePt->myTOFLField.at(i).myTOFL.at(j).TODR_AEO);
        }
        ax->plot(x, y);  // Add takeoff performance for current elevation to figure
    }
    ax->legend(plot_names);
    ax->legend()->location(matplot::legend::general_alignment::topleft);
    ax->xlabel("Gross mass, 1000 kg");
    ax->ylabel("TODR_{AEO} [m]");
    ax->title("Take Off Distance Required-Performance, Delta ISA = " + std::format("{:.1f}", myAcftPt->delta_isa));
    // Save plot
    std::string plot_file_path = rtIO->getPlotDir() + "/" + std::string(TOOL_NAME) + "_takeoff_performance.svg";
    matplot::save(plot_file_path);
}

void performance_assessment_output::generate_plot_landing_performance() {
    matplot::figure_handle fig = matplot::figure(true);
    fig->quiet_mode(true);
    fig->size(600, 500);
    matplot::axes_handle ax = fig->current_axes();
    matplot::hold(matplot::on);
    matplot::grid(matplot::on);
    vector<string> plot_names;
    for (unsigned int i(0); i < abstractLandingPerformancePt->myLDNField.size(); i++) {  // Iterates over elevations
        vector<double> x;
        vector<double> y;
        plot_names.push_back(std::format("{:.0f}", convertUnit(METER, FOOT, abstractLandingPerformancePt->myLDNField.at(i).elevation)) + " ft (" +
            std::format("{:.1f}", abstractLandingPerformancePt->myLDNField.at(i).elevation) + " m)");  // keeping only 2 decimal digits
        for (unsigned int j(0); j < abstractLandingPerformancePt->myLDNField.at(i).myLDN.size(); j++) {  // Iterates over masses
            x.push_back(abstractLandingPerformancePt->myLDNField.at(i).myLDN.at(j).grossMass / 1000);
            y.push_back(abstractLandingPerformancePt->myLDNField.at(i).myLDN.at(j).ALD_AEO);
        }
        ax->plot(x, y);  // Add takeoff performance for current elevation to figure
    }
    ax->legend(plot_names);
    ax->legend()->location(matplot::legend::general_alignment::topleft);
    ax->xlabel("Gross mass, 1000 kg");
    ax->ylabel("ALD_{AEO} [m]");
    ax->title("Available Landing Distance-Performance, Delta ISA = " + std::format("{:.1f}", myAcftPt->delta_isa));
    // Save plot
    std::string plot_file_path = rtIO->getPlotDir() + "/" + std::string(TOOL_NAME) + "_landing_performance.svg";
    matplot::save(plot_file_path);
}

void performance_assessment_output::generate_plot_balanced_field_length() {
    matplot::figure_handle fig = matplot::figure(true);
    fig->quiet_mode(true);
    fig->size(600, 500);
    matplot::axes_handle ax = fig->current_axes();
    matplot::hold(matplot::on);
    matplot::grid(matplot::on);
    vector<double> x_v_engine_failure;
    vector<double> y_takeoff_distance_required_all_engines_operational;
    vector<double> y_takeoff_distance_required_one_engine_inoperative;
    vector<double> y_accelerate_stop_distance_one_engine_inoperative;
    for (unsigned int i(0); i < abstractTakeOffPerformancePt->myBLFL.vEF.size(); i++) {
        x_v_engine_failure.push_back(abstractTakeOffPerformancePt->myBLFL.vEF.at(i));
        y_takeoff_distance_required_all_engines_operational.push_back(abstractTakeOffPerformancePt->myBLFL.TODR_AEO);
        y_takeoff_distance_required_one_engine_inoperative.push_back(abstractTakeOffPerformancePt->myBLFL.TODR_OEI.at(i));
        y_accelerate_stop_distance_one_engine_inoperative.push_back(abstractTakeOffPerformancePt->myBLFL.ASDR_OEI.at(i));
    }
    ax->plot(x_v_engine_failure, y_takeoff_distance_required_all_engines_operational);
    ax->plot(x_v_engine_failure, y_accelerate_stop_distance_one_engine_inoperative);
    ax->plot(x_v_engine_failure, y_takeoff_distance_required_one_engine_inoperative);
    ax->legend({"TODR_{AEO}", "ASDR_{OEI}", "TODR_{OEI}"});
    ax->legend()->location(matplot::legend::general_alignment::topright);
    ax->xlabel("Engine failure speed v_{EF} [m/s]");
    ax->ylabel("Takeoff run distance [m]");
    ax->title("Balanced Field Length-Performance, Delta ISA = " + std::format("{:.1f}", myAcftPt->delta_isa));
    // Save plot
    std::string plot_file_path = rtIO->getPlotDir() + "/" + std::string(TOOL_NAME) + "_balanced_field_length_performance.svg";
    matplot::save(plot_file_path);
}

void performance_assessment_output::generate_plot_v_n_diagram() {
    matplot::figure_handle fig = matplot::figure(true);
    fig->quiet_mode(true);
    fig->size(600, 500);
    matplot::axes_handle ax = fig->current_axes();
    matplot::hold(matplot::on);
    matplot::grid(matplot::on);
    // Gather data & plot
    vector<double> x_speed;
    vector<double> y_clean_upper;
    vector<double> y_clean_lower;
    vector<double> y_landing_upper;
    vector<double> y_landing_lower;
    vector<double> y_takeoff_upper;
    vector<double> y_takeoff_lower;
    for (unsigned int i(0); i < abstractVnDiagramPt->myCleanDiagram.size(); i++) {
        x_speed.push_back(abstractVnDiagramPt->myCleanDiagram.at(i).speed);
        y_clean_upper.push_back(abstractVnDiagramPt->myCleanDiagram.at(i).n_upperBound);
        y_clean_lower.push_back(abstractVnDiagramPt->myCleanDiagram.at(i).n_lowerBound);
    }
    ax->plot(x_speed, y_clean_upper)->color("blue");
    ax->plot(x_speed, y_clean_lower)->color("blue");
    x_speed.clear();
    for (unsigned int i(0); i < abstractVnDiagramPt->myLandingDiagram.size(); i++) {
        x_speed.push_back(abstractVnDiagramPt->myLandingDiagram.at(i).speed);
        y_landing_upper.push_back(abstractVnDiagramPt->myLandingDiagram.at(i).n_upperBound);
        y_landing_lower.push_back(abstractVnDiagramPt->myLandingDiagram.at(i).n_lowerBound);
    }
    ax->plot(x_speed, y_landing_upper)->color("magenta");
    ax->plot(x_speed, y_landing_lower)->color("magenta");
    x_speed.clear();
    for (unsigned int i(0); i < abstractVnDiagramPt->myTakeoffDiagram.size(); i++) {
        x_speed.push_back(abstractVnDiagramPt->myTakeoffDiagram.at(i).speed);
        y_takeoff_upper.push_back(abstractVnDiagramPt->myTakeoffDiagram.at(i).n_upperBound);
        y_takeoff_lower.push_back(abstractVnDiagramPt->myTakeoffDiagram.at(i).n_lowerBound);
    }
    ax->plot(x_speed, y_takeoff_upper)->color("yellow");
    ax->plot(x_speed, y_takeoff_lower)->color("yellow");
    x_speed.clear();

    ax->legend({"n_{clean}", "", "n_{takeoff}", "", "n_{landing}", ""});
    ax->legend()->location(matplot::legend::general_alignment::topleft);
    ax->xlabel("Equivalent airspeed [m/s]");
    ax->ylabel("load factor n [-]");
    ax->title("V-n-Diagram");
    // Save plot
    std::string plot_file_path = rtIO->getPlotDir() + "/" + std::string(TOOL_NAME) + "_Vn_diagram.svg";
    matplot::save(plot_file_path);
}

void performance_assessment_output::generate_plot_SAR_performance() {
// Best SAR performance plot
    matplot::figure_handle fig = matplot::figure(true);
    fig->quiet_mode(true);
    fig->size(600, 500);
    matplot::axes_handle ax = fig->current_axes();
    matplot::hold(matplot::on);
    matplot::grid(matplot::on);
    // Gather data
    vector<double>best_alt;
    vector<double>best_mass;
    vector<double>best_mach;
    for (unsigned int i(0); i < abstractSARPerformancePt->theSAR.size(); i++) {
        best_alt.push_back(abstractSARPerformancePt->theSAR.at(i).bestSARAlt);
        best_mass.push_back(abstractSARPerformancePt->theSAR.at(i).Mass / 1000);
        best_mach.push_back(abstractSARPerformancePt->theSAR.at(i).bestSARMach);
    }
    // Plot best SAR
    ax->plot(best_mass, best_alt);
    ax->plot(best_mass, best_mach)->use_y2(true);  // using second y-axis
    ax->xlabel("Mass [t]");
    ax->ylabel("Altitude [m]");
    ax->y2label("Mach number [-]");  // TODO(hansmann): Add offset to the left to right axis label as it is cut off
    ax->title("Best Specific Air Range (SAR)");
    // Save plot
    std::string plot_file_path = rtIO->getPlotDir() + "/" + std::string(TOOL_NAME) + "_best_SAR.svg";
    matplot::save(plot_file_path);
    matplot::cla();

    // Other SAR plots
    for (unsigned int i(0); i < abstractSARPerformancePt->theSAR.at(0).tabSAR.size(); i+=2) {  // iterates over every the altitudes of interest
        double current_altitude = abstractSARPerformancePt->theSAR.at(0).tabSAR.at(i).Altitude;
        double current_altitude_ft = convertUnit(METER, FOOT, current_altitude);
        // Generate new plot for every altitude of interest
        matplot::figure_handle fig = matplot::figure(true);
        fig->quiet_mode(true);
        fig->size(600, 500);
        matplot::axes_handle ax = fig->current_axes();
        matplot::hold(matplot::on);
        matplot::grid(matplot::on);
        ax->title("Altitude = " + std::format("{:.0f}", floor(current_altitude)) + " m (" + std::format("{:.0f}", floor(current_altitude_ft)) + " ft)");

        vector<std::string> curve_names;
        for (unsigned int j(0); j < abstractSARPerformancePt->theSAR.size(); j+=2) {  // iterates over every second mass
            vector<double> mach_number;
            vector<double> SAR;
            for (unsigned int k(0); k < abstractSARPerformancePt->theSAR.at(j).tabSAR.at(i).SAR.size(); k++) {  // iterates over mach & SAR
                if (abstractSARPerformancePt->theSAR.at(j).tabSAR.at(i).SAR.at(k) <= ACCURACY_MEDIUM && k != 0) {  // removes artifacts from unsucessfull SAR calculation attempts
                } else {
                    SAR.push_back(abstractSARPerformancePt->theSAR.at(j).tabSAR.at(i).SAR.at(k));
                    mach_number.push_back(abstractSARPerformancePt->theSAR.at(j).tabSAR.at(i).Mach.at(k));
                }
            }
            if (!SAR.empty()) {  // skips legend entry if no SAR could be calculated for current mass
                curve_names.push_back(std::format("{:.0f}", ceil(abstractSARPerformancePt->theSAR.at(j).Mass / 1000)) + " t");
            }
            ax->plot(mach_number, SAR);
        }
        ax->xlabel("Mach Number [-]");
        ax->ylabel("SAR [m/kg]");
        ax->xlim({0, 0.9});  // TODO(hansmann): Change to dynamic limits later
        ax->ylim({0, 700});
        ax->legend(curve_names);
        ax->legend()->location(matplot::legend::general_alignment::topleft);
        // Save plot
        std::string plot_file_path = rtIO->getPlotDir() + "/" + std::string(TOOL_NAME) + "_SAR_" + std::format("{:.0f}", floor(current_altitude)) + "m.svg";
        SAR_plot_names.push_back(std::string(TOOL_NAME) + "_SAR_" + std::format("{:.0f}", floor(current_altitude)) + "m.svg");  // saving plot names for html-report
        matplot::save(plot_file_path);
        matplot::cla();
    }
}

void performance_assessment_output::generate_plot_engine_performance() {
    for (unsigned int i(0); i < abstractEnginePerformancePt->theRatingEngine.size(); i++) {  // iterates over the ratings
        // New figure for thrust over mach at current engine rating
        matplot::figure_handle fig1 = matplot::figure(true);
        fig1->quiet_mode(true);
        fig1->size(600, 500);
        matplot::axes_handle ax1 = fig1->current_axes();
        matplot::hold(matplot::on);
        matplot::grid(matplot::on);
        // New figure for thrust over SFC for current engine rating
        matplot::figure_handle fig2 = matplot::figure(true);
        fig2->quiet_mode(true);
        fig2->size(600, 500);
        matplot::axes_handle ax2 = fig2->current_axes();
        matplot::hold(matplot::on);
        matplot::grid(matplot::on);
        // Gathering data
        double current_altitude(0.);
        vector<string> curve_names;
        std::string current_rating = abstractEnginePerformancePt->theRatingEngine.at(i).Rating;
        for (unsigned int j(0); j < abstractEnginePerformancePt->theRatingEngine.at(i).theParameter.size(); j++) {  // iterates over the altitudes
            vector<double> thrust;
            vector<double> mach;
            vector<double> sfc;
            current_altitude = abstractEnginePerformancePt->theRatingEngine.at(i).theParameter.at(j).Altitude;
            for (unsigned int k(0); k < abstractEnginePerformancePt->theRatingEngine.at(i).theParameter.at(j).Thrust.size(); k++) {
                thrust.push_back(abstractEnginePerformancePt->theRatingEngine.at(i).theParameter.at(j).Thrust.at(k) / 1000);  // [N]->[kN]
                mach.push_back(abstractEnginePerformancePt->theRatingEngine.at(i).theParameter.at(j).Mach.at(k));
                sfc.push_back(abstractEnginePerformancePt->theRatingEngine.at(i).theParameter.at(j).SFC.at(k) * 1000 * 1000);  // [kg/Ns] -> [mg/Ns]
            }
            ax1->plot(mach, thrust);
            ax2->plot(sfc, thrust);
            curve_names.push_back(std::format("{:.0f}", convertUnit(METER, FOOT, current_altitude)) + " ft");
        }
        ax1->xlabel("Mach");
        ax1->ylabel("Thrust per Engine [kN]");
        string rating_for_plot;
        if (current_rating == "takeoff") {
            rating_for_plot = "Takeoff";
        } else if (current_rating == "maximum_continuous") {
            rating_for_plot = "Maximum Continuous";
        } else if (current_rating == "cruise") {
            rating_for_plot = "Cruise";
        } else if (current_rating == "climb") {
            rating_for_plot = "Climb";
        } else {
            rating_for_plot = current_rating;
        }
        ax1->title("Rating: " + rating_for_plot);
        matplot::legend_handle lgd1 = ax1->legend();
        ax1->legend(curve_names);
        ax1->legend()->location(matplot::legend::general_alignment::topright);
        // lgd1->title(std::string("Altitude"));  //TODO(hansmann): Not sure about the syntax here. This results in a segmentation fault

        ax2->xlabel("SFC [mg/Ns]");
        ax2->ylabel("Thrust per Engine [kN]");
        ax2->title("Rating: " + rating_for_plot);
        matplot::legend_handle lgd2 = ax2->legend();
        ax2->legend(curve_names);
        ax2->legend()->location(matplot::legend::general_alignment::topright);
        // lgd2->title(std::string("Altitude"));

        // Save figures
        std::string plot_file_path_machplot = rtIO->getPlotDir() + "/" + std::string(TOOL_NAME) + "_engine_performance_" + current_rating + "_mach.svg";
        fig1->save(plot_file_path_machplot);
        engine_performance_plot_names.push_back(std::string(TOOL_NAME) + "_engine_performance_" + current_rating + "_mach.svg");  // saving plot names for html-report
        std::string plot_file_path_sfcplot = rtIO->getPlotDir() + "/" + std::string(TOOL_NAME) + "_engine_performance_" + current_rating + "_sfc.svg";
        engine_performance_plot_names.push_back(std::string(TOOL_NAME) + "_engine_performance_" + current_rating + "_sfc.svg");  // saving plot names for html-report
        fig2->save(plot_file_path_sfcplot);
    }
}
