/*
 * UNICADO - UNIversity Conceptual Aircraft Design and Optimization
 *
 * Copyright (C) 2025 UNICADO consortium
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 * Description:
 * This file is part of UNICADO.
 */

#ifndef SYSTEMS_DESIGN_SRC_COMMON_AIRCRAFTSYSTEMS_CONVENTIONALATA27_H_
#define SYSTEMS_DESIGN_SRC_COMMON_AIRCRAFTSYSTEMS_CONVENTIONALATA27_H_

#include <string>
#include <vector>
#include <memory>

#include <svl/SVL.h>
#include <aircraftGeometry2/airfoil_surface.h>
#include <aircraftGeometry2/processing/measure.h>
#include "../aircraftSystem.h"
#include "../powerProfile.h"
#include "../systemsIOData.h"                       // Import of header file from systemsIOData for accessing data_. Can't be included in
                                                    // aircraftSystem.h because that would lead to circular dependencies


class mission;
class highLiftSystem;

/** \brief contains all properties of the primary flight control system actuators
 */
class actuatorPFCS {
 public:
    std::string powerSource;/**< Name of the power source */
    int powerSourceID;/**< ID of the power source */
    std::string operationMode;/**< operation mode of the actuator e.g. active or standby */
    Vec3 refPoint;/**< [m] reference point of the actuator in global coordinates */
    double Mass;/**< [kg] component mass */
    double designPower;/**< [W] design power (max. power)*/
    double power;/**<[W] power demand */
    double efficiency; /**< [-] actuator efficiency */

    actuatorPFCS();
};


/** \brief Contains all properties of a control surface for the flight control system (ATA27)
 */
class controlSurfaceATA27 {
 public:
    double massControl;/**<[kg] mass of the control system according to [NASA76] */
    double c_S;/**<[m] Control surface mean local chord */

    std::string type;/**<control surface type*/
    double S; /**<[m^2] surface area*/

    double maxDefSpeed;/**<[rad/s] max. deflection speed**/
    double designPowerActuator;/**<[W] max. power of the individual actuators */
    double maxHingeMoment;/**<[Nm] max. hinge moment */
    double controlSurfaceDesignPower;/**<[W] max. power of the control surface */

    double S_surface;/**<[m^2] surface area of the wing, stabilizer, etc. */
    double MAC_surface;/**<[m^2] mean aerodynamic chord of the wing, stabilizer, etc. */
    double deltaC_H_0; /**<Changes in hinge moment coefficient due to airfoil type, [NASA76]*/

    int numberOfActuators;/**<Number of actuators per surface */

    //Actuators
    std::vector<actuatorPFCS> theActuators;/**<vector object of primary flight conrol system actuators */

    /** \brief Constructor
     */
    controlSurfaceATA27();
};

/** \brief Inherited class that contains all properties of a thsa control surface
 */
class thsaControlSurface: public controlSurfaceATA27 {
 public:
    /** \brief Constructor
     */
    thsaControlSurface();
};

/** \brief Inherited class that contains all properties of the detailed control surface
 */
class detailedControlSurface: public controlSurfaceATA27 {
 public:
    double phi_LE_CoSu; /**< Control Surface Leading Edge Sweep angle [deg]**/
    double phi_TE_CoSu; /**< Control Surface Trailing Edge Sweep angle [deg]**/

    double relPosOfContrSurfHingeLine; /**< Relative position of the hinge line of the control Surface (Distance LE 2 Hinge Line/ContrSurfChord) [-]**/
    double spanDistIBLE2CenterOfArea; /**< Distance in span direction from the ContrSurf inboard leading ldge to the ContrSurf center of area [m]**/
    double chordDistIBLE2CenterOfArea; /**< Distance in chord direction from the ContrSurf inboard leading ldge to the ContrSurf center of area [m]**/
    double distCoA2HingeLine;          /**< Minimum (perpendicular to hinge line) distance of center of area to the hinge line [m]**/
    double areaMomentAboutHingeLine;   /**< Area moment of the control surface about hinge line [m^3]**/

    double s_element;       /**< [m] absolute span of the control surface [m]**/
    double y_i_abs;         /**< [m] absolute span position of the inner element point (wrt. 'referencePty') **/
    double y_a_abs;         /**< [m] absolute span position of the outer element point (wrt. 'referencePty') **/
    double l_i_abs;         /**< [m] absolute inner element chord length **/
    double l_a_abs;         /**< [m] absolute outer element chord length **/
    double l_i_rel_TE;      /**< [-] relative position of the element from wing trailing edge to inner trailing edge of the element **/
    double l_a_rel_TE;      /**< [-] relative position of the element from wing trailing edge to outer trailing edge of the element **/

    double maxDeflection;   /**< [rad] maximum deflection angle */

    std::vector<Vec3> upperWingSurfaceAtContrSurfIB; /**< Upper profile section of the wing (or other horizontal surface) at the inboard position of the control surface **/
    std::vector<Vec3> upperWingSurfaceAtContrSurfOB; /**< Upper profile section of the wing (or other horizontal surface) at the outboard position of the control surface **/
    double localWingChordAtContrSurfIB; /**< chord length of the wing (or other horizontal surface) at the inboard position of the control surface **/
    double localWingChordAtContrSurfOB; /**< chord length of the wing (or other horizontal surface) at the ouboard position of the control surface **/

    detailedControlSurface();
};

/** \brief Inherited class that contains all properties for the vertical control surface
 */
class verticalControlSurface: public detailedControlSurface {
 public:
    verticalControlSurface();
};

/** \brief Inherited class that contains all properties for the horizontal control surface
 */
class horizontalControlSurface: public detailedControlSurface {
 public:
    horizontalControlSurface();
};

/** \brief main class of conventionalATA27. Contains all properties of the flight control system (ATA27)
 */
class conventionalATA27: public aircraftSystem {
 public:
    /** \brief special properties of the flight control system (ATA27)
     */
    class flightControlSystem {
     public:
        flightControlSystem();
        virtual ~flightControlSystem();

        bool flyMission;/**< True: Performance decrease is determined by flight mission in each step. False: Average performance decrease are assumed */

        bool electricalFCS;
        double roll_EFCSWeightFactor;     /**< Weight factor for electrical Flight Control System for roll controls**/
        double yaw_EFCSWeightFactor;      /**< Weight factor for electrical Flight Control System for yaw controls**/
        double pitch_EFCSWeightFactor;    /**< Weight factor for electrical Flight Control System for pitch controls**/
        double THSA_EFCSWeightFactor;     /**< Weight factor for electrical Flight Control System for THS controls**/
        double HL_TE_EFCSWeightFactor;    /**< Weight factor for electrical Flight Control System for High Lift Trailing Edge controls**/
        double Spoiler_EFCSWeightFactor;  /**< Weight factor for electrical Flight Control System for Spoiler controls**/
        double HL_LE_EFCSWeightFactor;    /**< Weight factor for electrical Flight Control System for High Lift Leading Edge controls**/

        double deploymentHighLiftTime;    /**< [s] **/
        double commonInst_FCS_Factor;     /**< Percentage of flight control system for common installation weight **/

        int numberOfSlatSystems;/**<Number of Slat Systems*/
        int numberOfFlapSystems;/**<Number of Flap Systems*/
        int numberOfAilerons;/**<Number of Ailerons*/
        int numberOfSpoilers;/**<Number of Spoiler*/
        int numberOfElevators;/**<Number of Elevator*/
        int numberOfRudders;/**<Number of Rudder*/
        int numberOfTHSAs;/**<Number of THSAs*/

        bool defaultAileronActuator;/**< Switch that enables the default layout for the aileron actuator architecture */
        bool defaultSpoilerActuator;/**< Switch that enables the default layout for the spoiler actuator architecture */
        bool defaultElevatorActuator;/**< Switch that enables the default layout for the elevator actuator architecture */
        bool defaultRudderActuator;/**< Switch that enables the default layout for the rudder actuator architecture */
        bool defaultTHSAActuator;/**< Switch that enables the default layout for the THSA actuator architecture */

        /** \brief Class for reading the actuator layout Settings
         */
        class Actuator {
         public:
            PowerSource powerSource;/**< Actuator Power Source */
            std::string operationMode;/**Operating mode in normal operation (e.g. "active", "standby", "damping")<*/
            double efficiency;/**< [-] actuator efficiency */
            double stdbyPower;/**[W] Standby Power e.g. due to leakage<*/
            double activePower;/**[W] Performance in active operation due to rudder deflections<*/

            Actuator()
            :
            operationMode(""),
            efficiency(1.),
            stdbyPower(0.),
            activePower(0.) {}
        };
        Actuator theDefaultActuator;//Default actuator as standard, if nothing else is known

        /** \brief Class for reading the settings for the control surfaces
         */
        class ControlDevice {
         public:
            std::string name;/**< Name of the control surface*/
            std::string Side;/**< Side of the aircraft the control device is mounted */
            double numberOfActuators;/**< Number of actuators*/
            double maxDefSpeed;/**< [rad/s] Maximum deflection rate*/

            ControlDevice()
            :
            name(""),
            Side(""),
            numberOfActuators(0),
            maxDefSpeed(0.) {}

            std::vector<Actuator> actuators;/** Vector containing the actuators of the control device */
        };

        /** \brief class for the high lift system properties
         */
        class HighLiftSystem {
         public:
            std::string type;/**< Type of high lift system */
            int numberOfPCUMotors;/**< [-] Number of PCU motors of the high lift system*/

            HighLiftSystem()
            :
            type(""),
            numberOfPCUMotors(0) {}

            std::vector<Actuator> PCU_Motors;/**< Vector containing all PCU motors of the high lift system*/
        };
        std::vector<HighLiftSystem> highLift;/**< Vector containing the high lift system control devices */
        std::vector<ControlDevice> ailerons;/**< Vector containing the aileron control devices */
        std::vector<ControlDevice> spoilers;/**< Vector containing the spoiler control devices */
        std::vector<ControlDevice> elevators;/**< Vector containing the elevator control devices */
        std::vector<ControlDevice> rudders;/**< Vector containing the rudder control devices */
        std::vector<ControlDevice> THSAs;/**< Vector containing the THSA control devices */
    };
    flightControlSystem flightControl;/**< Flight control system object */
    double rollFCMass;/**<[kg] mass of the roll control system  */
    double yawFCMass;/**<[kg] mass of the yqw control system */
    double pitchFCMass;/**<[kg] mass of the pitch control system */
    double THSA_FCMass;/**<[kg] mass of the trimmable horizontal stabilizer system */
    double FlapsFCMass;/**<[kg] mass of the flap control system */
    double SpoilersFCMass;/**<[kg] mass of the spoiler system */
    double SlatsFCMass;/**<[kg] mass of the slat control system */
    double commonInstFCMass;/**<[kg] mass of the common installations */
    double primaryFCSMass;/**<[kg] total mass of the primary flight control system */
    double secondaryFCSMass;/**<[kg] total mass of the secondary flight control system  */

    std::vector<horizontalControlSurface> Aileron;/**< Vector containing the aileron control surfaces */
    std::vector<horizontalControlSurface> Spoiler;/**< Vector containing the spoiler control surfaces */
    std::vector<horizontalControlSurface> Elevator;/**< Vector containing the elevator control surfaces */
    std::vector<verticalControlSurface> Rudder;/**< Vector containing the rudder control surfaces */
    std::vector<thsaControlSurface> THSA;/**< Vector containing the THSA control surfaces */
    std::vector<highLiftSystem> theHighLiftSystem;/**< Vector containing the high lift system surfaces */

    /** \brief calculates the power of the other systems
     *
     * \param allPreviousSystemsBleed powerProfile bleed air
     * \param allPreviousSystemsHydraulic vector<powerProfile> hydraulic
     * \param allPreviousSystemsElectric vector<powerProfile> electric
     * \param allPreviousSystemsHeat powerProfile heat load
     * \return void
     *
     */
    void calculateSystemsPower(const powerProfile& allPreviousSystemsBleed,
                               const std::vector<powerProfile>& allPreviousSystemsHydraulic,
                               const std::vector<powerProfile>& allPreviousSystemsElectric,
                               const powerProfile& allPreviousSystemsHeat) override;

    /** \brief calculates the mass of the flight control system
     *
     * \param sourceBleed powerProfile
     * \param sourceElectric vector<powerProfile>
     * \param sourceHydraulic vector<powerProfile>
     * \return void
     *
     */
    void calculateSinkMass(const powerProfile& sourceBleed,
                           const std::vector<powerProfile>& sourceElectric,
                           const std::vector<powerProfile>& sourceHydraulic) override;

    /** \brief calculates the power of the flight control system (ATA27)
     *
     * \return void
     *
     */
    void getPower();

    /** \brief calculates the mass of the flight control system (ATA27)
     *
     * \return double
     *
     */
    double getMass();

    /** \brief calculates the reference point of the flight control system
     *
     * \param aWing liftingSurface&
     * \param aHTP liftingSurface&
     * \param aVTP verticalsurface&
     * \return Vec3 reference point
     *
     */
    Vec3 getRefPoint();

    /** \brief Constructor of the flight control system
     *
     * \param systemName string
     * \param systemIsOperating bool
     * \param theSettings const systemsDesignSettings&
     * \param myAircraft aircraft
     *
     */
    conventionalATA27(const std::string& systemName, bool systemIsOperating, std::shared_ptr<systemsIOData> data, const RuntimeIO& rtIO);
    /** \brief Destructor of the flight control system
     *
     * \return virtual
     *
     */
    virtual ~conventionalATA27();

    /**
     * \brief function for reading the system specific parameters from the config
     * 
     * \param config const node&: systemsDesign configuration file
    */
    void readParameters(const node& config) override;
    /**
     * \brief function for updating the system specific values in the aircraft xml
    */
    void updateXML() override;
    /** \brief reads aileron information from config
     * \param config config nodes
     */
    void getAilerons(const node& config);
    /** \brief reads spoiler information from config
     * \param config config nodes
     */
    void getSpoilers(const node& config);
    /** \brief reads elevator information from config
     * \param config config nodes
     */
    void getElevators(const node& config);
    /** \brief reads rudder information from config
     * \param config config nodes
     */
    void getRudders(const node& config);
    /** \brief reads THSA information from config
     * \param config config nodes
     */
    void getTHSAs(const node& config);
    /** \brief reads flap system information from config
     * \param config config nodes
     */
    void getFlapSystems(const node& config);
    /** \brief reads slat system information from config
     * \param config config nodes
     */
    void getSlatSystems(const node& config);

    /** \brief checks user input in the configuration file for consistency
     * 
    */
    void checkUserInput() override;

 private:
    double MaximumGrossMass;/**< [kg] MTOM*/

    /** \brief sets the geometry of a horizontal control surface (Elevator, Aileron, Spoiler)
     *
     * \param geoContrSurf the geometry of the control surface
     * \param geoSurface the geometry of the parent aerodynamic surface
     * \param horizontalControlSurface the ATA27 control surface instance to write the values to
     * \param const flightControlSystem::ControlDevice &settingsControlDevice
     * \param posDeflection maximum positive deflection angle of the control surface
     * \param negDeflection maximum negative deflection angle of the control surface
     * \return void
     *
     */
    void setHorizontalContrSurfGeo(const geom2::MultisectionSurface<geom2::PolygonSection> &geoContrSurf,
                                   const geom2::MultisectionSurface<geom2::AirfoilSection> &geoSurface,
                                   horizontalControlSurface *CoSu,
                                   const flightControlSystem::ControlDevice &settingsControlDevice,
                                   double posDeflection, double negDeflection);

    /** \brief sets the temporary rudder geometry
     *
     * \param const verticalsurface::controlSurface &geoContrSurf
     * \param verticalsurface &geoSurface, string side
     * \param verticalControlSurface *contrSurf
     * \param const flightControlSystem::ControlDevice &settingsControlDevice
     * \return void
     *
     */
    void setRudderTemp(const geom2::MultisectionSurface<geom2::PolygonSection> &geoContrSurf,
                       const geom2::MultisectionSurface<geom2::AirfoilSection> &geoSurface, const std::string& side,
                       verticalControlSurface *contrSurf,
                       const flightControlSystem::ControlDevice &settingsControlDevice,
                       double posDeflection, double negDeflection);

    /** \brief sets the trimmable horizontal stabilizer as a control surface
     *
     * \param liftingSurface &geoSurface
     * \param thsaControlSurface *contrSurf
     * \param const flightControlSystem::ControlDevice &settingsControlDevice
     * \return void
     *
     */
    void setTHSA_AsControlSurface(const geom2::MultisectionSurface<geom2::AirfoilSection> &geoSurface, thsaControlSurface *contrSurf,
                                  const flightControlSystem::ControlDevice &settingsControlDevice);

    /** \brief sets the design power of the entire flight control system
     *
     * \param wingLoading double
     * \param VMO double: [m/s] maximum operating speed
     * \param MMO double: [-] maximum operating Mach number
     * \param myElectric vector<powerProfile>*
     * \param myHydraulic vector<powerProfile>*
     * \return void
     *
     */
    void setDesignPower_FCS(double wingLoading, double VMO, double MMO, std::vector<powerProfile> *myElectric, std::vector<powerProfile> *myHydraulic);

    /** \brief calculates the design power of the primary flight control system
     *
     * \param controlSurfaceATA27 *theControlSurface
     * \param double theWingLoading
     * \param theVMO double: [m/s] maximum operating speed
     * \param theMMO double: [-] maximum operating Mach number
     * \return void
     *
     */
    void getDesignPowerPFCS(controlSurfaceATA27 *theControlSurface,
                            double theWingLoading, double theVMO, double theMMO);

    /** \brief calculates the detailed mission power of the flight control system for the individual mission steps
     *
     * \param mission myMission
     * \param powerProfile *myBleedAir
     * \param vector<powerProfile> *myHydraulic
     * \param vector<powerProfile> *myElectric
     * \param powerProfile *myHeatLoad
     * \param double timeHighLift: [s] Time for high lift system deployment
     * \param vector<systemsIOData::Wing> thisWing
     * \return
     *
     */
    void getDetailedMissionPower(powerProfile *myBleedAir, std::vector<powerProfile> *myHydraulic, std::vector<powerProfile> *myElectric,
                                 powerProfile *myHeatLoad,
                                 double timeHighLift,
                                 std::vector<systemsIOData::Wing> thisWing);


    /** \brief calculates the average power over all mission steps
     *
     * \param mission myMission
     * \param powerProfile *myBleedAir
     * \param vector<powerProfile> *myHydraulic
     * \param vector<powerProfile> *myElectric
     * \param powerProfile *myHeatLoad
     * \param double timeHighLift: [s] time for high lift system deloyment
     * \param vector<systemsIOData::Wing> thisWing
     * \return
     *
     */
    void getAverageMissionPower(powerProfile *myBleedAir, std::vector<powerProfile> *myHydraulic, std::vector<powerProfile> *myElectric,
                                powerProfile *myHeatLoad,
                                double timeHighLift,
                                std::vector<systemsIOData::Wing> thisWing);

    /** \brief Calculation of leading edge and trailing edge sweep angle of the control surface
     *
     * \param myContrSurf detailedControlSurface*
     * \return void
     *
     */
    void calcContrSurfSweepAngles(const geom2::MultisectionSurface<geom2::PolygonSection> &surface, detailedControlSurface *myContrSurf);

    /** \brief Calculation of the position of the center of area of the control surface relative to its inboard leading edge
     *
     * \param myContrSurf detailedControlSurface*
     * \return void
     *
     */
    void calcContrSurfCenterOfArea(detailedControlSurface *myContrSurf);

    /** \brief Calculation of the Minimum (perpendicular to hinge line) distance of the center of area to the hinge line
     *
     * \param myContrSurf detailedControlSurface*
     * \return void
     *
     */
    void calcDistCoA2HingeLine(detailedControlSurface *myContrSurf);

    /** \brief calculates the hinge moment of the primary flight control system
     *
     * \param myControlSurface controlSurfaceATA27
     * \param q double: [N/m^2] dynamic pressure
     * \param theMach double
     * \return double
     *
     */
    double getHingeMomentPFCS(controlSurfaceATA27 myControlSurface, double q, double theMach);

    /** \brief calculates hinge moment at 0 deflection angle
     *
     * \param M double: Mach number
     * \return double
     *
     */
    double getC_H_0(double M);

    /** \brief calculates hinge moment with respect to deflection angle gamma
     *
     * \param M double: Mach number
     * \return double
     *
     */
    double getC_H_gamma(double M);

//    /** \brief calculates the power of the high lift system. CAUTION: OLD METHOD! HIGH LIFT OPWER IS NOW CALCULATED IN "highLiftSystem.cpp"
//     *
//     * \param theWings vector<liftingSurface>
//     * \param Speed double
//     * \param airDensity double
//     * \return double
//     *
//     */
//    double getHighLiftSystemPower(const std::vector<liftingSurface>& theWings, double Speed, double airDensity);

    /** \brief Calculates the maximum deflection angle in terms of amount (fabs(angle))
     *
     * \param minDeflection double: [deg] min deflection angle (usually negative)
     * \param maxDeflection double [deg] maximum deflection angle (usually positive)
     * \return double maxDefl: [rad] maximum deflection angle in terms of amount
     *
     */
    double getMaxDeflection(double minDeflection, double maxDeflection);

    /** \brief calculates the mass of the aileron control system
     *
     * \param myControlSurfaces vector<horizontalControlSurface>
     * \return double [kg]
     *
     */
    double getAileronControlMass(std::vector<horizontalControlSurface> myControlSurfaces);

    /** \brief calculates the mass of the rudder control system
     *
     * \param vector<verticalControlSurface> myControlSurfaces
     * \return double [kg]
     *
     */
    double getRudderControlMass(std::vector<verticalControlSurface> myControlSurfaces);

    /** \brief calculates the mass of the spoiler control system
     *
     * \param vector<horizontalControlSurface> myControlSurfaces
     * \return double [kg]
     *
     */
    double getSpoilerControlMass(std::vector<horizontalControlSurface> myControlSurfaces);

    /** \brief calculates the mass of the elevator control system
     *
     * \param vector<horizontalControlSurface> myControlSurfaces
     * \return double [kg]
     *
     */
    double getElevatorControlMass(std::vector<horizontalControlSurface> myControlSurfaces);


    /** \brief calculates the mass of the trimmable horizontal stabilizer control system
     *
     * \return double [kg]
     *
     */
    double getTHSAControlMass();
};

#endif // SYSTEMS_DESIGN_SRC_COMMON_AIRCRAFTSYSTEMS_CONVENTIONALATA27_H_
