/*
 * UNICADO - UNIversity Conceptual Aircraft Design and Optimization
 *
 * Copyright (C) 2025 UNICADO consortium
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 * Description:
 * This file is part of UNICADO.
 */

#ifndef SYSTEMS_DESIGN_SRC_COMMON_AIRCRAFTSYSTEMS_CONVENTIONALATA28_H_
#define SYSTEMS_DESIGN_SRC_COMMON_AIRCRAFTSYSTEMS_CONVENTIONALATA28_H_

/* The following statements import necessary libraries */
#include <string>                                                       // Import c++ standard library for using string
#include <vector>                                                       // Import c++ standard library for using vector
#include <memory>                                                       // Import c++ standard library for using memory
/* The following statements import necessary header files */
#include <svl/SVL.h>                                                    // Import of header file from src folder for SVL
#include "../aircraftSystem.h"                                          // Import of header file from src folder for aircraft system
#include "../powerProfile.h"                                            // Import of header file from src folder for power profile
#include "../systemsIOData.h"                                           // Import of header file from systemsIOData for accessing data_. Can't be included in
                                                                        // aircraftSystem.h because that would lead to circular dependencies


/** \class conventionalATA28
 * \brief class that contains all properties of the fuel system (ATA28)
 */
class conventionalATA28: public aircraftSystem {
 public:
    /**
     * \brief function for reading the system specific parameters from the config
     * 
     * \param config const node&: systemsDesign configuration file
    */
    void readParameters(const node& config) override;
    /**
     * \brief function for updating the system specific values in the aircraft xml
    */
    void updateXML() override;
    /** \brief calculates the system power [W]
     * \param allPreviousSystemsBleed powerProfile
     * \param allPreviousSystemsHydraulic vector<powerProfile>
     * \param allPreviousSystemsElectric vector<powerProfile>
     * \param allPreviousSystemsHeat powerProfile
     * \return void
     */
    void calculateSystemsPower(const powerProfile& allPreviousSystemsBleed,
                               const std::vector<powerProfile>& allPreviousSystemsHydraulic,
                               const std::vector<powerProfile>& allPreviousSystemsElectric,
                               const powerProfile& allPreviousSystemsHeat) override;

    /** \brief calculates the mass of the power sinks [W]
     * \param sourceBleed powerProfile
     * \param sourceElectric vector<powerProfile>
     * \param sourceHydraulic vector<powerProfile>
     * \return void
     */
    void calculateSinkMass(const powerProfile& sourceBleed,
                           const std::vector<powerProfile>& sourceElectric,
                           const std::vector<powerProfile>& sourceHydraulic) override;

    /** \brief calculates the required power of the fuel system (ATA28) [W]
     * \return void
     */
    void getPower();

    /** \brief calculates the mass of the fuel system (ATA28)
     * \return double: mass of the fuel system [kg]
     */
    double getMass();

    /** \brief calculates the reference point in x, y and z-direction
     * \param theWingCG Vec3: center of gravity of the wing
     * \return Vec3
     */
    Vec3 getRefPoint();

    /** \brief Constructor
     * \param systemName string
     * \param systemIsOperating bool: Switch whether system is operating (1) or not (0)
     * \param theSettings const systemsDesignSettings&
     * \param myAircraft aircraft
     */
    conventionalATA28(const std::string& systemName, bool systemIsOperating, std::shared_ptr<systemsIOData> data, const RuntimeIO& rtIO);
    virtual ~conventionalATA28();
};

#endif // SYSTEMS_DESIGN_SRC_COMMON_AIRCRAFTSYSTEMS_CONVENTIONALATA28_H_
