/*
 * UNICADO - UNIversity Conceptual Aircraft Design and Optimization
 *
 * Copyright (C) 2025 UNICADO consortium
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 * Description:
 * This file is part of UNICADO.
 */

#ifndef SYSTEMS_DESIGN_SRC_COMMON_AIRCRAFTSYSTEMS_HIGHLIFTSYSTEM_POWERCONTROLUNIT_H_
#define SYSTEMS_DESIGN_SRC_COMMON_AIRCRAFTSYSTEMS_HIGHLIFTSYSTEM_POWERCONTROLUNIT_H_

#include <string>
#include <vector>
#include <svl/SVL.h>

/** \brief contains all properties of the PCU
 */
class PowerControlUnit {
 public:
    double Mass;/**< [kg] mass of the PCU*/
    double installedPower;/**< [W] installed PCU power*/
    double efficiencyFactorMin;/**< [-] minimum efficiency factor from [Gro09]] p.89*/
    double efficiencyFactorMax;/**< [-] maximum efficiency factor from [Gro09]] p.89*/
    double safetyFactor;/**< [-] safety factor */
    double rotationSpeedPCU;/**< [1/s] PCU rotation speed from [Cha10] p. 46 */

    double hydraulicPressureMin;/**< [bar] minimum hydraulic pressure */
    double hydraulicPressureMax;/**< [bar] maximum hydraulic pressure */

    /**
     * \brief calculates installed power
     * 
     * \param theMTOM maximum takeoff weight [kg]
     * \param SrefWing reference area of the wing [m2]
    */
    void getInstalledPower(double theMTOM, double SrefWing);
    /**
     * \brief calculates mass of the power control unit
     * 
     * \param deviceType string
     * \param SrefWing reference area of the wing [m2]
    */
    void getMass(const std::string& deviceType, double SrefWing);

    /** \brief contains all properties of a motor inside a power control unit. There can be multiple motors in one PCU.
     */
    class PowerControlUnitMotor {
     public:
        std::string powerSource;/**< name of power source */
        int powerSourceID;/**< power source ID */
        Vec3 refPoint;/**< [m] reference point of the PCU motor */
        std::string operationMode;/**< operation mode */
        double power;/**< [W] power */

        PowerControlUnitMotor()
        :
        powerSource(""),
        powerSourceID(0),
        refPoint({0., 0., 0.}),
        operationMode(""),
        power(0.) {}
    };
    std::vector<PowerControlUnitMotor> PCU_Motor;/**< vector containing the PCU motors */

    explicit PowerControlUnit(const std::string& Type);
    virtual ~PowerControlUnit();
};

#endif // SYSTEMS_DESIGN_SRC_COMMON_AIRCRAFTSYSTEMS_HIGHLIFTSYSTEM_POWERCONTROLUNIT_H_
