/*
 * UNICADO - UNIversity Conceptual Aircraft Design and Optimization
 *
 * Copyright (C) 2025 UNICADO consortium
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 * Description:
 * This file is part of UNICADO.
 */

#include "highLiftSystem.h"

#include <runtimeInfo/runtimeInfo.h>
#include <aircraftGeometry2/processing/transform.h>


highLiftSystem::highLiftSystem(const systemsIOData::Wing& Wing, bool controlDeviceWarning, const std::vector<conventionalATA27::flightControlSystem::HighLiftSystem>& highLift,
                               double airDensity, double airSpeed, double theMTOM)
    :
    theControlDeviceWarning(controlDeviceWarning),
    theWing(&Wing) {
    sortControlDevices(); //function sorts control devices into leading and trailing edge
    getTEDevices(); //Creating the trailing Edge Device Panel
    getLEDevices(); //Creating the leading Edge Device Panel
    for (size_t n = 0; n < Systems.size(); n++) {
        double totalActuatorWeight(0.);
        for (size_t i = 0; i < Systems.at(n).Device.size(); i++) {
            //Determining the number of actuator stations per Device
            int numberStations(2);
            for (int j = 0; j <= numberStations - 1; j++) {
                //Determination of actuator and kinematics mass per device
                Systems.at(n).Device.at(i).ActuationSupport.push_back(ActuationAndSupport(Systems.at(n).Type,
                        Systems.at(n).Device.at(i).l_m_Device,
                        Systems.at(n).Device.at(i).Sref,
                        numberStations,
                        airDensity,
                        airSpeed));
                Systems.at(n).Device.at(i).ActuationSupport.back().getMass(Systems.at(n).Type, Systems.at(n).Device.at(i).Sref, numberStations);
                totalActuatorWeight += Systems.at(n).Device.at(i).ActuationSupport.back().Mass;
                //Determining the kinematics of each actuator
                if (Systems.at(n).Type == "Slat") {
                    Systems.at(n).Device.at(i).ActuationSupport.back().calcActuationMoments(Systems.at(n).Type);
                } else if (Systems.at(n).Type == "Fowler") {
                    Systems.at(n).Device.at(i).ActuationSupport.back().calcActuationMoments(Systems.at(n).Type);
                }
                //Determination of the work of each actuator per actuation
                Systems.at(n).Device.at(i).ActuationSupport.back().calcWorkLoad();
                //Determination of the power of each actuator per actuation
                Systems.at(n).Device.at(i).ActuationSupport.back().calcPower();
            }
        }
        // Determining the PCU mass
        Systems.at(n).PCU.push_back(PowerControlUnit(Systems.at(n).Type));
        double wingRefArea = geom2::measure::reference_area(theWing->aerodynamicSurface);
        Systems.at(n).PCU.back().getInstalledPower(theMTOM, wingRefArea);
        Systems.at(n).PCU.back().getMass(Systems.at(n).Type, wingRefArea);
        //Assignment of the PCU motors
        for (size_t m(0); m < highLift.at(n).PCU_Motors.size(); m++) {
            Systems.at(n).PCU.back().PCU_Motor.push_back(PowerControlUnit::PowerControlUnitMotor());
            Systems.at(n).PCU.back().PCU_Motor.back().refPoint[0] = theWing->aerodynamicSurface.origin[0] + 0.5 * theWing->aerodynamicSurface.sections.front().get_chord_length();
            Systems.at(n).PCU.back().PCU_Motor.back().refPoint[1] = theWing->aerodynamicSurface.origin[1];
            Systems.at(n).PCU.back().PCU_Motor.back().refPoint[2] = theWing->aerodynamicSurface.origin[2];
            Systems.at(n).PCU.back().PCU_Motor.back().powerSource = highLift.at(n).PCU_Motors.at(m).powerSource.type;
            Systems.at(n).PCU.back().PCU_Motor.back().powerSourceID = highLift.at(n).PCU_Motors.at(m).powerSource.sourceID;
            Systems.at(n).PCU.back().PCU_Motor.back().operationMode = highLift.at(n).PCU_Motors.at(m).operationMode;
            Systems.at(n).PCU.back().PCU_Motor.back().power += highLift.at(n).PCU_Motors.at(m).stdbyPower;
            if (Systems.at(n).PCU.back().PCU_Motor.back().operationMode == "active") {
                Systems.at(n).PCU.back().PCU_Motor.back().power += highLift.at(n).PCU_Motors.at(m).activePower;
            }
        }
        //Determining the WTB mass
        Systems.at(n).WTB.getMass(Systems.at(n).Type, Systems.at(n).PCU.back().installedPower);
        //Determining the gearbox masses
        Systems.at(n).Gearboxes.push_back(theGearboxes(Systems.at(n).Type));
        Systems.at(n).Gearboxes.back().getMass(Systems.at(n).Type, theMTOM, wingRefArea);
        //Determining the TorqueShaft masses
        Systems.at(n).TorqueShaft.push_back(TransmissionShaft(Systems.at(n).Type));
        Systems.at(n).TorqueShaft.back().getMass(Systems.at(n).Type, geom2::measure::span(theWing->aerodynamicSurface));
        //Determining the TorqueLimiter masses
        Systems.at(n).TorqueLimiters.push_back(theTorqueLimiters(Systems.at(n).Type));
        Systems.at(n).TorqueLimiters.back().getMass(Systems.at(n).Type, totalActuatorWeight);
        //Determining the total system mass
        Systems.at(n).systemMass = getSystemMass(n);
        //Determination of the required power of all actuators of this HL system
        Systems.at(n).requiredPowerActuators = getRequiredPowerActuators(n);
        //Determining the required power of the PCU
        Systems.at(n).requiredPowerPCU = getRequiredPowerPCU(n);
        //Determination of all maximum actuator loads
        Systems.at(n).requiredMomentActuatorsMax = getRequiredMomentActuatorsMax(n);
        //Determining the maximum PCU torque
        Systems.at(n).requriedMomentPCUMax = getRequriedMomentPCUMax(n);
        //Determining the maximum PCU power (dimensioning)
        Systems.at(n).maximumPowerPCU = getMaximumPowerPCU(n);
//        testOutput(n);
    }
    //ctor
}

highLiftSystem::~highLiftSystem() {
    //dtor
}

void highLiftSystem::sortControlDevices() {
    for (size_t i = 0; i < theWing->controlSurface.size(); i++) {
        /* check if control surface relative chord position is in front half of wing: leading edge device */
        geom2::Point_3 leadingEdgePoint = geom2::measure::left(theWing->controlSurface.at(i), theWing->controlSurface.at(i).sections.front().origin.z());
        if (leadingEdgePoint.x() >= 0. && leadingEdgePoint.x() < 0.5) {
            LEDevices.push_back(theWing->controlSurface.at(i));
        /* check if control surface relative chord position is in rear half of wing: trailing edge device */
        } else if (leadingEdgePoint.x() >= 0.5) {
            TEDevices.push_back(theWing->controlSurface.at(i));
        }
    }
}

// TODO(student#6#): Mantis ID 0000198: Actuator Massenbestimmung fuer MorphingDroopNose und MorphingTrailingEdge hinzufuegen
void highLiftSystem::getLEDevices() {
    // Leading Edge System
    if (LEDevices.size() > 0) {
        Systems.push_back(HLSystem());
        // TODO(student#1#): Mantis ID 0000214: Insert Krueger here
        for (size_t i = 0; i < LEDevices.size(); i++) {
            Systems.back().Edge = "LE";
            if (LEDevices.at(i).name == "Slat" || LEDevices.at(i).name == "slat") {
                Systems.back().Device.push_back(highLiftDevice());
                Systems.back().Type = "Slat";
                //control device has to be transformed to absolute coordinates for measurements
                geom2::MultisectionSurface<geom2::PolygonSection> absolute_device = geom2::transform::to_absolute(LEDevices.at(i), theWing->aerodynamicSurface);
                Systems.back().Device.back().Sref = geom2::measure::area(absolute_device);
                Systems.back().Device.back().l_m_Device = geom2::measure::width(absolute_device, absolute_device.sections.front().origin.z());
                //sum up the entire panel area
                Systems.back().systemSref += Systems.back().Device.back().Sref;
            } else if (LEDevices.at(i).name == "Special" || LEDevices.at(i).name == "special" ||
                       LEDevices.at(i).name == "Krueger" || LEDevices.at(i).name == "krueger" ||
                       LEDevices.at(i).name == "DroopNose" || LEDevices.at(i).name == "droop_nose") {
                if (theControlDeviceWarning) {
                    myRuntimeInfo->warn << "No specific method for LE High Lift System " << LEDevices.at(i).name << std::endl;
                    myRuntimeInfo->warn << "Device " << LEDevices.at(i).name << " is treated as Slat!" << std::endl;
                }
                Systems.back().Device.push_back(highLiftDevice());
                Systems.back().Type = "Slat";
                geom2::MultisectionSurface<geom2::PolygonSection> absolute_device = geom2::transform::to_absolute(LEDevices.at(i), theWing->aerodynamicSurface);
                Systems.back().Device.back().Sref = geom2::measure::area(absolute_device);
                double l_i = geom2::measure::width(absolute_device, absolute_device.sections.front().origin.z()); // width of first section
                double l_a = geom2::measure::width(absolute_device, absolute_device.sections.back().origin.z()); // width of second section
                double taper = l_a / l_i;
                double MAC = 2. / 3. * l_i * (1. + taper + taper * taper) / (1 + taper);
                Systems.back().Device.back().l_m_Device = MAC;
                //sum up the entire panel area
                Systems.back().systemSref += Systems.back().Device.back().Sref;
            } else if (LEDevices.at(i).name == "morphingDroopNose" || LEDevices.at(i).name == "morphing_droop_nose") {
                myRuntimeInfo->err << "No method for morphingDroopNose implemented yet!" << std::endl;
                myRuntimeInfo->err << "Abort!" << std::endl;
                exit(1);
            } else {
                myRuntimeInfo->err << "The LE-System is unknown: " << LEDevices.at(i).name << std::endl;
                myRuntimeInfo->err << "Abort!" << std::endl;
                exit(1);
            }
        }
        if (Systems.back().systemSref <= 0.0 && Systems.back().Device.size() > 0) {
            myRuntimeInfo->err << "The slat area cannot be smaller or equal to zero!" << std::endl;
            myRuntimeInfo->err << "Slat area = " << Systems.back().systemSref << std::endl;
            myRuntimeInfo->err << "Abort!" << std::endl;
            exit(1);
        }
    }
}

void highLiftSystem::getTEDevices() {
    // Trailing Edge System
    if (TEDevices.size() > 0) {
        Systems.push_back(HLSystem());
        for (size_t i = 0; i < TEDevices.size(); i++) {
            if (TEDevices.at(i).name != "Fowler" && TEDevices.at(i).name != "fowler" && TEDevices.at(i).name != "flap_fowler"
                    && TEDevices.at(i).name != "Aileron" && TEDevices.at(i).name != "aileron"
                    && TEDevices.at(i).name != "DroopAileron" && TEDevices.at(i).name != "droop_aileron"
                    && TEDevices.at(i).name != "SpoilerGround" && TEDevices.at(i).name != "spoiler_ground"
                    && TEDevices.at(i).name != "SpoilerAir" && TEDevices.at(i).name != "spoiler_air"
                    && TEDevices.at(i).name != "morphingTrailingEdge" && TEDevices.at(i).name != "morphing_trailing_edge") {
                if (theControlDeviceWarning) {
                    myRuntimeInfo->warn << "No specific method for TE High Lift System " << TEDevices.at(i).name << std::endl;
                    myRuntimeInfo->warn << "Device " << TEDevices.at(i).name << " is treated as Fowler!" << std::endl;
                }
            }
            Systems.back().Edge = "TE";
            if (TEDevices.at(i).name == "Fowler" || TEDevices.at(i).name == "fowler" || TEDevices.at(i).name == "flap_fowler"
                    || TEDevices.at(i).name == "Special" || TEDevices.at(i).name == "special"
                    || TEDevices.at(i).name == "DoubleFowler" || TEDevices.at(i).name == "double_fowler"
                    || TEDevices.at(i).name == "TripleFowler" || TEDevices.at(i).name == "triple_fowler"
                    || TEDevices.at(i).name == "Flap" || TEDevices.at(i).name == "flap"
                    || TEDevices.at(i).name == "ADHF"
                    || TEDevices.at(i).name == "SlottedFlap" || TEDevices.at(i).name == "slotted_flap") {
                Systems.back().Device.push_back(highLiftDevice());
                Systems.back().Type = "Fowler";
                geom2::MultisectionSurface<geom2::PolygonSection> absolute_device = geom2::transform::to_absolute(TEDevices.at(i), theWing->aerodynamicSurface);
                Systems.back().Device.back().Sref = geom2::measure::area(absolute_device);
                double l_i = geom2::measure::width(absolute_device, absolute_device.sections.front().origin.z()); // width of first section
                double l_a = geom2::measure::width(absolute_device, absolute_device.sections.back().origin.z()); // width of second section
                double taper = l_a / l_i;
                double MAC = 2. / 3. * l_i * (1. + taper + taper * taper) / (1 + taper);
                Systems.back().Device.back().l_m_Device = MAC;
                //Summieren der gesamten Panelflaeche
                Systems.back().systemSref += Systems.back().Device.back().Sref;
            }
            if (TEDevices.at(i).name == "morphingTrailingEdge" || TEDevices.at(i).name == "morphing_trailing_edge") {
                myRuntimeInfo->err << "No method for morphingDroopNose implemented yet!" << std::endl;
                myRuntimeInfo->err << "Abort!" << std::endl;
                exit(1);
            }
        }
        if (Systems.back().Device.size() == 0) {
            myRuntimeInfo->warn << "No 'Fowler' available!" << std::endl;
        }
        if (Systems.back().systemSref <= 0.0) {
            myRuntimeInfo->warn << "The flap area cannot be smaller or equal to zero!" << std::endl;
            myRuntimeInfo->warn << "flap area = " << Systems.back().systemSref << std::endl;
        }
    }
}

double highLiftSystem::getSystemMass(int SystemID) {
    double Mass(0.0);
    for (size_t i = 0; i < Systems.at(SystemID).Device.size(); i++) {
        for (size_t j = 0; j < Systems.at(SystemID).Device.at(i).ActuationSupport.size(); j++) {
            //Twice (once for each wing side)
            Mass += 2. * Systems.at(SystemID).Device.at(i).ActuationSupport.at(j).Mass;
        }
    }
    Mass +=   Systems.at(SystemID).PCU.back().Mass
              + Systems.at(SystemID).TorqueShaft.back().Mass
              + Systems.at(SystemID).Gearboxes.back().Mass
              + Systems.at(SystemID).TorqueLimiters.back().Mass
              + Systems.at(SystemID).WTB.Mass;
    return Mass;
}

double highLiftSystem::getRequiredPowerActuators(int SystemID) {
    double Power(0.0);
    for (size_t i = 0; i < Systems.at(SystemID).Device.size(); i++) {
        for (size_t j = 0; j < Systems.at(SystemID).Device.at(i).ActuationSupport.size(); j++) {
            Power += Systems.at(SystemID).Device.at(i).ActuationSupport.at(j).actuationPower
                     / Systems.at(SystemID).Device.at(i).ActuationSupport.at(j).efficiencyFactor;
        }
    }
    //Power of the entire shaft to the actuator (2 times)
    return 2 * Power;
}

double highLiftSystem::getRequiredPowerPCU(int SystemID) {
    double Power(0.0);
    Power = Systems.at(SystemID).requiredPowerActuators
            / Systems.at(SystemID).TorqueShaft.back().efficiencyFactor
            / Systems.at(SystemID).TorqueLimiters.back().efficiencyFactor
            / Systems.at(SystemID).Gearboxes.back().efficiencyFactor
            / Systems.at(SystemID).PCU.back().efficiencyFactorMax;
    // Power of the entire shaft
    return Power;
}

double highLiftSystem::getRequiredMomentActuatorsMax(int SystemID) {
    double Moment(0.0);
    for (size_t i = 0; i < Systems.at(SystemID).Device.size(); i++) {
        for (size_t j = 0; j < Systems.at(SystemID).Device.at(i).ActuationSupport.size(); j++) {
            Moment += Systems.at(SystemID).Device.at(i).ActuationSupport.at(j).maxActuationMoment
                      / (Systems.at(SystemID).Device.at(i).ActuationSupport.at(j).efficiencyFactor
                         * Systems.at(SystemID).TorqueShaft.back().rotationSpeedShaft
                         /  Systems.at(SystemID).Device.at(i).ActuationSupport.at(j).rotationSpeedActuator);
        }
    }
    // Moment of the entire actuator (2 times)
    return 2 * Moment;
}

double highLiftSystem::getRequriedMomentPCUMax(int SystemID) {
    double MomentMin(0.0);
    double MomentMax(0.0);
    //[Gro09], p. 20
    MomentMin = (Systems.at(SystemID).PCU.back().safetyFactor * Systems.at(SystemID).requiredMomentActuatorsMax)
                /
                (Systems.at(SystemID).TorqueShaft.back().efficiencyFactor
                 * Systems.at(SystemID).TorqueLimiters.back().efficiencyFactor
                 * Systems.at(SystemID).Gearboxes.back().efficiencyFactor
                 * Systems.at(SystemID).PCU.back().rotationSpeedPCU / Systems.at(SystemID).TorqueShaft.back().rotationSpeedShaft);
    MomentMax = Systems.at(SystemID).PCU.back().hydraulicPressureMax / Systems.at(SystemID).PCU.back().hydraulicPressureMin
                *
                Systems.at(SystemID).PCU.back().efficiencyFactorMax / Systems.at(SystemID).PCU.back().efficiencyFactorMin
                *
                MomentMin;
    return MomentMax;
}

double highLiftSystem::getMaximumPowerPCU(int SystemID) {
    double maxPower(0.0);
    maxPower = Systems.at(SystemID).PCU.back().rotationSpeedPCU / Systems.at(SystemID).PCU.back().efficiencyFactorMax
               * Systems.at(SystemID).requriedMomentPCUMax;
    return maxPower;
}

void highLiftSystem::testOutput(int SystemID) {
    myRuntimeInfo->out << " " << Systems.at(SystemID).Type << " - SYSTEM " << std::endl;
    myRuntimeInfo->out << "---------------------" << std::endl;
    myRuntimeInfo->out << "---------------------" << std::endl;
    myRuntimeInfo->out << "  Sref:                   " << Systems.at(SystemID).systemSref << " m^2" << std::endl;
    myRuntimeInfo->out << "  Mass:                   " << Systems.at(SystemID).systemMass << " kg" << std::endl;
    myRuntimeInfo->out << "  maximum Power PCU:      " << Systems.at(SystemID).maximumPowerPCU << " Watt" << std::endl;
    myRuntimeInfo->out << "  Power PCU (deployment): " << Systems.at(SystemID).requiredPowerPCU << " Watt" << std::endl;
    myRuntimeInfo->out << std::endl;
    myRuntimeInfo->out << std::endl;
    myRuntimeInfo->out << " Components " << std::endl;
    myRuntimeInfo->out << "---------------------" << std::endl;
    myRuntimeInfo->out << " Device:" << std::endl;
    for (size_t i = 0; i < Systems.at(SystemID).Device.size(); i++) {
        myRuntimeInfo->out << "      " << Systems.at(SystemID).Type << " " << i << std::endl;
        myRuntimeInfo->out << "          - Sref: " << Systems.at(SystemID).Device.at(i).Sref << " m^2" << std::endl;
        for (size_t j = 0; j < Systems.at(SystemID).Device.at(i).ActuationSupport.size(); j++) {
            myRuntimeInfo->out << "          - Actuation & Support Mass  " << j << " "
                               << Systems.at(SystemID).Device.at(i).ActuationSupport.at(j).Mass << " kg" << std::endl;
            myRuntimeInfo->out << "          - Actuation Max Torque      " << j << " "
                               << Systems.at(SystemID).Device.at(i).ActuationSupport.at(j).maxActuationMoment << " Nm" << std::endl;
            myRuntimeInfo->out << "          - Actuation Work Load       " << j << " "
                               << Systems.at(SystemID).Device.at(i).ActuationSupport.at(j).actuationWork << " Joule" << std::endl;
        }
    }
    myRuntimeInfo->out << " Power Control Unit:" << std::endl;
    myRuntimeInfo->out << "          - PCU Mass      " << Systems.at(SystemID).PCU.back().Mass << " kg" << std::endl;
    myRuntimeInfo->out << " Torque Shaft:" << std::endl;
    myRuntimeInfo->out << "          - Shaft Mass    " << Systems.at(SystemID).TorqueShaft.back().Mass << " kg" << std::endl;
    myRuntimeInfo->out << " Gearboxes:" << std::endl;
    myRuntimeInfo->out << "          - Gearbox Mass  " << Systems.at(SystemID).Gearboxes.back().Mass << " kg" << std::endl;
    myRuntimeInfo->out << " Torque Limiters:" << std::endl;
    myRuntimeInfo->out << "          - Limiter Mass  " << Systems.at(SystemID).TorqueLimiters.back().Mass << " kg" << std::endl;
    myRuntimeInfo->out << " Wing Tip Brake:" << std::endl;
    myRuntimeInfo->out << "          - WTB Mass  " << Systems.at(SystemID).WTB.Mass << " kg" << std::endl;
    myRuntimeInfo->out << "---------------------" << std::endl;
    myRuntimeInfo->out << "---------------------" << std::endl;
    myRuntimeInfo->out << std::endl;
    myRuntimeInfo->out << std::endl;
}
