# UNICADO - UNIversity Conceptual Aircraft Design and Optimization
#
# Copyright (C) 2025 UNICADO consortium
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#
# Description:
# This file is part of UNICADO.

"""Module containing functions for the preparation of geometrical data necessary for tank design."""
# Import standard libraries.
from math import pi as PI
from matplotlib import pyplot as plt
import sys
import numpy as np


def prepare_geometrical_data(paths_and_names, dict_ac_data, dict_tank_design, runtime_output):
    """Prepare geometrical data for further use.

    This function calls the according preparation function depending on the energy carrier.
    In case of KEROSENE, the function adds the following data to the 'dict_tank_design':
        dict_tank_design = {
            'tank_0': {...,
                'geometry':{
                    'inner_surface': value,
                    'outer_surface': value}},
            ...,
            'geometry':{
                'fuselage': {
                    'fuselage_width': value},
                'wing': {
                    'section_0': {
                        'length_chord': value,
                        'relative_position_front_spar': value,
                        'relative_position_rear_spar': value,
                        'y_position_: value,
                        'thickness_ratio': value},
                    ...},
                'horizontal_stabilizer': {}
            }
        }

    If LIQUID HYDROGEN is used, the following parameters are added to the dict 'dict_tank_design':
        dict_tank_design = {}

    :param dict paths_and_names: Dictionary containing system paths and ElementTrees
    :param dict dict_ac_data: Dict containing parameter and values from aircraft exchange and module configuration file
    :param dict dict_tank_design: Dict containing results from tank design
    :param logging.Logger runtime_output: Logging object used for capturing log messages in the module
    :return dict dict_tank_design: Dict containing results from tank design
    """

    match dict_ac_data['energy_carrier']:
        case 'kerosene':
            # Kerosene: Prepare geometrical data for integral wing tanks and trim/additional center tanks if selected.
            dict_tank_design = prepare_for_kerosene(dict_ac_data, dict_tank_design, runtime_output)
        case 'liquid_hydrogen':
            # Liquid hydrogen: Prepare geometrical data for cryogenic liquid hydrogen tanks.
            dict_tank_design = prepare_for_liquid_hydrogen(dict_ac_data, dict_tank_design, runtime_output)

    return dict_tank_design


"""Prepare data for kerosene tanks."""
def prepare_for_kerosene(dict_ac_data, dict_tank_design, runtime_output):
    """Prepare geometrical data for kerosene tanks.

    :param dict dict_ac_data: Dict containing parameter and values from aircraft exchange and module configuration file
    :param dict dict_tank_design: Dictionary containing tank design parameter
    :param logging.Logger runtime_output: Logging object used for capturing log messages in the module
    :return dict dict_tank_design: Dictionary containing tank design parameter
    """
    # Initialize empty list in dictionary.
    dict_tank_design['geometry'] = {}

    """Prepare tank values according to tank locations."""
     # Wing tanks.
    if 'wing' in dict_tank_design['tank_configuration']:
        dict_tank_design = prepare_wing_tanks(dict_ac_data, dict_tank_design, runtime_output)
    # Trim tank.
    if 'trim' in dict_tank_design['tank_configuration']:
        dict_tank_design = prepare_trim_tank(dict_ac_data, dict_tank_design, runtime_output)
    # Additional center tank.
    if 'additional_center' in dict_tank_design['tank_configuration']:
        dict_tank_design = prepare_additional_center_tank(dict_ac_data, dict_tank_design, runtime_output)

    # Debug print.
    runtime_output.debug('Debug: Geometrical data for all kerosene tanks prepared.')

    return dict_tank_design


def prepare_wing_tanks(dict_ac_data, dict_tank_design, runtime_output):
    """Extract and process wing values.
    
    This function adds the 'wing' sub dictionary to the 'geometry' sub dictionary:
        dict_tank_design = {
            ...,
            'geometry':{
                'wing': {
                    'fuselage_center_line': {},
                    'wing_root': {},
                    'kink': {}, -> only if wing planform is compound tapered
                    'wing_tip': {},
                    'wing_planform': (str, information about wing planform),
                    'wing_symmetry': (bool, information on wing symmetry),
                    'constant_leading_edge_sweep': (bool, information on constant leading edge sweep),
                    'absolute_x_position': (float, x coordinate of complete wing in m),
                    (*) interpolated value lists
                }
            }
        }
    The three/four sub dictionaries contain the following values (exemplified by the 'wing_tip' dictionary):
        'wing_tip': {
            'length_chord': (float, chord length of segment in m),
            'x_position': (float, relative x coordinate of foremost wing section point in m),
            'y_position': (float, relative y coordinate of foremost wing section point in m),
            'z_position': (float, relative z coordinate of foremost wing section point in m),
            'profile' : (str, name of wing section profile),
            'relative_position_front_spar': (float, relative position of front spar in percent),
            'relative_position_rear_spar': (float, relative position of rear spar in percent),
            'thickness_ratio': (float, profile thickness ratio, derived from profile name in percent)
            }
        }
    (*) There are lists with interpolated values for the section tank area, chord lengths, relative front spar
    positions, maximum tank lengths and heights, and the leading edge sweep. Those lists each contain one value per
    mm span.

    :param dict dict_ac_data: Dict containing parameter and values from aircraft exchange and module configuration file
    :param dict dict_tank_design: Dictionary containing tank design parameter
    :param logging.Logger runtime_output: Logging object used for capturing log messages in the module
    :return dict dict_tank_design: Dictionary containing tank design parameter
    """
    # TODO: Divide dict into 'left' and 'right'?

    """Extract wing information for further operations."""
    # Extract wing ID and information on wing symmetry.
    wing_id = next(key for key, value in dict_ac_data['wing_name'].items()
                if value == 'main_wing').split('ID')[-1]
    wing_symmetry = dict_ac_data['wing_symmetry']['wing_symmetry_ID' + str(wing_id)]
    if not wing_symmetry:
        # Not implemented yet.
        runtime_output.critical('Error: Asymmetrical wing not implemented yet! '
                                + 'Program aborted.')
        sys.exit('Exit information: Tank design not possible for asymmetrical wing!')

    wing_sections_from_aircraft_xml = [key for key in dict_ac_data['wing_section_chord_origin_x'].keys()
                                 if key.split('_')[-2] == f'ID{str(wing_id)}']
    number_of_wing_sections = len(wing_sections_from_aircraft_xml)
    wing_section_ids = ['ID' + str(wing_id) + '_ID' + wing_sections_from_aircraft_xml[i][-1]
                        for i in range(number_of_wing_sections)]

    # Prepare data for further operation depending on number of wing sections.
    if number_of_wing_sections < 3:
        runtime_output.critical('Error: Not enough wing sections given. Program aborted.')
        sys.exit('Exit information: Not enough wing sections given!')
    elif number_of_wing_sections == 3:
        # Single tapered wing w/o kink.
        wing_planform = 'single_tapered'
        wing_section_list = ['fuselage_center_line', 'wing_root', 'wing_tip']
    elif number_of_wing_sections == 4:
        # Compound tapered wing w/ kink.
        wing_planform = 'compound_tapered'
        wing_section_list = ['fuselage_center_line', 'wing_root', 'kink', 'wing_tip']
    else:
        # Not implemented yet.
        runtime_output.critical('Error: Too many wing sections! Program aborted.')
        sys.exit('Exit information: Too many wing sections!')

    """Create wing geometry dictionary."""
    # Automatic creation of wing geometry dictionary.
    parameter_list = ['chord_length', 'profile']
    position_parameters = ['x', 'y', 'z']
    base_keys = {
        'x': 'wing_section_chord_origin_x',
        'y': 'wing_section_chord_origin_y',
        'z': 'wing_section_chord_origin_z',
        'chord_length': 'wing_section_chord_length',
        'profile': 'wing_section_profile'
    }

    # Create tmp_dict with 'position' sub dictionary for x, y, z positions.
    tmp_dict = {
        'sections': {
            wing_section_list[i]: {
                'position': {
                    pos_param: dict_ac_data[base_keys[pos_param]][f"{base_keys[pos_param]}_{wing_section_ids[i]}"]
                    for pos_param in position_parameters
                },
                **{
                    param: dict_ac_data[base_keys[param]][f"{base_keys[param]}_{wing_section_ids[i]}"]
                    for param in parameter_list
                }
            }
            for i in range(len(wing_section_list))
        }
    }

    # Add further values to 'tmp_dict'.
    for i, section in enumerate(wing_section_list):
        # Add span index to section.
        tmp_dict['sections'][section]['span_index'] = abs(int(round(
            tmp_dict['sections'][section]['position']['y']*100, 0)))
        # Add thickness_ratio.
        tmp_dict['sections'][section]['thickness_ratio'] = dict_ac_data['wing_thickness_to_length'][i]

    # Add 'tmp_dict' to 'dict_tank_design'.
    dict_tank_design['geometry']['wing'] = tmp_dict

    """Interpolate necessary values."""
    # Calculate number of points for value interpolation (one point per mm).
    number_of_points = abs(int(round(dict_tank_design['geometry']['wing']['sections']['wing_tip']['position']['y']
                                     *100, 0)))
    # Write wing section span indexes (equal distances in mm) to list.
    section_span_indexes = [tmp_dict['sections'][section]['span_index'] for section in wing_section_list]

    # Interpolate spar positions.
    wing_span_in_mm = int(round(dict_tank_design['geometry']['wing']['sections']['wing_tip']['position']['y']*100, 0))
    front_spar_span_indexes = [
        int(dict_ac_data['spar_position_inner_spanwise']['spar_position_inner_spanwise_ID0_ID0']*wing_span_in_mm),
        int(dict_ac_data['spar_position_outer_spanwise']['spar_position_outer_spanwise_ID0_ID0']*wing_span_in_mm)]
    front_spar_relative_position = [
        dict_ac_data['spar_position_inner_chord']['spar_position_inner_chord_ID0_ID0'],
        dict_ac_data['spar_position_outer_chord']['spar_position_outer_chord_ID0_ID0']]
    relative_position_front_spar = np.interp(range(number_of_points + 1), front_spar_span_indexes,
                                             front_spar_relative_position)
    rear_spar_span_indexes = [
        int(dict_ac_data['spar_position_inner_spanwise']['spar_position_inner_spanwise_ID0_ID1']*wing_span_in_mm),
        int(dict_ac_data['spar_position_outer_spanwise']['spar_position_outer_spanwise_ID0_ID1']*wing_span_in_mm)]
    rear_spar_relative_position = [
        dict_ac_data['spar_position_inner_chord']['spar_position_inner_chord_ID0_ID1'],
        dict_ac_data['spar_position_outer_chord']['spar_position_outer_chord_ID0_ID1']]
    relative_position_rear_spar = np.interp(range(number_of_points + 1), rear_spar_span_indexes,
                                             rear_spar_relative_position)
    x_coordinate_leading_edge = [dict_tank_design['geometry']['wing']['sections'][section]['position']['x']
                                for section in wing_section_list]
    interpolated_x_coordinate_leading_edge = np.interp(range(number_of_points + 1), section_span_indexes,
                                                 x_coordinate_leading_edge)
    # Calculate absolute front spar position.
    absolute_position_front_spar_list = [(x_coordinate_leading_edge[idx]
                                        + dict_tank_design['geometry']['wing']['sections'][name]['chord_length']
                                        *relative_position_front_spar[idx])
                                        for idx, name in enumerate(wing_section_list)]
    absolute_position_rear_spar_list = [(x_coordinate_leading_edge[idx]
                                          + dict_tank_design['geometry']['wing']['sections'][name]['chord_length']
                                          *relative_position_rear_spar[idx])
                                         for idx, name in enumerate(wing_section_list)]
    # Interpolate values for absolute front/rear spar positions and tank length.
    interpolated_absolute_front_spar = np.interp(range(number_of_points + 1), section_span_indexes,
                                                        absolute_position_front_spar_list)
    interpolated_absolute_rear_spar = np.interp(range(number_of_points + 1), section_span_indexes,
                                                        absolute_position_rear_spar_list)
    interpolated_tank_lengths = interpolated_absolute_rear_spar - interpolated_absolute_front_spar
    # Tank height interpolation.
    tank_height_list = [(dict_tank_design['geometry']['wing']['sections'][section]['chord_length']
                         *dict_tank_design['geometry']['wing']['sections'][section]['thickness_ratio']
                         *dict_ac_data['a_to_d_factor']) for section in wing_section_list]
    interpolated_tank_heights = np.interp(range(number_of_points + 1), section_span_indexes, tank_height_list)
    # z coordinate interpolation.
    z_coordinate_leading_edge = [dict_tank_design['geometry']['wing']['sections'][section]['position']['z']
                                for section in wing_section_list]
    interpolated_z_coordinate_leading_edge = np.interp(range(number_of_points + 1), section_span_indexes,
                                                 z_coordinate_leading_edge)

    # Write data to 'dict_tank_design'.
    dict_tank_design['geometry']['wing']['interpolated_values'] = {}
    dict_tank_design['geometry']['wing']['interpolated_values']['x_coordinate_leading_edge'] = (
        interpolated_x_coordinate_leading_edge)
    dict_tank_design['geometry']['wing']['interpolated_values']['z_coordinate_leading_edge'] = (
        interpolated_z_coordinate_leading_edge)
    dict_tank_design['geometry']['wing']['interpolated_values']['x_coordinate_front_spar'] = (
        interpolated_absolute_front_spar)
    dict_tank_design['geometry']['wing']['interpolated_values']['x_coordinate_rear_spar'] = (
        interpolated_absolute_rear_spar)
    dict_tank_design['geometry']['wing']['interpolated_values']['tank_lengths'] = interpolated_tank_lengths
    dict_tank_design['geometry']['wing']['interpolated_values']['tank_heights'] = interpolated_tank_heights

    # Write maximum tank height and length to dictionary.
    for section in wing_section_list:
        dict_tank_design['geometry']['wing']['sections'][section]['maximum_tank_height'] = (
            float(interpolated_tank_heights[dict_tank_design['geometry']['wing']['sections'][section]['span_index']]))
        dict_tank_design['geometry']['wing']['sections'][section]['maximum_tank_length'] = (
            float(interpolated_tank_lengths[dict_tank_design['geometry']['wing']['sections'][section]['span_index']]))

    """Calculate leading edge sweep angle."""
    number_of_wing_segments = number_of_wing_sections - 1
    x_tmp, y_tmp, sweep_leading_edge = [], [], [0]
    # Store x and y coordinates to temporary lists.
    for section in dict_tank_design['geometry']['wing']['sections'].keys():
        x_tmp.append(dict_tank_design['geometry']['wing']['sections'][section]['position']['x'])
        y_tmp.append(dict_tank_design['geometry']['wing']['sections'][section]['position']['y'])
    # Calculate sweep angles at wing sections.
    for i in range(number_of_wing_segments):
        opposite_leg = x_tmp[i+1] - x_tmp[i]
        adjacent_leg = y_tmp[i+1] - y_tmp[i]
        phi = (opposite_leg/adjacent_leg) * 180/PI
        sweep_leading_edge.append(phi)
    # Set 'constant_leading_edge_sweep' parameter.
    if wing_planform == 'single_tapered':
        constant_leading_edge_sweep = True
    elif wing_planform == 'compound_tapered':
        constant_leading_edge_sweep = all(round(x, 3) == round(sweep_leading_edge[2], 3)
                                           for x in sweep_leading_edge[2:])

    # Prepare output.
    dict_tank_design['geometry']['wing']['information'] = {}
    dict_tank_design['geometry']['wing']['information']['wing_planform'] = wing_planform
    dict_tank_design['geometry']['wing']['information']['wing_symmetry'] = wing_symmetry
    dict_tank_design['geometry']['wing']['information']['wing_section_list'] = wing_section_list
    dict_tank_design['geometry']['wing']['position'] = {}
    dict_tank_design['geometry']['wing']['position']['x'] = dict_ac_data['wing_position_x']
    dict_tank_design['geometry']['wing']['position']['y'] = dict_ac_data['wing_position_y']
    dict_tank_design['geometry']['wing']['position']['z'] = dict_ac_data['wing_position_z']
    dict_tank_design['geometry']['wing']['information']['constant_leading_edge_sweep'] = constant_leading_edge_sweep

    # Debug print.
    runtime_output.debug('Debug: Geometrical data for wing tanks prepared.')

    return dict_tank_design


def prepare_trim_tank(dict_ac_data, dict_tank_design, runtime_output):
    """Extract horizontal stabilizer geometry for trim tank.

    :param _type_ dict_ac_data: _description_
    :param _type_ dict_tank_design: _description_
    :param _type_ runtime_output: _description_
    :return _type_: _description_
    """
    horizontal_stabilizer_id = next(key for key, value in dict_ac_data['empennage_name'].items()
            if value in ('horizontal_stabilizer', 'horizontal_stabiliser')).split('ID')[-1]
    symmetry = dict_ac_data['empennage_symmetry']['empennage_symmetry_ID'
                                                                                + str(horizontal_stabilizer_id)]
    if not symmetry:
        # Not implemented yet.
        runtime_output.critical('Error: Asymmetrical horizontal stabilizer not implemented yet! '
                                + 'Program aborted.')
        sys.exit('Exit information: Tank design not possible for asymmetrical horizontal stabilizer!')

    horizontal_stabilizer_sections_from_aircraft_xml = [key for key in
                                                        dict_ac_data['empennage_section_chord_origin_x'].keys()
                                                        if key.split('_')[-2] == f'ID{str(horizontal_stabilizer_id)}']
    number_of_horizontal_stabilizer_sections = len(horizontal_stabilizer_sections_from_aircraft_xml)
    horizontal_stabilizer_section_ids = ['ID' + str(horizontal_stabilizer_id) + '_ID'
                                         + horizontal_stabilizer_sections_from_aircraft_xml[i][-1]
                                         for i in range(number_of_horizontal_stabilizer_sections)]
    # Prepare data for further operation depending on number of wing sections.
    if number_of_horizontal_stabilizer_sections < 2:
        runtime_output.critical('Error: Not enough horizontal stabilizer sections given. Program aborted.')
        sys.exit('Exit information: Not enough horizontal stabilizer sections given!')
    elif number_of_horizontal_stabilizer_sections == 2:
        # Single tapered wing w/o kink.
        planform = 'single_tapered'
        section_list = ['fuselage_center_line', 'tip']
    elif number_of_horizontal_stabilizer_sections == 3:
        # Single tapered wing w/o kink.
        planform = 'single_tapered'
        section_list = ['fuselage_center_line', 'root', 'tip']
    else:
        # Not implemented yet.
        runtime_output.critical('Error: Too many horizontal stabilizer sections! Program aborted.')
        sys.exit('Exit information: Too many horizontal stabilizer sections!')

    """Create horizontal stabilizer geometry dictionary."""
    # Automatic creation of horizontal stabilizer geometry dictionary.
    parameter_list = ['chord_length', 'profile']
    position_parameters = ['x', 'y', 'z']
    base_keys = {
        'x': 'empennage_section_chord_origin_x',
        'y': 'empennage_section_chord_origin_y',
        'z': 'empennage_section_chord_origin_z',
        'chord_length': 'empennage_section_chord_length',
        'profile': 'empennage_section_profile'
    }

    # Create tmp_dict with 'position' sub dictionary for x, y, z positions.
    tmp_dict = {
        'sections': {
            section_list[i]: {
                'position': {
                    pos_param: dict_ac_data[base_keys[pos_param]][
                        f"{base_keys[pos_param]}_{horizontal_stabilizer_section_ids[i]}"]
                    for pos_param in position_parameters
                },
                **{
                    param: dict_ac_data[base_keys[param]][f"{base_keys[param]}_{horizontal_stabilizer_section_ids[i]}"]
                    for param in parameter_list
                }
            }
            for i in range(len(section_list))
        }
    }

    # Add further values to 'tmp_dict'.
    for i, section in enumerate(section_list):
        # Add span index to section.
        tmp_dict['sections'][section]['span_index'] = abs(int(round(
            tmp_dict['sections'][section]['position']['y']*100, 0)))
        # Add thickness_ratio.
        tmp_dict['sections'][section]['thickness_ratio'] = dict_ac_data['horizontal_stabilizer_thickness_to_length'][i]

    # Add 'tmp_dict' to 'dict_tank_design'.
    dict_tank_design['geometry']['horizontal_stabilizer'] = tmp_dict

    """Interpolate necessary values."""
    # Calculate number of points for value interpolation (one point per mm).
    number_of_points = abs(int(round(
        dict_tank_design['geometry']['horizontal_stabilizer']['sections']['tip']['position']['y']*100, 0)))
    # Write wing section span indexes (equal distances in mm) to list.
    section_span_indexes = [tmp_dict['sections'][section]['span_index']
                            for section in section_list]

    # Interpolate spar positions.
    horizontal_stabilizer_span_in_mm = number_of_points
    front_spar_span_indexes = [
        int(dict_ac_data['empennage_spar_position_inner_spanwise']
            ['empennage_spar_position_inner_spanwise_ID0_ID0']*horizontal_stabilizer_span_in_mm),
        int(dict_ac_data['empennage_spar_position_outer_spanwise']
            ['empennage_spar_position_outer_spanwise_ID0_ID0']*horizontal_stabilizer_span_in_mm)]
    front_spar_relative_position = [
        dict_ac_data['empennage_spar_position_inner_chord']['empennage_spar_position_inner_chord_ID0_ID0'],
        dict_ac_data['empennage_spar_position_outer_chord']['empennage_spar_position_outer_chord_ID0_ID0']]
    relative_position_front_spar = np.interp(range(number_of_points + 1), front_spar_span_indexes,
                                             front_spar_relative_position)
    rear_spar_span_indexes = [
        int(dict_ac_data['empennage_spar_position_inner_spanwise']
            ['empennage_spar_position_inner_spanwise_ID0_ID1']*horizontal_stabilizer_span_in_mm),
        int(dict_ac_data['empennage_spar_position_outer_spanwise']
            ['empennage_spar_position_outer_spanwise_ID0_ID1']*horizontal_stabilizer_span_in_mm)]
    rear_spar_relative_position = [
        dict_ac_data['empennage_spar_position_inner_chord']['empennage_spar_position_inner_chord_ID0_ID1'],
        dict_ac_data['empennage_spar_position_outer_chord']['empennage_spar_position_outer_chord_ID0_ID1']]
    relative_position_rear_spar = np.interp(range(number_of_points + 1), rear_spar_span_indexes,
                                             rear_spar_relative_position)
    x_coordinate_leading_edge = [
        dict_tank_design['geometry']['horizontal_stabilizer']['sections'][section]['position']['x']
        for section in section_list]
    interpolated_x_coordinate_leading_edge = np.interp(range(number_of_points + 1), section_span_indexes,
                                                 x_coordinate_leading_edge)
    # Calculate absolute front spar position.
    absolute_position_front_spar_list = [
        (x_coordinate_leading_edge[idx]
         + dict_tank_design['geometry']['horizontal_stabilizer']['sections'][name]['chord_length']
         *relative_position_front_spar[idx]) for idx, name in enumerate(section_list)]
    absolute_position_rear_spar_list = [
        (x_coordinate_leading_edge[idx]
         + dict_tank_design['geometry']['horizontal_stabilizer']['sections'][name]['chord_length']
         *relative_position_rear_spar[idx]) for idx, name in enumerate(section_list)]
    # Interpolate values for absolute front/rear spar positions and tank length.
    interpolated_absolute_front_spar = np.interp(range(number_of_points + 1), section_span_indexes,
                                                        absolute_position_front_spar_list)
    interpolated_absolute_rear_spar = np.interp(range(number_of_points + 1), section_span_indexes,
                                                        absolute_position_rear_spar_list)
    interpolated_tank_lengths = interpolated_absolute_rear_spar - interpolated_absolute_front_spar
    # Tank height interpolation.
    tank_height_list = [(dict_tank_design['geometry']['horizontal_stabilizer']['sections'][section]['chord_length']
                         *dict_tank_design['geometry']['horizontal_stabilizer']['sections'][section]['thickness_ratio']
                         *dict_ac_data['a_to_d_factor']) for section in section_list]
    interpolated_tank_heights = np.interp(range(number_of_points + 1), section_span_indexes, tank_height_list)
    # z coordinate interpolation.
    z_coordinate_leading_edge = [
        dict_tank_design['geometry']['horizontal_stabilizer']['sections'][section]['position']['z']
        for section in section_list]
    interpolated_z_coordinate_leading_edge = np.interp(range(number_of_points + 1), section_span_indexes,
                                                 z_coordinate_leading_edge)

    # Write data to 'dict_tank_design'.
    dict_tank_design['geometry']['horizontal_stabilizer']['interpolated_values'] = {}
    dict_tank_design['geometry']['horizontal_stabilizer']['interpolated_values']['x_coordinate_leading_edge'] = (
        interpolated_x_coordinate_leading_edge)
    dict_tank_design['geometry']['horizontal_stabilizer']['interpolated_values']['z_coordinate_leading_edge'] = (
        interpolated_z_coordinate_leading_edge)
    dict_tank_design['geometry']['horizontal_stabilizer']['interpolated_values']['x_coordinate_front_spar'] = (
        interpolated_absolute_front_spar)
    dict_tank_design['geometry']['horizontal_stabilizer']['interpolated_values']['x_coordinate_rear_spar'] = (
        interpolated_absolute_rear_spar)
    dict_tank_design['geometry']['horizontal_stabilizer']['interpolated_values']['tank_lengths'] = (
        interpolated_tank_lengths)
    dict_tank_design['geometry']['horizontal_stabilizer']['interpolated_values']['tank_heights'] = (
        interpolated_tank_heights)

    # Write maximum tank height and length to dictionary.
    for section in section_list:
        dict_tank_design['geometry']['horizontal_stabilizer']['sections'][section]['maximum_tank_height'] = (
            float(interpolated_tank_heights[
                dict_tank_design['geometry']['horizontal_stabilizer']['sections'][section]['span_index']]))
        dict_tank_design['geometry']['horizontal_stabilizer']['sections'][section]['maximum_tank_length'] = (
            float(interpolated_tank_lengths[
                dict_tank_design['geometry']['horizontal_stabilizer']['sections'][section]['span_index']]))

    """Calculate leading edge sweep angle."""
    number_of_horizontal_stabilizer_segments = number_of_horizontal_stabilizer_sections - 1
    x_tmp, y_tmp, sweep_leading_edge = [], [], [0]
    # Store x and y coordinates to temporary lists.
    for section in dict_tank_design['geometry']['horizontal_stabilizer']['sections'].keys():
        x_tmp.append(dict_tank_design['geometry']['horizontal_stabilizer']['sections'][section]['position']['x'])
        y_tmp.append(dict_tank_design['geometry']['horizontal_stabilizer']['sections'][section]['position']['y'])
    # Calculate sweep angles at wing sections.
    for i in range(number_of_horizontal_stabilizer_segments):
        opposite_leg = x_tmp[i+1] - x_tmp[i]
        adjacent_leg = y_tmp[i+1] - y_tmp[i]
        phi = (opposite_leg/adjacent_leg) * 180/PI
        sweep_leading_edge.append(phi)
    # Set 'constant_leading_edge_sweep' parameter.
    if planform == 'single_tapered':
        constant_leading_edge_sweep = True
    elif planform == 'compound_tapered':
        constant_leading_edge_sweep = all(round(x, 3) == round(sweep_leading_edge[2], 3)
                                           for x in sweep_leading_edge[2:])

    # Prepare output.
    dict_tank_design['geometry']['horizontal_stabilizer']['information'] = {}
    dict_tank_design['geometry']['horizontal_stabilizer']['information']['planform'] = planform
    dict_tank_design['geometry']['horizontal_stabilizer']['information']['symmetry'] = symmetry
    dict_tank_design['geometry']['horizontal_stabilizer']['information']['section_list'] = section_list
    dict_tank_design['geometry']['horizontal_stabilizer']['position'] = {}
    dict_tank_design['geometry']['horizontal_stabilizer']['position']['x'] = (
        dict_ac_data['empennage_position_x']
        + dict_ac_data['empennage_entity_position_x']['empennage_entity_position_x_ID'
                                                      + str(horizontal_stabilizer_id)])
    dict_tank_design['geometry']['horizontal_stabilizer']['position']['y'] = (
        dict_ac_data['empennage_position_y']
        + dict_ac_data['empennage_entity_position_y']['empennage_entity_position_y_ID'
                                                      + str(horizontal_stabilizer_id)])
    dict_tank_design['geometry']['horizontal_stabilizer']['position']['z'] = (
        dict_ac_data['empennage_position_z']
        + dict_ac_data['empennage_entity_position_z']['empennage_entity_position_z_ID'
                                                      + str(horizontal_stabilizer_id)])
    dict_tank_design['geometry']['horizontal_stabilizer']['information']['constant_leading_edge_sweep'] = (
        constant_leading_edge_sweep)

    # Debug print.
    runtime_output.debug('Debug: Geometrical data for trim tank prepared.')

    return dict_tank_design


def prepare_additional_center_tank(dict_ac_data, dict_tank_design, runtime_output):
    """Extract fuselage values for additional center tank.

    :param _type_ dict_ac_data: _description_
    :param _type_ dict_tank_design: _description_
    :param _type_ runtime_output: _description_
    :return _type_: _description_
    """
    # Extract wing position and index.
    position_wing = dict_ac_data['wing_position_x']
    idx_wing = int(round(position_wing*100, 0))
    # Extract fuselage length.
    max_id_key = max(dict_ac_data['fuselage_section_origin_x'].keys(), key=lambda x: int(x.split('_')[-1][2:]))
    fuselage_length = dict_ac_data['fuselage_section_origin_x'][max_id_key]
    # Interpolate fuselage widths to enable extraction of fuselage width at wing position.
    number_of_points_fuselage_length = int(round(fuselage_length*100, 0))
    indexes = [0] + [int(round(value * 100, 0)) for value in dict_ac_data['fuselage_section_origin_x'].values()]
    widths = [0] + list(dict_ac_data['fuselage_section_width'].values())
    interpolated_fuselage_widths_list = np.interp(range(number_of_points_fuselage_length + 1), indexes, widths)
    fuselage_width_at_wing = interpolated_fuselage_widths_list[idx_wing]
    # Extract cabin floor z coordinate and structural floor thickness.
    ids = [0, 1]
    z_coordinates = [0.0, -1.84]
    structural_floor_thicknesses = [0.21, 0.21]
    # Main passenger deck is always the first payload deck and cargo compartment always the last payload deck defined.
    main_deck_id = ids[0]
    cargo_deck_id = ids[-1]
    z_coordinate_cabin_floor = z_coordinates[main_deck_id]
    cabin_floor_structural_thickness = structural_floor_thicknesses[main_deck_id]
    z_coordinate_cargo_floor = z_coordinates[cargo_deck_id]
    cargo_compartment_height = (abs(z_coordinate_cargo_floor) - abs(z_coordinate_cabin_floor)
                                - abs(cabin_floor_structural_thickness))

    # Write to geometry sub dictionary.
    dict_tank_design['geometry']['fuselage'] = {
        'width_at_wing_position': fuselage_width_at_wing,
        'cargo_compartment_height': cargo_compartment_height,
        'z_coordinate_cargo_floor': z_coordinate_cargo_floor,
        'interpolated_values': {
            'width': interpolated_fuselage_widths_list
        }
    }

    # Debug print.
    runtime_output.debug('Debug: Geometrical data for additional center tank prepared.')

    return dict_tank_design


"""Prepare data for liquid hydrogen tanks."""
def prepare_for_liquid_hydrogen(dict_ac_data, dict_tank_design, runtime_output):
    """Prepare geometrical data for liquid hydrogen tanks.

    This function combines the following geometrical values from the aircraft xml to one list each:
        * Length of individual segments (1)
        * Widths of individual segments
        * Heights of individual segments
    (1) The cumulated lengths result in the x-coordinates of the respective segments (x_coord).
    Approach:
        1. Extract geometrical values for the fuselage, save to lists, and interpolate widths and heights.
        2. Calculate minimum diameter for every cm of fuselage length.
        3. Calculate maximum usable diameter for every cm of fuselage length.
    :param dict dict_ac_data: Dict containing parameter and values from aircraft exchange and module configuration file
    :param dict dict_tank_design: Dictionary containing tank design parameter
    :param logging.Logger runtime_output: Logging object used for capturing log messages in the module
    :return dict dict_tank_design: Dictionary containing tank design parameter
    """

    # Extract necessary data from data dict.
    x_coord_pressure_bulkhead = (
        dict_ac_data['fuselage_position_x']
        + dict_ac_data['fuselage_entity_position_x']['fuselage_entity_position_x_ID0']
        + dict_ac_data['fuselage_accommodation_position_x']['fuselage_accommodation_position_x_ID0']
        + dict_ac_data['fuselage_payload_tube_aft_position_x']['fuselage_payload_tube_aft_position_x_ID0_ID0'])
    idx_bulkhead = round(x_coord_pressure_bulkhead*100)
    # tail_length_before_bulkhead = 8.63  # est. length of section from start of tail to bulkhead (from CSR-02)
    factor_usable_diameter = dict_ac_data['factor_usable_diameter']
    # Initialize empty geometry dictionary.
    dict_tank_design['geometry'] = {}

    """Fuselage data."""
    # Initialize empty fuselage dictionary.
    dict_tank_design['geometry']['fuselage'] = {}

    # Get widths of fuselage.
    dict_tank_design['geometry']['fuselage']['widths'] = [0]
    dict_tank_design['geometry']['fuselage']['widths'].extend(dict_ac_data['fuselage_section_width'].values())

    # Get upper and lower heights of fuselage and sum up to height.
    upper_heights = [0]
    lower_heights = [0]
    upper_heights.extend(dict_ac_data['fuselage_section_upper_height'].values())
    lower_heights.extend(dict_ac_data['fuselage_section_lower_height'].values())
    dict_tank_design['geometry']['fuselage']['upper_heights'] = upper_heights
    dict_tank_design['geometry']['fuselage']['lower_heights'] = lower_heights
    dict_tank_design['geometry']['fuselage']['heights'] = [a + b for a, b in zip(upper_heights, lower_heights)]

    # Get z origin of fuselage sections.
    fuselage_section_origin_z = [0]
    fuselage_section_origin_z.extend(dict_ac_data['fuselage_section_origin_z'].values())

    # Generate a list with x coordinates.
    x_coord = [0]
    x_coord.extend(dict_ac_data['fuselage_section_origin_x'].values())
    # Extract fuselage length.
    length_fuselage = x_coord[-1]
    runtime_output.print(f"Total fuselage length: {float(length_fuselage):,.2f} m")
    # Calculate numbers of points (one point per cm).
    n_points = round(x_coord[-1]*100)
    x_coord_array = np.linspace(0, x_coord[-1], n_points)

    # Interpolate width and height.
    width_fuselage_external_interpolated = np.interp(x_coord_array, x_coord,
                                                     dict_tank_design['geometry']['fuselage']['widths'])
    upper_heights_external_interpolated = np.interp(x_coord_array, x_coord,
                                                    dict_tank_design['geometry']['fuselage']['upper_heights'])
    lower_heights_external_interpolated = np.interp(x_coord_array, x_coord,
                                                    dict_tank_design['geometry']['fuselage']['lower_heights'])
    height_fuselage_external_interpolated = np.interp(x_coord_array, x_coord,
                                                      dict_tank_design['geometry']['fuselage']['heights'])
    fuselage_section_origin_z_interpolated = np.interp(x_coord_array, x_coord, fuselage_section_origin_z)

    # Calculate minimum diameter for every coordinate (every cm).
    diameter_outside_max = ([min(width_fuselage_external_interpolated[idx],
                                    height_fuselage_external_interpolated[idx]) for idx in range(0, n_points)])
    # Calculate maximum usable diameter for every coordinate (every cm).
    diameter_usable = [diameter_outside_max[idx]*factor_usable_diameter for idx in range(0, n_points)]
    dict_tank_design['geometry']['fuselage']['diameter_usable'] = diameter_usable

    # Find index and x-coordinate of begin of tail section.
    idx_begin_tail_section = determine_begin_tail_section(width_fuselage_external_interpolated,
                                                            height_fuselage_external_interpolated)

    # Check if second iteration loop (tanks already written to acXML). If second iteration loop, the available length
    # in the tail cone is assumed to equal the total length of the tail cone tank.
    if dict_ac_data['tank_designator']['tank_designator_ID0'] is not None:
        tail_cone_tank_length = None

        for key, value in dict_ac_data['tank_designator'].items():

            if value == 'fuselage_tail_cone':
                # Get tank ID of tail cone tank.
                tail_cone_tank_ID = key.split('_')[-1]
                # Extract IDs.
                for cross_key, cross_value in dict_ac_data['tank_cross_section_name'].items():
                    ids = cross_key.split('_')[-2:]

                    if ids[0] == tail_cone_tank_ID and cross_value == 'rear_end_cap':
                        # Find rear end cap ID.
                        rear_end_cap_ID = ids[1]
                        # Generate position key.
                        position_key = f"tank_cross_section_position_{tail_cone_tank_ID}_{rear_end_cap_ID}"

                        if position_key in dict_ac_data['tank_cross_section_position']:
                            # Get tail cone tank length.
                            tail_cone_tank_length = dict_ac_data['tank_cross_section_position'][position_key]
                            # Generate index of bulkhead.
                            idx_bulkhead = idx_begin_tail_section + round(tail_cone_tank_length*100)
                            break

                if tail_cone_tank_length is not None:
                    break


    # Iterate through tank entities and extract necessary geometry.
    for tank_entity in dict_tank_design.keys():
        if tank_entity.startswith('tank_') and tank_entity[5:].isdigit():
            # Initialize empty dictionary and lists.
            dict_tank_design[tank_entity]['geometry'] = {}

            # Extract tank configuration.
            tank_designator = dict_tank_design[tank_entity]['designator']

            if tank_designator == 'fuselage_tail_cone':
                dict_tank_design[tank_entity]['geometry']['idx_front'] = idx_begin_tail_section
                dict_tank_design[tank_entity]['geometry']['idx_rear'] = idx_bulkhead
                runtime_output.debug('Preparing data for tank in fuselage tail cone.')
                dict_tank_design[tank_entity]['tank_shape'] = 'conical'
                lengths = x_coord_array[idx_begin_tail_section:idx_bulkhead]
                widths_external = width_fuselage_external_interpolated[idx_begin_tail_section:idx_bulkhead]
                heights_external = height_fuselage_external_interpolated[idx_begin_tail_section:idx_bulkhead]
                # Extract maximum dimensions (necessary for calculation).
                dict_tank_design[tank_entity]['geometry']['maximum_length'] = abs(lengths[-1] - lengths[0])
                # Assumption: Only the smaller height/width can be used.
                dict_tank_design[tank_entity]['geometry']['maximum_diameter_front'] = min(
                    widths_external[0], heights_external[0])
                dict_tank_design[tank_entity]['geometry']['maximum_diameter_rear'] = min(
                    widths_external[-1], heights_external[-1])

            elif tank_designator == 'fuselage_rear':
                dict_tank_design[tank_entity]['geometry']['idx_front'] = None
                if dict_tank_design['tank_configuration'] == 'fuselage_rear_tanks':
                    # Add extra space between tanks in rear and tail cone of fuselage.
                    # Attention: If this parameter is changed it has to be changed in 'calculateliquidhydrogentanks.py'
                    # as well!
                    extra_space = 0.2
                    idx_end_of_rear_tank = round(idx_begin_tail_section - extra_space*100)
                    dict_tank_design[tank_entity]['geometry']['idx_rear'] = idx_end_of_rear_tank
                else:
                    dict_tank_design[tank_entity]['geometry']['idx_rear'] = idx_begin_tail_section-1
                runtime_output.debug('Preparing data for tank in rear of fuselage.')
                dict_tank_design[tank_entity]['tank_shape'] = 'cylindrical'
                widths_external = width_fuselage_external_interpolated[idx_end_of_rear_tank]
                heights_external = height_fuselage_external_interpolated[idx_end_of_rear_tank]
                # Extract maximum dimensions (necessary for calculation).
                dict_tank_design[tank_entity]['geometry']['maximum_length'] = 0
                # Assumption: Only the smaller height/width can be used.
                dict_tank_design[tank_entity]['geometry']['maximum_diameter_front'] = min(
                    widths_external, heights_external)
                dict_tank_design[tank_entity]['geometry']['maximum_diameter_rear'] = (
                    dict_tank_design[tank_entity]['geometry']['maximum_diameter_front'])

            else:
                runtime_output.critical('Error: Tank combination "' + tank_designator + '" not implemented yet. '
                                        + 'Program aborted.')
                sys.exit('Exit information: Tank combination not implemented yet!')

    # Prepare output data.
    dict_tank_design['geometry']['fuselage']['interpolated_widths'] = width_fuselage_external_interpolated
    dict_tank_design['geometry']['fuselage']['interpolated_upper_heights'] = upper_heights_external_interpolated
    dict_tank_design['geometry']['fuselage']['interpolated_lower_heights'] = lower_heights_external_interpolated
    dict_tank_design['geometry']['fuselage']['interpolated_heights'] = height_fuselage_external_interpolated
    dict_tank_design['geometry']['fuselage']['interpolated_section_origin_z'] = fuselage_section_origin_z_interpolated

    # Debug print.
    runtime_output.debug('Debug: Geometrical data for liquid hydrogen tanks prepared.')

    return dict_tank_design


# Determine start of tail section.
def determine_begin_tail_section(widths, heights):
    """_summary_

    :param _type_ width: _description_
    :param _type_ height: _description_
    :return _type_: _description_
    """
    for idx in range(len(widths) - 1):
        # Check if the next width or height is smaller than the current
        smaller_width = round(widths[idx + 1],4) < round(widths[idx],4)
        smaller_height = round(heights[idx + 1], 4) < round(heights[idx], 4)
        if smaller_width or smaller_height:
            break
        else:
            idx = None
        
    return idx
