# UNICADO - UNIversity Conceptual Aircraft Design and Optimization
#
# Copyright (C) 2025 UNICADO consortium
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#
# Description:
# This file is part of UNICADO.

"""Module providing report functionalities for current calculation method."""
from yattag import Doc
from bs4 import BeautifulSoup
import os


def method_html_report(paths_and_names, routing_dict, data_dict, method_specific_output_dict, runtime_output):
    """HTML report function.

    This function is responsible for creating reports.
    [Add further information here...]

    :param dict paths_and_names: Dictionary containing system paths and ElementTrees
    :param dict routing_dict: Dictionary containing routing parameter
    :param dict data_dict: Dictionary containing results of module execution
    :return: None
    """
    runtime_output.print(f'HTML report is generated and saved...')

    """ Preparation."""
    # Extract tool and project name.
    tool_name = paths_and_names['tool_name']
    tool_version = paths_and_names['tool_version']
    report_title = (tool_name.replace('_', ' ')).capitalize()
    project_name = paths_and_names['name_of_project']
    root_of_module_config_tree = paths_and_names['root_of_module_config_tree']
    plot_switch = (eval(root_of_module_config_tree.find('.//plot_output/enable/value').text.capitalize()))

    if routing_dict['user_layer'] == 'kerosene':
        energy_information_data = [
            ("Tank configuration",
            data_dict['tank_configuration']),
            ("Volume required for mission",
            f"{data_dict['total_tank_parameters']['volume_required_for_mission']:.1f} L"),
            ("Volume provided in all tanks",
            f"{data_dict['total_tank_parameters']['volume_provided_in_all_tanks']:.1f} L"),
            ("Energy required for mission",
            f"{data_dict['total_tank_parameters']['energy_required_for_mission']/1E6:,.1f} MJ"),
            ("Energy provided in all tanks",
            f"{data_dict['total_tank_parameters']['energy_provided_in_all_tanks']/1E6:,.1f} MJ"),
        ]
        # Total energy information data.
        tmp_energy = sum(value['energy_available'] for key, value in data_dict.items() if key.startswith("tank_")
                        and key[5:].isdigit() and not value['energy_required_for_mission'])
        tmp_volume = sum(value['volume_available'] for key, value in data_dict.items() if key.startswith("tank_")
                        and key[5:].isdigit() and not value['energy_required_for_mission'])
        energy_missing = data_dict['total_tank_parameters'].get('energy_missing', 0)
        volume_missing = data_dict['total_tank_parameters'].get('volume_missing', 0)
        # Add some more information to data dict.
        energy_information_data.insert(2, (
            "Volume provided in necessary tanks",
            f"{data_dict['total_tank_parameters']['volume_provided_in_all_tanks'] - tmp_volume:.1f} L"))
        energy_information_data.insert(4, ("Volume missing", f"{volume_missing:.1f} L"))
        energy_information_data.insert(6, ("Energy provided in necessary tanks",
            f"{(data_dict['total_tank_parameters']['energy_provided_in_all_tanks'] - tmp_energy)/1E6:,.1f} MJ"))
        energy_information_data.insert(8, ("Energy missing", f"{energy_missing:,.1f} MJ"))
    elif routing_dict['user_layer'] == 'liquid_hydrogen':
        energy_information_data = [
            ("Tank configuration",
            data_dict['tank_configuration']),
            ("Volume required for mission",
            f"{data_dict['total_tank_parameters']['volume_required_for_mission']:.1f} m3"),
            ("Volume provided in all tanks",
            f"{data_dict['total_tank_parameters']['volume_provided_in_all_tanks']:.1f} m3"),
            ("Energy required for mission",
            f"{data_dict['total_tank_parameters']['energy_required_for_mission']/1E6:,.1f} MJ"),
            ("Energy provided in all tanks",
            f"{data_dict['total_tank_parameters']['energy_provided_in_all_tanks']/1E6:,.1f} MJ"),
        ]

    energy_information_title = "General information"
    energy_information_column_titles = ["Parameter", "Value"]
    energy_information_alignment = ["l", "r"]
    # Tank entity data.
    aircraft_tank_data = [
        (key[-1], value['designator'], value['energy_carrier_name'], float(round(value['volume_available'],2)),
         f"{round(value['energy_available']/1E6,1):,.1f}", value['energy_required_for_mission'])
        for key, value in data_dict.items()
        if key.startswith("tank_") and key[5:].isdigit()
    ]
    aircraft_tank_data_title = "Aircraft tank overview"
    if routing_dict['user_layer'] == 'kerosene':
        aircraft_tank_data_column_titles = ["Tank ID", "Designator", "Energy carrier", "Volume in L", "Energy in MJ",
                                            "Energy req. for mission?"]
    elif routing_dict['user_layer'] == 'liquid_hydrogen':
        aircraft_tank_data_column_titles = ["Tank ID", "Designator", "Energy carrier", "Volume in m3", "Energy in MJ",
                                            "Energy req. for mission?"]
    aircraft_tank_data_alignment = ["l", "l", "l", "r", "r", "m"]

    """Generate report."""
    # Start Yattag document.
    doc, tag, text = Doc().tagtext()

    # Generate header (without creating a new <html>).
    with tag('html', lang="en"):
        # Generate report header (adds <head>).
        generate_report_header(doc, report_title)

        # Generate body.
        with tag('body'):
            with tag('div', klass="logo"):
                text("")
            with tag('div', klass="content"):
                with tag('h1'):
                    text(f"Report - {report_title} of {project_name}")
                    with tag('br'):  # Line break for formatting
                        pass
                    with tag('span'):
                        with tag('font', size="2"):
                            text(f"All outputs of the program '{tool_name}' were created with version {tool_version}")
                            
                with tag('div', klass="container"):

                    # Data on the left side.
                    with tag('div', klass="box data"):
                        with tag('h2'):
                            text("Data")
                        # Add table using external function.
                        generate_table(doc, tag, text, energy_information_data, energy_information_title,
                                       energy_information_column_titles, energy_information_alignment)
                        # Add table using external function.
                        generate_table(doc, tag, text, aircraft_tank_data, aircraft_tank_data_title,
                                       aircraft_tank_data_column_titles, aircraft_tank_data_alignment)

                    # Plots on the right side.
                    # Ensure that plots are enabled before displaying them on report.
                    # if plot_switch:
                    #     with tag('div', klass="box plot"):
                    #         with tag('h2'):
                    #             text("Plot")
                    #         with tag('img', src="../plots/title", klass="image-plot"):
                    #             pass

    # Convert the document to HTML string.
    html_output = doc.getvalue()

    # Pretty HTML output with BeautifulSoup.
    soup = BeautifulSoup(html_output, 'html.parser')
    pretty_html = soup.prettify()

    """ Save report to current aircraft project directory. """
    # Save the report.
    project_directory = paths_and_names["project_directory"]
    report_directory = os.path.join(project_directory, 'reporting', 'report_html')
    os.makedirs(report_directory, exist_ok=True)
    output_path = os.path.join(report_directory, f"{tool_name}_report.html")

    # Write the HTML to the file.
    with open(output_path, "w", encoding="utf-8") as report:
        report.write(pretty_html)
    
    return


def generate_table(doc, tag, text, data, caption, column_titles, alignment=None):
    """Generate an HTML table.

    This function creates an HTML table from the provided data.
    Note: Options for the alignment types are "l" for left, "r" for right, and "m" for center. If not provided, columns
    are left-aligned by default.

    :param doc: Yattag Doc instance
    :param tag: Yattag tag function
    :param text: Yattag text function
    :param str caption: Table caption
    :param list column_titles: Titles of table columns
    :param list alignment: List of alignment types for ech column (defaults to None)
    :param list data: List of tuples containing table data
    """
    with tag('table', klass="content-table"):
        with tag('caption'):
            text(caption)
        with tag('thead'):
            with tag('tr'):
                for col in column_titles:
                    with tag('th'):
                        text(f"{col}")
        with tag('tbody'):
            row_number = 0
            for row in data:
                with tag('tr'):
                    column_number = 0
                    for col in row:
                        if col is None:
                            with tag('td'):
                                text('N.A.')
                        else:
                            if alignment is not None:
                                if alignment[column_number] == "m":
                                    with tag('td', style="text-align: center;"):
                                        text(col)
                                elif alignment[column_number] == "r":
                                    with tag('td', style="text-align: right;"):
                                        text(col)
                                elif alignment[column_number] == "l":
                                    with tag('td', style="text-align: left;"):
                                        text(col)
                                else:
                                    with tag('td'):
                                        text(col)
                            else:
                                with tag('td'):
                                    text(col)
                        column_number += 1
                row_number += 1
    
    return


def generate_report_header(doc, tool_name):
    """Generate HTML header.
    
    This function generates an HTML header with meta information, style sheets, and title.

    :param doc: Yattag Doc instance
    :param str tool_name: Name of the tool
    """
    with doc.tag('head'):
        doc.stag('meta', name="viewport", content="width=device-width, initial-scale=1.0")
        doc.stag('meta', charset="UTF-8")
        with doc.tag('title'):
            doc.text(tool_name)
        doc.stag('link', rel="stylesheet", href="style.css")
    
    return
