# UNICADO - UNIversity Conceptual Aircraft Design and Optimization
#
# Copyright (C) 2025 UNICADO consortium
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#
# Description:
# This file is part of UNICADO.

# Import standard libraries.
import numpy as np

# Import own modules.
import pyenergycarriers
from src.tube_and_wing.empirical.tank_design_tu_berlin.liquid_hydrogen.calculateparameterendcap import (
    calculate_parameter_end_cap)
from src.tube_and_wing.empirical.tank_design_tu_berlin.liquid_hydrogen.calculatethicknesswall import (
    calculate_thickness_wall)
from src.tube_and_wing.empirical.tank_design_tu_berlin.liquid_hydrogen.calculateparametercylinder import (
    calculate_parameter_cylinder)
from src.tube_and_wing.empirical.tank_design_tu_berlin.general.calculatecenterofgravity import (
    calculate_center_of_gravity)
from src.tube_and_wing.empirical.tank_design_tu_berlin.liquid_hydrogen.calculatezposition import (
    calculate_z_position)


def calculate_cylindrical_tank(tank_entity, dict_ac_data, dict_tank_design, runtime_output):
    """_summary_

    :param _type_ tank_entity: _description_
    :param float volume_required: _description_
    :param dict dict_ac_data: Dict containing parameter and values from aircraft exchange and module configuration file
    :param dict dict_tank_design: Dict containing results from tank design
    :param logging.Logger runtime_output: Logging object used for capturing log messages in the module
    :return dict dict_tank_design: Dictionary containing results from tank design for tube-and-wing aircraft
    """
    # Extract values from 'dict_ac_data'.
    mass_pumps = dict_ac_data['mass_pumps']
    mass_baffle = dict_ac_data['mass_baffle']
    type_end_cap = dict_ac_data['type_end_cap']
    density_wall = dict_ac_data['density_wall']
    material_wall = dict_ac_data['material_wall']
    internal_pressure = dict_ac_data['internal_pressure']
    density_insulation = dict_ac_data['density_insulation']
    mass_vapor_barrier = dict_ac_data['mass_vapor_barrier']
    factor_volume_allowance = dict_ac_data['factor_volume_allowance']
    tensile_stress_allowable = dict_ac_data['tensile_stress_allowable']
    thickness_wall_calculation_method = dict_ac_data['thickness_wall_calculation_method']

    energy_required = dict_tank_design[tank_entity]['energy_required']
    thickness_wall = 0.003
    thickness_insulation = 0.20
    thickness_wall_and_insulation = thickness_wall + thickness_insulation
    thickness_wall_minimum_aluminum = 0.003  # AD-2000: minimum wall thickness AL-2219 is 3 mmm.

    # Get density of liquid hydrogen.
    for key, value in dict_ac_data['energy_carrier_name'].items():
        if value == 'liquid_hydrogen':
            density_liquid_hydrogen = dict_ac_data['energy_carrier_density'].get(
                key.replace('energy_carrier_name', 'energy_carrier_density'))
            break
    liquid_hydrogen_volumetric_energy_density = (
        pyenergycarriers.EnergyCarrier('liquid_hydrogen', density_liquid_hydrogen).volumetric_energy_density)
    liquid_hydrogen_gravimetric_energy_density = (
        pyenergycarriers.EnergyCarrier('liquid_hydrogen', density_liquid_hydrogen).gravimetric_energy_density)

    volume_required = energy_required/liquid_hydrogen_volumetric_energy_density

    """Calculation. """
    # Note: The diameters are actual sizes. This means that the thicknesses for the tank wall and insulation have
    # already been taken into account.
    # Assumption: Maximum outer tank diameter equals maximum useable fuselage diameter minus twice the
    # thickness of insulation and wall. The calculated maximum outer diameter is assumed to be the actual tank
    # diameter.
    maximum_tank_diameter = (dict_tank_design[tank_entity]['geometry']['maximum_diameter_front']
                             - 2*thickness_wall_and_insulation)

    """Find position of end caps."""
    # Set iteration values.
    increment = 0.01
    boundary = 100
    distance = 0

    while boundary > distance:
        boundary = calculate_parameter_end_cap(runtime_output, maximum_tank_diameter, type_end_cap, thickness_wall,
                                               thickness_insulation, density_wall, density_insulation)[3]
        distance += increment

    """Calculate wall thickness."""
    maximum_diameter = maximum_tank_diameter
    thickness_wall = calculate_thickness_wall(runtime_output, maximum_diameter, internal_pressure,
                                              tensile_stress_allowable, thickness_wall_calculation_method)
    if (material_wall == 'aluminum') and (thickness_wall < thickness_wall_minimum_aluminum):
        thickness_wall = thickness_wall_minimum_aluminum

    """Calculate volumes, masses, and lengths of tank parts."""
    fuselage_geometry = dict_tank_design['geometry']['fuselage']
    # Note that the lengths of the end caps already take into account the tank wall and the insulation thickness.
    # 1) Front end cap (equals rear end cap).
    (volume_tank_end_cap, mass_wall_end_cap, mass_insulation_end_cap,
     length_end_cap) = calculate_parameter_end_cap(runtime_output, maximum_tank_diameter, type_end_cap, thickness_wall,
                                                   thickness_insulation, density_wall, density_insulation)

    # 2) Calculate how much volume is available in both end caps.
    volume_end_caps_available = 2*volume_tank_end_cap
    # energy_end_caps_available = volume_end_caps_available * dict_ac_data['liquid_hydrogen_volumetric_energy_density']
    # energy_remaining = energy_required - energy_end_caps_available
    volume_remaining = volume_required - volume_end_caps_available

    # 3) The remaining energy must be stored in the cylindrical part of the tank (iterative approach).
    length_cylindrical_section = 0
    volume_cylindrical_section = 0
    volume_cylindrical_section_usable = 0
    while volume_cylindrical_section_usable < volume_remaining:
        (volume_cylindrical_section, mass_wall_cylindrical_section,
         mass_insulation_cylindrical_section) = calculate_parameter_cylinder(
                    runtime_output, maximum_diameter, length_cylindrical_section, thickness_wall, thickness_insulation,
                    density_wall, density_insulation)
        volume_cylindrical_section_usable = volume_cylindrical_section/(1 + factor_volume_allowance)
        length_cylindrical_section += 0.01

    # 4) Total tank.
    volume_tank_available = volume_end_caps_available + volume_cylindrical_section_usable
    energy_hydrogen_available = volume_tank_available * liquid_hydrogen_volumetric_energy_density
    mass_hydrogen_available = energy_hydrogen_available/liquid_hydrogen_gravimetric_energy_density

    """Set dictionary values."""
    dict_tank_design[tank_entity]['geometry']['rear_end_cap'] = {
        'section_name': 'rear_end_cap',
        'shape': 'circular',
        'height': 0.0,
        'width': 0.0,
        'length': -length_end_cap,
        'global_position': {
            'x': dict_tank_design[tank_entity]['geometry']['idx_rear']/100,
            'y': 0.0,
            'z': None
        }
    }

    dict_tank_design[tank_entity]['geometry']['rear_tank_section'] = {
        'section_name': 'rear_tank_section',
        'shape': 'circular',
        'height': maximum_tank_diameter,
        'width': maximum_tank_diameter,
        'length': 0.0,
        'global_position': {
            'x': dict_tank_design[tank_entity]['geometry']['rear_end_cap']['global_position']['x'] - length_end_cap,
            'y': 0.0,
            'z': None
        }
    }
    idx_rear_tank_section = round(
        dict_tank_design[tank_entity]['geometry']['rear_tank_section']['global_position']['x']*100)
    dict_tank_design[tank_entity]['geometry']['rear_tank_section']['global_position']['z'] = calculate_z_position(
        fuselage_geometry, idx_rear_tank_section)
    # Set z-coordinate of rear end cap.
    dict_tank_design[tank_entity]['geometry']['rear_end_cap']['global_position']['z'] = (
        dict_tank_design[tank_entity]['geometry']['rear_tank_section']['global_position']['z'])

    dict_tank_design[tank_entity]['geometry']['front_tank_section'] = {
        'section_name': 'front_tank_section',
        'shape': 'circular',
        'height': maximum_tank_diameter,
        'width': maximum_tank_diameter,
        'length': 0.0,
        'global_position': {
            'x': (dict_tank_design[tank_entity]['geometry']['rear_tank_section']['global_position']['x']
                  - length_cylindrical_section),
            'y': 0.0,
            # Assumption: Cylindrical tank with base and top faces aligned vertically (right cylinder).
            'z': dict_tank_design[tank_entity]['geometry']['rear_tank_section']['global_position']['z']
        }
    }

    dict_tank_design[tank_entity]['geometry']['front_end_cap'] = {
        'section_name': 'front_end_cap',
        'shape': 'circular',
        'height': 0.0,
        'width': 0.0,
        'length': length_end_cap,
        'global_position': {
            'x': (dict_tank_design[tank_entity]['geometry']['front_tank_section']['global_position']['x']
                  - length_end_cap),
            'y': 0.0,
            'z': dict_tank_design[tank_entity]['geometry']['front_tank_section']['global_position']['z']
        }
    }

    """Calculate overall values."""
    # Calculate mass of tank wall.
    mass_wall = 2*mass_wall_end_cap + mass_wall_cylindrical_section
    # Calculate structural mass of tank.
    mass_tank_structure = mass_wall + mass_pumps + mass_baffle
    # Calculate mass of insulation.
    mass_insulation = mass_insulation_end_cap + mass_insulation_cylindrical_section
    mass_tank_insulation = mass_insulation + mass_vapor_barrier
    # Calculate overall tank mass.
    mass_tank_overall = mass_tank_structure + mass_tank_insulation
    # Calculate overall tank length.
    length_tank_overall = 2*length_end_cap + length_cylindrical_section
    # Calculate gravimetric efficiency tank.
    gravimetric_efficiency = mass_hydrogen_available/(mass_hydrogen_available + mass_tank_overall)
    # 
    dict_tank_design[tank_entity]['masses'] = {
        'masses_tank_structure': {
            'mass_wall': mass_wall,
            'mass_pumps': mass_pumps,
            'mass_baffle': mass_baffle
            },
        'masses_tank_insulation': {
            'mass_insulation': mass_insulation,
            'mass_vapor_barrier': mass_vapor_barrier
            },
        'mass_tank_overall': None
    }
    dict_tank_design[tank_entity]['masses']['mass_tank_overall'] = mass_tank_structure + mass_tank_insulation

    # Inertia (set inertia to zero).
    dict_tank_design[tank_entity]['geometry']['total'] = {}
    dict_tank_design[tank_entity]['geometry']['total']['inertia'] = {}  
    inertia_list = ['j_xx', 'j_yy', 'j_zz', 'j_xy', 'j_xz', 'j_yx', 'j_yz', 'j_zx', 'j_zy']
    for inertia in inertia_list:
        dict_tank_design[tank_entity]['geometry']['total']['inertia'][inertia] = 0.0

    # Calculate center of gravity.
    dict_tank_design[tank_entity]['geometry']['total']['center_of_gravity'] = calculate_center_of_gravity(
        runtime_output, dict_tank_design[tank_entity]['tank_shape'],
         dict_tank_design[tank_entity]['geometry']['maximum_diameter_front'],
         dict_tank_design[tank_entity]['geometry']['maximum_diameter_rear'],
         length_cylindrical_section)

    # Calculate position of tank entity (foremost point of tank).
    dict_tank_design[tank_entity]['geometry']['total']['global_position'] = {
        'x': dict_tank_design[tank_entity]['geometry']['front_end_cap']['global_position']['x'],
        'y': dict_tank_design[tank_entity]['geometry']['front_end_cap']['global_position']['y'],
        'z': dict_tank_design[tank_entity]['geometry']['front_end_cap']['global_position']['z'],
    }

    # Direction.
    dict_tank_design[tank_entity]['geometry']['total']['direction'] = {
        'x': 1,
        'y': 0,
        'z': 0
    }

    # Centroid.
    dict_tank_design[tank_entity]['geometry']['total']['centroid'] = {
        'x': dict_tank_design[tank_entity]['geometry']['total']['center_of_gravity']['x'],
        'y': dict_tank_design[tank_entity]['geometry']['total']['center_of_gravity']['y'],
        'z': dict_tank_design[tank_entity]['geometry']['total']['center_of_gravity']['z']
    }

    # Prepare data output.
    dict_tank_design[tank_entity]['energy_available'] = float(energy_hydrogen_available)
    dict_tank_design[tank_entity]['volume_available'] = float(volume_tank_available)
    dict_tank_design[tank_entity]['geometry']['length_tank_overall'] = float(length_tank_overall)
    dict_tank_design[tank_entity]['gravimetric_efficiency'] = float(gravimetric_efficiency)
    dict_tank_design[tank_entity]['energy_required_for_mission'] = True
    dict_tank_design[tank_entity]['volume_required_for_mission'] = True

    # Debug print.
    runtime_output.debug('Debug: The "calculate_cylindrical_tank" function was successfully executed.')

    return dict_tank_design
