# UNICADO - UNIversity Conceptual Aircraft Design and Optimization
#
# Copyright (C) 2025 UNICADO consortium
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#
# Description:
# This file is part of UNICADO.

""" Module containing functions for kerosene tank calculation. """
# Import standard libraries.

# Import own libraries.
import pyenergycarriers
from src.tube_and_wing.empirical.tank_design_tu_berlin.liquid_hydrogen.calculateconicaltank import calculate_conical_tank
from src.tube_and_wing.empirical.tank_design_tu_berlin.liquid_hydrogen.calculatecylindricaltank import calculate_cylindrical_tank


def calculate_liquid_hydrogen_tanks(dict_ac_data, dict_tank_design, runtime_output):
    """Calculate liquid hydrogen tanks of tube-and-wing aircraft.
    
    This function adds the following sub dictionaries to the 'dict_tank_design' dictionary:
    ...

    :param dict dict_ac_data: Dictionary containing data from aircraft exchange and module configuration file
    :param dict dict_tank_design: Dictionary containing tank design parameter
    :param logging.Logger runtime_output: Logging object used for capturing log messages in the module
    :return dict dict_tank_design: Dictionary containing tank design parameter
    """
    # Extract data and initialize variables.
    number_of_tanks = dict_tank_design['number_of_tanks']
    # Get density of liquid hydrogen.
    for key, value in dict_ac_data['energy_carrier_name'].items():
        if value == 'liquid_hydrogen':
            density_liquid_hydrogen = dict_ac_data['energy_carrier_density'].get(
                key.replace('energy_carrier_name', 'energy_carrier_density'))
            break
    liquid_hydrogen_volumetric_energy_density = (
        pyenergycarriers.EnergyCarrier('liquid_hydrogen', density_liquid_hydrogen).volumetric_energy_density)

    """Check if initial sizing loop and estimate energy demand if necessary."""
    initial_sizing_loop = all(value is None for value in dict_ac_data['mission_energy_amount'].values())
    if initial_sizing_loop:
        energy_demand = 13  # in L per passenger per 100 km
        runtime_output.print('Initial sizing loop! Energy demand is estimated (' + str(energy_demand) + ' liter LH2 per PAX per 100 km).')
        mission_fuel_amount = dict_ac_data['number_of_passengers']*(dict_ac_data['range']/1000)*energy_demand/100
        mission_energy_amount = (mission_fuel_amount/1000*liquid_hydrogen_volumetric_energy_density)
    else:
        mission_energy_amount = dict_ac_data['mission_energy_amount']['mission_energy_amount_ID0']

    """ Calculate tank volume. """
    # Extract all tank entities from 'dict_tank_design'.
    tank_entity_list = [key for key in dict_tank_design if key.startswith('tank_') and key[5:].isdigit()]
    # Check, if there is a tank in the tail cone of the aircraft.
    tail_cone_tank = next((key for key in tank_entity_list if dict_tank_design[key]['tank_position'] == 'tail_cone'),
                          None)
    rear_tank = next((key for key in tank_entity_list if dict_tank_design[key]['tank_position'] == 'rear'),
                          None)
    front_tank = next((key for key in tank_entity_list if dict_tank_design[key]['tank_position'] == 'front'),
                      None)
    # Sort list such that the key of the tank with the tail cone position is at the first list position. This ensures
    # that this tank is calculated first.
    if tail_cone_tank:
        tank_entity_list.remove(tail_cone_tank)
        tank_entity_list.insert(0, tail_cone_tank)

    # Iterate through tank entities (first entity is tail cone tank) and call functions from liquid hydrogen folder.
    tank_idx = 0
    energy_available = 0
    mass_tank_overall = 0
    length_tank_overall = 0
    tail_cone_tank_energy = 0
    mass_structure_overall = 0
    mass_insulation_overall = 0

    for tank_entity in tank_entity_list:
        # Call functions according to tank shape.
        match dict_tank_design[tank_entity]['tank_shape']:
            case 'conical':
                dict_tank_design = calculate_conical_tank(tank_entity, dict_ac_data, dict_tank_design, runtime_output)
            case 'cylindrical':
                dict_tank_design = calculate_cylindrical_tank(tank_entity, dict_ac_data, dict_tank_design,
                                                              runtime_output)

        energy_available += dict_tank_design[tank_entity]['energy_available']
        energy_remaining = mission_energy_amount - energy_available
        mass_tank_overall += dict_tank_design[tank_entity]['masses']['mass_tank_overall']
        mass_insulation_overall += sum(dict_tank_design[tank_entity]['masses']['masses_tank_insulation'].values())
        mass_structure_overall += sum(dict_tank_design[tank_entity]['masses']['masses_tank_structure'].values())
        length_tank_overall += dict_tank_design[tank_entity]['geometry']['length_tank_overall']
        
        # Set missing entries in 'dict_tank_design'.
        if tank_entity == tail_cone_tank:
            tail_cone_tank_energy = dict_tank_design[tank_entity]['energy_available']
            dict_tank_design[tank_entity]['energy_required'] = dict_tank_design[tank_entity]['energy_available']
            # Set required energy of remaining tanks.
            if dict_tank_design['tank_configuration'] == 'fuselage_rear_tanks':
                # Since this tank configuration contains only one tank in addition to the tail cone tank, the required
                # energy of this tank equals the 'energy_remaining'.
                dict_tank_design[tank_entity_list[-1]]['energy_required'] = energy_remaining
            elif dict_tank_design['tank_configuration'] == 'fuselage_front_and_rear_tanks':
                for tmp_tank in tank_entity_list[1:]:
                    if (dict_tank_design[tmp_tank]['tank_location'] == 'fuselage'
                        and dict_tank_design[tmp_tank]['tank_position'] == 'rear'):
                        # Tank at rear fuselage.
                        # Assumption: energy_share of rear tank contains energy of tail cone tank.
                        # energy_required = (mission_energy*energy_share) - energy_tail_cone_tank
                        dict_tank_design[tmp_tank]['energy_required'] = (
                            mission_energy_amount*dict_tank_design[tmp_tank]['energy_share']
                            - tail_cone_tank_energy)
                    elif (dict_tank_design[tmp_tank]['tank_location'] == 'fuselage'
                        and dict_tank_design[tmp_tank]['tank_position'] == 'front'):
                        # Tank at front fuselage.
                        dict_tank_design[tmp_tank]['energy_required'] = (
                            mission_energy_amount*dict_tank_design[tmp_tank]['energy_share'])

        # Print.
        tank_location = dict_tank_design[tank_entity]['tank_location']
        tank_position = dict_tank_design[tank_entity]['tank_position']
        if '_' in tank_position:
            print_position = (tank_location.capitalize() + ' ' + tank_position.split('_')[0] + ' '
                              + tank_position.split('_')[1] + ' ')
        else:
            print_position = tank_location.capitalize() + ' ' + tank_position + ' '
        runtime_output.print(
            print_position + 'tank (' + tank_entity + ') calculated: \n'
            + '                                  '
            + '- Volume (energy) available: '
            + f"{dict_tank_design[tank_entity]['volume_available']:,.2f}" + ' m3 ('
            + f"{dict_tank_design[tank_entity]['energy_available']/1e6:,.2f}"
            + ' MJ) \n'
            + '                                  '
            + '- Tank length: '
            + f"{dict_tank_design[tank_entity]['geometry']['length_tank_overall']:,.2f}" + ' m \n'
            + '                                  '
            + '- Tank mass (insulation, structure): '
            + f"{dict_tank_design[tank_entity]['masses']['mass_tank_overall']:,.2f}" + ' kg ('
            + f"{sum(dict_tank_design[tank_entity]['masses']['masses_tank_insulation'].values()):,.2f}" + ' kg, '
            + f"{sum(dict_tank_design[tank_entity]['masses']['masses_tank_structure'].values()):,.2f}" + ' kg) \n'
            + '                                  '
            + '- Gravimetric efficiency: '
            + f"{dict_tank_design[tank_entity]['gravimetric_efficiency']:,.2f}"
        )

        tank_idx += 1

    """Set total tank parameters."""
    # Calculate coordinates w.r.t. tank reference point (foremost middle point of foremost tank) and local reference 
    # points (foremost middle points of single tank entities).
    dict_tank_design['total_tank_parameters'] = {}
    inertia_list = ['j_xx', 'j_yy', 'j_zz', 'j_xy', 'j_xz', 'j_yx', 'j_yz', 'j_zx', 'j_zy']
    tank_section_list = ['front_end_cap', 'front_tank_section', 'rear_tank_section', 'rear_end_cap']

    if dict_tank_design['tank_configuration'] == 'fuselage_rear_tanks':
        dict_tank_design['total_tank_parameters']['position'] = {
            'x': dict_tank_design[rear_tank]['geometry']['total']['global_position']['x'],
            'y': dict_tank_design[rear_tank]['geometry']['total']['global_position']['y'],
            'z': dict_tank_design[rear_tank]['geometry']['total']['global_position']['z']
        }
        # Calculate positions of tank entities.
        tank_reference_point = dict_tank_design['total_tank_parameters']['position']
        tank_section_list = ['front_end_cap', 'front_tank_section', 'rear_tank_section', 'rear_end_cap']
        for tank_entity in [rear_tank, tail_cone_tank]:
            if tank_entity == rear_tank:
                # Local reference point equals tank reference point.
                local_reference_point = {key: 0.0 for key in ['x', 'y', 'z']}
            else:
                # Local reference point must be calculated.
                local_reference_point = {
                    'x': (dict_tank_design[tail_cone_tank]['geometry']['total']['global_position']['x']
                          - tank_reference_point['x']),
                    'y': (dict_tank_design[tail_cone_tank]['geometry']['total']['global_position']['y']
                          - tank_reference_point['y']),
                    'z': (dict_tank_design[tail_cone_tank]['geometry']['total']['global_position']['z']
                          - tank_reference_point['z'])
                }
            dict_tank_design[tank_entity]['geometry']['total']['position'] = local_reference_point

            # Iterate through tank sections.
            reference_point = {key: local_reference_point[key] + tank_reference_point[key]
                                for key in local_reference_point.keys() & tank_reference_point.keys()}
            for tank_section in tank_section_list:
                dict_tank_design[tank_entity]['geometry'][tank_section]['position'] = {
                    'x': (dict_tank_design[tank_entity]['geometry'][tank_section]['global_position']['x']
                        - reference_point['x']),
                    'y': 0.0,
                    'z': dict_tank_design[tank_entity]['geometry'][tank_section]['global_position']['z']
                        - reference_point['z']
                }
            
            # Set mass - just for correct output.
            dict_tank_design[tank_entity]['geometry']['total']['mass'] = (
                dict_tank_design[tank_entity]['masses']['mass_tank_overall'])

    # Inertia and mass.
    dict_tank_design['total_tank_parameters']['inertia'] = {inertia: 0.0 for inertia in inertia_list}
    dict_tank_design['total_tank_parameters']['mass'] = sum(
        dict_tank_design[f'tank_{i}']['masses']['mass_tank_overall'] for i in range(number_of_tanks))

    # Center of gravity (calculate weighted sums).
    # TODO: Check - especially z coordinates.
    cog_x = cog_y = cog_z = mass_total = 0
    for key, value in ((k, v) for k, v in dict_tank_design.items() if k.startswith('tank_') and isinstance(v, dict)):
        # Note: CoG is measured from lower left corner of front tank section, so some adaptions must be made.
        x = (value['geometry']['total']['center_of_gravity']['x'] + value['geometry']['front_end_cap']['length']
             + value['geometry']['total']['global_position']['x'])
        y = value['geometry']['total']['center_of_gravity']['y']
        if key == rear_tank and dict_tank_design[key]['tank_shape'] == 'cylindrical':
            # z position equals z position of front tank section.
            z = value['geometry']['front_tank_section']['global_position']['z']
            z_local = 0.0
        else:
            z = (value['geometry']['total']['center_of_gravity']['z']
                + (value['geometry']['front_tank_section']['global_position']['z']
                    - value['geometry']['front_tank_section']['height']*0.5))
            z_local = (value['geometry']['total']['center_of_gravity']['z']
                       - value['geometry']['front_tank_section']['height']*0.5)
        cog_x += value['masses']['mass_tank_overall'] * x
        cog_y += value['masses']['mass_tank_overall'] * y
        cog_z += value['masses']['mass_tank_overall'] * z
        mass_total += value['masses']['mass_tank_overall']
        # Set center of gravity in dictionary to corrected center of gravity w.r.t. local reference point.
        dict_tank_design[key]['geometry']['total']['center_of_gravity'] = {
            'x': value['geometry']['total']['center_of_gravity']['x'] + value['geometry']['front_end_cap']['length'],
            'y': y,
            'z': z_local}
        value['geometry']['total']['centroid'] = value['geometry']['total']['center_of_gravity']
        
    dict_tank_design['total_tank_parameters']['center_of_gravity'] = {
        'x': cog_x / mass_total,
        'y': cog_y / mass_total,
        'z': cog_z / mass_total
    }

    # Direction.
    dict_tank_design['total_tank_parameters']['direction'] = {
        'x': 1,
        'y': 0,
        'z': 0
    }

    # Since all tanks are dimensioned based on the required amount of energy, the total mission energy amount should be
    # covered after the tank calculation. However, a final energy check is implemented to ensure that.
    if energy_available > mission_energy_amount:
        volume_available = energy_available/liquid_hydrogen_volumetric_energy_density
        runtime_output.print('Energy check: Energy demand covered.')
        runtime_output.print('Tank design successful.')
    if energy_available < mission_energy_amount:
        energy_missing = mission_energy_amount - energy_available
        volume_missing = energy_missing/liquid_hydrogen_volumetric_energy_density
        runtime_output.print('Attention: Necessary amount of fuel cannot be stored in selected tanks! '
                        + str(int(round(energy_missing,0))) + 'J (' + str(round(volume_missing*1000,2)) +
                        'L) missing.')
    
    # Add extra space between tanks.
    if dict_tank_design['tank_configuration'] == 'fuselage_rear_tanks':
        space_between_tanks = (dict_tank_design['tank_0']['geometry']['total']['position']['x']
                               - dict_tank_design[rear_tank]['geometry']['rear_end_cap']['position']['x'])
        length_rear_tanks_overall = (dict_tank_design[rear_tank]['geometry']['length_tank_overall']
                                     + dict_tank_design[tail_cone_tank]['geometry']['length_tank_overall']
                                     + space_between_tanks)
        length_front_tank_overall = 0
        length_print = ('- Overall length of tanks in the rear of the fuselage: '
                        + f"{length_rear_tanks_overall:,.2f}" + ' m. \n')
    elif dict_tank_design['tank_configuration'] == 'fuselage_front_and_rear_tanks':
        space_between_tanks = (dict_tank_design['tank_0']['geometry']['total']['position']['x']
                               - dict_tank_design[rear_tank]['geometry']['rear_end_cap']['position']['x'])
        length_rear_tanks_overall = (dict_tank_design[rear_tank]['geometry']['length_tank_overall']
                                     + dict_tank_design[tail_cone_tank]['geometry']['length_tank_overall']
                                     + space_between_tanks)
        length_front_tank_overall = dict_tank_design[front_tank]['geometry']['length_tank_overall']
        length_print = ('- Overall length of tanks in the front of the fuselage: '
                        + f"{length_front_tank_overall:,.2f}" + ' m. \n'
                        + '                                  '
                        + '- Overall length of tanks in the rear of the fuselage: '
                        + f"{length_rear_tanks_overall:,.2f}" + ' m. \n')

    # Calculate additional fuselage length.
    # additional_fuselage_length_current = length_front_tank_overall + length_rear_tanks_overall
    # additional_fuselage_length_old = dict_ac_data['additional_fuselage_length']
    # if additional_fuselage_length_old is None:
    #     additional_fuselage_length = length_front_tank_overall + length_rear_tanks_overall
    # else:
    #     length_difference = additional_fuselage_length_old - additional_fuselage_length_current
    #     additional_fuselage_length = (-length_difference if length_difference != 0 
    #                                   else additional_fuselage_length_current)
    additional_fuselage_length = length_front_tank_overall + length_rear_tanks_overall

    # Results print.
    runtime_output.print('All tanks calculated: \n'
                         + '                                  '
                         + '- Volume (energy) available: '
                         + f"{volume_available:,.2f}" + ' L ('
                         + f"{energy_available/1e6:,.2f}"
                         + ' MJ) \n'
                         + '                                  '
                         + length_print
                         + '                                  '
                         + '- Tank mass (insulation, structure): '
                         + f"{mass_tank_overall:,.2f}" + ' kg ('
                         + f"{mass_insulation_overall:,.2f}" + ' kg, '
                         + f"{mass_structure_overall:,.2f}" + ' kg)')
    
    # Prepare data output.
    dict_tank_design['total_tank_parameters']['length_rear_tanks_overall'] = length_rear_tanks_overall
    dict_tank_design['total_tank_parameters']['length_front_tank_overall'] = length_front_tank_overall
    dict_tank_design['total_tank_parameters']['additional_fuselage_length'] = additional_fuselage_length
    dict_tank_design['total_tank_parameters']['energy_provided_in_all_tanks'] = energy_available
    dict_tank_design['total_tank_parameters']['volume_provided_in_all_tanks'] = volume_available
    dict_tank_design['total_tank_parameters']['energy_required_for_mission'] = mission_energy_amount
    dict_tank_design['total_tank_parameters']['volume_required_for_mission'] = (mission_energy_amount
                                                                                /liquid_hydrogen_volumetric_energy_density)

    # Debug print.
    runtime_output.debug('Debug: The "calculate_liquid_hydrogen_tanks" function was successfully executed.')
    
    return dict_tank_design
