# UNICADO - UNIversity Conceptual Aircraft Design and Optimization
#
# Copyright (C) 2025 UNICADO consortium
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#
# Description:
# This file is part of UNICADO.

"""Module providing calculation functions provided by the user."""
# Import standard modules.
import sys

# Import own modules.
from src.tube_and_wing.empirical.tank_design_tu_berlin.general.tankdesigntuberlin import tank_design_tu_berlin
from src.tube_and_wing.empirical.tank_design_tu_berlin.liquid_hydrogen.calculateliquidhydrogentanks import (
    calculate_liquid_hydrogen_tanks)


def method_liquid_hydrogen(paths_and_names, routing_dict, dict_ac_exchange, dict_mod_config, runtime_output):
    """Tank design method for liquid hydrogen (LH2)-powered tube-and-wing (TAW) aircraft.

    This function performs the tank design for liquid hydrogen-powered TAW aircraft configurations.
    [Add more information here...]
    The output dictionary 'liquid_hydrogen_output_dict' contains the results of the tank design and is structured
    according to the following scheme:
        liquid_hydrogen_output_dict = {'tank_0': {'parameter_1': value, 'parameter_2': value, ...},
                                       'tank_1': {'parameter_1': value, 'parameter_2': value, ...},
                                       ...,
                                       'number_of_tanks': int,
                                       'tank_configuration': str,
                                       'geometry': {...},
                                       'length_tank_overall': float}

    :param dict paths_and_names: Dictionary containing system paths and ElementTrees
    :param dict routing_dict: Dictionary containing routing parameters
    :param dict dict_ac_exchange: Dict containing parameters and according values from aircraft exchange file
    :param dict dict_mod_config: Dict containing parameters and according values from module configuration file
    :param logging.Logger runtime_output: Logging object used for capturing log messages in the module
    :return dict liquid_hydrogen_output_dict: Dictionary containing results from calculation for LH2-powered aircraft
    """

    """ Preparation. """
    # Check if any values from aircraft exchange file are missing.
    # Since no tank information is available in the first loop, these values are excluded from the check.
    values_to_exclude = ['tank_designator', 'tank_cross_section_name', 'tank_cross_section_position',
                         'additional_fuselage_length']
    for key, value in dict_ac_exchange.items():
        # Check, if any value is 'None'.
        if (key not in values_to_exclude) and (value is None):
            # If any value is 'None', the tank design is impossible and the program aborted.
            runtime_output.critical('Error: Tank design not possible due to missing parameter in aircraft exchange '
                                    + 'file (' + key + ')! '
                                    + 'Program aborted.')
            sys.exit(1)
    # Check if any necessary values from module configuration file are missing.
    necessary_values_for_liquid_hydrogen = ['type_end_cap', 'internal_pressure', 'type_insulation',
                                            'material_insulation', 'thickness_insulation', 'density_insulation',
                                            'material_wall', 'density_wall', 'thickness_wall_calculation_method',
                                            'tensile_stress_allowable', 'mass_baffle', 'mass_pumps',
                                            'mass_vapor_barrier', 'factor_usable_diameter', 'factor_volume_allowance']
    for key, value in dict_mod_config.items():
        # Check, if any value is 'None'.
        if key in necessary_values_for_liquid_hydrogen and value is None:
            # If any value is 'None', the tank design is impossible and the program aborted.
            runtime_output.critical('Error: Tank design not possible due to missing parameter in module configuration '
                                    + 'file (' + key + ')! '
                                    + 'Program aborted.')
            sys.exit(1)

    # Merge both data dictionaries.
    dict_design = dict_ac_exchange | dict_mod_config
    dict_design['energy_carrier'] = routing_dict['user_layer']

    """Calculate tanks."""
    # Tank design preparation.
    liquid_hydrogen_output_dict = tank_design_tu_berlin(paths_and_names, routing_dict, dict_design, runtime_output)
    # Tank calculation.
    liquid_hydrogen_output_dict = calculate_liquid_hydrogen_tanks(dict_design, liquid_hydrogen_output_dict,
                                                                  runtime_output)

    # Debug print.
    runtime_output.debug('Debug: The "method_liquid_hydrogen" function was successfully executed.')
    runtime_output.print('Tank design completed.')

    return liquid_hydrogen_output_dict
